/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:21 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "def.h"

extern boolean open_directory();
extern boolean err_directory();
extern boolean next_directory();
extern boolean is_ascii_file();
extern char *name_norm();
extern float *vector();

extern boolean verbose;


universe univ, *U = NULL;
int feat_dim = EMPTY;	/* if non-empty do not load dimension */

str100 inDataDir, dataFile, featNameFile, a_prioriFile;
str100 rootDataDir;
#define ROOT_DIR_ENV_VAR_NAME "TOOLDIAG_ROOT_DIR"
str100 outDataDir;
#define OUT_DIR_ENV_VAR_NAME "TOOLDIAG_DATA_DIR"
str100 gnuplot_exe;
#define GNUPLOT_EXE_ENV_VAR "GNUPLOT"


static str100 classFile;
static str80 linebuf, buf;
static char wheel[] = "|/-\\";
static boolean ascii_file = FALSE;

#define MINSAMPLE 2
#define MAXSTRLEN 100


#ifdef UNIX
static char slash = '/';
static char working_dir[] = "./";
#endif
#ifndef UNIX
static char slash = '\\';
static char working_dir[] = ".\\";
#endif


static void init_Class( C )
Clas *C;
{
 C->name[0] = '\0';
 C->numSampl = 0;
 C->S = NULL;
 init_extr_ms( &(C->stat) );
 C->a_priori_prob = 0.0;
}


static void init_universe()
{
 strcpy( U->name, "UNDEFINED" );
 U->nrClass = 0;
 U->sumSampl = 0;
 U->nrFeat = 0;
 init_extr_ms( &(U->stat) );
 U->nrSelFeat = 0;
 U->FSV = NULL;
 U->C = NULL;
 reset_norm();
}


void free_universe()
{
 int i;

 if( U == NULL )
   return;

 for( i = 0; i < U->nrClass; i++ )
 {
   free_extr_ms( &(U->C[i].stat) );
   FREE( U->C[i].S );
   init_Class( &(U->C[i]) );
 }
 free_extr_ms( &(U->stat) );
 FREE( U->FSV );
 FREE( U->C );
 init_universe();
 free_feat_names();
}


/* Add a slash at the end of the directory name if not already there */
static void add_slash( dirStr, maxlen )
char *dirStr;
int maxlen;
{
 int len;

 len = strlen( dirStr );
 if( dirStr[len-1] == slash )
   return;
 if( len-1 > maxlen )
   return;
 dirStr[len] = slash;
 dirStr[len+1] = '\0';
}


static void init_env()
{
 char *ev = NULL;

 ev = getenv(ROOT_DIR_ENV_VAR_NAME);
 /* if environment variable is not defined, use default value */
 if( ev == NULL )
 {
   strcpy( rootDataDir, working_dir );
   fprintf(stderr,"\n  Environment variable %s is not defined\n",
		ROOT_DIR_ENV_VAR_NAME );
   fprintf(stderr,"  Setting root directory to default \"%s\"\n",
		working_dir );
   fprintf(stderr,"  See the file 'setup.README' in the 'etc' directory\n");
   WAIT;
 }
 else
 {
   strcpy( rootDataDir, ev );
   add_slash( rootDataDir, 100 );
 }

 ev = getenv(OUT_DIR_ENV_VAR_NAME);
 /* if environment variable is not defined, use default value */
 if( ev == NULL )
 {
   strcpy( outDataDir, working_dir );
   fprintf(stderr,"\n  Environment variable %s is not defined\n",
		OUT_DIR_ENV_VAR_NAME );
   fprintf(stderr,"  Setting output data directory to default \"%s\"\n",
		working_dir );
   fprintf(stderr,"  See the file 'setup.README' in the 'etc' directory\n");
   WAIT;
 }
 else
 {
   strcpy( outDataDir, ev );
   add_slash( outDataDir, 100 );
 }
#ifndef UNIX
 /* read the environment variable for the gnuplot executable */
 ev = getenv( GNUPLOT_EXE_ENV_VAR );
 if( ev == NULL )
 {
   fprintf(stderr,"\n  Environment variable %s is not defined\n",
		GNUPLOT_EXE_ENV_VAR );
   fprintf(stderr,"  I need this variable to call the 'gnuplot' program" );
   WAIT;
 }
 else
   strcpy( gnuplot_exe, ev );
#endif
}


void init_all()
{
 init_random_generator();
 init_env();
 U = &univ;
 init_universe();
}


/*
 *  Given a class i, a sample j and a features selection vector FSV,
 *  fill in the variable 'smp' of length 'dim' with the respective
 *  values of the selected features
*/
void provideFV( smp, i, j, dim, FSV )
FeatVector smp;
int i, j, dim;
FeatSelectVector FSV;
{
 int k;

 for( k = 0; k < dim; k++ )
   smp[k] = U->C[i].S[j*U->nrFeat+FSV[k].rank];
}


static char sl[] = "------------------";
static char line[] =
        "----------------------------------------------------------------";


void showUniv( f )
FILE *f;
{
 int i; char *nn = NULL;

 nn = name_norm();
 fprintf( f, "\n%s     U N I V E R S E       %s\n", sl, sl );
 fprintf( f, "      Name                        : %25.25s\n", U->name );
 fprintf( f, "      Nr. of classes              : %25d\n", U->nrClass );
 fprintf( f, "      Nr. of all samples          : %25d\n", U->sumSampl );
 fprintf( f, "      Nr. of all      features    : %25d\n", U->nrFeat );
 if( nn != NULL )
   fprintf( f, "        Normalization of features : %25s\n", nn );
 fprintf( f, "      Nr. of selected features    : %25d\n", U->nrSelFeat );
 fprintf( f, "      %s%s%s\n", sl, sl, sl );
 fprintf( f, " CLASS                               NAME   Nr. of samples\n" );
 for( i = 0; i < U->nrClass; i++ )
   fprintf( f, " Nr. %3d : %30.30s   %8d\n",i+1,U->C[i].name,U->C[i].numSampl);
 fprintf( f, "%s\n\n", line );
}


/*
 * Load the a priori probabilities of the classes explicitly from a file
 */
#define SUM_EPSILON (0.01)
void load_a_priori_probs( fname )
char *fname;
{
 FILE *app = NULL;
 str100 buf;
 int i, nrClass;
 float p, sum = 0.0, *approb = NULL;

 app = fopen( fname, f_open_text_r );
 if( app == NULL )
   {printf("Cannot open %s! nothing done...", fname );gets(buf);return;}

 if( verbose )
   printf("\nLoading a priori probabilities from %s\n", fname );

 dataline( app, buf );
 sscanf( buf, "%d", &nrClass );
 if( nrClass != U->nrClass )
 {
  fprintf(stderr,"Nr. of classes %d\nin file %s differs from Nr. of classes %d",
	nrClass, fname, U->nrClass );
   fprintf(stderr,"\nIgnoring file %s", fname ); WAIT;
   fclose(app);
   return;
 }
 approb = vector( nrClass );
 for( i = 0; i < nrClass; i++ )
 {
   dataline( app, buf );
   sscanf( buf, "%f", &p );
   /* check plausibility */
   if( p <= 0.0 || p >=1.0 )
   {
     fprintf(stderr,"A priori probability %f for class %d\n", p, i+1);
     fprintf(stderr,"in file %s not plausible", fname );
     fprintf(stderr,"\nIgnoring file %s", fname ); WAIT;
     fclose(app); FREE(approb);
     return;
   }
   approb[i] = p;
   sum += p;
 }

 /* check plausibility */
 if( fabs(sum-1.0) > SUM_EPSILON )
 {
   fprintf(stderr,"Sum of a priori probabilities %f\n", sum );
   fprintf(stderr,"in file %s not plausible - Ignoring", fname ); WAIT;
   fclose(app); FREE(approb);
   return;
 }

 /* Copy the probabilities */
 for( i = 0; i < nrClass; i++ )
 {
   U->C[i].a_priori_prob = approb[i];
   if( verbose )
     printf("Explicit a priori probability is %6.2f%% for class %3d(%s)\n",
	100.0*approb[i], i+1, U->C[i].name );
 }

 FREE(approb);
 fclose( app );
}


boolean loadUnivDir()
{
 FILE *cf = NULL;
 int i, row, k, w = 0, dim, cptr, sumSampl = 0, offset = 1, strlen_w;
 boolean found, ok;
 str100 fileName;
 float value;

 if( verbose )
   { fprintf( stderr, "Searching sample file in %s\n", inDataDir ); }
 else
   { fprintf( stderr, "Searching directory...\n"); }
 /* first count the files */
 ok = open_directory( inDataDir );
 if( !ok || err_directory() )
   { printf("Cannot open directory %s ", inDataDir); WAIT; return(FALSE); }

 free_universe();
 strncpy( U->name, inDataDir, 100 );
 if( inDataDir[strlen(inDataDir)-1] != slash )
 {
   inDataDir[strlen(inDataDir)] = slash;
   inDataDir[strlen(inDataDir)+1] = '\0';
 }
 /* copy only the name of the data file to the universe name */
 i = strlen( inDataDir ) - 2;
 while( inDataDir[i] != slash && i > 0 )
   i--;
 if( i == 0 )
   offset = 0; 
 strncpy( U->name, &(inDataDir[i+offset]), strlen(&(inDataDir[i+offset]))-1 );
 U->name[strlen(&(inDataDir[i+offset]))-1] = '\0';

 strlen_w = strlen(wheel);
 do
 {
   found = next_directory( inDataDir, fileName );
   if( found )
     (U->nrClass)++;
 }
 while( found );
 close_directory();
 if( verbose )
   printf("Found %d files\n", U->nrClass );

 if( U->nrClass <= 0 || U->nrClass > MAXCLASS )
   { printf(" Number of classes invalid! Exitus...\n" ); exit(1); }
 /* allocate space for the universe */
 FREE( U->C );
 U->C = (Clas*)malloc( U->nrClass* sizeof( struct Clas_ ) );
 for( i = 0; i < U->nrClass; i++ )
   init_Class( &(U->C[i]) );

 /* scan through the files */
 ok = open_directory( inDataDir );
 if( !ok )
   { printf("Cannot open directory %s ", inDataDir); WAIT; return(FALSE); }
 i = 0;
 if( ! verbose )
   { fprintf( stderr, "Loading ... "); fflush( stderr ); }
 do
 {
   found = next_directory( inDataDir, fileName );
   if( found )
   {
     if( ! verbose )
       { fprintf( stderr,"%c\b", wheel[(w++) % strlen_w] ); fflush( stderr ); }

     strcpy( classFile, inDataDir ); 
     strcat( classFile, fileName );

     /* check the first file in the directory if all data is in ascii
  		1.) Everything ascii data
  		2.) Feature data in binary format
     */
     if( i == 0 )
       ascii_file = is_ascii_file( classFile );
     if( ascii_file == EMPTY )
       { printf("Cannot open %s! Exitus...\n", classFile ); exit(1); }
     if( ascii_file )
       cf = fopen( classFile, f_open_text_r ); 
     else
       cf = fopen( classFile, f_open_bin_r );
     if( cf == NULL )
       { printf("Cannot open %s! Exitus...\n", classFile ); exit(1); }
     if( verbose )
       printf("Scanning file: %s\n", fileName );
     /* read the name of the class */
     dataline( cf, linebuf ); sscanf( linebuf, "%s", U->C[i].name );
     if( linebuf[0] == '\0' ) EXIT("EMPTY CLASS NAME");
     /* groupname if exists */
     dataline( cf, linebuf );
     /* if the group name exits ignore it */
     cptr = 0;
     while( linebuf[cptr] == ' ' )
       cptr++;
     if( linebuf[cptr] >= '0' && linebuf[cptr] <= '9' )
       sscanf( linebuf, "%d", &dim );
     else
     {
       /* Feature vector dimension */
       if( feat_dim == EMPTY )
       {
         dataline( cf, linebuf ); sscanf( linebuf, "%d", &dim );
       }
       else
	 dim = feat_dim;
     } 
     /* check consistency */
     if( i == 0 )  /* first file ? */
     {
       U->nrFeat = dim;
     }
     else
       if( dim != U->nrFeat )
       {
         printf("Dim. of feature vector inconsistent: %d != %d! Exitus..\n",
		dim, U->nrFeat ); exit(1);
       }
     /* number of samples */
     dataline( cf, linebuf ); sscanf( linebuf, "%d", &(U->C[i].numSampl) );
     if( U->C[i].numSampl < MINSAMPLE )
     {
       printf("WARNING: class %d(%s) has only %d sample(s)",
                i+1, U->C[i].name, U->C[i].numSampl );
       printf(" - this might cause trouble!" ); WAIT;
     }
     sumSampl += U->C[i].numSampl;
     /* now allocate space for one class */
     dim = U->nrFeat * U->C[i].numSampl;
     FREE( U->C[i].S );
     U->C[i].S = (FeatVector) malloc(dim*sizeof(FeatVector*));
     CHKPTR(U->C[i].S);


     /* read the samples and update mean and standard deviation */
     for( row = 0; row < U->C[i].numSampl; row++ )
     {
       for( k = 0; k < U->nrFeat; k++ )
       {
         if( ascii_file )
           fscanf( cf, "%f", &value );
         else
           fread( &value, sizeof(float), 1, cf );
         /* printf("i=%d  row=%d  k=%d ascii=%d   value=%f\n",
		i, row, k, ascii_file, value ); DBG; /**/
         U->C[i].S[row*U->nrFeat+k] = value;
       }
     /* showFV( U->nrFeat, &(U->C[i].S[row*U->nrFeat]) ); /**/
     } 
     fclose( cf );
     i++; /* next class */
   }
 }
 while( found );
 if( ! verbose )
   printf("\n");
 
 U->sumSampl = sumSampl; 

 for( i = 0; i < U->nrClass; i++ )
 {
   U->C[i].a_priori_prob = (float)U->C[i].numSampl / (float)U->sumSampl;
   if( verbose )
     printf(" Estimated a priori probability of class %3d:  %5.2f%%\n",
		i+1, 100.0*U->C[i].a_priori_prob );
 }
 close_directory();
 set_all_samples_train();
 generate_stats( FALSE );
 return( TRUE );
}


boolean loadUnivFile()
{
 FILE *data = NULL;
 char **classNames = NULL, **buf1 = NULL;
 int *samplesPerClass = NULL, *buf2 = NULL;
 str100 nameBuf;
 int nrClass = 0, k, i, j, w = 0, c, dim, row, offset = 1, strlen_w;
 float dummy;
 FeatVector featBuf = NULL;

 /* look first type of file */
 ascii_file = is_ascii_file( dataFile );
 if( ascii_file == EMPTY )
   { printf("Cannot open %s ", dataFile ); WAIT; return(FALSE); }
 if( ascii_file )
   data = fopen( dataFile, f_open_text_r ); 
 else
   data = fopen( dataFile, f_open_bin_r );
 if( data == NULL )
   { printf("Cannot open %s ", dataFile ); WAIT; return(FALSE); }

 free_universe();

 strlen_w = strlen(wheel);
 /* copy only the name of the data file to the universe name */
 i = strlen( dataFile ) - 1;
 while( dataFile[i] != slash && i > 0 )
   i--;
 if( i == 0 )
   offset = 0;
 strncpy( U->name, &(dataFile[i+offset]), strlen(&(dataFile[i+offset])) );
 U->name[strlen(&(dataFile[i+offset]))] = '\0';

 fprintf( stderr, "Scanning file: %s\n", dataFile );
 /* feature vector dimension */
 if( feat_dim == EMPTY )
 {
   dataline( data, linebuf ); sscanf( linebuf, "%d", &dim );
 }
 else
   dim = feat_dim;
 /* count the classes */
 while( !feof( data ) )
 {
   k = 0; nameBuf[0] = '\0';
   while( k < dim && !feof( data ) )
   {
     if( ascii_file )
       fscanf( data, "%f", &dummy );
     else
     {
       fread( &dummy, sizeof(float), 1, data );/**/
       /* printf("dim=%d  dummy[%d]=>>>%f<<<", dim, k, dummy ); DBG;/**/
     }
     k++;
   }
   if( !feof( data ) )
   {
     if( ascii_file )
       fscanf( data, "%s", nameBuf );
     else
     {
       j = 0;
       do
       {
         fread( &(nameBuf[j]), sizeof(char), 1, data );
         /* printf("nameBuf[%d]=>>>%c<<<", j, nameBuf[j] ); DBG;/**/
         if( nameBuf[j] != '\n' )
           j++;
       }
       while( nameBuf[j] != '\n' && j < MAXSTRLEN );
       nameBuf[j] = '\0';
     }
     if( nameBuf[0] == '\0' ) EXIT("EMPTY CLASS NAME in 'loadUnivFile()'");
     /* check if the name already exist */
     c = 0;
     while( c < nrClass && !(strcmp(nameBuf,classNames[c]) == 0) )
       c++;
     if( c == nrClass )
     {
       /* new class */
       buf1 = (char**) malloc( (1+nrClass) * sizeof(char*) );
       buf2 = (int*) malloc( (1+nrClass) * sizeof(int) );
       /* copy old values */
       for( i = 0; i < nrClass; i++ )
       {
         buf1[i] = (char*) malloc( (1+strlen(classNames[i])) * sizeof(char) );
         strcpy( buf1[i], classNames[i] );
         FREE( classNames[i] );
         buf2[i] = samplesPerClass[i];
       }
       /* insert new value */
       buf1[nrClass] = (char*) malloc( (1+strlen(nameBuf)) * sizeof(char) );
       strcpy( buf1[nrClass], nameBuf );
       buf2[nrClass] = 1;
       FREE( classNames ); FREE( samplesPerClass );
       classNames = buf1; samplesPerClass = buf2; buf1 = NULL; buf2 = NULL;
       nrClass++;
     }
     else
       (samplesPerClass[c])++;
   }
 }
 fclose( data );
 for( i = 0; i < nrClass; i++ )
   U->sumSampl += samplesPerClass[i];
 U->nrFeat = dim;
 U->nrClass = nrClass;
 if( U->nrClass <= 0 || U->nrClass > MAXCLASS )
   { printf(" Number of classes invalid! Exitus...\n" ); exit(1); }
 /* allocate space for the universe */
 FREE( U->C );
 U->C = (Clas*)malloc( U->nrClass* sizeof( struct Clas_ ) );
 for( i = 0; i < U->nrClass; i++ )
 {
   init_Class( &(U->C[i]) );
   strcpy( U->C[i].name, classNames[i] );
   FREE( classNames[i] );

   /* now allocate space for one class */
   FREE( U->C[i].S );
   U->C[i].S =
      (FeatVector) malloc(U->nrFeat * samplesPerClass[i] * sizeof(FeatVector*));
   CHKPTR(U->C[i].S);
 }
 FREE( classNames );

 /* open file again to read the feature values */
 if( ascii_file )
   data = fopen( dataFile, f_open_text_r);
 else
   data = fopen(dataFile, f_open_bin_r);

 if( data == NULL )
   { printf("Cannot open %s! Exitus...\n", dataFile ); exit(1); }

 featBuf = (FeatVector) malloc(sizeof(FeatVector*) * U->nrFeat);
 /* feature vector dimension */
 if( feat_dim == EMPTY )
   { dataline( data, linebuf ); sscanf( linebuf, "%f", &dummy ); }

 fprintf( stderr, "Loading ... "); fflush( stderr );
 for( i = 0; i < U->sumSampl; i++ )
 {
   fprintf( stderr, "%c\b", wheel[(w++) % strlen_w] ); fflush( stderr );

   if( ascii_file )
   {
     for( k = 0; k < U->nrFeat; k++ )
       fscanf( data, "%f", &(featBuf[k]) );
     fscanf( data, "%s", nameBuf );
   }
   else
   {
     fread( featBuf, sizeof(float), U->nrFeat, data );
     j = 0;
     do
     {
       fread( &(nameBuf[j]), sizeof(char), 1, data );
       if( nameBuf[j] != '\n' )
         j++;
     }
     while( nameBuf[j] != '\n' && j < MAXSTRLEN );
     nameBuf[j] = '\0';
   }
   c = 0;
   while( c < U->nrClass && !(strcmp(nameBuf,U->C[c].name) == 0) )
     c++;
   if( c == U->nrClass )
     { printf("loadUnivFile> Cannot find class %s\n Exitus...\n",nameBuf );
       exit(1); }

   /* copy the sample buffer to the class */
   row = U->C[c].numSampl;
   for( k = 0; k < U->nrFeat; k++ )
     U->C[c].S[row*U->nrFeat+k] = featBuf[k];
   (U->C[c].numSampl)++;
 }
 fprintf( stderr, "\n");

 for( i = 0; i < U->nrClass; i++ )
 {
   U->C[i].a_priori_prob = (float)U->C[i].numSampl / (float)U->sumSampl;
   if( verbose )
     printf(" Estimated a priori probability of class %3d:  %5.2f%%\n",
		i+1, 100.0*U->C[i].a_priori_prob );
   if( U->C[i].numSampl != samplesPerClass[i] )
     { printf("loadUnivFile> Inconsistent data\n Exitus...\n");
       printf("U->C[%d].numSampl=%d != samplesPerClass[%d]=%d\n",
	i, U->C[i].numSampl, i, samplesPerClass[i] ); exit(1); }
   if( U->C[i].numSampl < MINSAMPLE )
   {
     printf("WARNING: class %d(%s) has only %d sample(s)",
                i+1, U->C[i].name, U->C[i].numSampl );
     printf(" - this might cause trouble! ..."); gets(buf);
   }
 }
 FREE( samplesPerClass );
 FREE( featBuf );
 fclose( data );
 set_all_samples_train();
 generate_stats( FALSE );
 return( TRUE );
}



void loadUniv()
{
 boolean ok = FALSE, load_fnames;
 char featNameFile[101];

 while( ! ok )
 {
   printf("Load data from (f)ile or from (d)irectory? ");
   gets( linebuf );
   switch( linebuf[0] )
   {
     case '\0' : return;
     case 'f': case 'F': printf("Load data from file? ");
       gets( dataFile ); ok = loadUnivFile(); break;
     case 'd': case 'D': printf("Load data from directory? ");
       gets( inDataDir ); ok = loadUnivDir(); break;
     default: printf("Unknown option - nothing done..."); gets(linebuf); break;
   }
   if( ok )
   {
     printf("Load file with the feature names?");
     get_boolean( TRUE, &load_fnames );
     if( load_fnames )
     {
       printf("Load feature names from? ");
       gets( featNameFile );
       if( featNameFile[0] != '\0' )
         load_feat_names( featNameFile );
     }
   }
 }
}
