/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:21 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include "def.h"
#include "ftsldef.h"

extern float select_univariate_minerr();
extern float selectMultivariate();



extern universe *U;


#define MIN_DISPLAY 20

static FeatSelectVector FSV = NULL;


static void selectUnivariateMAHALANOBIS( i, c, a, b )
int i;
float *c;
int a, b;
{
 float mean1, mean2, stddev1, stddev2, m, s, pd;

 mean1 = U->C[a].stat.mean[i]; stddev1 = U->C[a].stat.stddev[i]; 
 mean2 = U->C[b].stat.mean[i]; stddev2 = U->C[b].stat.stddev[i]; 
 if( mean1 == mean2 )
   pd = 0.0;
 else
 {
   m = ( mean1 - mean2 );
   s = (stddev1*stddev1 + stddev2*stddev2) / 2.0;
   pd = m * m / s;
 }
 *c = pd;
} 


static void selectUnivariateNORMMEANDIST( i, c, a, b )
int i;
float *c;
int a, b;
{
 float mean1, mean2, stddev1, stddev2, quot, pd;

 mean1 = U->C[a].stat.mean[i]; stddev1 = U->C[a].stat.stddev[i]; 
 mean2 = U->C[b].stat.mean[i]; stddev2 = U->C[b].stat.stddev[i]; 
 if( mean1 == mean2 )
 {
   printf("For feature %d the two classes %d and %d", i+1, a, b );
   printf(" have the same mean %f\n", mean1 );
   printf("Setting the criterion to -INFINITY...\n");
   pd = -INFINITY;
 }
 else
 {
   quot = (stddev1 + stddev2) / ( mean1 - mean2 );
   pd = 1.0-quot*quot;
 }
 *c = pd;
} 


static float selectUnivariateAverageClass( i, crit )
int i, crit;
{
 int a, b;
 float Jab, merit = 0.0;

 for( a = 0; a < U->nrClass-1; a++ )
 {
   for( b = a+1; b < U->nrClass; b++ )
   {
     switch( crit )
     {
       case MAHALANOBIS : selectUnivariateMAHALANOBIS( i, &Jab, a, b ); break;
       case NORMMEANDIST : selectUnivariateNORMMEANDIST( i, &Jab, a, b ); break;
     }
     merit += U->C[a].a_priori_prob * U->C[b].a_priori_prob * Jab;
   }
 }
 /* printf("Returning merit = %lf", merit ); DBG; /**/
 return( merit );
}


static float selectUnivariate( i, group, crit )
int i, group, crit;
{
 float c;
 feature feat;

 feat.rank = i; feat.crit = (float)EMPTY;

 switch( group )
 {
   case MIN_EST_ERR : c = select_univariate_minerr( &feat ); break;
   case INTER_CLASS_DISTANCE: ;
     c = selectMultivariate( group, crit, &feat, 1 ); break;
   case PROBABILISTIC_DISTANCE:
     switch( crit )
     {
       case NORMMEANDIST : c = selectUnivariateAverageClass( i, crit ); break;
       case MAHALANOBIS :  c = selectUnivariateAverageClass( i, crit ); break;
       case CHERNOFF :
       case BHATTACHARYYA :
       case BHATTACHARYYA_MATUSITA :
       case DIVERGENCE :
       case PATRICK_FISHER :
         c = selectMultivariate( group, crit, &feat, 1 );
         break;
       default: fprintf(stderr,"Unknown criterion- exit\n"); exit(1);
     }
     break;
   default: EXIT("Unknown criterion group");
 }
 return( c );
}


/* make it compatible to all compilers */
/* Note: this routine sorts in ascending order */
static int order_feats( feat1, feat2 )
#ifdef CONSTANT
const
#endif
void *feat1, *feat2;
{
 feature *f1, *f2;

 f1 = (feature*) feat1; f2 = (feature*) feat2;
 if( f1->crit < f2->crit ) return((int)1);
 if( f1->crit > f2->crit ) return((int)-1);
 return((int)0);
}


void selectFeaturesBF( group, crit )
int group, crit;
{
 int i, nrBest, nrSelectFeat; float unicrit;

 FSV = (feature*) malloc( sizeof(feature) * U->nrFeat ); CHKPTR( FSV );

 /* for each individual feature calculate the univariate criterion */
 for( i = 0; i < U->nrFeat; i++ )
 {
   printf("Analyzing feature nr. %4d of %4d\r", i+1, U->nrFeat);fflush(stdout);
   unicrit = selectUnivariate( i, group, crit );
   FSV[i].rank = i; FSV[i].crit = unicrit;
 }
 printf("\n\n");

 /* sort the features following the criterion */
 printf("Sorting..."); fflush( stdout );
 qsort( FSV, U->nrFeat, sizeof(feature), order_feats );
 printf(" done.\n");

 if( U->nrFeat > MIN_DISPLAY )
 {
   printf("Want to see the best ? features (1-%d): ", U->nrFeat );
   get_d_range( &nrBest, 1, U->nrFeat, LEFT_CLOSED__RIGHT_CLOSED );
   show_selected_feats( stdout, FSV, nrBest );
 }
 else
   show_selected_feats( stdout, FSV, U->nrFeat );
 printf("Want to select the best ? features (1-%d): ", U->nrFeat );
 get_d_range( &nrSelectFeat, 1, U->nrFeat, LEFT_CLOSED__RIGHT_CLOSED );
 U->nrSelFeat = nrSelectFeat;
 FREE( U->FSV );
 U->FSV = (feature*) malloc( U->nrSelFeat * sizeof(struct feature_) );
 for( i = 0; i < U->nrSelFeat; i++ )
   copy_feature( &(FSV[i]), &(U->FSV[i]) );
 show_selected_feats( stdout, U->FSV, U->nrSelFeat );
 FREE( FSV );
}
