/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:22 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "def.h"
#include "matrxdef.h"
#include "clasdef.h"

extern int induceMLP();
extern int induceMULGAUSS();

extern boolean is_ascii_file();
extern boolean is_directory();
extern int class1testNEAR_NEI();
extern int class1testLINMACH();
extern int class1testMULGAUSS();
extern int class1testRBF();
extern int class1testPARZEN();
extern int class1testQ_STAR();
extern int class1testMLP();
extern int class1testYUC();

extern int *int_vector();
extern universe *U;
extern int NNK;  /* The nearest neighbor K */
extern boolean *train_smp, *test_smp;
extern int nrSamplesTest, nrSamplesTrain;
extern int feat_dim;
extern str100 dataFile, outDataDir;


int num_classifier_model = NUM_CLASS_MODELS;
boolean verbose_reason = FALSE;
/* Allow rejection option in classification */
boolean rejection_opt = FALSE;

#define MAXSTRLEN 100
static str80 buf;
static str100 benchFile, nameBuf;
static char rejStr[17] = "REJECTED/UNKNOWN";

/* Classification protocol to file */
FILE *cpf = stdout;
/* For unclassified data one cannot calculate any error rates */
boolean have_label = TRUE;


static boolean show_confusion_matrix = FALSE, trace = FALSE;
static Matrix confMatRel, confMatAbs;
static int *nrSmpExpected = NULL;

static char nnStr[] = "-Nearest Neighbor, using raw data";
struct class_models_ class_model[] = {
  { NEAREST_NEIGHBOR,		"",					FALSE },
  { LINEAR_MACHINE,		"Linear Machine",			FALSE },
  { MULTIVARIATE_GAUSSIAN,	"Quadratic Gaussian Classifier",	FALSE },
  { RBF,			"Radial Basis Function Network",	FALSE },
  { PARZEN,			"Parzen window",			FALSE },
  { Q_STAR,			"Q* algorithm",				FALSE },
  { MLP,			"Multilayer Perceptron",                FALSE },
  { YUC,			"\"Your classifier\"",			FALSE }
};

int classifier_model = NEAREST_NEIGHBOR; /**/
/* int classifier_model = LINEAR_MACHINE; /**/
/* int classifier_model = MULTIVARIATE_GAUSSIAN; /**/
/* int classifier_model = RBF; /**/
/* int classifier_model = PARZEN; /**/
/* int classifier_model = Q_STAR; /**/
/* int classifier_model = MLP; /**/
/* int classifier_model = YUC; /**/


char *class_model_name()
{
 return( class_model[classifier_model].name );
}



void nameClassModel()
{
 str80 strBuf;

 sprintf( strBuf, "%d", NNK );
 strcat( strBuf, nnStr );
 strcpy( class_model[NEAREST_NEIGHBOR].name, strBuf );
}


void free_induced()
{
 switch( classifier_model )
 {
   case NEAREST_NEIGHBOR :	free_inducedNEAR_NEI(); break;
   case LINEAR_MACHINE :	free_inducedLINMACH(); break;
   case MULTIVARIATE_GAUSSIAN :	free_inducedMULGAUSS(); break;
   case RBF :			free_inducedRBF(); break;
   case PARZEN :		free_inducedPARZEN(); break;
   case Q_STAR :		free_inducedQ_STAR(); break;
   case MLP :			free_inducedMLP(); break;
   case YUC :			free_inducedYUC(); break;
   default: fprintf(stderr,"What the heck is classifier model %d? - exit...\n",
                classifier_model); exit(1);
 }
 class_model[classifier_model].induced = FALSE;
}

static char sl[] = "-------------";
static char line[] =
	"--------------------------------------------------------------";

void show_class_model( f )
FILE *f;
{
 if( classifier_model == MLP && U->nrSelFeat == 0 )
   return;

 nameClassModel();
 fprintf(f, "%s   C L A S S I F I E R  M O D E L   %s\n", sl, sl);
 fprintf( f, "\tNAME:\t\t%s\n", class_model[classifier_model].name );
 switch( classifier_model )
 {
   case NEAREST_NEIGHBOR :	show_NEAR_NEI( f ); break;
   case LINEAR_MACHINE :	show_LINMACH( f ); break;
   case MULTIVARIATE_GAUSSIAN :	show_MULGAUSS( f ); break;
   case RBF :			show_RBF( f ); break;
   case PARZEN :		show_PARZEN( f ); break;
   case Q_STAR :		show_Q_STAR( f ); break;
   case MLP :			show_MLP( f ); break;
   case YUC :			show_YUC( f ); break;
   default: fprintf(stderr,"What the heck is classifier model %d? - exit...\n",
                classifier_model); exit(1);
 }
 fprintf( f, "%s\n", line );
}


void set_class_model()
{
 int i, cm, old_classifier;
 boolean set;

 do
 {
   set = TRUE; nameClassModel();
   printf("%s   SPECIFICATION OF CLASSIFIER MODEL   %s\n", sl, sl);
   for( i = 0; i < NUM_CLASS_MODELS; i++ )
   {
     printf("(%d) %s", i+1, class_model[ i ].name );
     if( classifier_model == i )
       printf("\t\t\t<--- Current");
     printf("\n");
   }
   printf("Choice: "); gets(buf);
   if( isdigit(buf[0]) )
   {
     sscanf(buf,"%d", &cm );
     if( cm >= 1 && cm <= NUM_CLASS_MODELS )
     {
       old_classifier = classifier_model;
       free_induced();
       classifier_model = cm - 1;
       class_model[old_classifier].induced = FALSE;
       switch( classifier_model )
       {
         case NEAREST_NEIGHBOR :	initNEARNEI(); break;
         case LINEAR_MACHINE :		initLINMACH(); break;
         case MULTIVARIATE_GAUSSIAN :	initMULGAUSS(); break;
         case RBF :			initRBF(); break;
         case PARZEN :			initPARZEN(); break;
         case Q_STAR :			initQ_STAR(); break;
	 case MLP :			initMLP(); break;
         case YUC :			initYUC(); break;
         default:
           fprintf(stderr,"What the heck is classifier model %d? - exit...\n",
                  classifier_model); exit(1);
       }
       nameClassModel();
     }
     else
       printf("\n*** Unknown classifier option!\n\n");
   }
   else
   {
     switch( buf[0] )
     {
       case '\0' : ; break;
       case  '?' :  help( MENU_CLASSIFIER, buf ); set = FALSE; break;
       default: printf("\n*** Unknown option!\n\n");  break;
     }
   }
 }
 while( ! set );
}


void induce( nrSelFeat, FSV, interact )
int nrSelFeat;
FeatSelectVector FSV;
boolean interact;
{
 boolean ok = TRUE;

 if( class_model[classifier_model].induced )
 {
   free_induced();
   class_model[classifier_model].induced = FALSE;
 }
 switch( classifier_model )
 {
   case NEAREST_NEIGHBOR :	induceNEAR_NEI( nrSelFeat, FSV, interact );
     break;
   case LINEAR_MACHINE :	induceLINMACH( nrSelFeat, FSV, interact );
     break;
   case MULTIVARIATE_GAUSSIAN :	ok = induceMULGAUSS( nrSelFeat, FSV, interact );
     break;
   case RBF :			induceRBF( nrSelFeat, FSV, interact );
     break;
   case PARZEN :		inducePARZEN( nrSelFeat, FSV, interact );
     break;
   case Q_STAR :		induceQ_STAR( nrSelFeat, FSV, interact );
     break;
   case MLP :			ok = induceMLP( nrSelFeat, FSV, interact );
     break;
   case YUC :			induceYUC( nrSelFeat, FSV, interact );
     break;
   default: fprintf(stderr,"What the heck is classifier model %d? - exit...\n",
                classifier_model); exit(1);
 }
 class_model[classifier_model].induced = ok;
 /* debug_alloc_spy();	/**/
}


void induceDEMO()
{
 initQ_STAR();
 free_tt_lookups();
 set_all_samples_train();
 induceQ_STAR( U->nrSelFeat, U->FSV, TRUE );
 free_inducedQ_STAR(); 
 free_tt_lookups();
}


static int class1( nrSelFeat, FSV, smp, name )
int nrSelFeat;
FeatSelectVector FSV;
FeatVector smp;
char *name;
{
 int res_class;
 
 switch( classifier_model )
 {
   case NEAREST_NEIGHBOR:
	res_class = class1testNEAR_NEI(nrSelFeat, FSV, smp, name );
	break;
   case LINEAR_MACHINE:
	res_class = class1testLINMACH(nrSelFeat, FSV, smp, name );
	break;
   case MULTIVARIATE_GAUSSIAN:
	res_class = class1testMULGAUSS(nrSelFeat, FSV, smp, name );
	break;
   case RBF:
	res_class = class1testRBF(nrSelFeat, FSV, smp, name );
	break;
   case PARZEN:
	res_class = class1testPARZEN(nrSelFeat, FSV, smp, name );
	break;
   case Q_STAR:
	res_class = class1testQ_STAR( nrSelFeat, FSV, smp, name );
	break;
   case MLP:
	res_class = class1testMLP( nrSelFeat, FSV, smp, name );
	break;
   case YUC:
	res_class = class1testYUC( nrSelFeat, FSV, smp, name );
	break;
   default: fprintf(stderr,"What the heck is classifier model %d? - exit...\n",
	classifier_model); exit(1);
 }
 return( res_class );
}


int classify_one_testSample( i, j, s, nrSelFeat, FSV )
int i, j, s;	/* class, # sample of class, global # of sample */
int nrSelFeat;
FeatSelectVector FSV;
{
 FeatVector smp = NULL;
 char *name;	/* the expected name of the sample */
 int k, res_class;

 if( ! class_model[classifier_model].induced )
   EXIT("Classifier is not induced");

 /* EMPTY is used for the RESUBSTITUTION ESTIMATE */
 if( ! test_smp[s] && s != EMPTY )
   EXIT("Sample is no test sample");

 smp = (FeatVector)malloc(nrSelFeat*sizeof(FeatVector*)); CHKPTR(smp);
 for( k = 0; k < nrSelFeat; k++ )
   smp[k] = U->C[i].S[j*U->nrFeat+FSV[k].rank];
 name = U->C[i].name;

 res_class = class1(nrSelFeat, FSV, smp, name );

 FREE(smp);
 return( res_class );
}


static int classify_one_sample_from_file( rawSample, name )
FeatVector rawSample;
char *name;
{
 FeatVector sample = NULL;
 int k, res_class;

 if( ! class_model[classifier_model].induced )
   EXIT("Classifier is not induced");

 sample = (FeatVector) malloc(sizeof(FeatVector*) * U->nrSelFeat);
 CHKPTR(sample);

 for( k = 0; k < U->nrSelFeat; k++ )
   sample[k] = rawSample[ U->FSV[k].rank ];

 res_class = class1( U->nrSelFeat, U->FSV, sample, name );

 FREE(sample);
 return( res_class );
}


void supervised_learn_classifier( show )
boolean show;
{
 if( U->nrSelFeat == 0 )
   { printf("Select features first..."); gets( buf ); return; }

 if( U->nrClass < 2 )
   { printf("Induction only with two or more classes"); WAIT; return; }

 free_tt_lookups();
 set_all_samples_train();
 free_induced();
 if( show )
   { show_class_model( stdout ); printf("..."); gets(buf); }
 induce( U->nrSelFeat, U->FSV, TRUE );
}


void init_confMat()
{
 int i;

 init_Matrix( &confMatRel ); init_Matrix( &confMatAbs );
 confMatRel.row = U->nrClass;
 confMatRel.col = U->nrClass;
 if( rejection_opt )
   (confMatRel.col)++;
 Matrix_Alloc( &confMatRel );
 Reset_Matrix( &confMatRel );
 nrSmpExpected = int_vector( U->nrClass );
 for( i = 0; i < U->nrClass; i++ )
   nrSmpExpected[i] = 0;
}


static void free_confMat()
{
 Matrix_Free( &confMatRel ); Matrix_Free( &confMatAbs );
 FREE(nrSmpExpected);
}


/*
 *  Show a confusion matrix during a classification
 *   The rows are the expected classes, the columns the
 *   recognized classes
*/
void show_confMat( f )
FILE *f;
{
 int i, j, sumExp = 0, sum_detected;
 Matrix *Mabs, *Mrel;

 Copy_Matrix( &confMatRel, &confMatAbs );
 for( i = 0; i < confMatRel.row; i++ )
 {
   sum_detected = 0;
   sumExp += nrSmpExpected[i];
   for( j = 0; j < confMatRel.col; j++ )
     sum_detected += (int)confMatAbs.Elem[i*confMatRel.col+j];
   for( j = 0; j < confMatRel.col; j++ )
     confMatRel.Elem[i*confMatRel.col+j] /= (MatElem)sum_detected;
 }
   
 Mrel = &confMatRel; Mabs = &confMatAbs;

 fprintf(f,"\n  ~~~~~~  C O N F U S I O N  M A T R I X  ~~~~~~\n");
 fprintf(f," ROW: expected class  ---  COLUMN: detected class\n");
 fprintf(f,"# test samples: %d\nABSOLUTE:   ", sumExp );
 for( j = 0; j < Mabs->col; j++)
   fprintf(f,"%.10s  ", U->C[j].name );
 if( rejection_opt )
   fprintf( f, "%.18s", rejStr );
 fprintf(f,"\n");
 for( i = 0; i < Mabs->row; i++ )
 {
   fprintf(f,"%.10s", U->C[i].name );
   for( j = 0; j < Mabs->col; j++)
     fprintf(f,"%12d", (int)(Mabs->Elem[i*Mabs->col+j]));
   fprintf(f,"\n");
 }
 fprintf(f,"\nRELATIVE:   ");
 for( j = 0; j < Mrel->col; j++)
   fprintf(f,"%.10s  ", U->C[j].name );
 if( rejection_opt )
   fprintf( f, "%.18s", rejStr );
 fprintf(f,"\n");
 for( i = 0; i < Mrel->row; i++ )
 {
   fprintf(f,"%.10s", U->C[i].name );
   for( j = 0; j < Mrel->col; j++)
     fprintf(f,"%11.2f%%", 100.0*(float)Mrel->Elem[i*Mrel->col+j]);
   fprintf(f,"\n");
 }
 fprintf(f,"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
 free_confMat();
}



void update_confMat( res_class, nameBuf )
int res_class;
char *nameBuf;
{
 boolean found;
 int c, exp_class;

 /* get the index of the expected class */
 found = FALSE; c = 0;
 while( ! found && c < U->nrClass )
 {
   if( (strcmp(U->C[c].name,nameBuf)==0) )
   {
     found = TRUE;
     exp_class = c;
   }
   c++;
 }
 if( ! found )
 {
   fprintf(stderr,"WARNING: Never seen class \"%s\" in training set\n",nameBuf);
   fprintf(stderr,"\tConfusion matrix ignores this sample!!!"); WAIT;
   return;
 }

 (nrSmpExpected[exp_class])++;
 if( res_class == CORRECT_CLASS )
   (confMatRel.Elem[exp_class*confMatRel.col+exp_class])++;
 else
 {
   if( res_class == UNDEFINED_CLASS || res_class == REJECT_CLASS )
   {
     if( ! rejection_opt )
       EXIT("Reject/Unknown flag must be set");
     (confMatRel.Elem[exp_class*confMatRel.col+confMatRel.col-1])++;
   }
   else
     (confMatRel.Elem[exp_class*confMatRel.col+res_class])++;
 }
}



static boolean init_classify_from_file()
{
 boolean do_induce;
 char classProtFile[101], name[81];

 /* set all existing samples as training samples as test data is read from
	an outside file */
 if( ! class_model[classifier_model].induced )
 {
   show_class_model( stdout );
   printf("No classifier available - Induce it ?" );
   get_boolean( TRUE, &do_induce );
   if( do_induce )
     supervised_learn_classifier( FALSE );
   else
   {
     printf("Doing nothing..."); gets(buf);
     return( FALSE );
   }
 }
 if( ! class_model[classifier_model].induced )
   return( FALSE );

 strcpy( classProtFile, outDataDir );
 printf("Saving classification result to (<Enter> for screen):\n\t%s",
        classProtFile );
 gets( name );

 if( name[0] == '\0' )
   cpf = stdout;
 else
 {
   strcat( classProtFile, name );
   cpf = fopen( classProtFile, f_open_text_w );
   if( cpf == NULL )
   { printf("Cannot open \"%s\", output to screen.\n", classProtFile ); WAIT; }
 }

 printf("Verbose the reasoning process?" );
 get_boolean( verbose_reason, &verbose_reason );

 if( verbose_reason && cpf == stdout )
 {
   printf("Trace the reasoning process?" );
   get_boolean( trace, &trace );
 }

 if( have_label )
 {
   printf("Show confusion matrix?" );
   get_boolean( show_confusion_matrix, &show_confusion_matrix );

   if( show_confusion_matrix )
     init_confMat();
 }

 switch( classifier_model )
 {
   case NEAREST_NEIGHBOR : 
	fprintf(cpf,"\n#### Using %d-Nearest-Neighbor Classifier ####\n",NNK);
	break;
   case LINEAR_MACHINE : 
	fprintf(cpf,"\n#### Using Linear Machine ####\n" ); break;
   case MULTIVARIATE_GAUSSIAN :
	fprintf(cpf,"\n#### Using Bayes Classifier ####\n" ); break;
   case RBF :
	fprintf(cpf,"\n#### Using RBF Classifier ####\n" ); break;
   case PARZEN :
	fprintf(cpf,"\n#### Using Parzen Window Classifier ####\n" ); break;
   case Q_STAR :
	fprintf(cpf,"\n#### Using Q* Classifier ####\n" ); break;
   case MLP :
	fprintf(cpf,"\n#### Using Multilayer Perceptron Classifier ####\n" );
	break;
   case YUC :
	fprintf(cpf,"\n#### Using my own Classifier ####\n" ); break;
   default : fprintf(stderr,"What is classifier %d? -exit...\n",
		classifier_model ); exit(1);
 }
 return( TRUE );
}


static boolean close_classify_from_file()
{
 /* Closing the classification protocol file */
 if( cpf != stdout && cpf != NULL ) { fclose(cpf); cpf = NULL; }
 have_label = TRUE;
 cpf = stdout;
}


void classify_from_file()
{
 FILE *bench = NULL;
 boolean ascii_file;
 FeatVector featBuf = NULL;
 int j, k, featDim, nrSmp, nrErr, res_class;
 float err;

 if( U->nrSelFeat == 0 )
   { printf("Load universe or select features first..."); gets( buf ); return; }

 if( U->nrClass < 2 )
   { printf("Induction only with two or more classes"); WAIT; return; }


 printf("Load data to be classified from file? "); gets( benchFile );

 if( is_directory( benchFile ) )
 {
   printf("'%s' is a directory. Only accepting files...", benchFile );
   gets(buf); return;
 }

 ascii_file = is_ascii_file( benchFile );
 if( ascii_file )
   bench = fopen( benchFile, f_open_text_r ); 
 else
   bench = fopen( benchFile, f_open_bin_r );
 if( bench == NULL )
   { printf("Cannot open file '%s'!...", benchFile ); gets(buf); return; }

 printf("Do data samples have class labels, i.e. is the data pre-classified? ");
 get_boolean( have_label, &have_label );

 if( ! have_label )
 {
   printf("Be careful that your file with the test data really has");
   printf(" no labels.\nOtherwise inconsistent situations may ocurr.");
   WAIT;
 }

 if( ! init_classify_from_file() )
 {
   close_classify_from_file(); 
   return;
 }


 /* feature vector dimension */
 if( feat_dim == EMPTY )
 {
   dataline( bench, buf );
   sscanf( buf, "%d", &featDim );
   if( featDim != U->nrFeat )
     {printf("Universe and file have different feature dimensions: %d != %d...",
        U->nrFeat, featDim  ); gets( buf ); fclose( bench ); return; }
 }
 else
 {
   printf("\n\t");
   printf("Number of features was explicitly specified as %d for file\n\t%s\n",
	feat_dim, dataFile );
   printf("\tAssuming that test file is also without dimension in 1st line!\n");
 }

 featBuf = (FeatVector)malloc(sizeof(FeatVector*) * U->nrFeat); CHKPTR(featBuf);

 nrSmp = 0; nrErr = 0;
 while( !feof( bench ) )
 {
   for( k = 0; k < U->nrFeat; k++ )
   {
     if( ascii_file )
       fscanf( bench, "%f", &(featBuf[k]) );
     else
       fread( &(featBuf[k]), sizeof(float), 1, bench );
   }
   if( have_label )
   {
     if( ascii_file )
       fscanf( bench, "%s", nameBuf );
     else
     {
       j = 0;
       do
       {
         fread( &(nameBuf[j]), sizeof(char), 1, bench );
         if( nameBuf[j] != '\n' )
           j++;
       }
       while( nameBuf[j] != '\n' && j < MAXSTRLEN );
       nameBuf[j] = '\0';
     }
     if( nameBuf[0] == '\0' )   /* empty name, error */
        { fprintf( stderr, "Found an empty class name. Exit...\n"); exit(1); }
   }
   else
     nameBuf[0] = '\0';

   if( !feof( bench ) )
   {
     res_class = classify_one_sample_from_file( featBuf, nameBuf );
     nrSmp++;

     if( have_label )
     {
       if( show_confusion_matrix )
         update_confMat( res_class, nameBuf );
       if( res_class != CORRECT_CLASS )
         nrErr++;
       fprintf( cpf, "Sample nr. %d --- Current mean error = %7.2f%%",
		 nrSmp, (float)nrErr * 100.0 / (float)nrSmp ); 
       if( cpf != stdout )
         fprintf( cpf, "\n" );
       else
         fprintf( cpf, "\r" );
     }
     else
     {
       if( res_class != UNDEFINED_CLASS )
         fprintf( cpf, "Recognized sample # %d as class %d = %s\n",
		nrSmp, res_class+1, U->C[res_class].name );
       else
         fprintf( cpf, "Recognized sample # %d as UNKNOWN class\n", nrSmp );
     }
     if( verbose_reason && trace && cpf == stdout ) { NL; WAIT };
   }
 }
 BLANK_LINE; NL;
 fclose( bench );
 FREE( featBuf );

 if( have_label )
 {
   err = (float)nrErr / (float)nrSmp;
   fprintf(cpf,"\n --- RESULT for benchmark file %s:\n\n", benchFile );
   fprintf(cpf," ERROR RATE = %6.2f%% --- ACCURACY %6.2f%%",
	100.0 * err, 100.0 * (1.0-err) );
   fprintf(cpf," ---  Recognized %d of %d\n\n", nrSmp-nrErr, nrSmp );
   if( show_confusion_matrix )
   {
     show_confMat( cpf );
     NL;
   }
 }
 verbose_reason = FALSE;
 close_classify_from_file();
}


/**********  H E L P  *****************/

void helpMENU_CLASSIFIER( cmd )
char *cmd;
{
 int item;

 sscanf( &(cmd[1]), "%d", &item );
 item--;        /* discount one since in the menu it is one more */
 switch( item )
 {
   case NEAREST_NEIGHBOR:
     item_help("NEAREST_NEIGHBOR");
     item_bibRef( "DUDA73" );
     item_bibRef( "DEVIJVER82" );
     break;

   case LINEAR_MACHINE:
     item_help("LINEAR_MACHINE");
     item_bibRef( "DUDA73" );
     break;

   case MULTIVARIATE_GAUSSIAN:
     item_help("MULTIVARIATE_GAUSSIAN");
     item_bibRef( "DUDA73" );
     item_bibRef( "DEVIJVER82" );
     break;

   case RBF:
     item_help("RBF");
     item_bibRef( "HAYKIN94");
     item_bibRef( "POWELL87");
     item_bibRef( "POGGIO89");
     break;

   case PARZEN:
     item_help("PARZEN");
     item_bibRef( "PARZEN62");
     item_bibRef( "DUDA73");
     item_bibRef( "SCHALKOFF92");
     break;

   case Q_STAR:
     item_help("Q_STAR");
     item_bibRef( "RAUBER94A" );
     break;

   case MLP:
     item_help("MLP");
     item_bibRef( "HERTZ91" );
     item_bibRef( "DUDA98" );
     item_bibRef( "HAYKIN94" );
     item_bibRef( "SCHALKOFF92" );
     break;

   case YUC:
     item_help("YUC");
     break;
   default: ; break;
 }
}
