/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:22 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "def.h"

extern universe *U;
extern boolean verbose;
extern boolean feat_description;
extern char **feature_desc;


static str80 buf;


void corr( allClass, nrClass, classes, first, second )
boolean allClass;
int nrClass, *classes, first, second;
{
 int i, j, row, c, nrPoints = 0;
 float *x = NULL, *y = NULL;
 float Ex = 0.0, Ey = 0.0, Exy = 0.0;	/* expected values */
 float Sx = 0.0, Sy = 0.0;		/* standard deviation */
 float Cov, CorCoeff;	/* Covariance and linear correlation coefficient */

 if( allClass )
   nrPoints = U->sumSampl;
 else
   for( i = 0; i < nrClass; i++ )
   {
     c = classes[i]-1;
     if( U->C[c].numSampl < 2 )
     {
       fprintf(stderr,"corr> Not enough samples: %d; Exit...\n",
		 U->C[c].numSampl ); exit(1);
     }
     nrPoints += U->C[c].numSampl;
   }

 x = (float*) malloc( nrPoints * sizeof(float) ); CHKPTR(x);
 y = (float*) malloc( nrPoints * sizeof(float) ); CHKPTR(y);

 row = 0;
 for( i = 0; i < nrClass; i++ )
 {
   c = classes[i]-1;
   for( j = 0; j < U->C[c].numSampl; j++ )
   {
     x[row] = U->C[c].S[j*U->nrFeat+(first-1)];
     y[row] = U->C[c].S[j*U->nrFeat+(second-1)];
     /* printf("x[%d]=%f  y[%d]=%f\n", row, x[row], row, y[row] ); /**/
     row++;
   }
 }
 /* all x and y values have been determined now:
    calculate linear correlation coefficient 
 */
 for( i = 0; i < row; i++ )
 {
   Ex += x[i]; Ey += y[i]; Exy += x[i] * y[i];
 }
 Ex /= (float)row; Ey /= (float)row; Exy /= (float)row;
 Cov = Exy - Ex*Ey;

 for( i = 0; i < row; i++ )
 {
   Sx += (x[i]-Ex)*(x[i]-Ex); Sy += (y[i]-Ey)*(y[i]-Ey);
 }
 Sx = (float)sqrt( (double)(Sx/(float)(row-1)) );
 Sy = (float)sqrt( (double)(Sy/(float)(row-1)) );

 if( Sx == 0.0 || Sy == 0.0 )
 {
   fprintf(stderr,"corr> Cannot calculate correlation coefficient");
   fprintf(stderr," because standard deviation is 0; Exit...\n" ); exit(1);
 }
 CorCoeff = Cov / (Sx*Sy);

 printf("\n>>>----------------------------------------------------<<<\n");
 printf("         RESULT OF LINEAR CORRELATION ANALYSIS            \n");
 printf("            Feature nr. 1 = %d", first );
 if( feat_description ) printf( " (%s)", feature_desc[first-1] ); NL;
 printf("               Mean=%7.3f   Standard deviation=%7.3f\n", Ex, Sx );
 printf("            Feature nr. 2 = %d", second );
 if( feat_description ) printf( " (%s)", feature_desc[second-1] ); NL;
 printf("               Mean=%7.3f   Standard deviation=%7.3f\n\n", Ey, Sy );
 printf("            Covariance=%7.3f\n", Cov );
 printf("    ===>    Correlation coefficient=%7.3f   <===\n", CorCoeff );
 printf(">>>----------------------------------------------------<<<\n\n");

 GNU_linear_corr( CorCoeff, x, y, row, first, second, nrClass, classes );
 FREE( x ); FREE( y );
}


void correlation()
{
 boolean ok, allClass, already;
 int i, j, first, second, nrClass, *classes = NULL;

 printf(">>>--- Correlation analysis between two features ---<<<\n");
 printf("First feature (1-%d)? ",  U->nrFeat );
 get_d_range( &first, 1, U->nrFeat, LEFT_CLOSED__RIGHT_CLOSED );
 printf("Second feature (1-%d, except %d)? ", U->nrFeat, first );
 do
 {
   get_d_range( &second, 1, U->nrFeat, LEFT_CLOSED__RIGHT_CLOSED );
   ok = first != second;
   if( ! ok  )
     printf("Second must be different from first! Again ? ");
 }
 while( ! ok );
 printf("Correlation of all %d classes (y/n)?y\b", U->nrClass ); gets( buf );
 allClass = ((buf[0]=='y') || (buf[0]=='Y') || (buf[0]=='\0'));
 if( allClass )
 {
   nrClass = U->nrClass;
   classes = (int*) malloc( U->nrClass * sizeof(int) );
 }


 if( ! allClass )
 {
   if( U->nrClass > 2 )
   {
     printf("Number of classes to be analyzed (1-%d)? ", U->nrClass-1 );
     get_d_range( &nrClass, 1, U->nrClass-1, LEFT_CLOSED__RIGHT_CLOSED );
   }
   else
     nrClass = 1;
   classes = (int*) malloc( nrClass * sizeof(int) );
   for( i = 0; i < nrClass; i++ )
   {
     if( nrClass > 1 )
       printf("Class nr.%d ? ", i+1 );
     else
       printf("Class nr. (1-%d)? ", U->nrClass );
     ok = FALSE;
     do
     {
       get_d_range( &(classes[i]), 1, U->nrClass, LEFT_CLOSED__RIGHT_CLOSED );
       already = FALSE;
       for( j = 0; j < i; j++ )
         already = already || (classes[j] == classes[i]);

       if( already )
         printf("Class already selected! Again ? ");
       else
	 ok = TRUE;
     }
     while( ! ok );
   }
 }
 else
   for( i = 0; i < nrClass; i++ )
     classes[i] = i+1;
 corr( allClass, nrClass, classes, first, second );
 FREE( classes );
}
