/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:22 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>
#include "def.h"

extern universe *U;
extern char *name_norm();

static boolean done;
static str80 buf;


/***************************************/
/*****     MENU ITEMS              *****/
/***************************************/

/*   -----   MAIN MENU    -----        */
#define SET_CLASS       0
#define FEAT_SEL        1
#define FEAT_EXTR       2
#define INDUCE          3
#define ERR_EST         4
#define CLASS           5
#define SAMMON          6
#define MISC            7
#define INTFACE         8
#define LOAD            9
#define NORM            10

#define NUM_MAINMENU_ITEMS 11


/*******************************************************/
/*     Routines to handle an interrupt from <Ctrl>-C   */
/*******************************************************/
static jmp_buf jump_mainLoop; /* to save state in case of a SIGINT */

static void restart_from_ctrl_c( dummy ) /* for a SIGINT, restart in cmdloop */
int dummy;
{
 dummy = EMPTY;
 signal( SIGINT, restart_from_ctrl_c );
 printf( "\n\nProgram interrupted : (M)ain menu , (Q)uit without saving: " );
 gets( buf );
 if( buf[0] == 'q' || buf[0] == 'Q' ) exit(0);
 longjmp( jump_mainLoop, 1 );
}


static boolean welcome_shown = FALSE;
static int menuItem = EMPTY;

static char normStr1[] = "\tNormalize data";
static char normStr2[] = "----\tData is normalized";

static struct mainMenuItems {
        int index;
        str80 name;
} mainMenuItem[] = {
        { SET_CLASS,	"Classifier model" },
        { FEAT_SEL,	"Feature selection" },
        { FEAT_EXTR,	"Feature extraction" },
        { INDUCE,	"Classifier induction  alias  supervised learning " },
        { ERR_EST,	"Error estimation      alias  cross validation" },
        { CLASS,	"Classify from independent data" },
        { SAMMON,	"Sammon plot" },
        { MISC,		"Miscellaneous" },
        { INTFACE,	"Interfacing" },
        { LOAD,		"Load universe" },
        { NORM,		"" } };


static void execMenuItem()
{
 switch( menuItem )
 {
   case LOAD:		loadUniv(); showUniv( stdout ); break;
   case SET_CLASS:	set_class_model(); show_class_model(stdout); WAIT;break;
   case NORM:		normalize(); break;
   case FEAT_SEL:	featSelectMain(); break;
   case FEAT_EXTR:	featExtractMain(); break;
   case INDUCE:		supervised_learn_classifier( TRUE ); break;
   case ERR_EST:	errorEstimationMain(); break;
   case CLASS:		classify_from_file(); break;
   case SAMMON:		sammon_plot(); break;
   case MISC:		miscMain(); break;
   case INTFACE:	intface(); break;
   default: fprintf(stderr,"What the heck is item %d? - exit...\n", menuItem );
 }
}


static char bStr1[] =
">>>-------------------------------------------------------------<<<\n\
>>>                      TOOLDIAG 2.1                           <<<\n\
>>>               Pattern recognition toolbox                   <<<\n";

static char bStr2[] =
">>>        Copyright (C) 1992 - 1997 Thomas W. Rauber           <<<\n\
>>>  Universidade Federal do Espirito Santo, Vitoria-ES, Brazil <<<\n\
>>>                             &                               <<<\n\
>>>     Universidade Nova de Lisboa - Dept. Electrical Eng.     <<<\n\
>>>                             &                               <<<\n\
>>>           UNINOVA - Center for Intelligent Robotics         <<<\n\
>>>  E-Mail: thomas@inf.ufes.br  URL: www.inf.ufes.br/~thomas   <<<\n\
>>>          Contributions by Dirk Tilsner <dt@uninova.pt>      <<<\n\
>>>        (Bug reports, suggestions and comments welcome)      <<<\n";

static char bStr3[] =
">>>-------------------------------------------------------------<<<\n";

static char bStr4[] =
" --- Type '?' followed by the menu choice for help ---\n\
 --- Course material \"Pattern Recognition\" available at my home page\n";


void testing()
{
 printf("Test routine..."); WAIT;
}

void mainLoop()
{
 int i, d; char *str = NULL, *nn;

 /* Resume here */
 setjmp( jump_mainLoop );

 printf("%s", bStr1 );
 if( ! welcome_shown )
 printf("%s", bStr2 );
 printf("%s", bStr3 );
 if( ! welcome_shown )
     { welcome_shown = TRUE;  printf("%s", bStr4 ); }

 for( i = 0; i < NUM_MAINMENU_ITEMS; i++ )
 {
   if( i == NORM )
   {
     nn = name_norm();
     if( nn == NULL )
       printf("(%d)%s\n", i+1, normStr1 );
     else
       printf("%s: %s\n", normStr2, name_norm() );
   }
   else
   {
     str = mainMenuItem[i].name;
     printf("(%d)\t%s\n", i+1, str );
   }
 }
 /* printf("(T)EST\n\n");	/**/
 printf("(Q)uit\n\n");

 printf("Choice: ");
 gets(buf); done = FALSE;
 switch( buf[0] )
 {
   /* special cases */
   case '?': help( MENU_MAIN, buf ); break;
   case '\0': showUniv( stdout ); break;
   case 'q': case 'Q': done = TRUE; break;
   case 't': case 'T' : testing(); break;
   default:
     sscanf(buf,"%d", &d);
     if( d >= 1 && d <= NUM_MAINMENU_ITEMS )
     {
       menuItem = d-1;
       execMenuItem();
     }
 }
}


int main( argc, argv )
int argc;
char **argv;
{
 signal( SIGINT, restart_from_ctrl_c );
 init_all();
 parse_input( argc, argv );
 do
 {
   mainLoop();
 }
 while( !done );
 free_universe();
 return( 1 );
}


/**********  H E L P  *****************/

void helpMENU_MAIN( cmd )
char *cmd;
{
 int item;

 sscanf( &(cmd[1]), "%d", &item );
 item--;	/* discount one since in the menu it is one more */
 switch( item )
 {
   case LOAD: item_help("LOAD"); break; 
   case SET_CLASS: item_help("SET_CLASS"); break;
   case NORM: item_help("NORM"); item_bibRef("DUDA73"); break;
   case FEAT_SEL:
     item_help("FEAT_SEL");
     item_bibRef( "DEVIJVER82" ); break;
   case FEAT_EXTR:
     item_help("FEAT_EXTR");
     item_bibRef( "DEVIJVER82" ); break;
   case INDUCE: item_help("INDUCE"); break;
   case ERR_EST: item_help("ERR_EST"); break;
   case CLASS: item_help("CLASS"); break;
   case SAMMON: item_help("SAMMON"); item_bibRef("SAMMON69"); break;
   case MISC: item_help("MISC"); break;
   case INTFACE: item_help("INTFACE"); break;
   default: ; break;
 }
}
