/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:23 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "def.h"
#include "matrxdef.h"
#include "clasdef.h"

extern int classify_one_testSample();
extern boolean at_least2test();
extern float estimate_error_KNN();

extern universe *U;
extern boolean verbose, verbose_reason;
extern str100 outDataDir;
extern boolean *train_smp, *test_smp;
extern int nrSamplesTest, nrSamplesTrain;
extern int classifier_model;
extern struct class_models_ class_model[];

boolean error_estimation = FALSE;/* indicates that a error estimation is done */


#define PERCENT (100.0)
static boolean done;
static str80 buf;

static boolean show_confusion_matrix = FALSE;

#define NR_RUN_DEFAULT_RANDOM 5
#define NR_RUN_DEFAULT_BOOT 100
#define NR_RUN_MAX_BOOT 10000

static boolean random_split = TRUE, sequential_estimation = FALSE;
static boolean feature_selection, use_errest_default = FALSE;
static float percentTrain = 70.0;
static int nrRuns = NR_RUN_DEFAULT_RANDOM;
static int nrSubSetsROT = EMPTY;	/* into how many subsets */
static int OLD_classifier_model;
static int OLD_errest_strategy;

/* Store the parameters for averaging the estimated error over
	different runs */
typedef struct error_param_ {
		float mean, stddev, sqrsum;
		float min, max;
		float *err_run;
	} error_param;

static error_param err_par;

static boolean protocolFile;
static FILE *errpf = NULL;

static void init_error_param( nrRuns )
int nrRuns;
{
 if( feature_selection || ! protocolFile || nrRuns <= 1 ) return;

 /* Initialize the parameters of the error protocol */
 err_par.err_run = (float *) malloc(sizeof(float) * nrRuns);
 CHKPTR(err_par.err_run);
 err_par.mean = 0.0; err_par.stddev = 0.0; err_par.sqrsum = 0.0;
 err_par.min = INFINITY; err_par.max = -INFINITY;
}


static void update_error_param( m, nrRuns, errPercent )
int m, nrRuns;
float errPercent;
{
 float err;

 if( feature_selection || ! protocolFile || nrRuns <= 1 ) return;
 err = errPercent / PERCENT;
 err_par.err_run[m] = err;
 err_par.mean = 1.0/((float)(m+1))*(m*err_par.mean+err);
 err_par.sqrsum += err * err;
 err_par.stddev = (float)sqrt(fabs((1.0/(float)(m+1) *
	err_par.sqrsum - (err_par.mean * err_par.mean))));
 if( err < err_par.min )
   err_par.min = err;
 if( err > err_par.max )
   err_par.max = err;
}


static void close_error_param( nrRuns )
int nrRuns;
{
 int i;

 if( feature_selection || ! protocolFile || nrRuns <= 1 ) return;
 /* write the parameters to the file */
 fprintf( errpf, "\t--- AVERAGE STATISTICS OF ERROR ESTIMATION ---\n");
 fprintf( errpf, "Mean over %d runs has been taken\n", nrRuns );
 fprintf( errpf, "\tRun#\tError[%%]\n");
 for( i = 0; i < nrRuns; i++ )
   fprintf( errpf, "\t%4d\t%8.5f\n", i+1, PERCENT*err_par.err_run[i] );
 fprintf( errpf, "\tSTATISTICAL PARAMETERS OF THE ESTIMATED ERROR:\n");
 fprintf( errpf, "\tMEAN=%.4f%%\tSTANDARD DEVIATION=%.8f\tVARIANCE=%.8f\n",
	PERCENT*err_par.mean, err_par.stddev, err_par.stddev*err_par.stddev );
 fprintf( errpf, "\tMINIMUM=%.3f%%\tMAXIMUM=%.3f%%\n",
	PERCENT*err_par.min, PERCENT*err_par.max );
 FREE(err_par.err_run);
 fflush( errpf );
}


#define NUM_ERREST_STRATEGIES 5

#define RESUBSITUTION 0
#define HOLDOUT 1
#define LEAVE_ONE_OUT 2
#define ROTATION 3
#define BOOTSTRAP 4

/*--------------------------------------------------------------*/
static int num_errest_strategies = NUM_ERREST_STRATEGIES;

static struct strats {
	int index;
	char *name;
} errest_strat[] = {
	{ RESUBSITUTION,	"Resubstitution" },
	{ HOLDOUT,		"Holdout" },
	{ LEAVE_ONE_OUT,	"Leave-One-Out" },
	{ ROTATION,  "Rotation aka K-fold cross validation aka jackknifing" },
	{ BOOTSTRAP,		"Bootstrap" }
};
static int errest_strategy = LEAVE_ONE_OUT;	/* default */
/* static int errest_strategy = HOLDOUT;	/* default */
/*--------------------------------------------------------------*/

char *errest_strategy_name()
{
 return( errest_strat[errest_strategy].name );
}

static float globalErr, meanErr;
static int runErr, err;


static void estErrRESUB( nrSelFeat, FSV )
int nrSelFeat;
FeatSelectVector FSV;
{
 int i, j;

 set_all_samples_train();
 runErr = 0;
 if( show_confusion_matrix )
   init_confMat();

 if( ! class_model[classifier_model].induced || sequential_estimation )
 {
   if( ! sequential_estimation )
     printf("First inducing %s...\n", class_model[classifier_model].name );
   induce( nrSelFeat, FSV, FALSE );
 }
 for( i = 0; i < U->nrClass; i++ )
 {
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     err = classify_one_testSample( i, j, EMPTY, nrSelFeat, FSV );
     if( err != CORRECT_CLASS )
       runErr++;
     if( show_confusion_matrix )
       update_confMat( err, U->C[i].name );
   }
 }
 /* free_induced();	/**/
 globalErr = PERCENT*(float)runErr / (float)nrSamplesTrain;
 if( show_confusion_matrix )
   show_confMat( errpf );
}


static void estErrHOLDOUT( nrSelFeat, FSV )
int nrSelFeat;
FeatSelectVector FSV;
{
 int s, i, j, m; 
 float errAux;

 init_error_param( nrRuns );
 meanErr = 0.0;
 for( m = 1; m <= nrRuns; m++ )
 {
   printf("Error estimation by HOLDOUT: run %d of %d    \r", m, nrRuns );
   fflush(stdout);	/**/
   if( random_split )
     split_tt_random( U->sumSampl, percentTrain );

   /* take all training samples to train the classifier;
	then classify all test samples */
   induce( nrSelFeat, FSV, FALSE );
 
   if( show_confusion_matrix )
     init_confMat();

   runErr = 0; s = 0;
   for( i = 0; i < U->nrClass; i++ )
   {
     for( j = 0; j < U->C[i].numSampl; j++ )
     {
       if( test_smp[s] )
       {
         err = classify_one_testSample( i, j, s, nrSelFeat, FSV );
         if( err != CORRECT_CLASS )
           runErr++;
         /* if( err )
           printf("err=%d    runErr=%d  s=%d\n", err, runErr, s ); /**/
         if( show_confusion_matrix )
           update_confMat( err, U->C[i].name );
       }
       s++;
     }
   }
   errAux = PERCENT*(float)runErr / (float)nrSamplesTest;
   meanErr += errAux;
   /* printf("errAux=%f runErr=%d meanErr=%f nrSamplesTest=%d\n",
	errAux, runErr, meanErr, nrSamplesTest ); /**/

   if( nrRuns > 1 && ! feature_selection )
   {
     printf("Estimated error for run %d of %d was %6.2f%%", m, nrRuns, errAux );
     printf(" - Current mean is %6.3f%%  \r",meanErr/(float)m);fflush( stdout );
     update_error_param( m-1, nrRuns, errAux );
   }
   if( show_confusion_matrix )
     show_confMat( errpf );
   free_induced();
 }
 BLANK_LINE;
 globalErr = meanErr / (float)nrRuns;
 /* printf("globalErr=%f runErr=%d nrSamplesTest=%d\n",
	globalErr, runErr, nrSamplesTest ); /**/
 close_error_param( nrRuns );
}


static void estErrLEAVE( nrSelFeat, FSV )
int nrSelFeat;
FeatSelectVector FSV;
{
 int i, j, s = 0;
 float errAux;

 set_all_samples_train();
 runErr = 0;
 init_error_param( U->sumSampl );

 if( show_confusion_matrix )
   init_confMat();

 for( i = 0; i < U->nrClass; i++ )
 {
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     set_one_sample_test( s );
     induce( nrSelFeat, FSV, FALSE );
     err = classify_one_testSample( i, j, s, nrSelFeat, FSV );
     if( err != CORRECT_CLASS )
       runErr++;
     if( ! feature_selection )
     {
       printf("Current mean error for sample %d of %d is %6.3f%%\r  ",
         s+1, nrSamplesTrain+nrSamplesTest, PERCENT*(float)runErr/(float)(s+1));
       fflush( stdout );
       if( show_confusion_matrix )
         update_confMat( err, U->C[i].name );
       if( err == CORRECT_CLASS )
	 errAux = 0.0;
       else
	 errAux = PERCENT;
       update_error_param( s, U->sumSampl, errAux );
     }
     reset_one_sample_test( s );
     free_induced();
     s++;
   }
 }
 BLANK_LINE;
 globalErr = PERCENT*(float)runErr / (float)U->sumSampl;
 /* printf("globalErr=%f runErr=%d U->sumSampl=%d\n",
	globalErr, runErr, U->sumSampl ); /**/
 if( show_confusion_matrix )
   show_confMat( errpf );
 close_error_param( U->sumSampl );
}

static void estErrROTATION( nrSelFeat, FSV )
int nrSelFeat;
FeatSelectVector FSV;
{
 int eta, rest, r = 0, i, ptr = 0, offset, j, k, m, s;
 int *subSet = NULL;	/* The vector with the subsets */
 int *starts = NULL;	/* Where does the next subset begin */
 int *numElem = NULL;	/* and how many elements */
 float errAux;
 boolean split_ok;

 eta = U->sumSampl / nrSubSetsROT;
 rest = U->sumSampl % nrSubSetsROT;
 /* printf("#subSet=%d  eta=%d  rest=%d\n", nrSubSetsROT, eta, rest ); /**/
 subSet = (int*) malloc( U->sumSampl * sizeof(int) ); CHKPTR(subSet);
 starts = (int*) malloc( nrSubSetsROT * sizeof(int) ); CHKPTR(starts);
 numElem = (int*) malloc( nrSubSetsROT * sizeof(int) ); CHKPTR(numElem);

 gen_tt_ROTATION( subSet );
 /* mark the starts and number of elements */
 for( i = 0; i < nrSubSetsROT; i++ )
 {
   starts[i] = ptr;
   offset = 0;
   if( r < rest )
   {
     offset = 1; r++;
   }
   numElem[i] = eta + offset;
   ptr += numElem[i];
   /* printf("Subset %d has %d elements and starts at %d\n",
	i+1, numElem[i], starts[i] ); /**/
 }

 /* now rotate the subsets */
 set_all_samples_train();

 init_error_param( nrSubSetsROT );
 meanErr = 0.0;
 for( m = 0; m < nrSubSetsROT; m++ )
 {
   /* set the samples of the subset as test samples */
   for( k = 0; k < numElem[m]; k++ )
     set_one_sample_test( subSet[starts[m]+k] );

   /* Check the splitting - at least two training samples and one test */
   split_ok = at_least2test();
   split_ok = nrSamplesTest >= 1 && split_ok;
   if( ! split_ok )
   {
     fprintf(stderr,"estErrROTATION> Unable to split data such that");
     fprintf(stderr," each class\n\thas at least 2 training samples");
     fprintf(stderr," and at least one test sample !\n");
     fprintf(stderr,"\tMaybe there are very few samples per class?");
     EXIT("");
   }

   /* take all training samples to train the classifier;
	then classify all test samples */
   induce( nrSelFeat, FSV, FALSE );
 
   runErr = 0; s = 0;
   /* test the samples */

   if( show_confusion_matrix )
     init_confMat();

   for( i = 0; i < U->nrClass; i++ )
   {
     for( j = 0; j < U->C[i].numSampl; j++ )
     {
       if( test_smp[s] )
       {
         err = classify_one_testSample( i, j, s, nrSelFeat, FSV );
         if( err != CORRECT_CLASS )
           runErr++;
         /* printf("err=%d    runErr=%d  s=%d\n", err, runErr, s ); /**/
         if( show_confusion_matrix )
           update_confMat( err, U->C[i].name );
       }
       s++;
     }
   }
   free_induced();
   errAux = PERCENT*(float)runErr / (float)nrSamplesTest;
   meanErr += errAux;
   /* printf("\nerrAux=%f runErr=%d meanErr=%f  nrSamplesTest=%d\n\n",
	errAux, runErr, meanErr, nrSamplesTest ); /**/
   if( nrSubSetsROT > 1 && ! feature_selection )
   {
     printf("Estimated error for run %d of %d was %6.2f%%",
		m+1, nrSubSetsROT, errAux );
     printf(" - Current mean is %6.3f%%           \r", meanErr/(float)(m+1) );
     fflush( stdout );
     update_error_param( m, nrSubSetsROT, errAux );
   }
   /* reset the samples of the subset as test samples */
   for( k = 0; k < numElem[m]; k++ )
     reset_one_sample_test( subSet[starts[m]+k] );
   if( show_confusion_matrix )
     show_confMat( errpf );
 }
 BLANK_LINE;
 globalErr = meanErr / (float)nrSubSetsROT;
 /* printf("globalErr=%f runErr=%d nrSamplesTest=%d\n",
	globalErr, runErr, nrSamplesTest ); /**/
 close_error_param( nrSubSetsROT );
 FREE(subSet); FREE(starts); FREE(numElem);
}


static void estErrBOOTSTRAP( nrSelFeat, FSV )
int nrSelFeat;
FeatSelectVector FSV;
{
 int s, i, j, m; 
 float errAux;

 init_error_param( nrRuns );
 meanErr = 0.0;
 for( m = 1; m <= nrRuns; m++ )
 {
   split_tt_bootstrap( U->sumSampl );
   /* take all training samples to train the classifier;
	then classify all test samples */
   induce( nrSelFeat, FSV, FALSE );
 
   runErr = 0; s = 0;
   if( show_confusion_matrix )
     init_confMat();
   for( i = 0; i < U->nrClass; i++ )
   {
     for( j = 0; j < U->C[i].numSampl; j++ )
     {
       if( test_smp[s] )
       {
         err = classify_one_testSample( i, j, s, nrSelFeat, FSV );
         if( err != CORRECT_CLASS )
           runErr++;
         /* printf("err=%d    runErr=%d  s=%d\n", err, runErr, s ); /**/
         if( show_confusion_matrix )
           update_confMat( err, U->C[i].name );
       }
       s++;
     }
   }
   free_induced();
   errAux = PERCENT*(float)runErr / (float)nrSamplesTest;
   meanErr += errAux;
   /* printf("errAux=%f runErr=%d meanErr=%f\n", errAux, runErr, meanErr ); /**/
   if( nrRuns > 1 && ! feature_selection )
   {
     printf("Estimated error for run %d of %d was %6.2f%%", m, nrRuns, errAux );
     printf(" - Current mean is %6.3f%%  \r",meanErr/(float)m);fflush( stdout );
     update_error_param( m-1, nrRuns, errAux );
   }
   if( show_confusion_matrix )
     show_confMat( errpf );
 }
 BLANK_LINE;
 globalErr = meanErr / (float)nrRuns;
 close_error_param( nrRuns );
 /* printf("globalErr=%f runErr=%d nrSamplesTest=%d\n",
	globalErr, runErr, nrSamplesTest ); /**/
}


static void estErrStrat( nrSelFeat, FSV )
int nrSelFeat;
FeatSelectVector FSV;
{
 verbose_reason = FALSE;
 switch( errest_strategy )
 {
   case RESUBSITUTION :	estErrRESUB( nrSelFeat, FSV );		break;
   case HOLDOUT :	estErrHOLDOUT( nrSelFeat, FSV );	break;
   case LEAVE_ONE_OUT :	estErrLEAVE( nrSelFeat, FSV );		break;
   case ROTATION :	estErrROTATION( nrSelFeat, FSV );	break;
   case BOOTSTRAP :	estErrBOOTSTRAP( nrSelFeat, FSV );	break;
   default: fprintf(stderr,"What the heck is strategy %d? - exit...\n",
		errest_strategy); exit(1);
 }
}

static char sl[] = "-------";
static char line[] =
        "---------------------------------------------------------------------";

static void show_cross_validation( f )
FILE *f;
{
 fprintf(f, "%s%s    C R O S S  V A L I D A T I O N    %s%s\n", sl, sl, sl, sl);
 fprintf( f, "\tNAME:\t\t\t%s\n", errest_strat[errest_strategy].name );
 if( errest_strategy == HOLDOUT && random_split )
 {
   fprintf(f,"\t\tsplitting into %6.2f%% training data and %6.2f%% test data\n",
		percentTrain, PERCENT-percentTrain );
   if( nrRuns > 1 )
     fprintf(f,"\t\ttaking the mean over %d runs\n", nrRuns );
 }
 if( errest_strategy == ROTATION )
 {
   fprintf(f,"\t\tsplitting samples into %d subsets\n", nrSubSetsROT );
   fprintf(f,"\t\t(%d-fold cross-validation", nrSubSetsROT );
   fprintf(f,", Leave-%d-Out)\n", U->sumSampl/nrSubSetsROT );
 }

 if( sequential_estimation && ! feature_selection )
   fprintf(f,"\t\twith increasing number of features\n");
 fprintf(f, "\tCLASSIFIER MODEL:\t%s", class_model[classifier_model].name );
 fprintf( f, "\n%s\n", line );
}


static char strdelim[] = "=======================================";

static void estimateError()
{
 int i, nrSelFeat;
 float *errInfo = NULL;
 str80 errFile, name;

 strcpy( errFile, outDataDir );
 printf("Saving the error estimation protocol to file:\n\t\t%s", errFile );
 gets( name );
 protocolFile = name[0] != '\0';
 if( protocolFile )
 {
   strcat( errFile, name );
   errpf = fopen( errFile, f_open_text_w );
   if( errpf == NULL )
     { printf("Cannot open %s! Exitus...\n", errFile ); exit(1); }
   fprintf( errpf, "### E R R O R  E S T I M A T I O N  P R O T O C O L ###\n");
   fprintf( errpf, "Universe: %s\n", U->name );
   show_cross_validation( errpf );

   printf("Protocol also confusion matrix?" );
   get_boolean( FALSE, &show_confusion_matrix );
 }
 else
   printf("\t\tSaving nothing...\n");

 if( sequential_estimation )
 {
   errInfo = (float*) malloc( U->nrSelFeat * sizeof(float) );

   if( protocolFile )
     fprintf(errpf,"Estimation for increasing number of features!\n");
   for( i = 1; i <= U->nrSelFeat; i++ )
   {
     printf("Estimating error for %d features...\r", i ); fflush(stdout);
     if( protocolFile )
     {
       fprintf(errpf,"\n%s%s\n", strdelim, strdelim );
       fprintf(errpf," === Number of selected features: %3d ===\n", i );
     }
     nrSelFeat = i;
     estErrStrat( nrSelFeat, U->FSV );
     errInfo[i-1] = globalErr;
     printf(" ---> Error estimate = %5.2f%% using %4d of %4d features\n",
		globalErr, i, U->nrSelFeat ); /**/

     if( protocolFile )
     {
       fprintf( errpf,
     "Nr. Features: %3d - Estimated error:\t%5.2f\t%% - Accuracy:\t%5.2f\t%%\n",
                 i, globalErr, 100.0-globalErr );
       fprintf(errpf,"%s%s\n", strdelim, strdelim );
       fflush( errpf );
     }
   }
   GNU_error_plot( errInfo, U->nrSelFeat, errest_strat[errest_strategy].name );
   FREE( errInfo );
 }
 else
 {
   printf("Estimating error...\r"); fflush(stdout);
   nrSelFeat = U->nrSelFeat;
   estErrStrat( nrSelFeat, U->FSV );
   fprintf(stderr," --- ERROR RATE: %6.2f %%  =  ACCURACY: %6.2f%%",
	globalErr, PERCENT-globalErr );
   fprintf(stderr,"                              \n\n");
   if( protocolFile )
   {
     fprintf(errpf,"FINAL RESULT: ERROR RATE: %6.2f %%  =  ACCURACY: %6.2f%%\n",
	globalErr, PERCENT-globalErr );
   }
 }
 if( protocolFile )
 {
   fclose( errpf );
 }
 free_tt_lookups();
}


static void getParam()
{
 if( feature_selection )
   sequential_estimation = FALSE;
 if( U->nrSelFeat > 1 && ! feature_selection )
 {
  printf("Estimate error as a function of the sequence of ");
  printf("the selected features?" );
  get_boolean( sequential_estimation, &sequential_estimation );
 }

 if( sequential_estimation )
 {
   printf("\n");
   printf("\tError will be estimated for an increasing number of features.\n");
   printf("\tNote that this makes most sense only for those feature\n");
   printf("\tselection techniques which are sequential by nature, e.g.\n");
   printf("\t'Sequential Forward Search' or 'Sequential Backward Search',\n");
   printf("\tbut does make less sense for e.g. 'Branch and Bound'.\n\n");
 }

 switch( errest_strategy )
 {
   case HOLDOUT :
     printf("Split samples RANDOMLY into training and test data ? " );
     get_boolean( random_split, &random_split );
     if( errest_strategy == HOLDOUT )
     {
       if( random_split )
       {
         getPercentTrain( &percentTrain );
         printf("Take the mean over how many runs? %4d\b\b\b\b", nrRuns );
         gets( buf ); 
         if( buf[0] != '\0' )
           sscanf( buf, "%d", &nrRuns );
         if( nrRuns <= 0 || nrRuns > U->sumSampl )
         {
           printf("The number of runs %d is invalid - setting to default %d\n",
			nrRuns, NR_RUN_DEFAULT_RANDOM );
           nrRuns = NR_RUN_DEFAULT_RANDOM;
         }
       }
       else
       {
         printf("Splitting data explicitly into train set and test set\n");
         split_tt_explicit( U->sumSampl );
       }
     }
     break;
   case ROTATION:
     printf("ROTATION: Split %d samples randomly into how many subsets ? ",
		U->sumSampl );
     get_d_range( &nrSubSetsROT, 1, U->sumSampl, LEFT_CLOSED__RIGHT_CLOSED );
     if( nrSubSetsROT == U->sumSampl )
       { printf("This is equivalent to the %s strategy...",
		errest_strat[LEAVE_ONE_OUT].name ); gets(buf); }
     break;
   case RESUBSITUTION : case LEAVE_ONE_OUT: ; break;
   case BOOTSTRAP:
     printf("Take the mean over how many bootstrap runs (default=%d):",
		NR_RUN_DEFAULT_BOOT );
     nrRuns = NR_RUN_DEFAULT_BOOT;
     gets( buf ); 
     if( buf[0] != '\0' )
       sscanf( buf, "%d", &nrRuns );
     if( nrRuns <= 0 || nrRuns > NR_RUN_MAX_BOOT )
     {
       printf("The number of runs %d is invalid - setting to default %d\n",
		nrRuns, NR_RUN_DEFAULT_BOOT );
       nrRuns = NR_RUN_DEFAULT_BOOT;
     }
     break;
   default: fprintf(stderr,"What the heck is strategy %d? - exit...\n",
	errest_strategy); exit(1);
 }
 NL;
}


static void errEst()
{
 error_estimation = TRUE;
 nameClassModel();
 show_cross_validation( stdout );
 printf("\tIs this what you want? (y/n)y\b"); gets(buf);
 printf("\n");
 if( buf[0] == 'n' ) return;
 estimateError();
 error_estimation = FALSE;
}


static void set_cross_validation( mainLoop )
boolean mainLoop;
{
 int s; boolean set;

 do
 {
   set = TRUE;
   printf("%s   ERROR ESTIMATION (CROSS VALIDATION) METHOD   %s\n", sl, sl);
   for( s = 0; s < num_errest_strategies; s++ )
   {
    printf("(%d) %s", s+1, errest_strat[s].name );
    if( errest_strategy == s )
       printf("\t\t\t<--- Current");
    printf("\n");
   }
   if( mainLoop )
      printf("(Q)uit\n\n");
   printf("Choice: "); gets(buf);
   if( isdigit(buf[0]) )
   {
     sscanf( buf, "%d", &s );
     if( s >= 1 && s <= num_errest_strategies )
     {
       errest_strategy = errest_strat[s-1].index;
       getParam();
     }
   }
   else
   {
     switch( buf[0] )
     {
       case 'q': case 'Q': if( mainLoop ) done = TRUE; break;
       case '\0': ; break;
       case '?': help( MENU_CROSSVAL, buf ); set = FALSE; break;
       default: printf("\n*** Unknown option!\n"); set = FALSE; break;
     }
   }
 }
 while( ! set );
}


void initMIN_EST_ERR()
{
 boolean change;

 use_errest_default = TRUE; feature_selection = TRUE;
 printf("\n---  COMBINATION OF FEATURE SELECTION AND ERROR ESTIMATION  ---\n");
 printf("\nDefault for error estimation is:\n");
 printf("\tCLASSIFIER MODEL:\t\t1-Nearest Neighbor classifier\n");
 printf("\tCROSS VALIDATION TECHNIQUE:\t%s\n",errest_strat[LEAVE_ONE_OUT].name);
 printf("\nUse it ? (y/n)y\b" ); gets( buf ); printf("\n");
 use_errest_default = ! (buf[0] == 'n' || buf[0] == 'N');
 if( use_errest_default )
 {
   OLD_classifier_model = classifier_model;
   OLD_errest_strategy = errest_strategy;
   classifier_model = NEAREST_NEIGHBOR; nameClassModel();
   errest_strategy = LEAVE_ONE_OUT; 
   set_K( 1 );
 }
 else
 {
   show_class_model( stdout );
   printf("\nChange it?"); get_boolean( TRUE, &change );
   if( change )
     set_class_model();
   show_cross_validation( stdout );
   printf("\nChange cross validation?"); get_boolean( TRUE, &change );
   if( change )
   {
     set_cross_validation( FALSE );
     show_cross_validation( stdout );
   }
 }
}


void closeMIN_EST_ERR()
{
 if( use_errest_default )
 {
   /* Set classifier model and error estimation to old values */
   classifier_model = OLD_classifier_model; nameClassModel();
   errest_strategy = OLD_errest_strategy;
 }
 feature_selection = FALSE;
}


float select_univariate_minerr( feat )
feature *feat;
{
 float accuracy;

 feature_selection = TRUE;
 if( use_errest_default )
   accuracy = 1.0 - estimate_error_KNN( 1, feat );
 else
 {
   estErrStrat( 1, feat );
   accuracy = 1.0 - globalErr / PERCENT;
 }
 return( accuracy );
}


float select_multivariate_minerr( FSV, nrSelFeat )
FeatSelectVector FSV;
int nrSelFeat;
{
 float accuracy;

 feature_selection = TRUE;
 if( use_errest_default )
   accuracy = 1.0 - estimate_error_KNN( nrSelFeat, FSV );
 else
 {
   estErrStrat( nrSelFeat, FSV );
   accuracy = 1.0 - globalErr / PERCENT;
 }
 return( accuracy );
}




void errorEstimationMain()
{
 if( U->nrClass > 1 && U->nrSelFeat > 0 )
   do
   {
     done = FALSE;
     set_cross_validation( TRUE );
     if( ! done )
       errEst();
   }
   while( !done );
 else
 {
  if( U->nrClass < 2 )
    { printf("Error estimation only with two or more classes"); WAIT; }
  else
    { printf(" Please load universe or select features first !" ); WAIT; }
 }
}



/**********  H E L P  *****************/

void helpMENU_CROSSVAL( cmd )
char *cmd;
{
 int item;

 sscanf( &(cmd[1]), "%d", &item );
 item--;        /* discount one since in the menu it is one more */
 switch( item )
 {
   case RESUBSITUTION:
     item_help("RESUBSITUTION");
     item_bibRef( "DEVIJVER82" );
     break;

   case HOLDOUT:
     item_help("HOLDOUT");
     item_bibRef( "DEVIJVER82" );
     break;

   case LEAVE_ONE_OUT:
     item_help("LEAVE_ONE_OUT");
     item_bibRef( "DEVIJVER82" );
     break;

   case ROTATION:
     item_help("ROTATION");
     item_bibRef( "DEVIJVER82" );
     break;

   case BOOTSTRAP:
     item_help("BOOTSTRAP");
     item_bibRef( "EFRON79" );
     item_bibRef( "EFRON94" );
     break;
   default: ; break;
 }
}
