/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:23 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include "def.h"
#include "ftsldef.h"

extern float selectMultivariate();
extern double bin_coeff();

extern universe *U;
extern boolean verbose;
extern boolean return_log;


static str80 buf;
static boolean increasing = FALSE;
static FeatSelectVector FSV = NULL;
static int **metaFSV = NULL;	/* for increasing search stores all subsets */
static float *globalBest = NULL;
static int *pos = NULL, *bestPos = NULL;

static void alloc_all( meta_sel, left, len )
boolean meta_sel;
int left, len;
{
 int i;

 FSV = (feature*) malloc( sizeof(feature) * U->nrFeat ); CHKPTR(FSV);
 pos = (int*) malloc( U->nrFeat * sizeof(int) ); CHKPTR(pos);
 bestPos = (int*) malloc( U->nrFeat * sizeof(int) ); CHKPTR(bestPos);
 if( meta_sel )
 {
   globalBest = (float*) malloc( len * sizeof(float) ); CHKPTR(globalBest);
   metaFSV = (int**) malloc( len * sizeof(int*) ); CHKPTR(metaFSV);
   for( i = 0; i < len; i++ )
   {
     metaFSV[i] = (int*) malloc( (left+i) * sizeof(int) );
     CHKPTR(metaFSV[i]);
   }
 }
}


static void free_all( meta_sel, len )
boolean meta_sel;
int len;
{
 int i;

 if( meta_sel )
 {
   FREE(globalBest);
   for( i = 0; i < len; i++ )
     FREE(metaFSV[i]);
   FREE(metaFSV);
 }
 FREE(pos); FREE(bestPos); FREE(FSV);
}



static void initEXHAUST( group, crit )
int group, crit;
{
 switch( crit )
 {
   case PATRICK_FISHER:
     return_log = FALSE;
     printf("Return logaritm of selection criterion instead criterion?");
     printf(" (y/n)n\b"); gets( buf );
     if( buf[0] == 'y' ) return_log = TRUE;
     break;
 }
}


static boolean movePos( pos, p, k, c )
int *pos, p, k, c;
{
 int q;

 /* printf(">>> p=%d  k=%d  c=%d\n", p, k, c );	/**/
 if( pos[p] == p + c )
 {
   if( p == 0 )
     return( TRUE );
   else
     return( movePos( pos, p-1, k, c ) );
 }
 else
 {
   pos[p]++;
   for( q = p+1; q < k; q++ )
     pos[q] = pos[q-1]+1;
   return( FALSE );
 }
}


#define TOO_MUCH (100000.0)

void selectFeaturesEXHAUSTIVE( group, crit )
int group, crit;
{
 int i, j, nrSelectFeat, nsf;
 int left, right, len, maxFeat, minFeat;
 float merit, maxMerit;
 boolean exhausted, do_it;
 double exhaustive_selection, trials;	/* all possible subsets */
        /* binomial coefficient (D over d) */

 printf("Exhaustive search for increasing number of features (%d thru %d)?",
	1, U->nrFeat );
 get_boolean( increasing, &increasing );
 if( increasing )
 {
   printf("  --- THIS IS THE ULTIMATE META-SELECTION ---\n");
   minFeat = 1; maxFeat = U->nrFeat;
 }
 else
 {
   minFeat = 2; maxFeat = U->nrFeat-1;
 }

 if( maxFeat < minFeat )
 {
   printf("Inappropriate number of features for exhaustive selection -\n");
   printf("setting all features as selected"); WAIT;
   set_all_selected(); show_selected();
   return;
 }

 if( minFeat == maxFeat )
 {
   printf("For a minimum of %d and a total of %d features,\n an exhaustive",
	minFeat, U->nrFeat );
   printf(" search makes only sense for %d features", minFeat ); WAIT;
   increasing = FALSE;
   nrSelectFeat = minFeat;
 }

 if( ! increasing && minFeat != maxFeat )
 {
   printf("How many features do you wish to select ? (%d-%d): ",
	minFeat, maxFeat );
   get_d_range( &nrSelectFeat, minFeat, maxFeat, LEFT_CLOSED__RIGHT_CLOSED);
   left = nrSelectFeat; right = nrSelectFeat;
 }
 else
 {
   left = minFeat; right = maxFeat;
   len = right-left+1;
 }

 initEXHAUST( group, crit );
 alloc_all( increasing, left, len );

 for( nsf = left; nsf <= right; nsf++ )
 {
   nrSelectFeat = nsf;
   trials = 1.0;

   exhaustive_selection = bin_coeff( U->nrFeat, nrSelectFeat );
   NL;
   printf("Exhaustive search for %d features has %.0lf subsets (%d over %d).\n",
	nsf, exhaustive_selection, U->nrFeat, nrSelectFeat );
   if( exhaustive_selection > TOO_MUCH )
   {
     printf("Do you really want to try it?"); get_boolean( FALSE, &do_it );
     if( ! do_it )
     {
       free_all( increasing, len );
       return;
     }
   }

   for( i = 0; i < nrSelectFeat; i++ )
   {
     pos[ i ] = i; bestPos[ i ] = EMPTY;
   }
   exhausted = FALSE;
   maxMerit = -INFINITY;

   do
   {
     for( i = 0; i < nrSelectFeat; i++ )
       FSV[i].rank = pos[i];
     merit = selectMultivariate( group, crit, FSV, nrSelectFeat );
     if( merit > maxMerit )
     {
       maxMerit = merit;
       /* memorize the best set */
       for( i = 0; i < nrSelectFeat; i++ )
         bestPos[i] = pos[i];
     }
     printf("Trial # %.0lf of %.0lf --- Criterion=%f\tBest so far=%f        \r",
	trials, exhaustive_selection, merit, maxMerit ); fflush(stdout);
     trials += 1.0;
     exhausted = movePos( pos, nrSelectFeat-1,
			nrSelectFeat, U->nrFeat-nrSelectFeat );
   }
   while( ! exhausted );
   BLANK_LINE;
   printf("Best criterion for %d features = %f\n", nrSelectFeat, maxMerit );

   if( increasing )
   {
     for( i = 0; i < nrSelectFeat; i++ )
       metaFSV[nrSelectFeat-minFeat][i] = bestPos[i];
     globalBest[ nsf - left ] = maxMerit;
   }
 }

 if( increasing )
 {
   printf("\nRESULT OF INCREMENTAL EXHAUSTIVE SEARCH:\n");
   for( i = 0; i < len; i++ )
   {
     printf("Exhaustive search for %2d features gave best criterion %f\n for",
		left+i, globalBest[i] );
     printf("  set={ ");
     for( j = 0; j < i+1; j++ )
       printf("%d ", metaFSV[i][j]+1 );
     printf("}\n");
   }

   printf("How many features do you finally wish to select ? (%d-%d): ",
	minFeat, maxFeat );
   get_d_range( &nrSelectFeat, minFeat, maxFeat, LEFT_CLOSED__RIGHT_CLOSED);

   for( i = 0; i < nrSelectFeat; i++ )
     bestPos[i] = metaFSV[nrSelectFeat-minFeat][i];
 }


 U->nrSelFeat = nrSelectFeat;
 FREE( U->FSV );
 U->FSV = (feature*) malloc( U->nrSelFeat * sizeof(struct feature_) );
 for( i = 0; i < U->nrSelFeat; i++ )
 {
   U->FSV[i].rank = bestPos[i];
   U->FSV[i].crit = (float)EMPTY;       /* does not matter */
 }
 show_selected_feats( stdout, U->FSV, U->nrSelFeat );

 free_all( increasing, len );
}
