/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:23 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "def.h"
#include "matrxdef.h"


extern universe *U;
extern str100 outDataDir;
extern boolean verbose, feat_description;
extern char **feature_desc;
extern float **matrix_alloc();


static boolean done;
static str80 buf, name, lfxName, lvqFile, speFile;
static FILE *xf = NULL;
static str100 efn;	/* extracted features names */

static Matrix LFX = { 0, 0, NULL };	/* Linear feature extractor */

/*******************************************/
/*****      EXTRACTION METHODS         *****/
/*******************************************/
#define LDA 0
#define PCA 1
#define SAMMON 2
#define POLY 3

#define NUM_EXTRACTION_METHODS 4

struct feature_extraction_methods {
        int index;
        char *name;
} extraction_method[] = {
  { LDA,	"Linear discriminant analysis" },
  { PCA,  "Principal Component Analysis  alias  Karhunen-Loeve Expansion" },
  { SAMMON,  "Sammon mapping (a nonlinear method)" },
  { POLY,  "Higher-order combinations of existing features" }
};

/* int extr_method = LDA;	/* default */
/* int extr_method = PCA;	/* default */
/* int extr_method = SAMMON;	/* default */
int extr_method = POLY;		/* default */

/*--------------------------------------------------------------*/

static void show_linear_extractor( W )
Matrix *W;
{
 if( W->Elem == NULL || W->row == 0 || W->col == 0 )
 {
   printf("Feature extractor is empty..."); gets( buf ); return;
 }
 
 printf("\n\n *****    %d x %d FEATURE EXTRACTOR:    *****\n", W->row, W->col );
 print_Matrix( W );
 printf("..."); gets(buf);
}


static void save_linear_extractor( W )
Matrix *W;
{
 int i, j;

 if( W->Elem == NULL || W->row == 0 || W->col == 0 )
 {
   printf("Feature extractor is empty..."); gets( buf ); return;
 }

 strcpy( lfxName, outDataDir );
 printf("Saving linear feature extractor to: %s", lfxName );
 gets( name );
 if( name[0] != '\0' )
 {
   strcat( lfxName, name );
   xf = fopen( lfxName, f_open_text_w );
   if( xf == NULL )
     { printf("Cannot open %s for writing! ...\n", lfxName ); return; }

   fprintf( xf, "# Universe:\n%s\n", U->name );
   fprintf( xf, "# Feature extraction method:\n#   %s\n",
		extraction_method[extr_method].name );
   fprintf( xf, "# Row and column dimension of the linear feature extractor\n");
   fprintf( xf, "%d  %d\n", W->row, W->col );

   for( i = 0; i < W->row; i++ )
   {
     for( j = 0; j < W->col; j++)
       fprintf( xf, " %20.15f ", (float)W->Elem[i*W->col+j] );
     fprintf( xf, "\n");
   }
   fprintf( xf, "\n");
   fclose( xf );
 }
 else
   printf("\tNothing saved...\n");
}


static void load_extr( fname, W )
char *fname;
Matrix *W;
{
 int i, j;

 xf = fopen( fname, f_open_text_r );
 if( xf == NULL )
   {printf("Cannot open %s! nothing done...", fname );gets(buf);return;}

 if( verbose )
   printf("\nLoading feature extractor from\n\t%s", fname );
 dataline( xf, buf );
 if( strcmp( buf, U->name ) != 0 )
   {printf("Warning universe names different: %s --- %s\n", buf, U->name );}

 dataline( xf, buf );
 Matrix_Free( W );
 sscanf( buf, "%d %d", &(W->row), &(W->col) );
 if( U->nrFeat != W->row )
 {
   printf("\nERROR:\n");
   printf("Row size(%d) of the extractor is different", W->row );
   printf(" from nr. of features(%d)\n", U->nrFeat );
   printf("\tReturning..."); gets(buf); return;
 }
 Matrix_Alloc( W );
 for( i = 0; i < W->row; i++ )
   for( j = 0; j < W->col; j++)
     fscanf( xf, "%lf", &(W->Elem[i*W->col+j]) );
 fclose( xf );
 if( verbose )
   show_linear_extractor( W );
}


static void load_linear_extractor( W )
Matrix *W;
{
 strcpy( lfxName, outDataDir );
 printf("Loading linear feature extractor from:\n\t%s", lfxName );
 gets( name );
 if( name[0] == '\0' )
   return;
 strcat( lfxName, name );
 load_extr( lfxName, W );
}


static void save_extr_feat_names( dim )
int dim;
{
 int i;
 FILE *fn = NULL;

 strcpy( efn, outDataDir );
 printf("Saving the names of the extracted features to:\n\t%s", efn );
 gets( name );
 if( name[0] != '\0' )
 {
   strcat( efn, name );
   fn = fopen( efn, f_open_text_w );
   if( fn == NULL )
     { printf("Cannot open %s for writing! Exitus...\n", efn ); return; }

   /* determine the postfix */
   switch( extr_method )
   {
     case LDA : strcpy( name, "LDA" ); break;
     case PCA : strcpy( name, "PCA" ); break;
     case SAMMON : strcpy( name, "SAMMON" ); break;
     default: fprintf(stderr,"\n*** Unknown method %d\n", extr_method );
     exit(1); break;
   }
   fprintf( fn, "# Extracted features from universe: %s__%s\n", U->name, name );
   fprintf( fn, "%d\n", dim );
   for( i = 0; i < dim; i++ )
     fprintf( fn, "%s*%s_%d\n", U->name, name, i+1 );
   fclose( fn );
 }
 else
   printf("\tNothing saved...\n");
}

/*
 * Calculate the squared prediction error for a given sample.
 *
 * The SPE is defined as the square norm of the difference between the
 * original (observed) feature vector and the predicted (reconstructed)
 * value, using the linear model to fit the observed vector to the model
*/
static void calc_squared_prediction_error( W )
Matrix *W;
{
 int i, j, k, D, d, row;
 Matrix WT, WWT, Y, I, M, R;
 MatElem sqLen;
 FILE *spe = NULL;

 if( W->Elem == NULL || W->row == 0 || W->col == 0 )
 {
   printf("Feature extractor is empty..."); gets( buf ); return;
 }
 strcpy( speFile, outDataDir );
 printf("Saving the SQUARED PREDICTION ERROR in file:\n\t\t%s",
		speFile ); gets( name );

 if( name[0] == '\0' )
     { printf("Nothing saved...\n"); return; }

 strcat( speFile, name );
 spe = fopen( speFile, f_open_text_w );
 if( spe == NULL )
   { printf("Cannot open %s! Exitus...\n", speFile ); return; }
 if( verbose )
   printf("Generating SPE-File: %s\n", speFile );

 D = U->nrFeat;
 d = W->col;
 init_Matrix( &WT ); init_Matrix( &WWT );
 init_Matrix( &Y ); init_Matrix( &I ); init_Matrix( &M ); init_Matrix( &R );
 Transpose_Matrix( W, &WT );

 Y.row = D; Y.col = 1;
 Matrix_Alloc( &Y );

 Mult_Matrix( W, &WT, &WWT );
 Unit_Matrix( &I, D );
 Subtract_Matrix( &I, &WWT, &M );
 /* printf("WWT:\n"); print_Matrix( &WWT );
 printf("I - WWT:\n"); print_Matrix( &M ); DBG; /**/

 row = 1;
 fprintf( spe, "# Using %d principal components from originally %d features\n",
	d, D );
 fprintf( spe, "#\n# Sample Nr.\t\t Squared Prediction Error\n" );
 for( i = 0; i < U->nrClass; i++ )
 {
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     /* generate the vector to be mapped */
     for( k = 0; k < D; k++ )
       Y.Elem[k] = U->C[i].S[j*D+k]; 

     /* reconstruct */
     Mult_Matrix( &M, &Y, &R );

     /* Calculate the squared length of the D x 1 Matrix R */
     sqLen = 0.0;
     for( k = 0; k < D; k++ )
       sqLen += R.Elem[k] * R.Elem[k];
     fprintf( spe, "%4d\t%30.20f\n", row, (float)sqLen );
     row++;
   }
 }
 fclose( spe );
 printf("  ...done.\n");
 Matrix_Free( &WT ); Matrix_Free( &WWT );
 Matrix_Free( &Y ); Matrix_Free( &I ); Matrix_Free( &M ); Matrix_Free( &R );
}


static void extract_samplesLinear( W )
Matrix *W;
{
 int i, j, k, D, d;
 Matrix WT, Y, XI;
 FILE *lvq = NULL;

 if( W->Elem == NULL || W->row == 0 || W->col == 0 )
 {
   printf("Feature extractor is empty..."); gets( buf ); return;
 }
 strcpy( lvqFile, outDataDir );
 printf("Saving the extracted data in LVQ_PAK format in file:\n\t\t%s",
		lvqFile ); gets( name );

 if( name[0] == '\0' )
     { printf("Nothing saved...\n"); return; }

 strcat( lvqFile, name );
 lvq = fopen( lvqFile, f_open_text_w );
 if( lvq == NULL )
   { printf("Cannot open %s! Exitus...\n", lvqFile ); return; }
 if( verbose )
   printf("Generating LVQ-File: %s\n", lvqFile );

 D = U->nrFeat;
 d = W->col;
 init_Matrix( &WT ); init_Matrix( &Y ); init_Matrix( &XI );
 Transpose_Matrix( W, &WT );
 Y.row = D; Y.col = 1;
 Matrix_Alloc( &Y );

 fprintf( lvq, "%d\n", d );
 for( i = 0; i < U->nrClass; i++ )
 {
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     /* generate the vector to be mapped */
     for( k = 0; k < D; k++ )
       Y.Elem[k] = U->C[i].S[j*D+k]; 
     /* extract */
     Mult_Matrix( &WT, &Y, &XI );

     for( k = 0; k < d; k++ )
       fprintf( lvq, "%f ", (float)XI.Elem[k] );
     fprintf( lvq, "%s\n", U->C[i].name );
   }
 }
 fclose( lvq );
 printf("  ...done.\n");
 Matrix_Free( &WT ); Matrix_Free( &Y ); Matrix_Free( &XI );
 save_extr_feat_names( W->col );

 calc_squared_prediction_error( W );
}


static void featExtractLinear( method )
int method;
{
 Matrix_Free( &LFX );
 switch( method )
 {
   case LDA : featExtrDiscrimAnalysis( &LFX ); break;
   case PCA : featExtrPCA( &LFX ); break;
   default: fprintf(stderr,"\n*** Unknown method %d\n", method );exit(1); break;

 }
 /* Check if there was at least one non-zero eigenvalue */
 if( LFX.col != 0 )
 {
   show_linear_extractor( &LFX );
   save_linear_extractor( &LFX );

   /* extract the data to file */
   printf("Do you want to save the extracted samples (y/n)? y\b"); gets( buf );
   if( buf[0] == 'n' || buf[0] == 'N' )
     return;

   extract_samplesLinear( &LFX );
 }
}

/*-----------------------------------------------------------
	FEATURE EXTRACTION BY SAMMON MAPPING
  -----------------------------------------------------------*/
extern boolean map_sammon();

static void save_mapped_samples( map, d )
float **map;
int d;
{
 int i, j, k, row;
 FILE *lvq = NULL;

 strcpy( lvqFile, outDataDir );
 printf("Saving the extracted data in LVQ_PAK format in file:\n\t\t%s",
		lvqFile ); gets( name );

 if( name[0] == '\0' )
     { printf("Nothing saved...\n"); return; }

 strcat( lvqFile, name );
 lvq = fopen( lvqFile, f_open_text_w );
 if( lvq == NULL )
   { printf("Cannot open %s! Exitus...\n", lvqFile ); return; }
 if( verbose )
   printf("Generating LVQ-File: %s\n", lvqFile );

 fprintf( lvq, "%d\n", d );
 row = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     for( k = 0; k < d; k++ )
       fprintf( lvq, "%f ", map[row][k] );
     fprintf( lvq, "%s\n", U->C[i].name );
     row++;
   }
 }
 fclose( lvq );
 printf("  ...done.\n");
}

#define MF_DEFAULT      (0.2)
#define MF_MAX          (1.0)

#define NR_ITER_MAX 32000
#define NR_ITER_DEFAULT 300

#define DIM_DEFAULT 2

static float MF = MF_DEFAULT;           /* Learning rate */
static int dim = DIM_DEFAULT;
static int nrIter = NR_ITER_DEFAULT;


static void featExtrSAMMON()
{
 int N, L;
 float **X = NULL, **Y = NULL;
 int i, j, row;
 boolean ok, gen_file;

 L = U->nrFeat;
 N = U->sumSampl;

 if( L == 1 )
 {
   printf("Makes no sense to map 1-dimensional data"); WAIT;
   return;
 }

 printf("--- FEATURE EXTRACTION BY SAMMON MAPPING ---\n");
 if( L == 2 )
 {
   printf("\tMapping from dimension 2 to dimension 1 (have no other choice)" );
   WAIT;
   dim = 1;
 }
 else
 {
   if( dim > L-1 ) dim = L-1;
   printf("\tMap from dimension %d to dimension (1-%d)? %d\b", L, L-1, dim );
   get_d_range( &dim, 1, L-1, LEFT_CLOSED__RIGHT_CLOSED );
 }

 printf("\tNumber of iterations (actual=%d)? ", nrIter );
 get_d_range( &nrIter, 1, NR_ITER_MAX, LEFT_CLOSED__RIGHT_CLOSED );

 printf("\tLearning rate of Steepest Descent (actual=%4.2f)? ", MF );
 get_f_range( &MF, 0.0, MF_MAX, LEFT_OPEN__RIGHT_CLOSED );

 X = (float**) malloc( N * sizeof(float*) ); CHKPTR(X);
 Y = matrix_alloc( N, dim ); CHKPTR(Y);

 row = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     /* assign the j-th sample of the i-class to the sample pointers */
     X[row] = &(U->C[i].S[j*U->nrFeat]);
     /* showFV( L, X[row] );    /**/
     row++;
   }
 }

 ok = map_sammon( X, N, L, Y, dim, nrIter, MF );
 if( ok )
 {
   printf("\nSuccessfully extracted\nSave mapped samples to file");
   get_boolean( TRUE, &gen_file );
   if( gen_file )
   {
     save_mapped_samples( Y, dim );
     save_extr_feat_names( dim );
   }
 }
 else
   printf("Unable to generate Sammon map!\n\n");

 FREE(X);
 matrix_free( &Y, N );
}


static void plot_samplesLinear( W )
Matrix *W;
{
 int i, j, k, d, D, plot_dim, row;
 Matrix WT, Y, XI;
 float x, y, z;

 plot_dim = W->col;
 if( plot_dim != 2 &&  plot_dim != 3 ) return;

 GNU_init_gnu_file();
 GNU_scatter_init( U->nrClass, U->sumSampl, plot_dim );


 D = U->nrFeat;
 d = W->col;
 init_Matrix( &WT ); init_Matrix( &Y ); init_Matrix( &XI );
 Transpose_Matrix( W, &WT );
 Y.row = D; Y.col = 1;
 Matrix_Alloc( &Y );

 row = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   GNU_scatter_set_name( U->C[i].name );
   GNU_scatter_set_smp( i, U->C[i].numSampl );
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     /* generate the vector to be mapped */
     for( k = 0; k < D; k++ )
       Y.Elem[k] = U->C[i].S[j*D+k]; 
     /* extract */
     Mult_Matrix( &WT, &Y, &XI );
     x = (float)XI.Elem[0]; y = (float)XI.Elem[1];
     if( d == 3 ) z = (float)XI.Elem[2];

     GNU_scatter_set_point( x, y, z, row, plot_dim );
     GNU_scatter_set_index( row, i );
     row++;
   }
 }
 Matrix_Free( &Y );
 GNU_title( extraction_method[ extr_method ].name );
 GNU_xlabel( "1st Eigenaxis" );
 GNU_ylabel( "2nd Eigenaxis" );
 if( d == 3 ) GNU_zlabel( "3rd Eigenaxis" );
 GNU_scatter_plot();
 GNU_execute();
}



static void featExtract( method )
int method;
{
 switch( method )
 {
   case LDA : case PCA : featExtractLinear( method ); break;
   case SAMMON : featExtrSAMMON(); break;
   case POLY : featExtrPOLY(); break;
   default: fprintf(stderr,"\n*** Unknown method %d\n", method );exit(1); break;
 }
}

static char sl[] = "--------";

static void featExtractLoop()
{
 int method; boolean set;

 do
 {
   set = TRUE; NL;

   printf("\n%s  F E A T U R E  E X T R A C T I O N   %s\n\n", sl, sl );
   printf("%s%s M e t h o d %s%s\n", sl, sl, sl, sl );
   for( method = 0; method < NUM_EXTRACTION_METHODS; method++ )
   {
    printf("(%d) %s", method+1, extraction_method[method].name );
    printf("\n");
   }
   printf("\n%s  T o o l s  f o r  l i n e a r  e x t r a c t o r s  %s\n",
	sl, sl );
   printf("(S)ave linear feature extractor matrix to file\n");
   printf("(L)oad linear feature extractor matrix from file\n");
   printf("S(H)ow actual matrix\n");
   printf("(E)xtract samples in memory with actual linear extractor\n");
   if( LFX.col == 2 || LFX.col == 3 )
   printf("(P)lot %d-D extracted samples of actual linear extractor\n",LFX.col);
   printf("Choice: ");

   gets(buf); done = FALSE;
   switch( buf[0] )
   {
     case '\0': done = TRUE; break;
     case '?': help( MENU_FEATEXTRACT, buf ); set = FALSE; break;
     case 's': case 'S': save_linear_extractor( &LFX ); break;
     case 'l': case 'L': load_linear_extractor( &LFX ); break;
     case 'h': case 'H': show_linear_extractor( &LFX ); break;
     case 'e': case 'E': extract_samplesLinear( &LFX ); break;
     case 'p': case 'P': plot_samplesLinear( &LFX ); break;
     default:
       sscanf( buf, "%d", &method );
       if( method >= 1 && method <= NUM_EXTRACTION_METHODS )
       {
         extr_method = extraction_method[method-1].index;
         featExtract( extr_method );
       }
       break;
   }
 }
 while( ! set );
}


void featExtractMain()
{
 if( U->nrClass > 0 )
 {
   do
   {
     featExtractLoop();
   }
   while( !done );
   Matrix_Free( &LFX );
 }
 else
   { printf(" Please load universe first !..." ); gets( buf ); }
}


/**********  H E L P  *****************/

void helpMENU_FEATEXTRACT( cmd )
char *cmd;
{
 int item;

 sscanf( &(cmd[1]), "%d", &item );
 item--;        /* discount one since in the menu it is one more */
 switch( item )
 {
   case LDA:
     item_help("LDA");
     item_bibRef( "DEVIJVER82" );
     break;

   case PCA:
     item_help("PCA");
     item_bibRef( "DEVIJVER82" );
     break;

   case SAMMON:
     item_help("SAMMON");
     item_bibRef( "SAMMON69" );
     break;

   case POLY:
     item_help("POLY");
     item_bibRef( "DUDA73" );
     item_bibRef( "SCHUERMANN96" );
     break;
   default: ; break;
 }
}
