/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:24 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>
#include <string.h>
#include <ctype.h>
#include "def.h"
#include "ftsldef.h"


extern float select_multivariate_minerr();
extern float select_multivariate_prob_dist();
extern float select_multivariate_InterClassDist();
extern char *errest_strategy_name();
extern char *class_model_name();

extern universe *U;
extern boolean verbose;


/*
 * SIGNAL HANDLING FOR FLOATING POINT EXCEPTIONS
*/
static jmp_buf jmp;
static boolean selection_failed;

static void restart_from_fpe(  dummy )
int dummy;
{
 dummy = EMPTY;
 signal( SIGFPE, restart_from_fpe );
 fprintf( stderr, "\nFLOATING POINT EXCEPTION, probably caused by numbers");
 fprintf(  stderr," grown too large!" ); WAIT;
 selection_failed = TRUE;
 longjmp( jmp, 1 );
}



static boolean done;
static str80 buf;

/* Decides if the mutual distance between classes is calculated only
   between different classes or also within the same class
*/
boolean mutual_all = TRUE;


/*--------------------------------------------------------------*/
/* ##### EUCLIDEAN DISTANCES */
int num_euclid_distances = NUM_EUCLID_DISTANCES;

struct euclid_distances {
	int index;
	str80 name;
} euclid_distance[] = {
	{ J1,	"Scatter matrices distance   tr(Sw+Sb)" },
	{ J2,	"Scatter matrices distance   tr(Sw)/tr(Sb)" },
	{ J3,	"Scatter matrices distance   tr(inv(Sw)*Sb)" },
	{ J4,	"Scatter matrices distance   det(Sw+Sb)/det(Sw)" }
};
int euclid_dist = EMPTY;		/* default */
/*--------------------------------------------------------------*/


/*--------------------------------------------------------------*/
/*  ##### METRICS */

#define MAX_MINKOWSKI (100.0)
float minkowski_order;

struct distances {
	int index;
	str80 name;
} distance[] = {
	{ MINKOWSKI,		"Minkowski" },
	{ CITY_BLOCK,		"City block" },
	{ EUCLIDEAN,		"EUCLIDEAN DISTANCE" },
	{ CHEBYCHEV_DIST,	"Chebychev" },
	{ NONLINEAR,		"Nonlinear (Parzen & hyperspheric kernel)" }
};
int pattern_distance = EUCLIDEAN;

/*  ##### INTERCLASS DISTANCE SELECTION CRITERIA  */
int num_criteriaINTCLASS = NUM_DISTANCES;
/*--------------------------------------------------------------*/


/*--------------------------------------------------------------*/
/*  ##### SELECTION CRITERIA  GROUPS */
int num_criteria_groups = NUM_SEL_CRIT_GROUPS;

struct criteria_group {
        int index;
        str80 name;
} criterion_group[] = {
	{ MIN_EST_ERR,			"Estimated minimal error probability" },
	{ INTER_CLASS_DISTANCE,		"Inter-class distance" },
	{ PROBABILISTIC_DISTANCE,	"Probabilistic distance" }
};
int selection_criterion_group;
/*--------------------------------------------------------------*/


/*--------------------------------------------------------------*/
/*  ##### PROBABILISTIC DISTANCE SELECTION CRITERIA  */
int num_criteriaPROBDIST = NUM_PROB_DIST_CRITERIA;

boolean return_log = FALSE;	/* In order to obtain reasonable values */
float chernoff_parameter;

struct criteriaPROBDIST {
        int index;
        str80 name;
} criterionPROBDIST[] = {
	/* Multivariate normal distribution */
        { CHERNOFF,     "Chernoff" },
        { BHATTACHARYYA,  "Bhattacharyya distance" },
        { BHATTACHARYYA_MATUSITA,  "Matusita distance" },
        { DIVERGENCE,  "Divergence" },
        { MAHALANOBIS,  "Mahalanobis distance" },
        { PATRICK_FISHER,  "Patrick-Fisher" },
        { NORMMEANDIST,
   "Normalized distance between means\n\t\t\t\t = 1 - (s1+s2)^2 / (m1-m2)^2" }
};
/*--------------------------------------------------------------*/

int selection_criterion;


/*--------------------------------------------------------------*/
/*  ##### SEARCH STRATEGIES  */
int num_search_strategies = NUM_STRATEGIES;

struct search_strategies {
	int index;
	str80 name;
	int numCritGroups;	/* ACTUALIZE WHEN ADDING NEW !!! */
	int critGroup[NUM_SEL_CRIT_GROUPS];
	int actualGroup;
} search_strategy[] = {
  { BF,       "Best Features", 3,
	{MIN_EST_ERR, INTER_CLASS_DISTANCE, PROBABILISTIC_DISTANCE}, EMPTY },
  { SFS,  "Sequential Forward Selection", 3,
	{MIN_EST_ERR, INTER_CLASS_DISTANCE, PROBABILISTIC_DISTANCE}, EMPTY },
  { SBS,  "Sequential Backward Selection", 3,
	{MIN_EST_ERR, INTER_CLASS_DISTANCE, PROBABILISTIC_DISTANCE}, EMPTY },
  { B_AND_B,  "Branch and Bound", 2,
	{INTER_CLASS_DISTANCE, PROBABILISTIC_DISTANCE}, EMPTY },
  { EXHAUSTIVE,      "Exhaustive Search", 3,
	{MIN_EST_ERR, INTER_CLASS_DISTANCE, PROBABILISTIC_DISTANCE}, EMPTY }
};
int search_strat = SFS;
/*--------------------------------------------------------------*/


static void getCriterionGroup( strat, critGroup )
int strat, *critGroup;
{
 int i, g; char *critGroupName = NULL;
 int ng, actg;
 boolean chosen = FALSE;

 ng = search_strategy[strat].numCritGroups;
 actg = search_strategy[strat].actualGroup;
 do
 {
   printf("\n ### Specify selection criterion group ###\n" );
   for( i = 0; i < ng; i++ )
   {
     g = search_strategy[strat].critGroup[i];
     critGroupName = criterion_group[ g ].name;
     printf("(%d) %s", i+1, critGroupName );
     if( ((i+1)==actg) && actg >=1 && actg <= ng ) printf("\t\t\t<--- Current");
     printf("\n");
   }
   printf("Choice: "); gets(buf);
   switch( buf[0] )
   {
     case '?': help( MENU_FEATSEL_CRIT_GROUP, buf ); break;
     case '\0':
       if( actg >=1 && actg <= ng )
       {
	 chosen = TRUE;
	 g = actg;
       }
       break;
     default:
       sscanf(buf,"%d", &g);
       if( g >= 1 && g <= search_strategy[strat].numCritGroups )
       {
	 chosen = TRUE;
         search_strategy[strat].actualGroup = g;
       }
       else
       {
	 printf("Invalid criterion group! Specify again ? ");
       }
   }
 }
 while( ! chosen );
 /* printf("strat=%d  g=%d group=%d",
	 strat, g, search_strategy[strat].critGroup[g-1] );DBG; /**/
 *critGroup = search_strategy[strat].critGroup[g-1];
 selection_criterion_group = *critGroup;
}



static void getCriterionPROBDIST( strat, crit )
int strat, *crit;
{
 int i, c, offset; float f; char *critName = NULL;

 offset = 0;
 if( strat != BF )
   offset = 1;

 printf("\n ### Specify probabilistic distance selection criterion ###\n");
 for( i = 0; i < NUM_PROB_DIST_CRITERIA-offset; i++ )
 {
   critName = criterionPROBDIST[ i ].name;
   printf("(%d) %s\n", i+1, critName );
 }
 printf("Choice: ");
 get_d_range( &c, 1, NUM_PROB_DIST_CRITERIA-offset,LEFT_CLOSED__RIGHT_CLOSED);
 *crit = criterionPROBDIST[c-1].index;
 if( *crit == CHERNOFF )
 {
   printf(" Enter parameter of Chernoff probabilistic distance [0,1]: ");
   get_f_range( &f, 0.0, 1.0, LEFT_CLOSED__RIGHT_CLOSED );
   chernoff_parameter = f;
 }
}


void get_euclidean_dist()
{
 int i, ed = 1;

 printf("\n --- Specify Euclidan distance metric ---\n" );
 printf("\tLegend:\tSw = Within-classes scatter matrix\n" );
 printf("\t\tSb = Between-classes scatter matrix\n" );
 printf("\t\ttr = trace,   inv = inverse,   det = determinant\n" );
 for( i = 0; i < num_euclid_distances; i++ )
   printf("(%d) %s\n", i+1, euclid_distance[ i ].name );
 printf("Choice: %d\b", ed );
 get_d_range( &ed, 1, num_euclid_distances, LEFT_CLOSED__RIGHT_CLOSED );
 euclid_dist = ed - 1;
}


static void getDistance( strat, dist )
int strat, *dist;
{
 int i, d; float f;

 euclid_dist = EMPTY;
 mutual_all = TRUE;

 printf("\n ### Specify distance metric for strategy %d = %s ###\n",
       strat+1, search_strategy[strat].name );
 for( i = 0; i < NUM_DISTANCES; i++ )
   printf("(%d) %s\n", i+1, distance[ i ].name );
 printf("Choice: ");
 get_d_range( &d, 1, NUM_DISTANCES, LEFT_CLOSED__RIGHT_CLOSED );
 *dist = distance[d-1].index;
 pattern_distance = *dist;
 if( *dist == MINKOWSKI )
 {
   printf(" Order of Minkowski metric: ");
   get_f_range( &f, 0.0, MAX_MINKOWSKI, LEFT_OPEN__RIGHT_CLOSED );
   minkowski_order = f;
 }
 if( *dist == EUCLIDEAN )
 {
   printf("Calculate Euclidean distance from scatter matrices\n");
   printf("\ti.e. criteria J1 to J4? (y/n)n\b");
   gets( buf );
   if( buf[0] == 'y' )
     get_euclidean_dist();
 }
 if( *dist == NONLINEAR )
   init_nonlinear_dist();
 if( euclid_dist == EMPTY )
 {
   printf("Calculate distance only between different classes (1)\n");
   printf("              or also within the same class itself(2)?2\b");
   gets(buf);
   if( buf[0] == '1' )
     mutual_all = FALSE;
 }
}


static void selectFeatures( strat, group, crit )
int strat, group, crit;
{
 switch( strat )
 {
   case BF      : selectFeaturesBF( group, crit ); break;
   case SFS     : selectFeaturesSFS( group, crit ); break;
   case SBS     : selectFeaturesSBS( group, crit ); break;
   case B_AND_B : selectFeaturesB_AND_B( group, crit ); break; 
   case EXHAUSTIVE : selectFeaturesEXHAUSTIVE( group, crit ); break; 
 }
}


static char sl[] = "--------------------";

static char line[] =
 "-----------------------------------------------------------------------";

/*
 * Write a comment character at the beginning of the line
 * when writing to a file
*/
#define CC if(f!=stdout&&f!=stderr){fprintf(f,"#");}

void show_featSelect( f )
FILE *f;
{
 int strat, gr, crit, dist;

 strat = search_strat; gr = selection_criterion_group;
 crit = selection_criterion; dist = pattern_distance;

 CC;fprintf(f,"\n");
 CC;fprintf(f,"%s  F E A T U R E  S E L E C T I O N   %s\n", sl, sl );
 CC;fprintf(f,"\tSearch strategy:\t\t%s\n", search_strategy[strat].name );
 CC;fprintf(f,"\tSelection criterion group:\t%s\n", criterion_group[gr].name );

 switch( gr )
 {
   case MIN_EST_ERR:
     CC;fprintf(f,"\t - Classifier:\t\t\t%s\n", class_model_name() );
     CC;fprintf(f,"\t - Cross validation:\t\t%s\n", errest_strategy_name() );
     break;
   case INTER_CLASS_DISTANCE:
     CC;fprintf(f,"\t\tDistance metric:\t%s", distance[crit].name );
     if( dist == MINKOWSKI )
       fprintf(f," of order s=%f", minkowski_order );
     fprintf(f,"\n");
     if( dist == EUCLIDEAN )
     {
       if( euclid_dist != EMPTY )
       {
         fprintf(f,"\t   based on criterion:  %s\n",
         euclid_distance[ euclid_dist ].name );
       }
     }
     if( mutual_all )
       {CC;fprintf(f,"\tCalculating mutual distance between ALL classes\n");}
     else
     {
       CC;fprintf(f,"\tCalculating mutual distance only between");
       fprintf(f," DIFFERENT classes\n");
     }
     break;

   case PROBABILISTIC_DISTANCE:
     CC;fprintf(f,"\t\tDistance metric:\t%s", criterionPROBDIST[crit].name );
     if( crit == CHERNOFF )
       fprintf(f," with s = %.3f", chernoff_parameter );
     fprintf(f,"\n");
     break;
 }
 CC;fprintf(f,"%s\n", line );
}


static void featSelect( strat )
int strat;
{
 int crit, crit_group, dist;

 getCriterionGroup( strat, &crit_group );
 switch( crit_group )
 {
   case MIN_EST_ERR:
     initMIN_EST_ERR();
     break;
   case INTER_CLASS_DISTANCE:
     getDistance( strat, &dist );
     crit = dist;
     break;
   case PROBABILISTIC_DISTANCE:
     getCriterionPROBDIST( strat, &crit ); 
     break;
 }
 selection_criterion = crit;

 show_featSelect( stdout );
 printf("\tIs this what you want? (y/n)y\b"); gets(buf);
 if( buf[0] == 'n' ) return;

 if( crit_group != MIN_EST_ERR )
   set_all_samples_train();
 selectFeatures( strat, crit_group, crit );
 free_tt_lookups();
 save_selected_feat();
 closeMIN_EST_ERR();
}


void featSelectDEMO()
{
 euclid_dist = EMPTY;
 mutual_all = TRUE;

 printf("\n--- FEATURE SELECTION:\n\tSearch strategy = Sequential forward\n");
 printf("\tSelection criterion = Euclidean distance\n");
 printf("\t\tDistance is calculated between all classes.\n\n");
 selectFeatures( SFS, INTER_CLASS_DISTANCE, EUCLIDEAN );
 save_selected_feat();
}


static void set_feature_selection()
{
 int s; boolean set;

 do
 {
   set = TRUE; NL;
   printf("---------     F E A T U R E  S E L E C T I O N      ---------\n\n");
   printf("%s   Search strategy   %s\n", sl, sl );
   for( s = 0; s < num_search_strategies; s++ )
   {
    printf("(%d) %s", s+1, search_strategy[s].name );
    if( search_strat == s )
      printf("\t\t\t<--- Current");
    NL;
   }
   printf("%s%s%s\n(T)ools and old algorithms\n", sl, sl, sl );
   printf("(Q)uit\n\n");
   printf("Choice: "); gets(buf);
   if( isdigit(buf[0]) )
   {
     sscanf( buf, "%d", &s );
     if( s >= 1 && s <= num_search_strategies )
     {
       search_strat = search_strategy[s-1].index;
       featSelect( search_strat );
     }
   }
   else
   {
     switch( buf[0] )
     {
       case 't': case 'T': featSelectToolsLoop(); break;
       case 'q': case 'Q': done = TRUE; break;
       case '\0': featSelect( search_strat ); break;
       case '?': help( MENU_FEATSEL_SEARCH, buf ); set = FALSE; break;
       default: printf("\n*** Unknown option!\n"); set = FALSE; break;
     }
   }
 }
 while( ! set );
 close_nonlinear_dist();
}


float selectMultivariate( group, crit, FSV, len )
int group, crit;
FeatSelectVector FSV;
int len;
{
 float merit;

 switch( group )
 {
   case MIN_EST_ERR : merit = select_multivariate_minerr( FSV, len ); break;
   case INTER_CLASS_DISTANCE :
        merit = select_multivariate_InterClassDist( crit, FSV, len ); break;
   case PROBABILISTIC_DISTANCE :
        merit = select_multivariate_prob_dist( crit, FSV, len ); break;
   default: fprintf(stderr,"SMV: What the heck is crit %d? - exit...\n", crit );
        exit(1);
 }
 return( merit );
}


void featSelectMain()
{
 selection_failed = FALSE;

 /* Prepare for floating point exception */
 signal( SIGFPE, restart_from_fpe );

 /* In case of a floating point exception program re-enters here */
 setjmp( jmp );
 if( selection_failed )
 {
   FREE( U->FSV );
   U->nrSelFeat = 0;
   return;
 }

 if( U->nrFeat == 1 )
 {
   printf("Only one feature. Selection is nonsense. Setting selected"); WAIT;
   set_all_selected();
   return;
 }

 if( U->nrClass > 1 )
   do
   {
     done = FALSE;
     if( ! done )
       set_feature_selection();
   }
   while( !done );
 else
   { printf(" Feature selection only with two or more classes!" ); WAIT; }
}



/**********  H E L P  *****************/

void helpMENU_FEATSEL_SEARCH( cmd )
char *cmd;
{
 int item;

 if( cmd[1] == 't' || cmd[1] == 'T' )
 {
   item_help("OLD_ALG");
   item_bibRef( "TOOLDIAG" );
   return;
 }
 sscanf( &(cmd[1]), "%d", &item );
 item--;        /* discount one since in the menu it is one more */
 switch( item )
 {
   case BF:
     item_help("BF");
     item_bibRef( "DEVIJVER82" );
     break;

   case SFS:
     item_help("SFS");
     item_bibRef( "DEVIJVER82" );
     break;

   case SBS:
     item_help("SBS");
     item_bibRef( "DEVIJVER82" );
     break;

   case B_AND_B:
     item_help("B_AND_B");
     item_bibRef( "NARENDRA77" );
     item_bibRef( "DEVIJVER82" );
     break;

   case EXHAUSTIVE:
     item_help("EXHAUSTIVE");
   default: ; break;
 }
}


void helpMENU_FEATSEL_CRIT_GROUP( cmd )
char *cmd;
{
 int item;

 sscanf( &(cmd[1]), "%d", &item );
 item--;        /* discount one since in the menu it is one more */
 switch( item )
 {
   case MIN_EST_ERR:
     item_help("MIN_EST_ERR");
     item_bibRef( "DEVIJVER82" );
     break;

   case INTER_CLASS_DISTANCE:
     item_help("INTER_CLASS_DISTANCE");
     item_bibRef( "DEVIJVER82" );
     break;

   case PROBABILISTIC_DISTANCE:
     item_help("PROBABILISTIC_DISTANCE");
     item_bibRef( "DEVIJVER82" );
     break;

   default: printf("Invalid item"); break;
 }
}
