/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:24 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "def.h"
#include "matrxdef.h"

extern universe *U;
extern float Euclidian_Distance();
extern float vector_len_sqr();
extern float *vector();
extern boolean feat_description;
extern char **feature_desc;
extern str100 outDataDir;

extern boolean *train_smp, *test_smp;
extern int nrSamplesTest, nrSamplesTrain;


static Matrix CoVar = { 0, 0, NULL },
	aux1 = { 0, 0, NULL }, aux2 = { 0, 0, NULL };
static Matrix CorrelMat = { 0, 0, NULL };

static float correlation( f1, f2 )
int f1, f2;
{
 int i, j, s, row;
 float *x = NULL, *y = NULL, bias;
 float Ex = 0.0, Ey = 0.0, Exy = 0.0;   /* expected values */
 float Sx = 0.0, Sy = 0.0;              /* standard deviation */
 float Cov, CorCoeff;   /* Covariance and linear correlation coefficient */

 if( nrSamplesTrain == 0 )
   EXIT("ERROR: No training samples in 'correlation()'");

 x = vector( nrSamplesTrain ); y = vector( nrSamplesTrain );
 s = 0; row = 0;
 for( i = 0; i < U->nrClass; i++ )
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     if( train_smp[s] )
     {
       x[row] = U->C[i].S[j*U->nrFeat+f1];
       y[row] = U->C[i].S[j*U->nrFeat+f2];
       /* printf("f1=%d  f2=%d  x[%d,%d]=%f  y[%d,%d]=%f\n",
		f1, f2, i+1, j+1, x[row], i+1, j+1, y[row] );	/**/
       row++;
     }
     s++;
   }
 
 /* all x and y values have been determined now:
    calculate linear correlation coefficient 
 */
 for( i = 0; i < row; i++ )
 {
   Ex += x[i]; Ey += y[i]; Exy += x[i] * y[i];
 }
 Ex /= (float)row; Ey /= (float)row; Exy /= (float)row;
 Cov = Exy - Ex*Ey;

 for( i = 0; i < row; i++ )
 {
   Sx += (x[i]-Ex)*(x[i]-Ex);
   Sy += (y[i]-Ey)*(y[i]-Ey);
 }
 Sx /= (float)row;
 Sy /= (float)row;
 /* compensate for the bias of the standard deviation */
 if( nrSamplesTrain > 1 )
 {
   bias = (float)(nrSamplesTrain)/(float)(nrSamplesTrain-1);
   Sx *= bias; Sy *= bias;
 }
 Sx = (float)sqrt( (double)(Sx) );
 Sy = (float)sqrt( (double)(Sy) );

 CorCoeff = Cov / (Sx*Sy);

#ifdef xxx
 printf("\n>>>----------------------------------------------------<<<\n");
 printf("         RESULT OF LINEAR CORRELATION ANALYSIS            \n");
 printf("            Feature nr. 1 = %d", f1+1 );
 if( feat_description ) printf( " (%s)", feature_desc[f1] ); NL;
 printf("               Mean=%7.3f   Standard deviation=%7.3f\n", Ex, Sx );
 printf("            Feature nr. 2 = %d", f2+1 );
 if( feat_description ) printf( " (%s)", feature_desc[f2] ); NL;
 printf("               Mean=%7.3f   Standard deviation=%7.3f\n\n", Ey, Sy );
 printf("            Covariance=%7.3f\n", Cov );
 printf("    ===>    Correlation coefficient=%7.3f   <===\n", CorCoeff );
 printf(">>>----------------------------------------------------<<<\n\n");
#endif

 FREE( x ); FREE( y );
 return( CorCoeff );
}


/* Generate a global Correlation matrix of the data */
static void stats_globalCorrel()
{
 int i, j, dim;
 float cor;

 dim = U->nrFeat;

 Matrix_Free( &CorrelMat );
 CorrelMat.row = dim; CorrelMat.col = dim;
 Matrix_Alloc( &CorrelMat );

 for( i = 0; i < dim; i++ )
   for( j = i; j < dim; j++ )
   {
     cor = correlation( i, j );
     CorrelMat.Elem[i*dim+j] = cor;
     if( i != j )
       CorrelMat.Elem[j*dim+i] = cor;
   }
}


/* Generate a global Covariance matrix of the data */
static void stats_globalCoVar()
{
 FeatSelectVector FSV = NULL;
 int i;

 FSV = (feature*) malloc( U->nrFeat * sizeof(struct feature_) );
 for( i = 0; i < U->nrFeat; i++ )
 {
   FSV[i].rank = i; FSV[i].crit = (float)EMPTY;
 }
 Matrix_Free( &CoVar );
 scatter_mats_dev_kit( FSV, U->nrFeat, &CoVar, &aux1, &aux2 );
 Matrix_Free( &aux1 ); Matrix_Free( &aux2 );
 FREE( FSV );
}


void init_extr_ms( s )
statistics *s;
{
 s->min = NULL; s->max = NULL;
 s->mean = NULL; s->stddev = NULL;
}


void free_extr_ms( s )
statistics *s;
{
 FREE( s->min ); FREE( s->max );
 FREE( s->mean ); FREE( s->stddev );
 init_extr_ms( s );
}


static void alloc_extr_ms( s, dim )
statistics *s;
int dim;
{
 int k;

 free_extr_ms( s );
 s->min = (FeatVector) malloc(sizeof(FeatVector*) * dim ); CHKPTR(s->min);
 s->max = (FeatVector) malloc(sizeof(FeatVector*) * dim ); CHKPTR(s->max);
 s->mean = (FeatVector) malloc(sizeof(FeatVector*) * dim ); CHKPTR(s->mean);
 s->stddev = (FeatVector) malloc(sizeof(FeatVector*) * dim ); CHKPTR(s->stddev);
 for( k = 0; k < dim; k++ )
 {
   s->min[k] = INFINITY; s->max[k] = -INFINITY;
   s->mean[k] = 0.0; s->stddev[k] = 0.0;
 }
}


/*
 * generate statistical parameters of the training data:
 * minima, maxima, means, standard deviations (local for
 * each class and global
*/
void stats_extr_ms()
{
 int s, sC, nrTrainC;
 int i, j, k;
 boolean cont_local = FALSE, cont_global = FALSE;
 float val, mean, variance, min, max;
 float aux, meanC, varianceC, minC, maxC, stddev, stddevC, bias;

 if( nrSamplesTrain == 0 )
   EXIT("ERROR: No training samples in 'stats_extr_ms()'");

 alloc_extr_ms( &(U->stat), U->nrFeat );
 for( i = 0; i < U->nrClass; i++ )
   alloc_extr_ms( &(U->C[i].stat), U->nrFeat );

 /* do it for each feature */
 for( k = 0; k < U->nrFeat; k++ )
 {
   s = 0;
   mean = 0.0, variance = 0.0, min = INFINITY, max = -INFINITY;
   for( i = 0; i < U->nrClass; i++ )
   {
     meanC = 0.0; varianceC = 0.0; minC = INFINITY, maxC = -INFINITY;
     sC = 0; nrTrainC = 0;
     /* mean */
     for( j = 0; j < U->C[i].numSampl; j++ )
     {
       if( train_smp[s] )
       {
         val = U->C[i].S[j*U->nrFeat+k];
         meanC += val;
         mean += val;
         if( val < min ) min = val; if( val > max ) max = val;
         if( val < minC ) minC = val; if( val > maxC ) maxC = val;
	 nrTrainC++;
       }
       sC++; s++;
     }
     meanC /= (float)nrTrainC;

     /* variance */
     s -= sC; sC = 0;
     for( j = 0; j < U->C[i].numSampl; j++ )
     {
       if( train_smp[s] )
       {
         val = U->C[i].S[j*U->nrFeat+k];
         aux = val - meanC;
         varianceC += aux * aux;
       }
       sC++;
     }
     varianceC /= (float) nrTrainC;
     /* compensate for the bias of the standard deviation */
     if( nrTrainC > 1 )
     {
	bias = (float)(nrTrainC)/(float)(nrTrainC-1);
	if( varianceC == 0.0 )
	{
         fprintf(stderr,"WARNING: Standard deviation is zero for class %d",i+1);
         fprintf(stderr," and feature %d\n", k+1 );
         if( ! cont_local )
         {
           fprintf(stderr," This might cause some trouble. Continue?");
           get_boolean( TRUE, &cont_local );
	   if( ! cont_local )
               { printf("\n...Exitus...\n"); exit(1); }
         }
        }
        varianceC *= bias;
     }
     stddevC = (float) sqrt(varianceC);
     /* write statistical parameters for this class */
     U->C[i].stat.min[k] = minC; U->C[i].stat.max[k] = maxC;
     U->C[i].stat.mean[k] = meanC; U->C[i].stat.stddev[k] = stddevC;
   }

   s = 0;
   /* calculate global statistical parameters */
   mean  /= (float)nrSamplesTrain;
   for( i = 0; i < U->nrClass; i++ )
     for( j = 0; j < U->C[i].numSampl; j++ )
     {
       if( train_smp[s] )
       {
         val = U->C[i].S[j*U->nrFeat+k];
         aux = val - mean;
         variance += aux * aux;
       }
       s++;
     }
   variance /= (float)nrSamplesTrain;
   /* compensate for the bias of the standard deviation */
   if( nrSamplesTrain > 1 )
   {
     bias = (float)(nrSamplesTrain)/(float)(nrSamplesTrain-1);
     if( variance == 0.0 )
     {
       fprintf(stderr,"WARNING: Global standard deviation is zero for" );
       fprintf(stderr," feature %d\n", k+1 );
       if( ! cont_global )
       {
         fprintf(stderr," This might cause some trouble. Continue?");
         get_boolean( TRUE, &cont_global );
         if( ! cont_global )
             { printf("\n...Exitus...\n"); exit(1); }
       }
      }
      variance *= bias;
   }
   stddev = (float) sqrt(variance);
   /* write global statistical parameters */
   U->stat.min[k] = min; U->stat.max[k] = max;
   U->stat.mean[k] = mean; U->stat.stddev[k] = stddev;
 }
}


static float *inertia = NULL;
static float IG, IW, IB;
static Matrix DispersionMat = { 0, 0, NULL };
static float Fisher_criterion;


static void free_inertia_dispersion()
{
 FREE( inertia );
 Matrix_Free( &DispersionMat );
 Fisher_criterion = (float)INFINITY;
}


static void alloc_inertia_dispersion()
{
 /* class-specific, global, within-class and between-class */
 free_inertia_dispersion();
 inertia = (float*) malloc(sizeof(float) * U->nrClass ); CHKPTR(inertia);
 DispersionMat.row = U->nrClass; DispersionMat.col = U->nrClass;
 Matrix_Alloc( &DispersionMat );
 Fisher_criterion = (float)INFINITY;
}


/*
 * generate statistical parameters of the training data:
 * INERTIA: global, within-class and between-class
*/
static void stats_inertia()
{
 int s, nrTrainC;
 int i, j, dim;
 float Iwi;
 float dist, dist_sqr, len_xi_sqr, len_mi_sqr;

 if( nrSamplesTrain == 0 )
   EXIT("ERROR: No training samples in 'stats_inertia()'");

 dim = U->nrFeat;
 alloc_inertia_dispersion();

 IW = 0.0; IG = 0.0;
 s = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   Iwi = 0.0; nrTrainC = 0;
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     if( train_smp[s] )
     {
       /* calculate the squareed norm between the sample and its class-specific
		mean */
       /* printf("x[%d]: ", i+1); showFV( dim, &(U->C[i].S[j*dim]) );
       printf("m[%d]: ", i+1); showFV( dim, U->C[i].stat.mean );	/**/
       dist = Euclidian_Distance( &(U->C[i].S[j*dim]), U->C[i].stat.mean, dim );
       dist_sqr = dist * dist;
       len_xi_sqr = vector_len_sqr( &(U->C[i].S[j*dim]), dim );
       /* printf("dist_sqr=%f  len_xi_sqr=%f", dist_sqr, len_xi_sqr ); DBG; /**/
       Iwi += dist_sqr;
       IG += len_xi_sqr;
       nrTrainC++;
     }
     s++;
   }
   Iwi /= (float) nrTrainC;
   /* printf("Iwi[%d]=%f", i+1, Iwi ); DBG;	/**/
   inertia[i] = Iwi;
   IW += (float) nrTrainC * Iwi;
 }

 /* Between-class */
 IB = 0.0;
 s = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   nrTrainC = 0;
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     if( train_smp[s] )
       nrTrainC++;
     s++;
   }
   len_mi_sqr = vector_len_sqr( U->C[i].stat.mean, dim );
   /* printf("nrTrainC[%d]=%d  len_mi_sqr[%d]=%f",
		i+1, nrTrainC, i+1, len_mi_sqr ); DBG; /**/
   IB += (float) nrTrainC * len_mi_sqr;
 }

 IG /= (float) nrSamplesTrain;
 IW /= (float) nrSamplesTrain;
 IB /= (float) nrSamplesTrain;
}


/*
 * The dispersion of two classes:
 * disp = Euclidean_Norm( Mean(class_1) - Mean(class_2) )
		/ square_root(Inertia(class_2))
 *
*/
static float dispersion( i, j, dim )
{
 float disp;

 disp = Euclidian_Distance( U->C[i].stat.mean, U->C[j].stat.mean, dim ) /
		(float)sqrt( inertia[j] );
 /* printf("Dispersion(%d,%d)=%f", i+1, j+1, disp );	DBG; /**/
 return( disp );
}


/*
 * generate statistical parameters of the training data:
 * DISPERSION and OVERLAPPING
*/
void stats_disp_overlap()
{
 int i, j, dim;

 dim = U->nrFeat;
 Fisher_criterion = IB / IW;
 if( DispersionMat.row == 0 || DispersionMat.col == 0 )
 {
   printf("WARNING: Cannot write (%d,%d) dispersion matrix",
		DispersionMat.row, DispersionMat.col );
   WAIT; return;
 }
 for( i = 0; i < U->nrClass; i++ )
   DispersionMat.Elem[i*U->nrClass+i] = 0.0;

 for( i = 0; i < U->nrClass; i++ )
   for( j = 0; j < U->nrClass; j++ )
   {
     if( i == j )
       DispersionMat.Elem[i*U->nrClass+i] = 0.0;
     else
       DispersionMat.Elem[i*U->nrClass+j] = dispersion( i, j, dim );
   }
}


void generate_stats( calc_all )
boolean calc_all;  /* Calculate the complete set */
{
 printf("Generating statistical information of the universe...\n");
 stats_extr_ms();

 if( ! calc_all )
   return;

 stats_globalCoVar();
 stats_globalCorrel();
 stats_inertia();
 stats_disp_overlap();
}


static void show_stat_matrix( M, f )
Matrix *M;
FILE *f;
{
 int i, j;

 if( M->row == 0 || M->col == 0 )
 {
   fprintf(f,"WARNING: Cannot display (%d,%d) matrix", M->row, M->col );
   WAIT; return;
 }
 for( i = 0; i < M->row; i++ )
 {
   for( j = 0; j < M->col; j++)
     fprintf(f," %10.5f ", (float)M->Elem[i*M->col+j]);
   fprintf(f,"\n");
 }
 fprintf(f,"\n");
}


/*
 * Write the statistical parameters to a file 
*/

static char tstr[] = "~~~~~~~~~~~~~~~~~~~~";

void show_stats_extr_ms( f )
FILE *f;
{
 int i, j;

 fprintf( f, "\n%s EXTREMA, MEAN AND STANDARD DEVIATION %s\n", tstr, tstr );
 fprintf( f, "\tGLOBALLY (for all classes)\n");
 fprintf( f, "%15s%15s%15s%15s\n", "Min", "Max", "Mean", "Std.Dev.");
 for( i = 0; i < U->nrFeat; i++ )
 {
   fprintf( f, "%15.5f%15.5f%15.5f%15.5f\t",
	U->stat.min[i], U->stat.max[i], U->stat.mean[i], U->stat.stddev[i] );
   fprintf( f, "# %4d", i+1 );
   if( feat_description )
     fprintf( f, "\t%s", feature_desc[i] );
   fprintf( f, "\n" );
 }

 fprintf( f, "\n\tLOCALLY (for each particular class)\n");
 for( j = 0; j < U->nrClass; j++ )
 {
   fprintf( f, " ~~~ CLASS # %d = %s ~~~~~\n", j+1, U->C[j].name );
   fprintf( f, "%15s%15s%15s%15s\n", "Min", "Max", "Mean", "Std.Dev.");
   for( i = 0; i < U->nrFeat; i++ )
   {
     fprintf( f, "%15.5f%15.5f%15.5f%15.5f\t",
	U->C[j].stat.min[i], U->C[j].stat.max[i],
	U->C[j].stat.mean[i], U->C[j].stat.stddev[i] );
     fprintf( f, "# %4d", i+1 );
     if( feat_description )
       fprintf( f, "\t%s", feature_desc[i] );
     fprintf( f, "\n" );
   }
 }
}


static void show_stats_CoVar( f )
FILE *f;
{
 fprintf( f, "\n%s GLOBAL COVARIANCE MATRIX %s\n", tstr, tstr );
 show_stat_matrix( &CoVar, f );
 Matrix_Free( &CoVar );
}


static void show_stats_Correl( f )
FILE *f;
{
 fprintf( f, "\n%s GLOBAL CORRELATION MATRIX %s\n", tstr, tstr );
 show_stat_matrix( &CorrelMat, f );
 Matrix_Free( &CorrelMat );
}


static void show_stats_inertia( f )
FILE *f;
{
 int i;

 fprintf( f, "\n%s INERTIA %s\n", tstr, tstr );
 for( i = 0; i < U->nrClass; i++ )
   fprintf( f, "\t%10.5f for class %d(%s)\n",
	inertia[i], i+1, U->C[i].name );
 fprintf( f, "\tGLOBAL INERTIA:\t\t%10.5f\n", IG );
 fprintf( f, "\tWITHIN-CLASS INERTIA:\t%10.5f\n", IW );
 fprintf( f, "\tBETWEEN-CLASS INERTIA:\t%10.5f\n", IB );
}


void show_stats_disp_overlap( f )
FILE *f;
{
 fprintf( f, "\n%s DISPERSION AND OVERLAPPING %s\n", tstr, tstr );
 fprintf( f, "\tFISHER'S CRITERION: %10.5f\n", Fisher_criterion );
 fprintf( f, "\tDISPERSION-MATRIX:\n");
 show_stat_matrix( &DispersionMat, f );
 free_inertia_dispersion();
}


void show_stat_param( f )
FILE *f;
{
 generate_stats( TRUE );
 fprintf( f, "\t--- S T A T I S T I C S  O F  T H E  F E A T U R E S  ---\n" );
 show_stats_extr_ms( f );
 show_stats_CoVar( f );
 show_stats_Correl( f );
 show_stats_inertia( f );
 show_stats_disp_overlap( f );
 fprintf( f, "%s%s%s%s\n", tstr, tstr, tstr, tstr );
}

void drop_stat_param()
{
 FILE *sf = NULL;
 char statFile[101], name[81];

 set_all_samples_train();
 strcpy( statFile, outDataDir );
 printf("Saving statistical parameters to (<Enter> for screen):\n\t%s",
	statFile );
 gets( name );

 if( name[0] == '\0' )
 {
   NL;
   show_stat_param( stdout );
   return;
 }

 strcat( statFile, name );
 sf = fopen( statFile, f_open_text_w );
 if( sf == NULL )
   { printf("Cannot open \"%s\"!\n", statFile ); WAIT; return; }

 show_stat_param( sf );
 fclose( sf );
}
