/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:24 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


 /*   I N T E R F A C E  T O  T H E  G N U P L O T  P R O G R A M   */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "def.h"


extern universe *U;

extern float *vector();
extern str100 outDataDir;
extern str100 gnuplot_exe;
extern boolean verbose;
extern boolean feat_description;
extern char **feature_desc;


static char cmd[201], str_gnu[201];
static str100 gp_file_name, class_file_name, err_file_name, corr_file_name;
static FILE *gf = NULL;
static str20 buf;

#ifndef UNIX
static FILE *tmpf = NULL;
static char gp_tmp_file[] = "_tmp.gnu";
#endif
static char gp_file_default[] = "tooldiag.gnu";
static char err_file_default[] = "error";
static char corr_file_default[] = "lin_corr";


void GNU_init_gnu_file()
{
 strcpy( gp_file_name, outDataDir );
 strcat( gp_file_name, gp_file_default );
 gf = fopen( gp_file_name, f_open_text_w );
 if( gf == NULL )
   { printf("Cannot open %s! Exitus...\n", gp_file_name ); exit(1); }

 /* Standard header */
 fprintf( gf, "#\n#  Batch file for gnuplot\n" );
 fprintf( gf, "#  Generated automatically !\n#\n" );
 fprintf( gf, "#\n" );
 fprintf( gf, "# FRAMEMAKER INTERCHANGE FORMAT (gnuplot 3.5 and up)\n#\n");
 fprintf( gf, "# set term mif\n# set output \"out.mif\"\n# set size 0.5,0.5\n");
 fprintf( gf, "#\n" );
 fprintf( gf, "# Pure LaTeX\n#\n");
 fprintf( gf, "# set term latex\n# set output \"out.tex\"\n# set size 0.8,0.8\n");
 fprintf( gf, "#\n#\n" );
}


void GNU_close_gnu_file()
{
 fprintf( gf, "pause -1 \"Hit return to exit...\"" );
 fclose( gf );
 gf = NULL;
}


void GNU_execute()
{
 GNU_close_gnu_file();
#ifdef UNIX
 sprintf( cmd, "cd %s\n\texec gnuplot %s", outDataDir, gp_file_name );
 if( verbose )
   printf("\n --- Execute:\n\t%s\n", cmd );
 system( cmd );
#endif

#ifndef UNIX
 /*
    generate a temporary file in the ACTUAL directory and call the
    gnuplot executable
 */
 tmpf = fopen( gp_tmp_file, f_open_text_w );
 if( tmpf == NULL )
 {
   fprintf( stderr, "I need write permission in the actual directory\n" );
   fprintf( stderr, " in order to call gnuplot.\n");
   return;
 }
 /* Generate local file */
 fprintf( tmpf, "#\n# Temporary gnuplot file, do not edit\n#\n" );
 fprintf( tmpf, "cd \"%s\"\n", outDataDir );
 fprintf( tmpf, "load \"%s\"\n", gp_file_default );
 fclose( tmpf );

 /* Do the system call to gnuplot */
 sprintf( cmd, "%s %s", gnuplot_exe, gp_tmp_file );
 system( cmd );
 /* unlink( gp_tmp_file ); */
#endif
}



void GNU_comment( str )
char *str;
{ fprintf( gf, "#  %s\n", str ); }


void GNU_title( title )
char *title;
{ fprintf( gf, "set title \"-TOOLDIAG- %s\"\n", title ); }


void GNU_xlabel( xlabel )
char *xlabel;
{ fprintf( gf, "set xlabel \"%s\"\n", xlabel ); }


void GNU_ylabel( ylabel )
char *ylabel;
{ fprintf( gf, "set ylabel \"%s\"\n", ylabel ); }


void GNU_zlabel( zlabel )
char *zlabel;
{ fprintf( gf, "set zlabel \"%s\"\n", zlabel ); }


void GNU_xrange( min, max )
float min, max;
{ fprintf( gf, "set xrange [%f:%f]\n", min, max ); }


void GNU_yrange( min, max )
float min, max;
{ fprintf( gf, "set yrange [%f:%f]\n", min, max ); }


void GNU_zrange( min, max )
float min, max;
{ fprintf( gf, "set zrange [%f:%f]\n", min, max ); }


/*----------------------------------------------------* 
	S C A T T E R  P L O T
 *----------------------------------------------------*/


/* For scatter plots store the points, name of classes, and indices */
/* The classes are not necessarily equal to the 'univerrse' classes */

static int plot_dim = EMPTY;
static float *x = NULL, *y = NULL, *z = NULL;

static int nr_class = EMPTY;		/* how many plotting classes */
static int sumSampl = EMPTY;		/* how many samples */
static int class_ptr = EMPTY;		/* actual class */
static int *smp_per_class = NULL;	/* how many samples per class */
static int *class_indx = NULL;		/* the class of each sample */
static char **class_nam = NULL;		/* the names of each class */



#define OFFSET (1.0)
#define TOL (10.0)
static void adapt_coord( min, max )
float *min, *max;
{
 float d, delta;
 d = *max - *min;
 if( d == 0.0 )
   delta = OFFSET;
 else
   delta = d / TOL;
 *min -= delta; *max += delta;
}


void GNU_scatter_init( nc, ns, dim )
int nc, ns;	/* number of classes, number of samples */
int dim;	/* 2-D or 3-D plot */
{
 class_ptr = 0;
 sumSampl = 0;
 nr_class = nc;
 sumSampl = ns;
 plot_dim = dim;
 if( plot_dim != 2 && plot_dim != 3 ) EXIT("GNU_scatter_init> Dim-Problems");
 smp_per_class = (int*) malloc( nc * sizeof(int)); CHKPTR(smp_per_class);
 class_indx = (int*) malloc( ns * sizeof(int)); CHKPTR(class_indx);
 class_nam = (char**) malloc( nc * sizeof(char*) ); CHKPTR(class_nam);
 x = vector( ns ); CHKPTR(x);
 y = vector( ns ); CHKPTR(y);
 if( dim == 3 )
 {
   z = vector( ns ); CHKPTR(z);
 }
}


void GNU_scatter_close()
{
 int i;

 FREE(smp_per_class);
 FREE(class_indx);
 for( i = 0; i < nr_class; i++ )
   FREE(class_nam[i]);
 FREE(class_nam);
 FREE(x); FREE(y); FREE(z);
 class_ptr = EMPTY;
 nr_class = EMPTY;
 plot_dim = EMPTY;
}


void GNU_scatter_set_name( name )
char *name;
{
 int i;
 char **new_class_nam = NULL;

 if( nr_class == 0 ) EXIT("GNU_scatter_set_name> Problems");

 new_class_nam = (char**) malloc((1+class_ptr)*sizeof(char*));
 CHKPTR(new_class_nam);

 for( i = 0; i < class_ptr; i++ )
 {
   new_class_nam[i] = (char*) malloc( (1+strlen(class_nam[i])) * sizeof(char) );
   CHKPTR(new_class_nam[i]);
   strcpy( new_class_nam[i], class_nam[i] );
   FREE( class_nam[i] );
 }

 new_class_nam[class_ptr] = (char*) malloc( (1+strlen(name)) * sizeof(char) );
 CHKPTR(new_class_nam[class_ptr]);
 strcpy(new_class_nam[class_ptr], name );

 FREE( class_nam );
 class_nam = new_class_nam;

 class_ptr++;
}


void GNU_scatter_set_index( indx, c )
int indx, c;
{
 class_indx[ indx ] = c;
 if( indx+1 > sumSampl )
   sumSampl = indx+1;
}


void GNU_scatter_set_point( x_coord, y_coord, z_coord, row, dim )
float x_coord, y_coord, z_coord;
int row, dim;
{
 if( row < 0 || row >= sumSampl ) EXIT("GNU_scatter_set_point> wrong index");
 x[ row ] = x_coord;
 y[ row ] = y_coord;
 if( dim == 3 )
   z[ row ] = z_coord;
}


void GNU_scatter_set_smp( c, smp )
int c, smp;
{ smp_per_class[ c ] = smp; }


static void GNU_scatter_bug_warning()
{
 /* ----------------------------------------------------------------- */
 /*  There seem to be bugs in gnuplot (version 3.5),                  */
 /*  in relation to the 'splot' command                               */
 /*  3-D plots only work properly from version 3.6 on (hopefully)     */
 /* ----------------------------------------------------------------- */
 fprintf( stderr,"\nWarning: Gnuplot 3-D plots bug-free only from");
 fprintf( stderr," version 3.6 and up\n");
 fprintf( stderr, "\t3-D plot might not be correct !!!\n");
}


void GNU_scatter_plot()
{
 int i, j, row, dummy;
 float min, max;
 FILE *cf = NULL;

 if( class_ptr != nr_class ) EXIT("gnu_scatter_plot> Problems");
 if( plot_dim != 2 && plot_dim != 3 ) EXIT("GNU_scatter_plot> Dim-Problems");

 /* set ranges */

 getMin( x, sumSampl, &min, &dummy ); getMax( x, sumSampl, &max, &dummy );
 adapt_coord( &min, &max );
 GNU_xrange( min, max );

 getMin( y, sumSampl, &min, &dummy ); getMax( y, sumSampl, &max, &dummy );
 adapt_coord( &min, &max );
 GNU_yrange( min, max );

 if( plot_dim == 3 )
 {
   getMin( z, sumSampl, &min, &dummy ); getMax( z, sumSampl, &max, &dummy );
   adapt_coord( &min, &max );
   GNU_zrange( min, max );
 }

 /* Copy the data to gnuplot data files: one for each class */
 row = 0;
 for( i = 0; i < nr_class; i++ )
 {
   strcpy( class_file_name, outDataDir );
   strcat( class_file_name, class_nam[i] );
   cf = fopen( class_file_name, f_open_text_w );
   if( cf == NULL )
     { printf("Cannot open %s! Exitus...\n", class_file_name ); exit(1); }

   if( verbose )
     printf("Dumping %s\n", class_file_name );
   fprintf( cf, "# %s\n", class_nam[i] );
   for( j = 0; j < smp_per_class[i]; j++ )
   {
     fprintf( cf, "%f\t%f", x[row], y[row] );
     if( plot_dim == 3 )
       fprintf( cf, "\t%f", z[row] );
     fprintf( cf, "\n");
     row++;
   }
   fclose( cf );
 }

 if( plot_dim == 2 ) fprintf( gf, "plot ");
 if( plot_dim == 3 ) fprintf( gf, "splot ");

  for( i = 0; i < nr_class-1; i++ )
    fprintf( gf, "\"%s\", ", class_nam[i] );
  fprintf( gf, "\"%s\"\n", class_nam[nr_class-1] );
 if( plot_dim == 3 )
   GNU_scatter_bug_warning();
 GNU_scatter_close();
}



#define ERR_TOL (10.0)
void GNU_error_plot( errInfo, nrSelFeat, errEstStrat )
float *errInfo;
int nrSelFeat;
char *errEstStrat;
{
 int i;
 float maxErr, minErr, dE;
 FILE *cf = NULL;

 GNU_init_gnu_file();

 strcpy( err_file_name, outDataDir );
 strcat( err_file_name, err_file_default );
 cf = fopen( err_file_name, f_open_text_w );
 if( cf == NULL )
   { printf("Cannot open %s! Exitus...\n", err_file_name ); exit(1); }

 maxErr = 0.0; minErr = INFINITY;
 for( i = 0; i < nrSelFeat; i++ )
 {
   fprintf( cf, "%3d %7.5f\n", i+1, errInfo[i] );
   if( errInfo[i] > maxErr )
     maxErr = errInfo[i];
   if( errInfo[i] < minErr )
     minErr = errInfo[i];
 }
 fclose( cf );


 dE = maxErr - minErr;

 if( maxErr == 0.0 )
 {
   maxErr = 1.0;
   minErr = -1.0;
 }
 else
 {
   minErr -= dE / ERR_TOL;
   if( minErr < 0.0 )
     minErr = 0.0;
   maxErr += dE / ERR_TOL;
   if( maxErr > 100.0 )
     maxErr = 100.0;
 }
 minErr = 0.0;	/* start always from zero */


 sprintf( str_gnu, "# Universe %s\n#\n", U->name );
 GNU_comment( str_gnu );

 if( nrSelFeat > 1 )
   fprintf( gf, "set xrange [1:%d]\n", nrSelFeat );
 else
   fprintf( gf, "set xrange [0.9:1.1]\n" );
 fprintf( gf, "set xtics 0,1,%d\n", nrSelFeat );
 fprintf( gf, "set xtics (" );
 for( i = 0; i < nrSelFeat; i++ )
 {
   fprintf( gf, "\"%d", U->FSV[i].rank+1 );
   if( feat_description )
     fprintf( gf, "(%.20s)", feature_desc[U->FSV[i].rank] );
   fprintf( gf, "\" %d", i+1 );
   if( i != nrSelFeat-1 )
     fprintf( gf, ", " );
 }
 fprintf( gf, ")\n" );
 fprintf( gf, "set yrange [%d:%d]\n", (int)minErr, (int)(maxErr)+1 );
 fprintf( gf, "set title \"ERROR ESTIMATION = %s\"\n", errEstStrat );
 fprintf( gf, "set xlabel \"Sequence of selected features (accumulated)\"\n");
 fprintf( gf, "set ylabel \"Estimated error [%%]\"\n");
 fprintf( gf, "plot \"%s\" with linespoints\n", err_file_default );
 GNU_execute();
#ifdef UNIX
 (void)sleep(1);
#endif
}



void GNU_linear_corr( corr, x, y, dim, first, second, nrClass, classes )
float corr, *x, *y;
int dim;
int nrClass, *classes;
{
 int i, dummy;
 FILE *cf = NULL;
 float min, max;

 GNU_init_gnu_file();

 GNU_comment("Batch file to visualize linear correlation");
 sprintf( str_gnu, "Universe %s\n#\n", U->name );
 GNU_comment( str_gnu );

 strcpy( corr_file_name, outDataDir );
 strcat( corr_file_name, corr_file_default );
 cf = fopen( corr_file_name, f_open_text_w );
 if( cf == NULL )
   { printf("Cannot open %s! Exitus...\n", corr_file_name ); exit(1); }

 for( i = 0; i < dim; i++ )
   fprintf( cf, "%f  %f\n", x[i], y[i] );
 fclose( cf );

 getMin( x, dim, &min, &dummy ); getMax( x, dim, &max, &dummy );
 adapt_coord( &min, &max );
 GNU_xrange( min, max );

 getMin( y, dim, &min, &dummy ); getMax( y, dim, &max, &dummy );
 adapt_coord( &min, &max );
 GNU_yrange( min, max );


 strcpy( str_gnu, "LINEAR CORRELATION FOR CLASS" );
 if( nrClass > 1 )
   strcat( str_gnu, "ES" );
 strcat( str_gnu, " " );
 for( i = 0; i < nrClass; i++ )
 {
   sprintf( buf, "%d ", classes[i] );
   strcat( str_gnu, buf );
 }
 sprintf( buf, " is %.3f", corr );
 strcat( str_gnu, buf );

 GNU_title( str_gnu );

 fprintf( gf, "set xlabel \"First feature = %d",  first );
 if( feat_description )
   fprintf( gf, " (%s)", feature_desc[first-1] );
 fprintf( gf, "\"\n" );
 fprintf( gf, "set ylabel \"Second feature = %d", second );
 if( feat_description )
   fprintf( gf, " (%s)", feature_desc[second-1] );
 fprintf( gf, "\"\n" );

 fprintf( gf, "plot \"%s\"\n", corr_file_default );
 GNU_execute();
#ifdef UNIX
 (void)sleep(1);
#endif
}
