/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:31 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "def.h"

extern universe *U;
extern float **matrix_alloc();
extern float *vector();

extern int nrSamplesTrain;
extern int *train_vec, *train_vecClass, *train_vecSmp;

static str80 buf;

#define MAX_ITER	1000000
#define MAX_ETA		(100.0)
#define MAX_EPS		(1000000.0)
#define MAX_WEIGHT	INFINITY

static float eta = 0.001;		/* Learning Rate */
static float eps = 0.00001;	/* Similarity between two successive steps */
static int max_iter = 1000;	/* Maximum number of iterations */
static boolean stoch_learn =FALSE;	/* Stochastic (per pattern) learning or batch */

void get_linlearn_param( ask_batch, ask_similar )
boolean ask_batch, ask_similar;
{
 printf("Learning rate of Gradient Descent (actual=%.7f)? ", eta );
 get_f_range( &eta, 0.0, MAX_ETA, LEFT_OPEN__RIGHT_OPEN );

 printf("Maximum number of learning steps (actual=%d)? ", max_iter );
 get_d_range( &max_iter, 1, MAX_ITER, LEFT_CLOSED__RIGHT_CLOSED );

 if( ask_similar )
 {
   printf("Similarity of weights between two successive steps (actual=%.7f)? ",
	 eps );
   get_f_range( &eps, 0.0, MAX_EPS, LEFT_OPEN__RIGHT_OPEN );
 }

 if( ask_batch )
 {
   printf("Stochastic learning (on a per pattern basis)");
   get_boolean( stoch_learn, &stoch_learn );
 }
}


void show_linlearn_param( f, show_batch, show_near )
FILE *f;
boolean show_batch, show_near;
{
 fprintf(f,"\t\tGradient Descent Learning\n");
 if( show_batch )
   if( stoch_learn )
     fprintf(f,"\t\tusing stochastic weight updating\n");
   else
     fprintf(f,"\t\tusing batch weight updating\n");
 fprintf(f,"\t\tLearning rate: %.5f\n", eta );
 fprintf(f,"\t\tStopping after %d iterations\n", max_iter );
 if( show_near )
   fprintf(f,"\t\t  or if two successive weight matrices are closer than %f\n",
	eps );
}


boolean gradient_descent( W, d, nsf, FSV, verbose )
float ***W;			/* The weight matrix to be updated */
float (*d)(int, float*);	/* The Discriminant Function */
int nsf;
FeatSelectVector FSV;
boolean verbose;
{
 int d1, i, j, k, p, indx, c, clss, iter = 0;
 boolean random_order = TRUE, done = FALSE;
 FeatVector x = NULL;		/* Augmented sample vector */
 float batch_error;
 float nabla_ij;
 float *delta_p = NULL;
 double diff_sum, diff;

 c = U->nrClass;
 d1 = nsf + 1;

 delta_p = vector( c );
 x = vector( d1 ); x[0] = 1.0;

 NL;
 while( ! done )
 {
   diff_sum = 0.0;
   if( stoch_learn )
   {
     random_order = TRUE;
     gen_train_vec( random_order );
     for( p = 0; p < nrSamplesTrain; p++ )
     {
       indx = train_vec[p];
       clss = train_vecClass[ indx ];
       k = train_vecSmp[ indx ];

       provideFV( &(x[1]), clss, k, nsf, FSV );
       /* printf("\nSample # %d which is %d of class %d(%s)\n",
		p+1, k+1, clss+1, U->C[clss].name );
       printf("The augmented sample is: "); showFV( d1, x );	/**/

       /* Calculate the deltas:
	target is 0 for i != clss and 1 for i == clss */
       for( i = 0; i < c; i++ )
         delta_p[i] = - d( i, x );
       delta_p[clss] += 1.0;	/* The real class */
       /* for( i = 0; i < c; i++ )
         printf("delta[%d] = %f\n", i, delta_p[i] );  DBG;  /**/
       for( i = 0; i < c; i++ )
       {
         for( j = 0; j < d1; j++ )
         {
           nabla_ij = delta_p[i] * x[j];
           (*W)[i][j] += eta * nabla_ij;
           if( (*W)[i][j] > MAX_WEIGHT )
           {
             printf("\nWeight for class %d and component %d is getting too big",
		i+1, j );
             printf(" at iteration %d\n", iter );
             printf("Check learning rate! Current = %.7f\n\n", eta );
             return( FALSE );
           }
           /* printf("CLASS %d: nabla_ij=%f W:%f--->%f", i+1,
		nabla_ij, (*W)[i][j]-eta * nabla_ij,(*W)[i][j] ); NL;	/**/
           diff = (double)(nabla_ij * nabla_ij);
           diff_sum += diff;
         }
       }
     }
     diff_sum = (float)sqrt(c*d1*eta*eta*diff_sum / (float)nrSamplesTrain );
   }

   else	/* BATCH LEARNING */
   {
     random_order = FALSE;
     gen_train_vec( random_order );
     for( i = 0; i < c; i++ )
     {
       for( j = 0; j < d1; j++ )
       {
         batch_error = 0.0;
         for( p = 0; p < nrSamplesTrain; p++ )
         {
           indx = train_vec[p];
           clss = train_vecClass[ indx ];
           k = train_vecSmp[ indx ];

           provideFV( &(x[1]), clss, k, nsf, FSV );
           /* printf("\nSample # %d which is %d of class %d(%s)\n",
		p+1, k+1, clss+1, U->C[clss].name );
           printf("The augmented sample is: "); showFV( d1, x );	/**/
           delta_p[i] = - d( i, x );
           if( i == clss )
             delta_p[clss] += 1.0;	/* The real class */
           nabla_ij = delta_p[i] * x[j];
           batch_error += nabla_ij;
         }
         batch_error *= eta;
         diff = (double)(batch_error * batch_error);
         diff_sum += diff;
         (*W)[i][j] += batch_error;
         if( (*W)[i][j] > MAX_WEIGHT )
         {
           printf("\nWeight for class %d and component %d is getting too big",
		i+1, j );
           printf(" at iteration %d\n", iter );
           printf("Check learning rate! Current = %.7f\n\n", eta );
           return( FALSE );
         }
         /* printf("Batch_error=%f W:%f--->%f", batch_error,
		(*W)[i][j]-batch_error, (*W)[i][j] ); NL;   /**/
       }
     }
     diff_sum = (float)sqrt(c * d1 * diff_sum / (float)nrSamplesTrain );
   }

   iter++;
   done = iter >= max_iter || diff_sum <= eps;

   if( verbose )
   {
     printf("GRADIENT DESCENT: Iteration = %d  Approximation error = %lf\r",
	iter, diff_sum ); fflush(stdout);
     if( iter >= max_iter )
       printf("\nEnd of learning: Reaching maximum of %d iterations\n", iter );
     if( diff_sum <= eps )
       printf("\nEnd of learning: Reaching approximation error limit of %f\n",
		eps );
   }
   /* NL; matrix_print( (*W), c, nsf + 1 ); /* WAIT;	/**/
 }
 FREE(x); FREE( delta_p );
 return( TRUE );
}


boolean perceptron_learning( W, d, nsf, FSV, verbose )
float ***W;			/* The weight matrix to be updated */
int (*d)(int, float*);		/* The Discriminant Function */
int nsf;
FeatSelectVector FSV;
boolean verbose;
{
 int not_learned, error, t_ip, d_ip, d1, i, j, k, p, indx, c, clss, iter = 0;
 boolean random_order = TRUE, done = FALSE;
 FeatVector x = NULL;		/* Augmented sample vector */
 float aux;

 c = U->nrClass;
 d1 = nsf + 1;

 x = vector( d1 ); x[0] = 1.0;

 NL;
 while( ! done )
 {
   not_learned = 0;
   random_order = TRUE;
   gen_train_vec( random_order );
   for( p = 0; p < nrSamplesTrain; p++ )
   {
     indx = train_vec[p];
     clss = train_vecClass[ indx ];
     k = train_vecSmp[ indx ];

     provideFV( &(x[1]), clss, k, nsf, FSV );
     /* printf("\nSample # %d which is %d of class %d(%s)\n",
	p+1, k+1, clss+1, U->C[clss].name );
     printf("The augmented sample is: "); showFV( d1, x );	/**/
     for( i = 0; i < c; i++ )
     {
       d_ip = d( i, x );
       /* Set the target value */
       if( i == clss )
         t_ip = 1;
       else
         t_ip = -1;

       error = t_ip*d_ip;
       if( error == -1 )
         not_learned++;
       aux = eta * (float)(1 - error) * (float)t_ip;
       for( j = 0; j < nsf; j++ )
         (*W)[i][j] += aux * x[j];
     }
   }

   iter++;
   done = iter >= max_iter || not_learned == 0;

   if( verbose )
   {
     printf("PERCEPTRON LEARNING: Iteration = %d", iter );
     printf("    Patterns currently not learned: %d   \r", not_learned );
     fflush(stdout);
     if( iter >= max_iter )
     {
       printf("\nEnd of learning: Reaching maximum of %d iterations\n", iter );
     }
     if( done && not_learned == 0 )
     {
       printf("\nPatterns successfully separated after %d steps\n", iter );
       WAIT;
     }
   }
 }
 FREE(x);
 return( TRUE );
}
