/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:25 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "def.h"

extern boolean is_ascii_file();

extern universe *U;
extern boolean verbose;
extern boolean *train_smp, *test_smp;
extern int nrSamplesTest, nrSamplesTrain;
extern str100 outDataDir;


static boolean done;
static str80 buf;
static str100 linebuf;
static char defNameTrain[] = "_train", defNameTest[] = "_test";
static int menuItem = EMPTY;


#define MAXSTRLEN 100

#define RANDOM 1
#define EXPLICIT 2
#define BOOTSTRAP 3

static int split_method = RANDOM;


static void merge_data_files()
{
 FILE *in1 = NULL, *in2 = NULL, *out = NULL;
 str80 in1Nam, in2Nam, outNam;
 str80 classNam1, classNam2;
 boolean in1Ascii, in2Ascii, outAscii;
 FeatVector featVec1 = NULL, featVec2 = NULL, featVec3 = NULL;
 int j, k1, k2, dim1, dim2, dim3;

 printf("\n--- Merging two data files (all filenames with full path) ---\n");
 printf("\tMerge file: "); gets( in1Nam );
 if( in1Nam[0] == '\0' ) { printf("File name empty..."); gets(buf); return; }
 in1Ascii = is_ascii_file( in1Nam );
 switch( in1Ascii )
 {
   case TRUE : in1 = fopen( in1Nam, f_open_text_r ); break;
   case FALSE : in1 = fopen( in1Nam, f_open_bin_r ); break;
   case EMPTY : printf("File not found..."); gets(buf); return;
   default: fprintf(stderr, "Error in merge_data_files, exit...\n"); exit(1);
 }
 printf("\t  and file: "); gets( in2Nam );
 if( in2Nam[0] == '\0' ) { printf("File name empty..."); gets(buf); return; }
 if( strcmp(in1Nam,in2Nam)==0 ) { printf("File names identical...");
                 gets(buf); return; }
 in2Ascii = is_ascii_file( in2Nam );
 switch( in2Ascii )
 {
   case TRUE : in2 = fopen( in2Nam, f_open_text_r ); break;
   case FALSE : in2 = fopen( in2Nam, f_open_bin_r ); break;
   case EMPTY : printf("File not found..."); gets(buf); fclose(in1); return;
   default: fprintf(stderr, "Error in merge_data_files, exit...\n"); exit(1);
 }
 printf("\t   to file: "); gets( outNam );
 if( outNam[0] == '\0' ) { printf("File name empty..."); gets(buf); return; }
 if( strcmp(in1Nam,outNam)==0 ) { printf("File names identical...");
                gets(buf); return; }
 if( strcmp(outNam,in2Nam)==0 ) { printf("File names identical...");
                gets(buf); return; }
 outAscii = FALSE;
 printf(" Write %s in b)inary or a)scii mode ?b\b", outNam ); gets( buf );
 if( buf[0] == 'a' )
   outAscii = TRUE;
 if( outAscii )
   out = fopen( outNam, f_open_text_w );
 else
   out = fopen( outNam, f_open_bin_w );
 if( out == NULL )
 {
   printf("Cannot open %s! Exitus...\n", outNam );
   fclose(in1); fclose(in2);
 }

 if( verbose )
   printf("\tMerging %s and %s to %s\n", in1Nam, in2Nam, outNam );

 dataline( in1, linebuf ); sscanf( linebuf, "%d", &dim1 );
 dataline( in2, linebuf ); sscanf( linebuf, "%d", &dim2 );
 dim3 = dim1 + dim2;
 featVec1 = (FeatVector) malloc(sizeof(FeatVector*) * dim1);
 featVec2 = (FeatVector) malloc(sizeof(FeatVector*) * dim2);
 featVec3 = (FeatVector) malloc(sizeof(FeatVector*) * dim3);
 fprintf( out, "%d\n", dim3 );

 /* scan the input files line by line and write the result to the output */
 while( !feof( in1 ) && !feof( in2 ) )
 {
   k1 = 0; k2 = 0;
   while( k1 < dim1 && !feof( in1 ) )
   {
     if( in1Ascii )
       fscanf( in1, "%f", &(featVec1[k1]) );
     else
       fread( &(featVec1[k1]), sizeof(float), 1, in1 );
     k1++;
   }
   while( k2 < dim2 && !feof( in2 ) )
   {
     if( in2Ascii )
       fscanf( in2, "%f", &(featVec2[k2]) );
     else
       fread( &(featVec2[k2]), sizeof(float), 1, in2 );
     k2++;
   }
   /* get the first name */
   if( !feof( in1 ) )
   {
     if( in1Ascii )
       fscanf( in1, "%s", classNam1 );
     else
     {
       j = 0;
       do
       {
         fread( &(classNam1[j]), sizeof(char), 1, in1 );
         if( classNam1[j] != '\n' )
           j++;
       }
       while( classNam1[j] != '\n' && j < MAXSTRLEN );
       classNam1[j] = '\0';
     }
   }
   /* get the second name */
   if( !feof( in2 ) )
   {
     if( in2Ascii )
       fscanf( in2, "%s", classNam2 );
     else
     {
       j = 0;
       do
       {
         fread( &(classNam2[j]), sizeof(char), 1, in2 );
         if( classNam2[j] != '\n' )
           j++;
       }
       while( classNam2[j] != '\n' && j < MAXSTRLEN );
       classNam2[j] = '\0';
     }
   }
   if( !feof( in1 ) && !feof( in2 ) )
   {
     if( strcmp(classNam1,classNam2) == 0 )
     {
       /* write the new line */
       if( outAscii )
       {
         for( k1 = 0; k1 < dim1; k1++ )
           fprintf( out, "%f ", featVec1[k1] );
         for( k2 = 0; k2 < dim2; k2++ )
           fprintf( out, "%f ", featVec2[k2] );
         fprintf( out, "%s\n", classNam1 );
       }
       else
       {
         fwrite( featVec1, sizeof(FeatVector*), dim1, out );
         fwrite( featVec2, sizeof(FeatVector*), dim2, out );
         fprintf( out, "%s\n", classNam1 );
       }
     }
     else
     {
       printf("ERROR: Detected different class names in files %s and %s\n",
     	   in1Nam, in2Nam );
       fprintf(stderr,"\t>>>%s<<<  ----  >>>%s<<< ...", classNam1, classNam2 );
       gets( buf );
       FREE( featVec1 ); FREE( featVec2 ); FREE( featVec3 );
       fclose(in1); fclose(in2); fclose(out); return;
     }
   }
 }
 FREE( featVec1 ); FREE( featVec2 ); FREE( featVec3 );
 fclose(in1); fclose(in2); fclose(out);
}


static void getSplitMethod()
{
 printf("\n\t\t----  S P L I T T I N G   M E T H O D  ----\n");
 printf("\n\tSplit RANDOMLY(1), EXPLICITLY(2) or by BOOTSTRAPPING(3)? %d\b",
	split_method );
 get_d_range( &split_method, RANDOM, BOOTSTRAP, LEFT_CLOSED__RIGHT_CLOSED );
}


static FILE *open_file( name, outAscii )
char *name;
boolean outAscii;
{
 FILE *f = NULL;

 if( outAscii )
   f =  fopen( name, f_open_text_w );
 else
   f = fopen( name, f_open_bin_w );
 if( f == NULL )
   { fprintf( stderr, "Cannot open %s", name ); WAIT; }
 return( f );
}


static boolean ok_out_files( f1, f2 )
FILE *f1, *f2;
{
 if( f1 == NULL || f2 == NULL )
 {
   if( f1 != NULL ) fclose( f1 );
   if( f2 != NULL ) fclose( f2 );
   return FALSE;
 }
 return TRUE;
}


#define NR_ITER_MAX 100

static void split_train_test()
{
 FILE *outTrain = NULL, *outTest = NULL, *out;
 str80 outNam1, outNam2, tmpStr, outNam1Root, outNam2Root;
 boolean outAscii;
 float percentTrain, disc_space;
 int i, s, k, c, smp, nrLoops, nrIter = 5;

 printf("\n--- Splitting the actual data set into training and test ---\n");
 printf("\tTraining file name: %s\b\b\b\b\b\b", defNameTrain ); gets( linebuf );
 strcpy( outNam1Root, outDataDir );
 if( linebuf[0] != '\0' )
   strcpy( tmpStr, linebuf );
 else
   strcpy( tmpStr, defNameTrain );
 strcat( outNam1Root, tmpStr );

 strcpy( outNam2Root, outDataDir );
 printf("\t    Test file name: %s\b\b\b\b\b", defNameTest ); gets( linebuf );
 if( linebuf[0] != '\0' )
 {
   if( strcmp( tmpStr, linebuf ) == 0 )
   {
      printf("File names identical, using default.\n");
      strcpy( outNam1Root, outDataDir );
      strcat( outNam1Root, defNameTrain );
      strcpy( tmpStr, defNameTest );
   }
   else
     strcpy( tmpStr, linebuf );
 }
 else
   strcpy( tmpStr, defNameTest );
 strcat( outNam2Root, tmpStr );


 outAscii = FALSE;
 printf(" Write in b)inary or a)scii mode ?b\b" ); gets( buf );
 if( buf[0] == 'a' )
   outAscii = TRUE;

 getSplitMethod();
 if( split_method == RANDOM )
   getPercentTrain( &percentTrain );

 if( split_method != EXPLICIT )
 {
   printf(" Repeat splitting how many times (actual=%d)? ", nrIter );
   get_d_range( &nrIter, 1, NR_ITER_MAX, LEFT_CLOSED__RIGHT_CLOSED );
   disc_space = (float) (nrIter * U->nrFeat * sizeof(float) * U->sumSampl);
   printf(" This means more than %.0f bytes of disc space \n", disc_space );
   printf("\tfor the binary format and much more for the ascii format !!!\n");
   printf(" Continue (y/n)?y\b"); gets( buf );
   if( buf[0] == 'n' ) return;
   nrLoops = nrIter;
 }
 else
   nrLoops = 1;

 strcpy( outNam1, outNam1Root );
 strcpy( outNam2, outNam2Root );

 for( i = 0; i < nrLoops; i++ )
 {
   if( split_method != EXPLICIT && nrLoops > 1 )
   {
     sprintf( outNam1, "%s.%d", outNam1Root, i+1 );
     sprintf( outNam2, "%s.%d", outNam2Root, i+1 );
   }

   outTrain = open_file( outNam1, outAscii );
   outTest = open_file( outNam2, outAscii );
   if( ! ok_out_files( outTrain, outTest ) )
     return;

   switch( split_method )
   {
     case RANDOM:
       split_tt_random( U->sumSampl, percentTrain ); break; 
     case EXPLICIT:
       split_tt_explicit( U->sumSampl ); break; 
     case BOOTSTRAP:
       split_tt_bootstrap( U->sumSampl ); break; 
     default : EXIT("Trouble with splitting method");
   }
   fprintf( stderr, "R E S U L T  O F  S P L I T ");
   if( nrLoops > 1 )
     fprintf( stderr, " --- Train-Test pair %3d", i+1 );
   fprintf( stderr, "\n # training samples=%d(%7.2f%%)",
	nrSamplesTrain, 100.0*(float)nrSamplesTrain/(float)U->sumSampl );
   fprintf( stderr, "  # test samples=%d(%7.2f%%)  All=%d\n",
    nrSamplesTest, 100.0*(float)nrSamplesTest/(float)U->sumSampl, U->sumSampl );

   fprintf( outTrain, "%d\n", U->nrFeat );
   fprintf( outTest, "%d\n", U->nrFeat );

   c = 0; smp = 0;
   for( s = 0; s < U->sumSampl; s++ )
   {
     if( train_smp[s] )
       out = outTrain;
     else if ( test_smp[s] )
       out = outTest;
     else { fprintf(stderr,"Sample neither train nor test. Exit...\n");exit(1);}

     if( outAscii )
       for( k = 0; k < U->nrFeat; k++ )
         fprintf( out, "%f ", U->C[c].S[smp*U->nrFeat+k] );
     else
       fwrite( &(U->C[c].S[smp*U->nrFeat]), sizeof(FeatVector*),
		U->nrFeat, out );
     fprintf( out, "%s\n", U->C[c].name );
     /* fprintf( stdout, "out=%d  s=%d  smp=%d  train=%d  test=%d  %s\n",
	(int)out, s, smp, train_smp[s], test_smp[s], U->C[c].name );	/**/
     smp++;
     if( smp == U->C[c].numSampl )
     {
       c++;
       smp = 0;
     }
   }
   fclose(outTrain); fclose(outTest); out = NULL;
 }
}

/***************************************/
/*****     MENU ITEMS              *****/
/***************************************/

/*   -----   INTERFACE MENU    -----        */
#define LVQ		0
#define SNNS		1
#define MERGE		2
#define SPLIT		3

#define NUM_INTFACEMENU_ITEMS 4


static struct intFaceMenuItems {
        int index;
        str80 name;
} intFaceMenuItem[] = {
        { LVQ,		"Learning Vector Quantization (LVQ_PAK)" },
        { SNNS,		"Stuttgart Neural Network Simulator (SNNS)" },
        { MERGE,	"Merge two data files to a single data file" },
        { SPLIT,	"Split a data file into two files (train & test)" } };




static void execMenuItem()
{
 switch( menuItem )
 {
   case LVQ:	lvqMenu(); break;
   case SNNS:	snnsMenu(); break;
   case MERGE:	merge_data_files(); break;
   case SPLIT:	split_train_test(); break;
   default: fprintf(stderr,"What the heck is item %d? - exit...\n", menuItem );
 }
}

static char sl[] =
     "-----------------";


void intFaceMain()
{
 int i, d; char *str = NULL;

 done = FALSE;
 do
 {
   printf("\n%s  I N T E R F A C I N G   %s\n", sl, sl );
   for( i = 0; i < NUM_INTFACEMENU_ITEMS; i++ )
   {
     str = intFaceMenuItem[i].name;
     printf("(%d)\t%s\n", i+1, str );
   }

   printf("Choice: "); gets(buf);
   switch( buf[0] )
   {
     /* special cases */
     case '?': help( MENU_INTFACE, buf ); break;
     case '\0': done = TRUE; showUniv( stdout ); break;
     default:
       sscanf(buf,"%d", &d);
       if( d >= 1 && d <= NUM_INTFACEMENU_ITEMS )
       {
         menuItem = d-1;
         execMenuItem();
       }
   }
 }
 while( ! done );
}


void intface()
{
 if( U->nrClass >= 1 )
   do
   {
     intFaceMain();
   }
   while( !done );
 else
   { printf("Please load universe first !...");gets( buf );}
}


/**********  H E L P  *****************/

void helpMENU_INTFACE( cmd )
char *cmd;
{
 int item;

 sscanf( &(cmd[1]), "%d", &item );
 item--;	/* discount one since in the menu it is one more */
 switch( item )
 {
   case LVQ:	item_help("LVQ"); break;
   case SNNS:	item_help("SNNS"); break;
   case MERGE:	item_help("MERGE"); break;
   case SPLIT:	item_help("SPLIT"); break;
   default: ; break;
 }
}

