/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:25 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "def.h"
#include "matrxdef.h"

extern universe *U;
extern boolean verbose;
extern boolean num_recip_error;


/* Feature Extraction based on

	P R I N C I P A L  C O M P O N E N T  A N A L Y S I S
				alias
	K A R H U N E N - L O E V E  E X P A N S I O N
*/

#define EPSILON_WR (0.001)
#define EPSILON_WI (0.001)

/*
 *  the column vectors of the feature extractor 'W' are the Eigenvectors
 *  of the global covariance matrix 'T'
*/
static void feat_extract_PCA( T, W )
Matrix *T, *W;
{
 float maxLambda, *wr = NULL, *wi = NULL, *sortEigVal = NULL, *eigVec = NULL;
 float sumEigVal, sumVariance;
 Matrix aux, T_orig;		/* the original matrix */
 int i, j, dim, maxI, nrEigVal, maxIndex;
 boolean found;

 /* save the original matrix to calculate later the eigenvectors */
 init_Matrix( &T_orig );
 Copy_Matrix( T, &T_orig );

 dim = T->row;

 if( dim == 1 )
 {
   W->row = dim; W->col = dim; Matrix_Alloc( W );
   W->Elem[0] = 1.0;
   return;
 }

 wr = (float*) malloc( dim * sizeof(float) ); CHKPTR(wr);
 wi = (float*) malloc( dim * sizeof(float) ); CHKPTR(wi);
 sortEigVal = (float*) malloc( dim * sizeof(float) ); CHKPTR(sortEigVal);
 eigVec = (float*) malloc( dim * sizeof(float) ); CHKPTR(eigVec);

 printf("Calculating eigenvalues...\n");
 if( dim == 1 )
 {
   wr[0] = T->Elem[0]; wi[0] = 0.0;
 }
 else
 {
   num_recip_error = FALSE;
   balanc( T ); /* printf("balanced:\n"); print_Matrix( T ); /**/
   /* If an error occured during the numerical calculus, reset dim */
   if( num_recip_error ) dim = EMPTY;
   elmhes( T ); /* printf("elmhes:\n"); print_Matrix( T ); /**/
   if( num_recip_error ) dim = EMPTY;
   hqr( T, wr, wi );  /**/
   if( num_recip_error ) dim = EMPTY;
 }
 for( i = 0; i < dim; i++ )
 {
   printf("Calculating eigenvalue nr. %d\r", i+1 ); fflush(stdout);
   /* printf("wr[%d]=%32.30f  wi[%d]=%32.30f\n", i, wr[i], i, wi[i] ); /**/
   /* Is there an imaginary part of a resonably large real eigenvalue? */
   if( fabs(wi[i]) > EPSILON_WI && fabs(wr[i]) > EPSILON_WR )
   {
     printf("Eigenvalue %3d had some imaginary part: %20.5f\n", i+1, wi[i] );
   }
 }

 /* Eigenvalues are determined - Calaculate corresponding eigenvectors */
 /* Sort in descending order */

 /* Analyse the eigenvalues first */
 maxIndex = dim;
 /* printf("Maxindex=%d", maxIndex );DBG;/**/
 nrEigVal = 0; found = FALSE; sumEigVal = 0.0; sumVariance = 0.0;
 while( ! found )
 {
   maxLambda = -INFINITY;
   for( i = 0; i < maxIndex; i++ )
   {
     if( wr[i] > maxLambda )
     {
       maxLambda = wr[i];
       maxI = i;
     }
   }
   if( maxLambda <= EPSILON_WR )
     found = TRUE;
   else
   {
     sortEigVal[ nrEigVal ] = maxLambda;
     sumEigVal += maxLambda;
     nrEigVal++;
     wr[ maxI ] = -INFINITY;
     if( nrEigVal == maxIndex )
       found = TRUE;
   }
 }
 printf("\nN O N - Z E R O  E I G E N V A L U E S   -   ");
 if( nrEigVal == 0 )
 {
  printf("none\n Sorry, this data seems inappropriate for this");
  printf(" feature extraction algorithm"); WAIT;
 }
 else
 {
   printf("Reconstruction error\n");
   for( i = 0; i < nrEigVal; i++ )
   {
     sumVariance += sortEigVal[i];
     printf("Eigenvalue %3d of %3d = %15.6f        %10.4f %%\n",
	i+1, nrEigVal, sortEigVal[i], 100.0 * (1.0 - sumVariance/sumEigVal) );
   }

   /* Ask user for new dimension */
   printf("\n");

   printf(" - DETERMINATION OF DIMENSION OF THE REDUCED FEATURE SPACE -\n\n");
   printf("As the minimum of the original dimension of the feature space(%d)",
	dim );
   printf("\nand the number of non-zero eigenvalues is %d\n", nrEigVal );
   printf("choose a value for the reduced dimension between %d and %d: ",
	1, nrEigVal );
   get_d_range( &nrEigVal, 1, nrEigVal, LEFT_CLOSED__RIGHT_CLOSED );

   /* Create matrix with dimensions D x d */
   W->row = dim; W->col = nrEigVal;
   Matrix_Alloc( W );
   Reset_Matrix( W );

   for( i = 0; i < nrEigVal; i++ )
   {
     printf("Calculating eigenvector for eigenvalue nr. %d\r", i+1 );
     fflush(stdout);
     calcEigVec( &T_orig, &(sortEigVal[i]), eigVec, i );
     for( j = 0; j < W->row; j++ )
       W->Elem[j*nrEigVal+i] = eigVec[j];
   }
   printf("\nNew values from the inverse iteration procedure:\n");
   for( i = 0; i < nrEigVal; i++ )
     printf("Eigenvalue nr. %3d of %d = %15.6f\n",i+1, nrEigVal, sortEigVal[i]);
 } 


 /* DEBUG-VERIFY
 if( nrEigVal > 0 )
 {
   printf("\nResult of multiplication of T_orig and feature extractor:\n");
   init_Matrix( &aux );
   Mult_Matrix( &T_orig, W, &aux );
   print_Matrix( &aux );
   printf("Each column divided by respective eigenvalue\n");
   printf(" (should be identical to the feature extractor below):\n");
   for( i = 0; i < nrEigVal; i++ )
     for( j = 0; j < W->row; j++ )
       aux.Elem[j*nrEigVal+i] /= sortEigVal[i];
   print_Matrix( &aux ); Matrix_Free( &aux );
 }
 /* END DEBUG-VERIFY */


 Matrix_Free( &T_orig );
 FREE(wr); FREE(wi); FREE(sortEigVal); FREE(eigVec);
}



void featExtrPCA( W )
Matrix *W;
{
 int j, dim;
 Matrix T, SB, SW;
 FeatSelectVector FSV = NULL;

 dim = U->nrFeat;

 FSV = (feature*) malloc( sizeof(feature) * dim ); CHKPTR(FSV);
 for( j = 0; j < dim; j++ )
 {
   FSV[ j ].rank = j;
   FSV[ j ].crit = (float)EMPTY;
 }

 init_Matrix( &T ); init_Matrix( &SB ); init_Matrix( &SW );
 printf("\n  --- FEATURE EXTRACTION BASED ON ---\n");
 printf("     PRINCIPAL COMPONENT ANALYSIS    \n");
 printf("      (KARHUNEN-LOEVE-EXPANSION)    \n\n");
 printf("Calculating %d x %d global covariance matrix...",
	dim, dim ); fflush(stdout);
 scatter_mats_dev_kit( FSV, dim, &T, &SB, &SW );	/**/
 NL;

 feat_extract_PCA( &T, W );

 FREE(FSV);
 Matrix_Free( &T ); Matrix_Free( &SB ); Matrix_Free( &SW );
}
