/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:31 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "def.h"
#include "matrxdef.h"
#include "clasdef.h"

extern universe *U;
extern Matrix *Means, MeanGlobal, CovarianceGlobal;
extern boolean verbose_reason;
extern float **matrix_alloc();
extern float *vector();
extern int *int_vector();
extern boolean rejection_opt;

extern FILE *cpf;               /* Classification protocol file */
extern boolean have_label;      /* Only deliver the classified class for
        pre-classified data */


/* The weight matrix
   number of rows: c = number of classes
   number of columns: d+1 number of features+1 */
static int c, d, d1;
static float **W = NULL;
static float *decision_func = NULL, *x = NULL;
static int *decision_func_perceptron = NULL;
static boolean old_rejection_opt;


#define DELTA 0
#define DETERMINISTIC_LMS 1
#define PERCEPTRON 2

#define NUM_TRAIN_ALGOS 3

struct trainAlgos {
        int index;
        str80 name;
} trainAlgo[] = {
  { DELTA, "Delta-Rule (a.k.a Widrow-Hoff rule, LMS rule)" },
  { DETERMINISTIC_LMS,
	"Deterministic Least-Mean-Square rule (a.k.a Pseudoinverse)" },
  { PERCEPTRON, "Perceptron learning rule" }
};
/* static int training_algo = DELTA;	/**/
/* static int training_algo = DETERMINISTIC_LMS;	/**/
static int training_algo = PERCEPTRON;	/**/


static void get_training_algo()
{
 int i, j;

 printf("\n ### Specify training algorithm for the linear machine ###\n" );
 for( i = 0; i < NUM_TRAIN_ALGOS; i++ )
   printf("(%d) %s\n", i+1, trainAlgo[i].name );
 printf("Choice: ");
 get_d_range( &j, 1, NUM_TRAIN_ALGOS, LEFT_CLOSED__RIGHT_CLOSED );
 training_algo = trainAlgo[j-1].index;
 if( training_algo == DELTA )
   get_linlearn_param( TRUE, TRUE );
 if( training_algo == PERCEPTRON )
   get_linlearn_param( FALSE, FALSE );
}


void show_LINMACH( f )
FILE *f;
{
 fprintf( f, "\tLEARNING ALGORITHM: %s\n", trainAlgo[training_algo].name );
 switch( training_algo )
 {
   case DELTA: show_linlearn_param( f, TRUE, TRUE ); break;
   case DETERMINISTIC_LMS: ; break;
   case PERCEPTRON: show_linlearn_param( f, FALSE, FALSE ); break;
   default:
     EXIT("Unknown training algorithm");
 } 
}


void initLINMACH()
{
 free_induced();
 /* Ask for the training algorithm */
 get_training_algo();
}


static void prepareLINMACH()
{
 int i;

 free_induced();
 c = U->nrClass;
 d = U->nrSelFeat;
 d1 = U->nrSelFeat + 1;

 W = matrix_alloc( c, d1 ); 
 x = vector( d1 ); x[0] = 1.0;	/* Augmented sample vector */
 decision_func = vector( c );

 if( training_algo == DELTA || training_algo == PERCEPTRON )
 {
   if( training_algo == PERCEPTRON )
   {
      decision_func_perceptron = int_vector( c );
      old_rejection_opt = rejection_opt;
      rejection_opt = TRUE;
   }
   /* Set the initial weights to random values */
   for( i = 0; i < c; i++ )
   {
     random_unit_vector( W[i], d1 );
     /* printf("Random init W[%3d]=",i ); showFV( d1, W[i] ); DBG;   /**/
   }
 }
}


static float LDF( i, x )
int i;
float *x;
{
 int j;
 float di = 0.0;

 for( j = 0; j < d1; j++ )
 {
   di += W[i][j] * x[j];
   /* printf("W[%d][%d]=%f x[%d]=%f\n", i,j, W[i][j], j, x[j] ); DBG; /**/
 }

 return( di );
}


/* Perceptron Threshold Function */
static int PTF( i, x )
int i;
float *x;
{
 return( SGN( LDF( i, x ) ) );
}


extern boolean gradient_descent();
extern boolean perceptron_learning();

static boolean deterministicLMS( W, nsf, FSV, verbose )
float ***W;             /* The weight matrix to be updated */
int nsf;
FeatSelectVector FSV;
boolean verbose;
{
 MatElem dummy;
 int i, j;
 Matrix mat1, mat2, mat3;

 init_Matrix( &mat1 ); init_Matrix( &mat2 ); init_Matrix( &mat3 );

 init_means_covar( FSV, nsf );
 gen_means( FSV, nsf );
 gen_global_mean_and_covar( FSV, nsf );

 /* printf("Global mean vector:\n" );
 print_Matrix( &MeanGlobal );
 printf("Global Covariance matrix:\n" );
 print_Matrix( &CovarianceGlobal ); DBG; /**/

 Invert_Matrix( &CovarianceGlobal, &dummy );

 for( i = 0; i < c; i++ )
 {
   /* printf(" *** mean vector of %d:\n", i+1 );
   print_Matrix( &(Means[i]) );	/**/

   Subtract_Matrix( &(Means[i]), &MeanGlobal, &mat1 ); /* mat1 = mi-m */ 
   Mult_Matrix( &CovarianceGlobal, &mat1, &mat2 ); /* mat2 = Cov^-1 * (mi-m) */
   /* print_Matrix( &mat1 );
   print_Matrix( &mat2 ); DBG;	/**/
   for( j = 0; j < nsf; j++ )
     (*W)[i][j+1] = U->C[i].a_priori_prob * (float)mat2.Elem[j];

   Transpose_Matrix( &MeanGlobal, &mat1 ); /* mat1 = m ^T */
   Mult_Matrix( &mat1, &mat2, &mat3 );	/* mat3 = m ^T *  Cov^-1 * (mi-m) */
   (*W)[i][0] = U->C[i].a_priori_prob * (1.0 - (float)mat3.Elem[0] );
 }

 Matrix_Free( &mat1 ); Matrix_Free( &mat2 ); Matrix_Free( &mat3 );
 free_global_mean_and_covar();
 return( TRUE );
}


/*
 * Induce a classifier with the actual set of training samples
 * The dimension of the vector is 'nsf'
 * 'FSV' contains the information of the actual subset of features
 * that are used.
 * 'iact' tells if interaction if desired. For instance in error
 * estimation this is not the case
*/
void induceLINMACH( nsf, FSV, iact )
int nsf;
FeatSelectVector FSV;
boolean iact;
{
 boolean ok, interact, verbose_induction = FALSE;

 interact = iact;
 if( interact )
 {
   printf("\n --- Inducing a %d-dimensional Linear Machine ---\n\n", nsf);
   printf("Verbose learning of decision structure?");
   get_boolean( FALSE, &verbose_induction );
 }
 prepareLINMACH();

 switch( training_algo )
 {
   case DELTA: ok = gradient_descent( &W, LDF, nsf, FSV, verbose_induction );
	break;
   case DETERMINISTIC_LMS:
	ok = deterministicLMS( &W, nsf, FSV, verbose_induction ); break;
   case PERCEPTRON:
	ok = perceptron_learning( &W, PTF, nsf, FSV, verbose_induction );
	break;
   default:
     EXIT("Unknown training algorithm");
 } 
 if( ! ok )
 {
   if( training_algo == DELTA )
     {fprintf(stderr, "Unable to perform gradient descent."); WAIT;}
   if( training_algo == PERCEPTRON )
     {fprintf(stderr, "Unable to do perceptron learning."); WAIT;}
 }
 else
   if( verbose_induction )
   {
     printf("\nLEARNED %d x %d WEIGHT MATRIX:\n", U->nrClass, nsf+1 );
     matrix_print( W, U->nrClass, nsf+1 );
   }
 verbose_induction = FALSE;
}


/*
 *  Free all the structures that were dynamically allocated to store
 *  the classifier, for instance matrices of prototypes
*/
void free_inducedLINMACH()
{
 matrix_free( &W, c );
 FREE( decision_func );
 FREE( x );
 if( training_algo == PERCEPTRON )
 {
    FREE( decision_func_perceptron );
    rejection_opt = old_rejection_opt;
 }
}


static int class1PERCEPTRON( x, dim, name )
FeatVector x;
int dim;
char *name;
{
 int i, poscnt, best_class, resulting_class;

 poscnt = 0;
 for( i = 0; i < c; i++ )
 {
   decision_func_perceptron[i] = PTF( i, x );
   if( verbose_reason )
     fprintf( cpf, " * Perceptron decision function of class %d(%s) is %d\n",
	i+1, U->C[i].name, decision_func_perceptron[i] );
   if( decision_func_perceptron[i] == 1 )
   {
     best_class = i;
     poscnt++;
   }
 }

 /* Only one positive class (with value 1) --> decision possible */
 if( poscnt == 1 )
 {
   if( verbose_reason )
     fprintf( cpf, "   * Best class was %d(%s)\n",
		best_class+1, U->C[best_class].name );

   if( ! have_label )
     return( best_class );

   /* 
      First case: correct classification:
      The name of the best class corresponds to the expected class
      return the value 'CORRECT_CLASS'
   */
   if( (strcmp(U->C[best_class].name,name)==0) )
     resulting_class = CORRECT_CLASS;
   else
   {
     /* 
        Second case: wrong classification:
        return the value of the best class: 'best_class'
     */
     resulting_class = best_class;
     if( verbose_reason )
       fprintf( cpf, " *** Confused %s with expected class %s\n",
		U->C[best_class].name, name );
   }
 }
 /* Several positive classes --> decision impossible */
 else
 {
   resulting_class = UNDEFINED_CLASS;
   if( verbose_reason && have_label )
	fprintf( cpf, " *** Confused UNKNOWN class with expected class %s\n",
		name );
 }

 return( resulting_class );
}


/*
 *  Classify a single sample 'smp' of length 'nsf'
 *  The name of the expected class is stored in 'name'
 *  'FSV' is not needed here, but for compatibility reasons passed
*/
int class1testLINMACH( nsf, FSV, smp, name )
int nsf;
FeatSelectVector FSV;
FeatVector smp;
char *name;
{
 int d1, i, j, resulting_class, best_class;
 float maxfunc;

 d1 = nsf + 1;
 copy_FV( smp, &(x[1]), nsf );
 if( verbose_reason )
 {
   fprintf( cpf,"\n\n ---- DECISION FUNCTIONS FOR UNKNOWN SAMPLE:\n");
   fshowFV( cpf, nsf, smp );
 }

 if( training_algo == PERCEPTRON )
   return( class1PERCEPTRON( x, d1, name ) );


 for( i = 0; i < c; i++ )
 {
   decision_func[i] = LDF( i, x );
   if( verbose_reason )
     fprintf( cpf, "   * Decision function of class %d(%s) is %f\n",
	i+1, U->C[i].name, decision_func[i] );
 }

 /* decide the best class */
 best_class = 0; maxfunc = -INFINITY;
 for( i = 0; i < c; i++ )
   if( decision_func[i] > maxfunc )
   {
     maxfunc = decision_func[i];
     best_class = i;
   }

 if( verbose_reason )
   fprintf( cpf, "   * Best class was %d(%s)\n",
		best_class+1, U->C[best_class].name );

 if( ! have_label )
   return( best_class );

 /* 
    First case: correct classification:
    The name of the best class corresponds to the expected class
    return the value 'CORRECT_CLASS'
 */
 if( (strcmp(U->C[best_class].name,name)==0) )
   resulting_class = CORRECT_CLASS;
 else
 {
   /* 
      Second case: wrong classification:
      return the value of the best class: 'best_class'
   */
   resulting_class = best_class;
   if( verbose_reason )
     fprintf( cpf, " *** Confused %s with expected class %s\n",
		U->C[best_class].name, name );
 }
 return( resulting_class );
}
