/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:25 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "def.h"

extern universe *U;
extern boolean verbose;
extern str100 outDataDir;
extern boolean feat_description;
extern char **feature_desc;


static str80 buf, name, lvqFile, featNameFile;
static boolean done;
static int menuItem = EMPTY;

#define MAX_PRECISION 30
static int save_precision = 5;
static char formatStr[11];


static void set_save_precision()
{
 printf("Save values with precision (actual=%d)? ", save_precision );
 get_d_range( &save_precision, 0, MAX_PRECISION, LEFT_CLOSED__RIGHT_CLOSED );
 sprintf( formatStr, "%%.%df ", save_precision );
 /* printf("formatStr=>>>%s<<<", formatStr ); WAIT;     /**/
}


void gen_lvq_data_file()
{
 int i, j, k;
 FILE *lvq = NULL, *feats = NULL;

 if( U->nrSelFeat == 0 )
   { printf(" Select features first please!..." ); gets( buf ); return; }

 strcpy( lvqFile, outDataDir );
 printf("Saving the data in LVQ format in file:\n\t\t%s", lvqFile );
 gets( name );
 if( name[0] != '\0' )
 {
   strcat( lvqFile, name );
   lvq = fopen( lvqFile, f_open_text_w );
   if( lvq == NULL )
     { printf("Cannot open %s! Exitus...\n", lvqFile ); exit(1); }
   if( verbose )
     printf("Generating LVQ-File: %s\n", lvqFile );

   set_save_precision();

   fprintf( lvq, "%d\n", U->nrSelFeat );
   for( i = 0; i < U->nrClass; i++ )
   {
     for( j = 0; j < U->C[i].numSampl; j++ )
     {
       for( k = 0; k < U->nrSelFeat; k++ )
         fprintf( lvq, formatStr, U->C[i].S[j*U->nrFeat+U->FSV[k].rank] );
       fprintf( lvq, "%s\n", U->C[i].name );
     }
   }
   fclose( lvq );
 }

 if( ! feat_description )
   return;

 strcpy( featNameFile, outDataDir );
 printf("Saving the feature names in file:\n\t\t%s", featNameFile );
 gets( name );
 if( name[0] != '\0' )
 {
   strcat( featNameFile, name );
   feats = fopen( featNameFile, f_open_text_w );
   if( feats == NULL )
     { printf("Cannot open %s! Exitus...\n", featNameFile ); exit(1); }
   if( verbose )
     printf("Generating feature description-File: %s\n", featNameFile );

   fprintf( feats, "%d\n", U->nrSelFeat );
   for( k = 0; k < U->nrSelFeat; k++ )
     fprintf( feats, "%s\n", feature_desc[U->FSV[k].rank] );
   fclose( feats );
 }
}


static void filter_selected_lvq()
{
 str80 raw, filtered;
 FILE *r, *f;
 str100 nameBuf;
 int featDim, k, i;
 FeatVector featBuf = NULL;

 if( U->nrSelFeat == 0 )
   { printf(" Select features first please!..." ); gets( buf ); return; }

 printf("Name of the file to be filtered? "); gets( raw );
 r = fopen( raw, f_open_text_r );
 if( r == NULL )
   { printf(" Cannot open %s...", raw ); gets( buf ); return; }

 strcpy( filtered, outDataDir );
 printf("Name of the output file with filtered features?\n\t\t%s", filtered );
 gets( buf );
 if( (strcmp(raw,buf) == 0) )
   { printf(" Sorry file names are the same!..." ); gets( buf ); return; }
 strcat( filtered, buf );
 f = fopen( filtered, f_open_text_w );
 if( f == NULL )
   { printf(" Cannot open %s...", raw ); gets( buf ); return; }

 if( verbose )
   printf("Filtering\n\t\t%s\nTO\n\t\t%s\n", raw, filtered );

 /* try to read the dimension */
 dataline( r, buf ); sscanf( buf, "%d", &featDim );
 if( featDim != U->nrFeat )
   { printf("Universe and file have different feature dimensions: %d != %d...",
	U->nrFeat, featDim  ); gets( buf ); fclose( r ); fclose( f ); return; }

 featBuf = (FeatVector) malloc(sizeof(FeatVector*) * U->nrFeat);CHKPTR(featBuf);

 set_save_precision();

 fprintf( f, "%d\n", U->nrSelFeat );
 while( !feof( r ) )
 {
   for( k = 0; k < U->nrFeat; k++ )
     fscanf( r, "%f", &(featBuf[k]) );
   fscanf( r, "%s", nameBuf );
   if( !feof( r ) )
   {
     for( i = 0; i < U->nrSelFeat; i++ )
       fprintf( f, formatStr, featBuf[ U->FSV[i].rank ] );
     fprintf( f, "%s\n", nameBuf );
   }
 }
 FREE( featBuf );
 fclose( r ); fclose( f );
}



#define GEN_DATA_FILE_LVQ	0
#define FILTER_LVQ		1

#define NUM_LVQ_ITEMS 2


static struct lvqMenuItems {
        int index;
        str80 name;
} lvqMenuItem[] = {
        { GEN_DATA_FILE_LVQ,	"Generate data file" },
        { FILTER_LVQ,		"Filter only selected features from a file" } };


static void execMenuItem()
{
 switch( menuItem )
 {
   case GEN_DATA_FILE_LVQ:	gen_lvq_data_file(); break;
   case FILTER_LVQ:		filter_selected_lvq(); break;
   default: fprintf(stderr,"What the heck is item %d? - exit...\n", menuItem );
 }
}

static char sl[] =
     "-----------------";



void lvqMenu()
{
 int i, d; char *str = NULL;

 do
 {
   printf("\n%s  LVQ_PAK DATA FORMAT  %s\n", sl, sl );
   for( i = 0; i < NUM_LVQ_ITEMS; i++ )
   {
     str = lvqMenuItem[i].name;
     printf("(%d)\t%s\n", i+1, str );
   }
   printf("Choice: ");
   gets(buf); done = FALSE;
   switch( buf[0] )
   {
     /* special cases */
     case '?': help( MENU_LVQ, buf ); break;
     case '\0': done = TRUE; break;
     default:
       sscanf(buf,"%d", &d);
       if( d >= 1 && d <= NUM_LVQ_ITEMS )
       {
         menuItem = d-1;
         execMenuItem();
       }
   }
 }
 while (! done);
}


/**********  H E L P  *****************/

void helpMENU_LVQ( cmd )
char *cmd;
{
 int item;

 sscanf( &(cmd[1]), "%d", &item );
 item--;        /* discount one since in the menu it is one more */
 switch( item )
 {
   case GEN_DATA_FILE_LVQ:	item_help("GEN_DATA_FILE_LVQ"); break;
   case FILTER_LVQ:		item_help("FILTER_LVQ"); break;
   default: ; break;
 }
}
