/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:25 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include "def.h"
#include "matrxdef.h"

static char buf[80];


/*----------------------------------------------------------
	Two alternative definitions for vectors and matrices
------------------------------------------------------------*/
/*  F I R S T  A L T E R N A T I V E */


/* allocate a float vector of length 'dim' */
float *vector( dim )
int dim;
{
 float *v = NULL;

 /* printf("Allocating %d bytes     \r", sizeof(float) * dim );
 fflush(stdout);        /**/
 v = (float *) malloc(sizeof(float) * dim );
 if( v == NULL )
 {
   printf("\nCould not allocate memory for %d-dimensional vector !!!\n", dim );
 }
 return(v);
}


/* allocate a double vector of length 'dim' */
double *dbl_vector( dim )
int dim;
{
 double *v = NULL;

 /* printf("Allocating %d bytes     \r", sizeof(double) * dim );
 fflush(stdout);        /**/
 v = (double *) malloc(sizeof(double) * dim );
 if( v == NULL )
 {
   printf("\nCould not allocate memory for %d-dimensional vector !!!\n", dim );
 }
 return(v);
}


/* allocate an integer vector of length 'dim' */
int *int_vector( dim )
int dim;
{
 int *v = NULL;

 /* printf("Allocating %d bytes     \r", sizeof(int) * dim );
 fflush(stdout);        /**/
 v = (int *) malloc(sizeof(int) * dim );
 if( v == NULL )
 {
   printf("\nCould not allocate memory for %d-dimensional vector !!!\n", dim );
 }
 return(v);
}


float **matrix_alloc( row, col )
int row, col;
{
 float **a = NULL;
 int i, j;

 a = (float**) malloc( row * sizeof(float*) );
 if( a == NULL )
 {
   printf("\nCould not allocate memory for %d-dimensional row vector !!!\n",
	row );
   return NULL;
 }
 for( i = 0; i < row; i++ )
 {
   a[i] = (float*) malloc( col * sizeof(float) );
   if( a[i] == NULL )
   {
     printf("\nCould not allocate %d-dimensional column vector at row %d!!!\n",
		col, row );
     for( j = 0; j < i; j++ )
       FREE( a[j] );
     FREE( a );
     return NULL;
   }
 }
 return a;
}


void matrix_free( M, row )
float ***M;
int row;
{
 int i;

 if( (*M) == NULL )
   return;

 for( i = 0; i < row; i++ )
   FREE( (*M)[i] );
 FREE( (*M) );
}


void matrix_print( M, row, col )
float **M;
int row, col;
{
 int i, j;

 if( M == NULL )
   return;

 for( i = 0; i < row; i++ )
 {
   for( j = 0; j < col; j++)
     printf(" %10.5f ", (float)M[i][j]);
   printf("\n");
 }
 printf("\n");
}


/* Set all element of a matrix to zero */
void matrix_zero( M, row, col )
float ***M;
int row, col;
{
 int i, j;

 if( (*M) == NULL )
   return;

 for( i = 0; i < row; i++ )
   for( j = 0; j < col; j++)
     (*M)[i][j] = 0.0;
}






/*------------------------------------------------------------*/

/*  S E C O N D  A L T E R N A T I V E */

/*
 *  Print Matrix
*/
void print_Matrix( M )
Matrix *M;
{ 
 int i, j;

 if( M->row == 0 && M->col == 0 )
 {
   printf("EMPTY MATRIX!!!\n");
   return;
 }
 for( i = 0; i < M->row; i++ )
 {
   for( j = 0; j < M->col; j++)
     printf(" %10.5f ", (float)M->Elem[i*M->col+j]);
   printf("\n");
 }
 printf("\n");
}


/*
 *  Allocate memory for the elements of the matrix
 *   the row and column values must be set
*/
void Matrix_Alloc( M )
Matrix *M;
{
 if( (M->row == 0) || (M->col == 0) )
 {
   fprintf(stderr,
     "Warning: trying allocate space for an invalid (%d x %d) matrix!",
		M->row, M->col );
   WAIT; return;
 } 
 if( M->Elem != NULL )
   fprintf(stderr,
	"Warning: trying allocate space for a non-empty matrix pointer!\n");
 M->Elem = (MatElem*) malloc(M->row * M->col * sizeof(MatElem));
 if( M->Elem == NULL )
 {
   fprintf(stderr,"No memory allocated for matrix with dimensions %d x %d",
                    M->row, M->col);
   WAIT; return;
 }
}


/*
 *  Free the memory for the elements of a matrix
*/
void Matrix_Free( M )
Matrix *M;
{ if( M->Elem != NULL ) {free(M->Elem); M->Elem = NULL; M->row = 0; M->col = 0;}
  /* else fprintf(stderr,"Warning: trying to free already free space !\n");/**/
}


/*
 *  Initializiation of a matrix: must be called before any use
*/
void init_Matrix( M )
Matrix *M;
{ M->row = 0; M->col = 0; M->Elem = NULL; }

/*
 * Return the unit matrix of dimension 'dim'
*/
void Unit_Matrix( M, dim )
Matrix *M;
int dim;
{
 int i, j;

 Matrix_Free( M );
 M->row = dim; M->col = dim;
 Matrix_Alloc( M );
 for( i = 0; i < M->row; i++ )
   for( j = 0; j < M->col; j++)
     if( i == j )
       M->Elem[i*M->col+j] = 1.0;
     else
       M->Elem[i*M->col+j] = 0.0;
}

/*
 * Copy A to B
*/
void Copy_Matrix( A, B )
Matrix *A, *B;
{
 int i, j;

 Matrix_Free( B );
 B->row = A->row; B->col = A->col;
 Matrix_Alloc( B );
 for( i = 0; i < B->row; i++ )
   for( j = 0; j < B->col; j++)
     B->Elem[i*B->col+j] = A->Elem[i*B->col+j];
}

/*
 *  Zero all elements of a matrix
*/
void Reset_Matrix( M )
Matrix *M;                         /*  M=0  */
{
 int i, j;

 for( i = 0; i < M->row; i++ )
   for( j = 0; j < M->col; j++)
     M->Elem[i*M->col+j] = 0.0;
}

/*
 *  Adding two matrices
*/
void Add_Matrix( A, B, C )
Matrix *A, *B, *C;                  /*  C=A+B  */
{
 int i, j;

 if( (A->row != B->row) || (A->col != B->col) )
 {
   fprintf( stderr, "Add_Matrix> Cannot add matrices !!!..."); gets(buf);
 }
 else
 {
   Matrix_Free( C );
   C->row = A->row; C->col = A->col;
   Matrix_Alloc( C );
   for( i = 0; i < C->row; i++ )
     for( j = 0; j < C->col; j++)
       C->Elem[i*C->col+j] = A->Elem[i*C->col+j] + B->Elem[i*C->col+j];
 }
}


/*
 *  Subtracting two matrices
*/
void Subtract_Matrix( A, B, C )
Matrix *A, *B, *C;                  /*  C=A-B  */
{
 int i, j;

 if( (A->row != B->row) || (A->col != B->col) )
 {
   fprintf( stderr, "Subtract_Matrix> Cannot subtract matrices !!!...");
   gets(buf);
 }
 else
 {
   Matrix_Free( C );
   C->row = A->row; C->col = A->col;
   Matrix_Alloc( C );
   for( i = 0; i < C->row; i++ )
     for( j = 0; j < C->col; j++)
       C->Elem[i*C->col+j] = A->Elem[i*C->col+j] - B->Elem[i*C->col+j];
 }
}


/*
 *  Multiplying a matrix with a scalar
*/
void ScalarMult_Matrix( Scalar, M )
MatElem Scalar;
Matrix *M;
{
 int i, j;

 for( i = 0; i < M->row; i++ )
   for( j = 0; j < M->col; j++)
     M->Elem[i*M->col+j] *= Scalar;
}


/*
 *  Multiply two matrices
*/
void Mult_Matrix( A, B, C )
Matrix *A, *B, *C;                  /*  C=A*B  */
{
 int i, j, k;

 if( A->col != B->row )
 {
   fprintf( stderr, "Mult_Matrix> Cannot multiply matrices");
   fprintf( stderr, " (%d,%d) x (%d,%d) ...",
	A->row, A->col, B->row, B->col ); gets(buf);
 }
 else
 {
   Matrix_Free( C );
   C->row = A->row; C->col = B->col;
   Matrix_Alloc( C );
   /* Multiply */
   for( i = 0; i < C->row; i++ )
   {
     for( j = 0; j < C->col; j++)
     {
       C->Elem[i*C->col+j] = 0.0;
       for( k = 0; k < A->col; k++ )
       {
         C->Elem[i*C->col+j] += A->Elem[i*A->col+k] * B->Elem[k*B->col+j];
       }
     }
   }
 }
}

/*
 *  Transpose Matrix
*/
void Transpose_Matrix( M, T )
Matrix *M, *T;
{
 int i, j;

 Matrix_Free( T );
 T->row = M->col; T->col = M->row;
 Matrix_Alloc( T );
 /* Transpose */
 for( i = 0; i < M->row; i++ )
 {
   for( j = 0; j < M->col; j++) {
     T->Elem[j*T->col+i] = M->Elem[i*M->col+j];
   }
 }
}


/*
 *  Splitting a square matrix in four parts
 *                  p               q
 *		(         |			)
 *  	p	(   M11   |   M12		)
 *		(         |			)
 *		(   ------|---------------------)
 *		(         |			)
 *		(         |			)
 *      q	(         |			)
 *		(   M21   |    M22		)
 *		(         |			)
 *		(         |			)
 *		(         |			)
 *		(         |			)
 *		(         |			)
*/
void Split_Matrix( M, M11, M12, M21, M22, p, q )
Matrix *M, *M11, *M12, *M21, *M22;
int p, q;
{
 int i, j;

 if( (M->row != M->col) || (M->row != (p+q)) )
 {
   fprintf( stderr, "Split_Matrix> problems !!!..."); gets(buf);
 }
 else
 {
   Matrix_Free( M11 ); M11->row = p; M11->col = p; Matrix_Alloc( M11 );
   Matrix_Free( M12 ); M12->row = p; M12->col = q; Matrix_Alloc( M12 );
   Matrix_Free( M21 ); M21->row = q; M21->col = p; Matrix_Alloc( M21 );
   Matrix_Free( M22 ); M22->row = q; M22->col = q; Matrix_Alloc( M22 );

   for( i = 0; i < M11->row; i++ )
     for( j = 0; j < M11->col; j++)
       M11->Elem[i*M11->col+j] = M->Elem[i*M->col+j];

   for( i = 0; i < M12->row; i++ )
     for( j = 0; j < M12->col; j++)
       M12->Elem[i*M12->col+j] = M->Elem[i*M->col+j+p];

   for( i = 0; i < M21->row; i++ )
     for( j = 0; j < M21->col; j++)
       M21->Elem[i*M21->col+j] = M->Elem[(i+p)*M->col+j];

   for( i = 0; i < M22->row; i++ )
     for( j = 0; j < M22->col; j++)
       M22->Elem[i*M22->col+j] = M->Elem[(i+p)*M->col+j+p];
 }
}


void set_Matrix();

/*
 *  Inversion of a square matrix.
 *  Calculates also the determinant of the matrix
*/
/*  Scale the matrix before inversion if its mean value is too far from 1
 *  invert then and rescale again
*/
/* #define SCALE_INVERSION		/**/
void Invert_Matrix( M, Det )
Matrix *M;
MatElem *Det;   /* Determinant */
{
 int k, j, i, n;
 MatElem D = 1.0;

#ifdef SCALE_INVERSION
 Matrix S;
 MatElem meanVal = 0.0, invMeanVal, tolerance = 1000.0;

 Copy_Matrix( M, &S );
 if( S.row != S.col)
     {fprintf(stderr,"Invert_Matrix> not square matrix !. Exit...\n"); exit(1);}

 for( i = 0; i < S.row; i++ )
   for( j = 0; j < S.col; j++)
     meanVal += S.Elem[i*S.col+j];

 meanVal /= (MatElem)(S.row*S.col);
 invMeanVal = 1.0 / meanVal;

 if( meanVal > tolerance || invMeanVal > tolerance )
   ScalarMult_Matrix( invMeanVal, M );

 /* printf("meanVal=%lf  invMeanVal=%lf", meanVal, invMeanVal ); NL;
 print_Matrix( M ); DBG;	/**/
#endif



 if( M->row != M->col)
     {fprintf(stderr,"Trying to invert non-square matrix ! Exit...\n");exit(1);}

 n = M->row;
 for(k=0; k<n; k++)
 {
   for(i=0; i<n; i++)
     for(j=0; j<n; j++)
       if( i!=k && j!=k )
         M->Elem[i*n+j] -= M->Elem[k*n+j]*M->Elem[i*n+k]/M->Elem[k*n+k];

   for(i=0; i<n; i++)
     for(j=0; j<n; j++)
       if( i==k && i!=j )
         M->Elem[i*n+j]= -M->Elem[i*n+j]/M->Elem[k*n+k];

   for(i=0; i<n; i++)
     for(j=0; j<n; j++)
       if( j==k && i!=j )
         M->Elem[i*n+j]= M->Elem[i*n+j]/M->Elem[k*n+k];
   D *= M->Elem[k*n+k];
   M->Elem[k*n+k]= 1.0/M->Elem[k*n+k];
 }

#ifdef SCALE_INVERSION
 /*----------------------------------------*/
 /*    R U L E S                           */
 /*      det( c * A ) = c^n * det(A)       */
 /*      det(inverse(A)) = 1 / det(A)      */
 /*----------------------------------------*/

 if( meanVal > tolerance || invMeanVal > tolerance )
 {
   ScalarMult_Matrix( invMeanVal, M );
   for(i=0; i<n; i++)
     D *= meanVal;
 }
#endif
 *Det = D; 
}

/*
 * Calculate the trace of a square matrix, i.e. the sum of all diagonal
 * elements of the matrix
*/
MatElem trace( M )
Matrix *M;
{
 MatElem tr = 0.0;
 int j;

 if( M->row != M->col)
     {fprintf(stderr,"trace> not square matrix !. Exit...\n"); exit(1);}

 for( j = 0; j < M->col; j++)
   tr += M->Elem[j*M->col+j];

 return( tr );
}


/*
 *  Read Matrix from stdin
*/
void set_Matrix( M )
Matrix *M;
{
 int i, j;
 char buf[81];

 printf("Rows= "); gets(buf); sscanf( buf, "%d", &(M->row) ); 
 printf("Colums= "); gets(buf); sscanf( buf, "%d", &(M->col) );
 Matrix_Alloc( M );

 for( i = 0; i < M->row; i++ )
   for( j = 0; j < M->col; j++ )
   {
     printf("M[%d][%d]= ", i+1, j+1);
     gets(buf); sscanf( buf, "%lf",  &(M->Elem[i*M->col+j]) );
   }
}





/* #define MATRIX_TEST	/**/
#ifdef MATRIX_TEST
/**************************************   TEST */
static Matrix T1, T2, T3, T4, T5, T6;

void main()
{
 MatElem Det;
 int p,q;

 init_Matrix( &T1 ); init_Matrix( &T2 ); init_Matrix( &T3 );
 init_Matrix( &T4 ); init_Matrix( &T5 ); init_Matrix( &T6 );

/*
 set_Matrix( &T1 );
*/
 T1.row = 2; T1.col = 2;
 Matrix_Alloc( &T1 );
 T1.Elem[0]=0.03; T1.Elem[1]= 0.04; T1.Elem[2]= 0.05; T1.Elem[3]= 0.08; /**/
 /* T1.Elem[0]=30.0; T1.Elem[1]= 40.0; T1.Elem[2]= 50.0; T1.Elem[3]= 80.0;
 print_Matrix( &T1 );
/*
 set_Matrix( &T2 );
 print_Matrix( &T2 );
*/

 Invert_Matrix( &T1, &Det );
 printf("DET=%lf\n", Det );

/*
 print_Matrix( &T2 );
 Mult_Matrix( &T2, &T1, &T3 );
 print_Matrix( &T3 );
*/


 Matrix_Free( &T1 );
 Matrix_Free( &T2 );
 Matrix_Free( &T3 );
 Matrix_Free( &T4 );
 Matrix_Free( &T5 );
 Matrix_Free( &T6 );
}
/*************************************/
#endif
