/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:30 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/* Uncomment if you have this include file */
/* #include <values.h>		/**/
#include "def.h"
#include "matrxdef.h"
#include "clasdef.h"

extern universe *U;
extern boolean verbose_reason;
extern float **matrix_alloc();
extern float *vector();
extern boolean rejection_opt;
extern str100 outDataDir;

extern int nrSamplesTrain;
extern int *train_vec, *train_vecClass, *train_vecSmp;

extern FILE *cpf;               /* Classification protocol file */
extern boolean have_label;      /* Only deliver the classified class for
        pre-classified data */


/* The weight matrices
   (in case of momentum learning we need to buffer old differences)
	WI_delta, WH_delta)
   WI: Hidden to Output weights
   number of rows: c = number of classes
   number of columns: h number of neurons in the hidden layer

   WH: Input to hidden weights
   number of rows: c = number of neurons in the hidden layer
   number of columns: h number of features (+1 for augmented feature vector)
*/



static int c, d, d1, H = EMPTY, H1;
static float **WI = NULL, **WH = NULL, **WI_delta = NULL, **WH_delta = NULL;
static float **WI_err = NULL, **WH_err = NULL;
/* Storing the output of the hidden and output layer */
static float *gout = NULL, *ghidden = NULL;
static float *decision_func = NULL, *x = NULL;

static boolean augmented_hidden = TRUE; /* use an additional bias term in
		the hidden layer also */
static int aug_bias = 1;	/* if augmented_hidden==TRUE then 1 */

/* Load the weights from file instead of learning them */
#define MAX_PRECISION 30
static boolean load_weights = FALSE;
static int save_precision = 5;
static str80 wfname, name, buf;


#ifdef UNIX
#ifdef MAXINT
#define MAX_ITER	MAXINT
#endif
#ifndef MAXINT
#define MAX_ITER	100000000
#endif
#endif
#ifdef NON_UNIX		/* Only for 16-bit integer */
#define MAX_ITER	10000
#endif

#define MAX_ETA		(1000.1)
#define MAX_MOMPARAM	(1.0)	/* moment should be between 0.0 and 1.0 */
#define MAX_EPS		(1000000.0)
#define MAX_WEIGHT	(10000.0)
#define MAX_NEURON	10000
#define MAX_ERR		(100.0)	/* Stop when this error is reached */

static float eta = 0.2;		/* Learning Rate */
static float eps = 0.0000001;	/* Similarity between two successive steps */
static float err_limit = 0.0001;	/* Stop when this error is reached */
static int max_iter = 1000;	/* Maximum number of iterations */
static boolean stoch_learn =FALSE;	/* Stochastic (per pattern) learning or batch */
static boolean momentum = FALSE;  /* use momentum learning */
static float momparam = 0.9;	/* Momentum parameter */


/*
 * Do not use K&R C here, to keep it compatible in the definition of 'actfuncs'
*/
static float sigmoid( float z )
{
  /* Have to check the argument in order to avoid floating exception on DEC */

  if( z < -50.0 ) return( 0.0 );
  if( z > 50.0 ) return( 1.0 );

  return( 1.0 / ( 1.0 + (float)(exp((double)(-z)))) );
}

static float d_sigmoid( float g )
{
 return( g * (1.0 - g) );
}


static float tanh_act( float z )
{
  /* Have to check the argument in order to avoid floating exception on DEC */

  if( z < -50.0 ) return( -1.0 );
  if( z > 50.0 ) return( 1.0 );

  return( (float)tanh((double)z) );
}


static float d_tanh_act( float g )
{
 return( 1.0 - g * g );
}


/* TYPES OF ACTIVATION FUNCTIONS */
#define SIGMOID			0
#define TAN_H			1

static struct actfuncs {
	int index;
	str80 name;
        float (*act_func)(float);		/* The Activation Function */
        float (*deriv_act_func)(float);	/* ... and its derivative */
} actfunc[] = {
	{ SIGMOID,	"SIGMOID",		sigmoid,	d_sigmoid },
	{ TAN_H,	"HYPERBOLIC TANGENT",	tanh_act,	d_tanh_act }
};

static int activation = SIGMOID;


void get_backprop_param()
{
 printf("Number of input variables: %d+1=%d\n", d, d1 );
 printf("Number of output neurons: %d\n", c );
 printf("Number of neurons in the hidden layer (suggestion=%d)? ", H );
 get_d_range( &H, 1, MAX_NEURON, LEFT_CLOSED__RIGHT_CLOSED );

 printf("Use also a bias term in the hidden layer");
 get_boolean( augmented_hidden, &augmented_hidden );
 if( augmented_hidden )
 {
   H1 = H + 1;
   aug_bias = 1;
 }
 else
 {
   H1 = H;
   aug_bias = 0;
 }

 activation++;
 printf("Activation function SIGMOID(1) or TANH(2)? %d\b",
	activation);
 get_d_range( &activation, SIGMOID+1, TAN_H+1,
	LEFT_CLOSED__RIGHT_CLOSED );
 activation--;



 printf("Learning rate of Gradient Descent (actual=%.7f)? ", eta );
 get_f_range( &eta, 0.0, MAX_ETA, LEFT_OPEN__RIGHT_OPEN );

 printf("Maximum number of learning steps (actual=%d)? ", max_iter );
 get_d_range( &max_iter, 1, MAX_ITER-1, LEFT_CLOSED__RIGHT_CLOSED );

 printf("Similarity of weights between two successive steps (actual=%.7f)? ",
	eps );
 get_f_range( &eps, 0.0, MAX_EPS, LEFT_CLOSED__RIGHT_OPEN );

 printf("Stop training when this mean squared error is reached (actual=%.7f)? ",
		 err_limit );
 get_f_range( &err_limit, 0.0, MAX_ERR, LEFT_CLOSED__RIGHT_OPEN );

 printf("Stochastic learning (on a per pattern basis)");
 get_boolean( stoch_learn, &stoch_learn );
 if( ! stoch_learn )
 {
   printf("Use momentum parameter");
   get_boolean( momentum, &momentum );
   if( momentum )
   {
     printf("Momentum parameter (actual=%.7f)? ", momparam );
     get_f_range( &momparam, 0.0, MAX_MOMPARAM, LEFT_CLOSED__RIGHT_CLOSED );
   }
 }
}


void show_MLP( f )
FILE *f;
{
 fprintf(f,"\tTRAINING ALGORITHM: Error backpropagation\n");
 if( H == EMPTY )
 {
   fprintf(f,"... Not yet initialized ...\n");
   return; 
 }
 fprintf(f,"\tHIDDEN LAYER:\n");
 fprintf(f,"\t\tNumber of neurons: %d", H );
 if( augmented_hidden )
 {
   fprintf(f,"+1\n\t\tAugmented vector\n");
   fprintf(f,"\t\t  (i.e. one extra neuron with constant value 1)\n");
 }
 else
   NL;
 fprintf(f,"\tACTIVATION FUNCTION:\t%s\n", actfunc[activation].name );


 fprintf(f,"\tWEIGHT UPDATING: ");
 if( stoch_learn )
   fprintf(f,"stochastic (on a per pattern basis)\n");
 else
   fprintf(f,"batch (present all patterns, then update)\n");
 if( ! stoch_learn && momentum )
   fprintf(f,"\t\tBATCH LEARNING: Momentum parameter: %.5f\n", momparam );
 fprintf(f,"\tLEARNING RATE: %.5f\n", eta );
 fprintf(f,"\tSTOPPING RULE: ");
 fprintf(f,"after %d iterations\n", max_iter );
 fprintf(f,"\t\tor if two successive weight sets are closer than %f\n",
        eps );
 fprintf(f,"\t\tor if learning error reaches limit of %f\n", err_limit );
}


void initMLP()
{
 if( U->nrSelFeat == 0 )
   { printf(" Select features first please!..." ); gets( buf ); return; }

 free_induced();

 c = U->nrClass;
 d = U->nrSelFeat;
 d1 = U->nrSelFeat + 1;
 /* Suggestion for the number of hidden neurons: (2 * number_of_features + 1)
	(Kolmogorov Mapping Neural Network Existence Theorem) */
 if( H == EMPTY )
 {
   H = 2 * d + 1; H1 = H;
   if( augmented_hidden )
   {
     aug_bias = 1;
     H1 = H + 1; 
   }
   else
     aug_bias = 0;
 }
 get_backprop_param();
}


/* Set the initial weights to random values */
static void random_init_weights()
{
 int i, h, j;

 for( i = 0; i < c; i++ )
 {
   random_unit_vector( WI[i], H1 );
   /* printf("Random init WI[%3d]=", i ); showFV( H1, WI[i] ); /**/
 }
 for( h = 0; h < H; h++ )
 {
   random_unit_vector( WH[h], d1 );
   /* printf("Random init WH[%3d]=", h ); showFV( d1, WH[h] ); DBG;   /**/
 }
 /* matrix_print( WI, c, H1 ); matrix_print( WH, H, d1 ); DBG; /**/
}


static void prepareMLP()
{
 int i, h, j;

 free_induced();

 x = vector( d1 ); x[0] = 1.0;	/* Augmented sample vector */

 WI = matrix_alloc( c, H1 );
 WH = matrix_alloc( H, d1 );
 if( ! stoch_learn )
 {
   WI_err = matrix_alloc( c, H1 );
   WH_err = matrix_alloc( H, d1 );
   if( momentum )
   {
     WI_delta = matrix_alloc( c, H1 );
     matrix_zero( &WI_delta, c, H1 );
     WH_delta = matrix_alloc( H, d1 );
     matrix_zero( &WH_delta, H, d1 );
   }
 }

 decision_func = vector( c );
 gout = vector( c );
 ghidden = vector( H1 );

 if( augmented_hidden )
   ghidden[0] = 1.0;
}


/*
 *  Free all the structures that were dynamically allocated to store
 *  the classifier, for instance matrices of prototypes
*/
void free_inducedMLP()
{
 FREE( x );
 matrix_free( &WI, c ); matrix_free( &WH, H );
 if( ! stoch_learn )
 {
   matrix_free( &WI_err, c ); matrix_free( &WH_err, H );
   if( momentum )
   {
     matrix_free( &WI_delta, c ); matrix_free( &WH_delta, H );
   }
 }
 FREE( decision_func );
 FREE( gout ); FREE( ghidden );
}




static void feedforward( verbose )
boolean verbose;
{
 int i, j, h;
 float Sh, Si, act;

 if( verbose )
 {
   printf("The augmented sample is: "); showFV( d1, x );	/**/
   printf("\n------> FEEDFORWARD...\n--- HIDDEN LAYER:\n");	/**/
 }

 /* INPUT ---> HIDDEN */
 for( h = 0; h < H; h++ )
 {
   Sh = 0.0;
   for( j = 0; j < d1; j++ )
   {
     Sh += WH[h][j] * x[j];
     /* printf("WH[%d][%d]=%f x[%d]=%f\n", h, j, WH[h][j], j, x[j] ); /**/
   }
   act = actfunc[ activation ].act_func( Sh );
   ghidden[h+aug_bias] = act;
   if( verbose )
     printf("ghidden[%d] = g(%f) = %f\n", h+aug_bias, Sh, act );

 }
 if( verbose && augmented_hidden )
     printf("ghidden[%d]=%f\n", 0, ghidden[0] ); /**/

 if( verbose )
   printf("--- OUTPUT LAYER:\n");       /**/
 /* HIDDEN ---> OUTPUT */
 for( i = 0; i < c; i++ )
 {
   Si = 0.0;
   for( h = 0; h < H1; h++ ) 
   {
     Si += WI[i][h] * ghidden[ h ];
     /* printf("WI[%d][%d]=%f ghidden[%d]=%f\n",
	i, h, WI[i][h], h, ghidden[h] ); /**/
   }
   act = actfunc[ activation ].act_func( Si );
   gout[ i ] = act;
   if( verbose )
     printf("gout[%d] = g(%f) = %f\n", i, Si, act );
   decision_func[i] = gout[ i ];
 } 
}



static boolean error_backpropagation( nsf, FSV, verbose )
int nsf;
FeatSelectVector FSV;
boolean verbose;
{
 int d1, i, j, k, p, h, indx, c, clss, nr_weights, iter = 0;
 boolean random_order = TRUE, done = FALSE;
 float batch_error, mean_squared_error;
 float nabla_ih, nabla_hj, delta_h;
 float *diff_pat = NULL, *delta_i = NULL, *gderiv = NULL, gderivi;
 float delta_WI_ih, delta_WH_hj, delta_W, aux;
 float weight_diff, normfac1;

 random_init_weights();

 c = U->nrClass;
 d1 = nsf + 1;
 nr_weights = d1 * H + H1 * c;

 diff_pat = vector( c );
 delta_i = vector( c );
 gderiv =  vector( H );
 x = vector( d1 ); x[0] = 1.0;

 if( verbose )
 {
   fprintf(stdout,"Learning weights by error backpropagation...\r");
   fflush(stdout);
 }
 normfac1 = (float)(nr_weights*nrSamplesTrain);
 while( ! done )
 {
   weight_diff = 0.0; mean_squared_error = 0.0;
   if( stoch_learn )
   {
     random_order = TRUE;
     gen_train_vec( random_order );
     for( p = 0; p < nrSamplesTrain; p++ )
     {
       indx = train_vec[p];
       clss = train_vecClass[ indx ];
       k = train_vecSmp[ indx ];

       provideFV( &(x[1]), clss, k, nsf, FSV );
       /* printf("\nSample # %d which is %d of class %d(%s)\n",
		p+1, k+1, clss+1, U->C[clss].name );
       printf("The augmented sample is: "); showFV( d1, x );	/**/

       /* Forward the sample through the net */
       feedforward( FALSE );

       /* Calculate the differences between target and calculated :
	target is 0 for i != clss and 1 for i == clss */
       for( i = 0; i < c; i++ )
         diff_pat[i] = - decision_func[i];
       diff_pat[clss] += 1.0;	/* The real class */
       for( i = 0; i < c; i++ )
       {
         mean_squared_error += (diff_pat[i] * diff_pat[i]);
         /* printf("diff_pat[%d] = %f\n", i, diff_pat[i] );  DBG;  /**/
       }

       /* HIDDEN to OUTPUT: DELTA RULE */

       for( i = 0; i < c; i++ )
       {
         gderivi = actfunc[ activation ].deriv_act_func( gout[i] );
         delta_i[i] = diff_pat[i] * gderivi;
         for( h = 0; h < H1; h++ )
         {
           nabla_ih = 2.0 * delta_i[i] * ghidden[h];
           delta_W = eta * nabla_ih;
           WI[i][h] += delta_W;

           /* printf("nabla[%d][%d]=2.0*delta[%d]*ghidden[%d]=2*%.20f*%f= %f\n",
		i, h, i, h, delta_i[i], ghidden[h], nabla_ih );
           printf("WI[%d][%d]:  %.20f--->%.20f\n",
		i, h, WI[i][h]-delta_W, WI[i][h] );WAIT;	/**/
           /* accumulate the difference */
           weight_diff += delta_W * delta_W;
         }
       }

       /* INPUT to HIDDEN: GENERALIZED DELTA RULE */
       for( h = 0; h < H; h++ )
       {
         gderiv[h] = actfunc[ activation ].deriv_act_func(ghidden[h+aug_bias]);

         /* aux stores   SUM i=1 to c ( delta_i * W_ih )  */
         aux = 0.0;
         for( i = 0; i < c; i++ )
           aux += delta_i[i] * WI[i][h+aug_bias];
         /* Calcualte deltas for the hidden layer */
         delta_h = gderiv[h] * aux;
         /* printf("delta_h[%d]=gderiv[%d]*aux= %f*%f= %f\n",
		h, h, gderiv[h], aux, delta_h );	/**/
         for( j = 0; j < d1; j++ )
         {
           nabla_hj = 2.0 * delta_h * x[j];
           delta_W = eta * nabla_hj;
           WH[h][j] += delta_W;

           /* printf("nabla[%d][%d]=2.0*delta_h[%d] * x[%d] =2*%.20f*%f= %f\n",
		h, j, h, j, delta_h, x[j], nabla_hj );
           printf("WH[%d][%d]:  %.20f--->%.20f\n",
		h, j, WH[h][j]-delta_W, WH[h][j] );WAIT;	/**/
           /* accumulate the difference */
           weight_diff += delta_W * delta_W;
           /* printf("weight_diff=%.20f\n", weight_diff );	/**/
         }
       }
     }	/* next pattern */

     /* Defining the MSE as the mean over all classes and patterns */
     mean_squared_error /= (float)( nrSamplesTrain * c );
     /* printf("Normalized weight_diff=%f/(%d*%d)=%.20f\n",weight_diff,
	nr_weights, nrSamplesTrain, sqrt(weight_diff)/normfac1 );	/**/
     weight_diff = (float)sqrt((double)weight_diff) / normfac1;
   }

   else	/* BATCH LEARNING */
   {
     random_order = FALSE;
     gen_train_vec( random_order );
     matrix_zero( &WI_err, c, H1 );
     matrix_zero( &WH_err, H, d1 );

     for( p = 0; p < nrSamplesTrain; p++ )
     {
       indx = train_vec[p];
       clss = train_vecClass[ indx ];
       k = train_vecSmp[ indx ];

       provideFV( &(x[1]), clss, k, nsf, FSV );
       /* printf("\nSample # %d which is %d of class %d(%s)\n",
		p+1, k+1, clss+1, U->C[clss].name );
       /* printf("The augmented sample is: "); showFV( d1, x );	/**/

       /* Forward the sample through the net */
       feedforward( FALSE );

       /* Calculate the differences between target and calculated :
		target is 0 for i != clss and 1 for i == clss */
       for( i = 0; i < c; i++ )
         diff_pat[i] = - decision_func[i];
       diff_pat[clss] += 1.0;   /* The real class */
       for( i = 0; i < c; i++ )
       {
         mean_squared_error += (diff_pat[i] * diff_pat[i]);
         /* printf("diff_pat[%d] = %f\n", i, diff_pat[i] );  DBG;  /**/
       }

       /* Accumulate the error for each weight */
       /* HIDDEN to OUTPUT */
       for( i = 0; i < c; i++ )
       {
         gderivi = actfunc[ activation ].deriv_act_func( gout[i] );
         delta_i[i] = diff_pat[i] * gderivi;
         for( h = 0; h < H1; h++ )
         {
           nabla_ih = delta_i[i] * ghidden[h];
           WI_err[i][h] += nabla_ih;

           /* printf("nabla[%d][%d]=2.0*delta[%d]*ghidden[%d]=2*%.20f*%f= %f\n",
		i, h, i, h, delta_i[i], ghidden[h], nabla_ih );
           printf("WI_err[%d][%d]:  %f--->%f\n",
		i, h, WI_err[i][h] - nabla_ih, WI_err[i][h]);     WAIT; /**/
         }
       }

       /* INPUT to HIDDEN */
       for( h = 0; h < H; h++ )
       {
         gderiv[h] = actfunc[ activation ].deriv_act_func(ghidden[h+aug_bias]);

         /* printf("gderiv[%d]=%.20f\n", h, gderiv[h] );	/**/
         /* aux stores   SUM i=1 to c ( delta_i * W_ih )  */
         aux = 0.0;
         for( i = 0; i < c; i++ )
           aux += delta_i[i] * WI[i][h+aug_bias];
         /* Calculate deltas for the hidden layer */
         delta_h = gderiv[h] * aux;
         for( j = 0; j < d1; j++ )
         {
           nabla_hj = delta_h * x[j];
           WH_err[h][j] += nabla_hj;

           /* printf("nabla[%d][%d]=2.0*delta_h[%d] * x[%d] =2*%.20f*%f= %f\n",
		h, j, h, j, delta_h, x[j], nabla_hj );
           printf("WH_err[%d][%d]:  %.20f--->%.20f\n",
		h, j, WH_err[h][j] - nabla_hj, WH_err[h][j]); WAIT;     /**/
         }
       }
     }	/* done for all samples, now update the weights */

     /* WEIGHT UPDATING */

     /* HIDDEN to OUTPUT: DELTA RULE */
     for( i = 0; i < c; i++ )
     {
       for( h = 0; h < H1; h++ )
       {
         delta_WI_ih = 2.0 * eta * WI_err[i][h] / (float)nrSamplesTrain;
         weight_diff += delta_WI_ih * delta_WI_ih;
         if( momentum )
         {
           WI[i][h] += delta_WI_ih + momparam * WI_delta[i][h];
           WI_delta[i][h] = delta_WI_ih;
           /* printf("WI[%d][%d]:  %f--->%f\n", i, h,
		WI[i][h]+delta_WI_ih+momparam*WI_delta[i][h], WI[i][h] );/**/
         }
         else
         {
           WI[i][h] += delta_WI_ih;
           /* printf("WI[%d][%d]:  %f--->%f\n",
		i, h, WI[i][h]-delta_WI_ih, WI[i][h] );/**/
         }
       }
     }

     /* INPUT to HIDDEN: GENERALIZED DELTA RULE */
     for( h = 0; h < H; h++ )
     {
       for( j = 0; j < d1; j++ )
       {
         delta_WH_hj = 2.0 * eta * WH_err[h][j] / (float)nrSamplesTrain;
         weight_diff += delta_WH_hj * delta_WH_hj;
         if( momentum )
         {
           WH[h][j] += delta_WH_hj + momparam * WH_delta[h][j];
           WH_delta[h][j] = delta_WH_hj; 
           /* printf("WH[%d][%d]:  %f--->%f\n", h, j,
		WH[h][j]-(delta_WH_hj+momparam*WH_delta[h][j]),WH[h][j]);/**/
         }
         else
         {
           WH[h][j] += delta_WH_hj;
           /* printf("WH[%d][%d]:  %f--->%f\n", h, j,
		WH[h][j]-delta_WH_hj, WH[h][j]);/**/
         }
       }
     }
     weight_diff = (float)sqrt((double)weight_diff / (double)nr_weights );
     mean_squared_error /= (float)( nrSamplesTrain * c );
   } /* end BATCH */

   /* WAIT;	/**/

   iter++;
   done = iter >= max_iter || weight_diff <= eps ||
		mean_squared_error <= err_limit;

   if( verbose )
   {
     printf("Iteration = %d  Weight similarity = %f", iter, weight_diff );
     printf(" Mean squared error = %.8f \r", mean_squared_error ); /**/
     fflush(stdout); /**/
     if( iter >= max_iter )
       printf("\nEnd of learning: Reaching maximum of %d iterations\n", iter );
     if( weight_diff < eps )
     {
       printf("\nEnd of learning: Reaching similarity limit of %f\n", eps );
       printf("Could have been caused by network paralysis,\n");
       printf("especially with few iterations."); WAIT;
     }
     if( mean_squared_error < err_limit )
       printf("\nEnd of learning: Reaching error limit of %f\n", err_limit );
   }
 }
 FREE( diff_pat ); FREE( gderiv );
 return( TRUE );
}


static void save_weights_to_file()
{
 FILE *f = NULL;
 char formatStr[11];
 int i, j;

 if( WI == NULL || WH == NULL )
 {
   printf("Cannot save empty weight matrix ..."); gets( buf ); return;
 }

 strcpy( wfname, outDataDir );
 printf("Saving weights to: %s", wfname );
 gets( name );
 if( name[0] == '\0' )
 {
   printf("\tNothing saved...\n");
   return;
 }
 
 strcat( wfname, name );
 f = fopen( wfname, f_open_text_w );
 if( f == NULL )
     { printf("Cannot open %s for writing!...\n", wfname ); return; }

 printf("Save weights with precision (actual=%d)? ", save_precision );
 get_d_range( &save_precision, 0, MAX_PRECISION, LEFT_CLOSED__RIGHT_CLOSED );
 sprintf( formatStr, "%%.%df\n", save_precision );
 /* printf("formatStr=>>>%s<<<", formatStr ); WAIT;	/**/

 time_string( buf );
 fprintf( f, "# Weight file saved on %s", buf );
 fprintf( f, "#\n# Dimensions: INPUT --- HIDDEN --- OUTPUT\n%d %d %d\n",
	d1, H1, c );
 fprintf( f, "#\n# bias term in the hidden layer: 1=Yes  0=No\n%d\n",
	(int)augmented_hidden );

 for( i = 0; i < c; i++ )
   for( j = 0; j < H1; j++)
     fprintf( f, formatStr, (float)WI[i][j] );

 for( i = 0; i < H; i++ )
   for( j = 0; j < d1; j++)
     fprintf( f, formatStr, (float)WH[i][j] );
 fclose( f );
}


static boolean load_weights_from_file()
{
 FILE *f = NULL;
 int i, j, d1_tmp, H_tmp, H1_tmp, c_tmp, aug_tmp;
 float **WI_tmp = NULL, **WH_tmp = NULL;
 float w;

 strcpy( wfname, outDataDir );
 printf("Loadind weights from: %s", wfname );
 gets( name );
 if( name[0] == '\0' )
 {
   printf("\tNothing loaded...\n");
   return( FALSE );
 }
 
 strcat( wfname, name );
 f = fopen( wfname, f_open_text_r );
 if( f == NULL )
   {printf("Cannot open %s! nothing done", wfname );WAIT; return(FALSE);}

 dataline( f, buf );
 /* fprintf(stderr,"buf=>>>%s<<<\n", buf );	/**/
 sscanf( buf, "%d %d %d", &d1_tmp, &H1_tmp, &c_tmp );
 /* check the plausibility and consistency of the dimensions */
 if( d1_tmp != U->nrSelFeat+1 )
 {
   fprintf(stderr, "Inconsistent number of input elements:\n\t");
   fprintf(stderr, "Number of selected features + 1 = %d, but loaded %d\n",
	U->nrSelFeat+1, d1_tmp );
   fclose( f ); WAIT; return( FALSE );
 }

 if( c_tmp != U->nrClass )
 {
   fprintf(stderr, "Inconsistent number of output elements:\n\t");
   fprintf(stderr, "Number of selected classes = %d, but loaded %d\n",
	U->nrClass, c_tmp );
   fclose( f ); WAIT; return( FALSE );
 }

 if( H1_tmp <=0 || H1_tmp > MAX_NEURON ) 
 {
   fprintf(stderr, "Implausible number of hidden elements: %d\n", H1 );
   fclose( f ); WAIT; return( FALSE );
 }

 dataline( f, buf );
 sscanf( buf, "%d", &aug_tmp );

 if( aug_tmp != 0 && aug_tmp != 1 )
 {
   fprintf(stderr, "Implausible variable for hidden bias: %d\n", aug_tmp );
   fclose( f ); WAIT; return( FALSE );
 }
 H_tmp = H1_tmp - aug_tmp;

 /* Allocate the matrices */
 WI_tmp = matrix_alloc( c_tmp, H1_tmp );
 WH_tmp = matrix_alloc( H_tmp, d1_tmp );

 /* Now load the weights */
 for( i = 0; i < c_tmp; i++ )
   for( j = 0; j < H1_tmp; j++)
   {
     fscanf( f, "%f", &w );
     if( ! feof(f) )
     {
       WI_tmp[i][j] = w;
     }
     else
     {
       fprintf(stderr, "Found unexpected end of file in '%s'\n", wfname );
       fclose( f ); matrix_free( &WI_tmp, c_tmp );
       matrix_free( &WH_tmp, H_tmp ); WAIT; return( FALSE );
     }
   }

 for( i = 0; i < H_tmp; i++ )
   for( j = 0; j < d1_tmp; j++)
   {
     fscanf( f, "%f", &w );
     if( ! feof(f) )
     {
       WH_tmp[i][j] = w;
     }
     else
     {
       fprintf(stderr, "Found unexpected end of file in '%s'\n", wfname );
       fclose( f ); matrix_free( &WI_tmp, c_tmp );
       matrix_free( &WH_tmp, H_tmp ); WAIT; return( FALSE );
     }
   }
 fclose( f );

 /* Now copy the variables */
 d1 = d1_tmp; H = H_tmp; H1 = H1_tmp; c = c_tmp;
 prepareMLP();
 matrix_free( &WI, c ); matrix_free( &WH, H );
 WI = WI_tmp; WH = WH_tmp;
 printf("... Loaded successfully!\n");
 return( TRUE );
}


int induceMLP( nsf, FSV, iact )
int nsf;
FeatSelectVector FSV;
boolean iact;
{
 boolean ok, interact, verbose_induction = FALSE;

 interact = iact;
 if( interact )
 {
   printf("Do not learn weights, do load instead");
   get_boolean( load_weights, &load_weights );
   if( load_weights )
   {
     ok = load_weights_from_file();
     if( ! ok )
       return( FALSE );
     return( TRUE );
   }
 }

 if( H == EMPTY )
 {
   fprintf(stderr,"Specify parameters of MLP first!"); WAIT; NL;
   return( FALSE );
 }

 prepareMLP();
 if( interact )
 {
   printf("\n --- Inducing a %d-%d-%d Two-layer Perceptron ---\n\n",
	nsf+1, H1, c );
   printf("Verbose learning of decision structure?");
   get_boolean( verbose_induction, &verbose_induction );
 }

 ok = error_backpropagation( nsf, FSV, verbose_induction );

 if( ! ok )
 {
   fprintf(stderr, "Unable to do error backpropagation. Try again"); WAIT;
   return( FALSE );
 }
 else
 {
   if( verbose_induction )
   {
     printf("\nLEARNED WEIGHT MATRICES:\n\tInput ---> Hidden\n");
     matrix_print( WH, H, nsf+1 );
     printf("\tHidden ---> Output\n");
     matrix_print( WI, c, H1 );
   }
   if( interact )
     save_weights_to_file();
 }
 verbose_induction = FALSE;
 return( TRUE );
}



/*
 *  Classify a single sample 'smp' of length 'nsf'
 *  The name of the expected class is stored in 'name'
 *  'FSV' is not needed here, but for compatibility reasons passed
*/
int class1testMLP( nsf, FSV, smp, name )
int nsf;
FeatSelectVector FSV;
FeatVector smp;
char *name;
{
 int d1, i, j, resulting_class, best_class;
 float maxfunc;

 d1 = nsf + 1;
 copy_FV( smp, &(x[1]), nsf );
 if( verbose_reason )
 {
   fprintf(cpf, "\n\n ---- DECISION FUNCTIONS FOR UNKNOWN SAMPLE:\n");
   fshowFV( cpf, nsf, smp );
 }

 feedforward( FALSE );

 /* decide the best class */
 best_class = 0; maxfunc = -INFINITY;
 for( i = 0; i < c; i++ )
   if( decision_func[i] > maxfunc )
   {
     maxfunc = decision_func[i];
     best_class = i;
   }

 if( verbose_reason )
 {
   for( i = 0; i < c; i++ )
     fprintf( cpf, "   * Decision function of class %d(%s) is %f\n",
	i+1, U->C[i].name, decision_func[i] );

   fprintf( cpf, "   * Best class was %d(%s)\n",
		best_class+1, U->C[best_class].name );
 }

 if( ! have_label )
   return( best_class );

 /* 
    First case: correct classification:
    The name of the best class corresponds to the expected class
    return the value 'CORRECT_CLASS'
 */
 if( (strcmp(U->C[best_class].name,name)==0) )
   resulting_class = CORRECT_CLASS;
 else
 {
   /* 
      Second case: wrong classification:
      return the value of the best class: 'best_class'
   */
   resulting_class = best_class;
   if( verbose_reason )
     fprintf( cpf, " *** Confused %s with expected class %s\n",
		U->C[best_class].name, name );
 }
 return( resulting_class );
}
