/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:26 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <setjmp.h>
#include "def.h"
#include "matrxdef.h"
#include "clasdef.h"

extern universe *U;
extern boolean *train_smp, *test_smp;
extern int nrSamplesTest, nrSamplesTrain;
extern boolean verbose_reason;

extern FILE *cpf;               /* Classification protocol file */
extern boolean have_label;      /* Only deliver the classified class for
        pre-classified data */


Matrix *Means = NULL;
Matrix *Covariances = NULL;
Matrix MeanGlobal = { 0, 0, NULL };
Matrix CovarianceGlobal = { 0, 0, NULL };

static double *auxPar = NULL, *decisionFunc = NULL;
static Matrix commonCoVar, matSmp;
static boolean verbose_induction, interact = FALSE, identical_covar = FALSE;
static Matrix mat1, mat2, mat3, mat4;
static FeatVector smp = NULL;

static char bn[2][4] = {"NO", "YES"};

/* #define PEEPMAT(M) printf("M=\n");print_Matrix(&M);DBG;      /**/
#define PEEPMAT(M)      /**/


void show_MULGAUSS( f )
FILE *f;
{
 fprintf( f, "\tSIMPLIFICATIONS:\n");
 fprintf( f, "\t  * Having identical covariance matrices: %s\n",
	bn[identical_covar] );
}


void init_means_covar( FSV, len )
FeatSelectVector FSV;
int len;
{
 /* --- MEANS AND COVARIANCES --- */

 /* allocate space for means and covariances of all classes */
 Means = (Matrix*) malloc( U->nrClass * sizeof( Matrix ) );
 Covariances = (Matrix*) malloc( U->nrClass * sizeof( Matrix ) );
 CHKPTR(Means); CHKPTR(Covariances);
 smp = (FeatVector) malloc( len * sizeof( FeatVector* ) ); CHKPTR(smp);
 
 init_Matrix( &mat1 ); mat1.col = 1; mat1.row = len;
 Matrix_Alloc( &mat1 ); Reset_Matrix( &mat1 );
 init_Matrix( &mat2 ); init_Matrix( &mat3 ); init_Matrix( &mat4 );
 if( verbose_induction ) printf("\n---- MEANS AND COVARIANCES ----\n\n");
}


void close_means_and_covar()
{
 Matrix_Free( &mat1 ); Matrix_Free( &mat2 );
 Matrix_Free( &mat3 ); Matrix_Free( &mat4 );
 FREE(smp);
 if( verbose_induction ) printf("\n\n");
}



/*
 *   Generate the mean vectors for each class
 *   by estimation of the maximum likelihood estimator
*/
void gen_means( FSV, len )
FeatSelectVector FSV;
int len;
{
 float mean, scalar;
 int i, j, k, s, feat, a, b, nrTrainClass;
 Matrix *M = NULL;

 /* --- MEANS --- */

 for( i = 0; i < U->nrClass; i++ )
 {
   init_Matrix( &(Means[i]) );
   init_Matrix( &(Covariances[i]) );
   Means[i].col = 1; Means[i].row = len;
   Matrix_Alloc( &(Means[i]) ); Reset_Matrix( &(Means[i]) );
   Covariances[i].col = len; Covariances[i].row = len;
   Matrix_Alloc( &(Covariances[i]) ); Reset_Matrix( &(Covariances[i]) );
 }

 s = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   /* generate mean for class i */
   if( nrSamplesTrain == U->sumSampl )	/* all samples are training */
     for( k = 0; k < len; k++ )
     {
       feat = FSV[k].rank;
       mean = U->C[i].stat.mean[feat];
       Means[i].Elem[k] = (MatElem)mean;
     }
   else
   {
     nrTrainClass = 0;
     for( j = 0; j < U->C[i].numSampl; j++ )
     {
       if( train_smp[s] )
       {
         provideFV( smp, i, j, len, FSV );
         for( k = 0; k < len; k++ ) mat1.Elem[k] = (MatElem)smp[k];
         Add_Matrix( &mat1, &(Means[i]), &mat2 );
         Copy_Matrix( &mat2, &(Means[i]) );
         nrTrainClass++;
       }
       s++;
     } 
     if( nrTrainClass == 0 )
     {
       fprintf(stderr, "No training samples for class %d(%s)\n",
		i+1, U->C[i].name );
       fprintf(stderr, "\tunable to generate mean vector. Exitus...\n");
       exit(1);
     }
     scalar = 1.0 / (MatElem)nrTrainClass;
     ScalarMult_Matrix( scalar, &(Means[i]) );
   }

   if( verbose_induction )
   {
     printf(" *** Mean vector of class %d(%s):\n", i+1, U->C[i].name );
     print_Matrix( &(Means[i]) ); /**/
   }
 }
}

/*
 *   Generate the covariance matrices for each class
 *   by estimation of the maximum likelihood estimator
*/
void gen_covar( FSV, len )
FeatSelectVector FSV;
int len;
{
 float mean, scalar;
 int i, j, k, s, feat, a, b, nrTrainClass;
 Matrix *M = NULL;

 s = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   nrTrainClass = 0;
   /* generate covariance matrix for class i */
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     if( train_smp[s] )
     {
       provideFV( smp, i, j, len, FSV );
       for( k = 0; k < len; k++ ) mat1.Elem[k] = (MatElem)smp[k];
       Subtract_Matrix( &mat1, &(Means[i]), &mat2 );
       Transpose_Matrix( &mat2, &mat3 );
       Mult_Matrix( &mat2, &mat3, &mat4 );
       Add_Matrix( &mat4, &(Covariances[i]), &mat2 );
       Copy_Matrix( &mat2, &(Covariances[i]) );
       nrTrainClass++;
     }
     s++;
   } 
   if( nrTrainClass-1 == 0 ) EXIT("Unable to estimate covariance");
   scalar = 1.0 / (MatElem)(nrTrainClass - 1);
   ScalarMult_Matrix( scalar, &(Covariances[i]) );

   /* Check the covariances, it may not have a zero element */
   M = &(Covariances[i]);
   for( a = 0; a < M->row; a++ )
     for( b = 0; b < M->col; b++ )
       if( M->Elem[a*M->col+b] == 0.0 )
       {
         fprintf(stderr,"\n\nHad some zero element in the covariance matrix ");
         fprintf(stderr,"of class %d - Exitus...\n", i+1 ); exit(1);
       }
   if( verbose_induction )
   {
     printf("     Covariance matrix of class %d(%s):\n", i+1, U->C[i].name );
     print_Matrix( &(Covariances[i]) ); /**/
   }
 }
}


/*
 *   Generate the mean vectors and covariance matrices for each class
 *   by estimation of the maximum likelihood estimator
*/
void gen_means_and_covar( FSV, len )
FeatSelectVector FSV;
int len;
{
 init_means_covar( FSV, len );
 gen_means( FSV, len );
 gen_covar( FSV, len );
 close_means_and_covar();
}


/*
 *   Generate the global mean vector and covariance matrix
 *   by estimation of the maximum likelihood estimator
*/
void gen_global_mean_and_covar( FSV, len )
FeatSelectVector FSV;
int len;
{
 float mean, scalar;
 int i, j, k, s, feat, a, b;
 Matrix *M = NULL;
 Matrix mat1, mat2, mat3, mat4;
 FeatVector smp = NULL;

 smp = (FeatVector) malloc( len * sizeof( FeatVector* ) ); CHKPTR(smp);
 
 init_Matrix( &mat1 ); mat1.col = 1; mat1.row = len;
 Matrix_Alloc( &mat1 ); Reset_Matrix( &mat1 );
 init_Matrix( &mat2 ); init_Matrix( &mat3 ); init_Matrix( &mat4 );

 init_Matrix( &MeanGlobal ); init_Matrix( &CovarianceGlobal );
 MeanGlobal.col = 1; MeanGlobal.row = len;
 Matrix_Alloc( &MeanGlobal ); Reset_Matrix( &MeanGlobal );
 CovarianceGlobal.col = len; CovarianceGlobal.row = len;
 Matrix_Alloc( &CovarianceGlobal ); Reset_Matrix( &CovarianceGlobal );

 s = 0;

 if( nrSamplesTrain == U->sumSampl )	/* all samples are training */
   for( k = 0; k < len; k++ )
   {
     feat = FSV[k].rank;
     mean = U->stat.mean[feat];
     MeanGlobal.Elem[k] = (MatElem)mean;
   }
 else
 {
   for( i = 0; i < U->nrClass; i++ )
   {
     for( j = 0; j < U->C[i].numSampl; j++ )
     {
       if( train_smp[s] )
       {
         provideFV( smp, i, j, len, FSV );
         for( k = 0; k < len; k++ ) mat1.Elem[k] = (MatElem)smp[k];
         Add_Matrix( &mat1, &MeanGlobal, &mat2 );
         Copy_Matrix( &mat2, &MeanGlobal );
       }
       s++;
     } 
     if( nrSamplesTrain == 0 ) EXIT("Unable to generate mean vector");
     scalar = 1.0 / (MatElem)nrSamplesTrain;
     ScalarMult_Matrix( scalar, &MeanGlobal );
   }
 }
 /* printf(" *** Global mean vector:\n" );
 print_Matrix( &MeanGlobal ); DBG; /**/

 s = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     if( train_smp[s] )
     {
       provideFV( smp, i, j, len, FSV );
       for( k = 0; k < len; k++ ) mat1.Elem[k] = (MatElem)smp[k];
       Subtract_Matrix( &mat1, &MeanGlobal, &mat2 );
       Transpose_Matrix( &mat2, &mat3 );
       Mult_Matrix( &mat2, &mat3, &mat4 );
       Add_Matrix( &mat4, &CovarianceGlobal, &mat2 );
       Copy_Matrix( &mat2, &CovarianceGlobal );
     }
     s++;
   } 
 }
 if( nrSamplesTrain-1 == 0 ) EXIT("Unable to estimate covariance");
 scalar = 1.0 / (MatElem)(nrSamplesTrain - 1);
 ScalarMult_Matrix( scalar, &CovarianceGlobal );

 /* Check the covariances, it may not have a zero element */
 M = &CovarianceGlobal;
 for( a = 0; a < M->row; a++ )
   for( b = 0; b < M->col; b++ )
     if( M->Elem[a*M->col+b] == 0.0 )
     {
       fprintf(stderr,"\n\nHad some zero element in the covariance matrix ");
       fprintf(stderr,"of class %d - Exitus...\n", i+1 ); exit(1);
     }
 /* printf("     Global Covariance matrix:\n" );
 print_Matrix( &CovarianceGlobal ); DBG; /**/

 Matrix_Free( &mat1 ); Matrix_Free( &mat2 );
 Matrix_Free( &mat3 ); Matrix_Free( &mat4 );
 FREE(smp);
}


/* Free means and covariances */
void free_means_and_covar()
{
 int i;

 if( Means == NULL )
   return;
 for( i = 0; i < U->nrClass; i++ )
   Matrix_Free( &(Means[i]) );
 FREE( Means );

 if( Covariances == NULL )
   return;
 for( i = 0; i < U->nrClass; i++ )
   Matrix_Free( &(Covariances[i]) );
 FREE( Covariances );
}


void free_global_mean_and_covar()
{
 Matrix_Free( &MeanGlobal );
 Matrix_Free( &CovarianceGlobal );
}


void initMULGAUSS()
{
 free_induced();
 printf("\n #### Specifications for the Multivariate Gaussian classifier: \n");
 printf("\nSimplification: Identical covariance matrices for all classes?");
 get_boolean( identical_covar, &identical_covar ); printf("\n");
}


static double MAHALANOBIS_DISTANCE( x, mean, invCoVar )
Matrix *x, *mean, *invCoVar;
{
 double dist;
 Matrix mat1, mat2, mat3;

 init_Matrix( &mat1 ); init_Matrix( &mat2 ); init_Matrix( &mat3 );

 Subtract_Matrix( x, mean, &mat1 ); /* mat1 = x-m */
 PEEPMAT(mat1);
 Transpose_Matrix( &mat1, &mat2 );   /* mat2 = (m1 - m2) T  */
 PEEPMAT(mat2);
 Mult_Matrix( &mat2, invCoVar, &mat3 );
 Mult_Matrix( &mat3, &mat1, &mat2 );
 PEEPMAT(mat2);

 dist = (double)mat2.Elem[0];
 Matrix_Free( &mat1 ); Matrix_Free( &mat2 ); Matrix_Free( &mat3 );
 if( dist < 0 )
   { printf("Negative Mahalanobis distance: %lf", dist ); WAIT; }
 return( dist );
}


/*
 * SIGNAL HANDLING FOR FLOATING POINT EXCEPTIONS
*/
static jmp_buf jmp;
static boolean induction_failed;

static void restart_from_fpe(  dummy )
int dummy;
{
 dummy = EMPTY;
 signal( SIGFPE, restart_from_fpe );
 fprintf( stderr, "\nFLOATING POINT EXCEPTION, probably caused by inversion\n");
 fprintf(  stderr,"of a singular matrix. Classifier induction failed!" ); WAIT;
 induction_failed = TRUE;
 longjmp( jmp, 1 );
}



int induceMULGAUSS( nsf, FSV, iact )
int nsf;
FeatSelectVector FSV;
boolean iact;
{
 int i;
 MatElem Det;
 Matrix mat1, mat2;

 /* Prepare for floating point exception */
 signal( SIGFPE, restart_from_fpe );

 induction_failed = FALSE;
 /* In case of a floating point exception program re-enters here */
 setjmp( jmp );
 if( induction_failed )
   return( FALSE );

 interact = iact;
 if( interact )
 {
   printf("\n --- Inducing a %d-variate Gaussian model ---\n\n", nsf);
   printf("Verbose learning of decision functions?");
   get_boolean( FALSE, &verbose_induction );
 }

 gen_means_and_covar( FSV, nsf );

 init_Matrix( &mat1 ); mat1.col = nsf; mat1.row = nsf;
 Matrix_Alloc( &mat1 ); Reset_Matrix( &mat1 );
 init_Matrix( &mat2 );

 init_Matrix( &commonCoVar );
 init_Matrix( &matSmp );
 matSmp.row = nsf; matSmp.col = 1; Matrix_Alloc( &matSmp );

 auxPar = (double*) malloc( U->nrClass * sizeof( double ) ); CHKPTR(auxPar);
 decisionFunc = (double*) malloc( U->nrClass * sizeof( double ) );
 CHKPTR(decisionFunc);
 for( i = 0; i < U->nrClass; i++ )
   auxPar[i] = 0.0; 


 if( identical_covar )
 {
   for( i = 0; i < U->nrClass; i++ )
   {
     /* print_Matrix( &(Covariances[i]) ); print_Matrix( &mat1 ); DBG; /**/
     Add_Matrix( &(Covariances[i]), &mat1, &mat2 );
     Copy_Matrix( &mat2, &mat1 );
   }
   Copy_Matrix( &mat1, &commonCoVar );
   ScalarMult_Matrix( 1.0/(MatElem)U->nrClass, &commonCoVar );
   if( verbose_induction )
   {
     printf("\n\tCommon covariance matrix:\n");
     print_Matrix( &commonCoVar ); /**/
   }
   Invert_Matrix( &commonCoVar, &Det );
   for( i = 0; i < U->nrClass; i++ )
     auxPar[i] = log( (double)U->C[i].a_priori_prob ) - 0.5*log(Det);
 }
 else
 {
   for( i = 0; i < U->nrClass; i++ )
   {
     Invert_Matrix( &(Covariances[i]), &Det );
     auxPar[i] = log( (double)U->C[i].a_priori_prob ) - 0.5*log(Det);
   }
 }

 if( verbose_induction )
 {
   printf("---- AUXILIARY VARIABLES:\n");
   printf("\tlog(P(class)) - 1/2 * log(Determinant(Covariance(class)))\n" );
   printf("\tand INVERSE COVARIANCE MATRICES ----\n\n");
   for( i = 0; i < U->nrClass; i++ )
   {
     printf("Class %3d  -  %20s:\t%lf\n", i+1, U->C[i].name, auxPar[i] );
     if( ! identical_covar )
       print_Matrix( &(Covariances[i]) ); /**/
   }
   if( identical_covar )
   {
     printf("\n\tInverse common covariance matrix:\n");
     print_Matrix( &commonCoVar ); /**/
   }
 }
 Matrix_Free( &mat1 ); Matrix_Free( &mat2 );
 verbose_induction = FALSE;
 return( TRUE );
}


/*
 *  Free all the means, covariances and auxiliary variables
*/
void free_inducedMULGAUSS()
{
 free_means_and_covar();
 Matrix_Free( &commonCoVar ); Matrix_Free( &matSmp );
 FREE( auxPar ); FREE( decisionFunc );
}


int class1testMULGAUSS( nsf, FSV, smp, name )
int nsf;
FeatSelectVector FSV;
FeatVector smp;
char *name;
{
 int i, maxClass, res_class;
 double mahaDist, maxDist;

 for( i = 0; i < nsf; i++ )
   matSmp.Elem[i] = (MatElem)(smp[i]);
 
 if( verbose_reason )
 {
   fprintf( cpf,"\n\n ---- DECISION FUNCTIONS FOR UNKNOWN SAMPLE:\n");
   fshowFV( cpf, nsf, smp );
 }
 for( i = 0; i < U->nrClass; i++ )
 {
   if( identical_covar )
     mahaDist = MAHALANOBIS_DISTANCE( &matSmp, &(Means[i]), &(commonCoVar) );
   else
     mahaDist = MAHALANOBIS_DISTANCE( &matSmp, &(Means[i]), &(Covariances[i]) );
   decisionFunc[i] = auxPar[i] - 0.5 * mahaDist;
   if( verbose_reason )
   {
     fprintf( cpf, "\tClass %3d --- %20s = %.5lf\n",
		i+1, U->C[i].name, decisionFunc[i] ); /**/
   }
 }
 
 /* find the class with the highest value of the decision function */
 maxDist = -INFINITY;
 for( i = 0; i < U->nrClass; i++ )
   if( decisionFunc[i] > maxDist )
   {
     maxDist = decisionFunc[i];
     maxClass = i;
   }
 if( ! have_label )
   return( maxClass );


 if( (strcmp(U->C[maxClass].name,name)==0) )
   res_class = CORRECT_CLASS;
 else
 {
   res_class = maxClass;
   if( verbose_reason )
     fprintf( cpf, " *** Confused %s with expected class %s\n",
		U->C[maxClass].name, name );
 }
 return( res_class );
}
