/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:26 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "def.h"
#include "matrxdef.h"
#include "clasdef.h"

extern float Euclidian_Distance();
extern float *vector();

extern universe *U;
extern boolean verbose, verbose_reason;
extern str100 outDataDir;
extern boolean feat_description;
extern char **feature_desc;
extern boolean *train_smp, *test_smp;

extern FILE *cpf;		/* Classification protocol file */
extern boolean have_label;	/* Only deliver the classified class for
	pre-classified data */


#define MIN_DISPLAY 20

#define MAX_K 10
#define K_DEFAULT 1
int NNK = K_DEFAULT;  /* The nearest neighbor K */


typedef struct NearNeibor_ {
          float dist;
          int classIndex;
        } NearNeibor;

static NearNeibor *NN = NULL;
static float *classDists = NULL;	/* The minimum distance to each CLASS */
static str80 buf;


void set_K( K )
int K;
{
 int knn;

 if( K == EMPTY )
 {
   NNK = K_DEFAULT;
   printf( "Nearest neighbor K=%d\b", NNK ); gets( buf );
   if( buf[0] != '\0' )
   {
     sscanf( buf, "%d", &knn );
     if( knn <= 0 || knn > MAX_K )
     {
       printf("Value %d for K is invalid. Setting to default %d...",
            knn, K_DEFAULT );
       knn = K_DEFAULT;
       gets( buf );
     }
     NNK = knn;
   }
 }
 else
 {
   knn = K;
   if( knn <= 0 || knn > MAX_K )
     knn = K_DEFAULT;
   NNK = knn;
 }
 /* printf("K=%d",NNK);DBG;	/**/
}


void show_NEAR_NEI( f )
FILE *f;
{
}


static void alloc_NN()
{
 int i;

 /* Allocate memory for each K */
 NN = (NearNeibor*)malloc( NNK*sizeof(struct NearNeibor_)); CHKPTR(NN);

 /* Also register the minimum distance to each class */
 classDists = vector(U->nrClass);
 for( i = 0; i < U->nrClass; i++ )
   classDists[i] = INFINITY;
}


static void free_NN()
{
 FREE( NN );
 FREE( classDists );
}


void initNEARNEI()
{
 free_induced();
 set_K( EMPTY );
}


static void init_NN( K )
{
 int j;

 for( j = 0; j < K; j++ )
 {
   NN[j].dist = INFINITY;   /* infinity */
   NN[j].classIndex = EMPTY;
 }
}


static void fill_NN_slot( NN, K, slot, classIndex, distance )
NearNeibor *NN;
int K, slot, classIndex;
float distance;
{
 int b;

 /* found a distance smaller than existing one */
 /* shift downward */
 for( b = K-1; b > slot; b-- )
 {
   NN[b].dist = NN[b-1].dist;
   NN[b].classIndex = NN[b-1].classIndex;
 }
 NN[slot].dist = distance;
 NN[slot].classIndex = classIndex; 

 /* printf("  fill_NN_slot: \n");
 for( b = 0; b < K; b++ )
   printf("K[%d] = %d with %f\n...",b, NN[b].classIndex+1, NN[b].dist );
 gets(buf); /**/
}


/*
 *  evaluate the NN table
 *  return the class with the highest frequency
*/
static int evaluate_KNN( NN, K )
NearNeibor *NN;
int K;
{
 int i, a, maxFreq, Freq, maxIndex, eqi;
 int *FreqC = NULL, *equalFreq = NULL, nr_equal_freq;
 float max_apriori;

 if( verbose_reason )
 {
   fprintf(cpf,"\n\n--- Classification result of %d-Nearest Neighbor ---\n",K ); 
   for( a = 0; a < K; a++ )
   {
     i = NN[a].classIndex;
     if( i >= 0 && i < U->nrClass )
       fprintf(cpf," %3d. Neighbor = %20s with distance %.6f\n",
               a+1, U->C[i].name, NN[a].dist);
   }
 }

 /* look for class with maximum fequency */
 maxFreq = 0; maxIndex = EMPTY;
 if( K == 1 )
 {
   for( i = 0; i < U->nrClass; i++ )
   {
     Freq = 0;
     for( a = 0; a < K; a++ )
       if( NN[a].classIndex == i )
         Freq++;
     if( Freq > maxFreq )
     {
       maxFreq = Freq; maxIndex = i;
     }
   }
 }
 else	/* K > 1 */
 {
   FreqC = (int*) malloc(U->nrClass * sizeof(int)); CHKPTR(FreqC);
   equalFreq = (int*) malloc(U->nrClass * sizeof(int)); CHKPTR(equalFreq);
   for( i = 0; i < U->nrClass; i++ )
   {
     FreqC[i] = 0;
     for( a = 0; a < K; a++ )
       if( NN[a].classIndex == i )
         (FreqC[i])++;
     if( FreqC[i] > maxFreq )
     {
       maxFreq = FreqC[i]; maxIndex = i;
     }
   }
   /* Check out if there are more than one class with the same frequency
      if yes, choose the one with the highest a priori probability
   */
   nr_equal_freq = 0; 
   for( i = 0; i < U->nrClass; i++ )
     if( FreqC[i] == maxFreq )
     {
       equalFreq[nr_equal_freq] = i;
       nr_equal_freq++;
     }

   if( nr_equal_freq > 1 )
   {
     if( verbose_reason )
     {
       fprintf(cpf,"There were %d classes with %d nearest neighbors,\n",
		nr_equal_freq, maxFreq );
       fprintf(cpf,"\tsearching highest a priori probability...\n");
     }
     max_apriori = 0.0;
     for( i = 0; i < nr_equal_freq; i++ )
     {
       eqi = equalFreq[i];
       if( verbose_reason )
       {
	 fprintf(cpf,"Class %d(%s) had equal frequency %d",
		eqi+1, U->C[eqi].name, FreqC[eqi]);
	 fprintf(cpf," and a priori probability %6.2f%%\n",
			100.0*U->C[eqi].a_priori_prob );
       }
       if( U->C[eqi].a_priori_prob > max_apriori )
       {
	 max_apriori = U->C[eqi].a_priori_prob; maxIndex= eqi;
       }
     }
     if( verbose_reason )
       fprintf(cpf,"Class %d(%s) had maximum a priori probability %f\n",
		maxIndex+1, U->C[maxIndex].name, max_apriori );
   }
   FREE(FreqC); FREE(equalFreq);
 }
 if( maxIndex == EMPTY )
   EXIT("evaluate_KNN> Problems");
 return( maxIndex );
}


static int est_err_KNN( nrSelFeat, K, leave_out, leave_out_Sample, FSV )
int nrSelFeat, K, leave_out;
FeatVector leave_out_Sample;
FeatSelectVector FSV;
{
 int i, j, k, a, actual_sample_index;
 boolean smaller;
 FeatVector Sample;
 float distance;

 /* if 'leave_out' has the value EMPTY do nothing */
 set_one_sample_test( leave_out );
 Sample = (FeatVector)malloc(nrSelFeat*sizeof(FeatVector*)); CHKPTR(Sample);
 actual_sample_index = 0;
 init_NN( K );

 for( i = 0; i < U->nrClass; i++ )
 {
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     if( train_smp[ actual_sample_index ] )
     {
       for( k = 0; k < nrSelFeat; k++ )
         Sample[k] = U->C[i].S[j*U->nrFeat+FSV[k].rank];
       distance = Euclidian_Distance(leave_out_Sample, Sample, nrSelFeat);
       /* if( actual_sample_index == 0 )
       {printf("Euclidian_Distance=%f between:\n", distance );
       showFV( nrSelFeat, leave_out_Sample); showFV( nrSelFeat, Sample);}/**/

       /* compare the distance with the distances already calculated */
       a = 0; smaller = FALSE;
       while( !smaller && (a < K) )
       {
         smaller = ( distance < NN[a].dist );
         a++;
       } 
       if( smaller )
         fill_NN_slot( NN, K, a-1, i, distance );
     }
     actual_sample_index++;
   }
 }
 FREE( Sample );
 reset_one_sample_test( leave_out );
 /* distances are filled in ; evaluate */
 return( evaluate_KNN( NN, K ) );
}


/*
 *    ERROR ESTIMATION FOR K_NEAREST NEIGHBOR CLASSIFIER
 *       BY METHOD: LEAVE-ONE-OUT
*/
float estimate_error_KNN( nrSelFeat, FSV )
int nrSelFeat;
FeatSelectVector FSV;
{
 float global_error;
 int nrSampl, i, j, k, leave_out, id_class;
 FeatVector Sample;
 FeatVector FV = NULL;

 Sample = (FeatVector)malloc(nrSelFeat*sizeof(FeatVector*)); CHKPTR(Sample);
 FV = (FeatVector)malloc(nrSelFeat*sizeof(FeatVector*)); CHKPTR(FV);
 alloc_NN();

 set_all_samples_train();

 nrSampl = 0;
 for( i = 0; i < U->nrClass; i++ )
   nrSampl += U->C[i].numSampl;
 leave_out = 0;
 global_error = 0.0;

 if( verbose )
 {
   BLANK_LINE;
   printf("# Features: %2d - K: %2d - Total number of samples: %7d\n",
       nrSelFeat, NNK, nrSampl);
   printf("                             Samples to leave out:");
 }
 for( i = 0; i < U->nrClass; i++ )
 {
   if( U->C[i].numSampl < NNK )
     { printf("Warning!: Class %d = %s has less samples than K = %d...",
             i+1, U->C[i].name, NNK ); gets( buf ); }

   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     for( k = 0; k < nrSelFeat; k++ )
       Sample[k] = U->C[i].S[j*U->nrFeat+FSV[k].rank];
     id_class = est_err_KNN( nrSelFeat, NNK, leave_out, Sample, FSV );

     /* write result into matrix */

     if( i != id_class )
       global_error += 1.0;
     leave_out++;
     if( verbose )
     {
       printf("%7d\b\b\b\b\b\b\b", nrSampl-leave_out); fflush(stdout);
     }
   }
 }
 if( verbose ) printf("\n");
 global_error /= (float)nrSampl;
 FREE( FV ); FREE( Sample ); free_NN();

 free_tt_lookups();
 /* printf("Returning %f\n",global_error); /**/
 return( global_error );
}


int class1testNEAR_NEI( nrSelFeat, FSV, sample, name )
int nrSelFeat;
FeatSelectVector FSV;
FeatVector sample;
char *name;
{
 int i, expected, recognized;

 /* EMPTY means no sample is left out here: compare to every sample */
 recognized = est_err_KNN( nrSelFeat, NNK, EMPTY, sample, FSV );

 if( ! have_label )
   return( recognized );

 if( verbose_reason )
 {
   expected = EMPTY;
   for( i = 0; i < U->nrClass; i++ )
     if( (strcmp(U->C[i].name,name)==0) )
       expected = i;
   if( expected == EMPTY || expected >= U->nrClass )
     { printf("\nWarning: Class \"%s\" in test set never seen in training set",
	name ); WAIT; }
   if( recognized != expected )
   {
     fprintf( cpf, " *** Confused %s with expected class %s\n",
	U->C[recognized].name, name );
   }
 }
 if( (strcmp(U->C[recognized].name,name)==0) )
   return( CORRECT_CLASS );

 return( recognized );
}


void est_err_nearneib()
{
 int i;
 boolean protocolFile;
 float err, *errInfo = NULL;
 FILE *ef = NULL;
 str80 errFile, name;

 if( U->nrSelFeat == 0 )
   { printf("Select features first..."); gets( buf ); return; }

 set_K( EMPTY );

 strcpy( errFile, outDataDir );
 printf("Saving the error estimation protocol to file:\n\t\t%s", errFile );
 gets( name );
 protocolFile = name[0] != '\0';
 if( protocolFile )
 {
   strcat( errFile, name );
   ef = fopen( errFile, f_open_text_w );
   if( ef == NULL )
     { printf("Cannot open %s! Exitus...\n", errFile ); exit(1); }
   fprintf( ef, "Universe: %s\n", U->name );
 }
 else
   printf("\t\tNothing saved...\n");

 errInfo = (float*) malloc( U->nrSelFeat * sizeof(float) );
 printf("- Estimating error using leave-one-out\n");
 for( i = 1; i <= U->nrSelFeat; i++ )
 {
   err = estimate_error_KNN( i, U->FSV );
   errInfo[i-1] = 100.0*err;
   printf(" ---> Error estimate = %5.2f%% using %4d of %4d features\n",
	 100.0*err, i, U->nrSelFeat ); /**/

   if( protocolFile )
     fprintf( ef,
     "Nr. Features: %3d - Estimated error:\t%5.2f\t%% - Accuracy:\t%5.2f\t%%\n",
		 i, 100.0*err, 100.0*(1.0-err) );
 }
 if( protocolFile )
   fclose( ef );
 GNU_error_plot( errInfo, U->nrSelFeat, "Leave-One-Out" );
 FREE( errInfo );
}


void induceNEAR_NEI( nsf, FSV, interact )
int nsf;
FeatSelectVector FSV;
boolean interact;
{
 alloc_NN();
 init_NN( NNK );

 if( interact )
   printf("That's easy. Nothing to do, just keeping all raw samples!\n");
}


void free_inducedNEAR_NEI()
{
 free_NN();
 /* There is no classifier structure to be de-allocated in this case */
}



/* --- old algorithm --- */

void featSelectMinErr_SFS()
{
 int i, j, nrSelectFeat, feat, minFeat;
 boolean protocolFile, *already = NULL;
 float err, *errInfo = NULL, minErr;
 FILE *ef = NULL;
 str80 errFile, name;
 FeatSelectVector FSV = NULL;

 set_K( EMPTY );

 strcpy( errFile, outDataDir );
 printf("Saving the error estimation protocol to file:\n\t\t%s", errFile );
 gets( name );
 protocolFile = name[0] != '\0';
 if( protocolFile )
 {
   strcat( errFile, name );
   ef = fopen( errFile, f_open_text_w );
   if( ef == NULL )
     { printf("Cannot open %s! Exitus...\n", errFile ); exit(1); }
   fprintf( ef, "Universe: %s\n", U->name );
 }
 else
   printf("\t\tNothing saved...\n");

 printf("Maximum number of features to be selected (1-%d)? ", U->nrFeat);
 get_d_range( &nrSelectFeat, 1, U->nrFeat, LEFT_CLOSED__RIGHT_CLOSED );

 already = (boolean*) malloc( U->nrFeat * sizeof(boolean) ); CHKPTR(already);
 for( i = 0; i < U->nrFeat; i++ )
   already[i] = FALSE;

 errInfo = (float*) malloc( U->nrFeat * sizeof(float) ); CHKPTR(errInfo);
 FSV = (feature*) malloc( U->nrFeat * sizeof(struct feature_) ); CHKPTR(FSV);

 feat = 0;
 for( i = 0; i < nrSelectFeat; i++ )
 {
   /* for all features not yet considered test the error rate */
   minErr = 1.0;
   for( j = 0; j < U->nrFeat; j++ )
   {
     if( ! already[j] )
     {
       FSV[ feat ].rank = j;
       printf("Considering feature nr. %4d of %4d-%4d as candidate nr. %4d\r",
	j+1, U->nrFeat, feat, i+1 ); fflush( stdout );
       if( verbose ) printf("\n" );
       err = estimate_error_KNN( feat+1, FSV );
       if( err < minErr )
       {
         minErr = err;
         minFeat = j;
       }
     }
   }
   /* which one had the best performance ? */
   FSV[ feat ].rank = minFeat; FSV[ feat ].crit = minErr;
   already[ minFeat ] = TRUE;
   errInfo[ feat ] = 100.0*minErr;
   feat++;
   printf("\n ---> Error estimate = %5.2f %%\n", 100.0*minErr ); /**/
   printf("For feature vector:\n\t( ");
   for( j = 0; j < feat; j++ )
     printf("%d ", FSV[j].rank+1 );
   printf(")\n\n");

   if( protocolFile )
     fprintf( ef,
	"Nr. Features: %3d - Estimated error: %5.2f %% - Accuracy: %5.2f %%\n",
                      i+1, 100.0*minErr,100.0*(1.0-minErr) );
 }
 U->FSV = (feature*) malloc( feat * sizeof(struct feature_) );
 printf("\n--- SELECTED FEATURES --- : %d\n", feat);
 for( j = 0; j < feat; j++ )
 {
   U->FSV[j].rank = FSV[j].rank;
   U->FSV[j].crit = FSV[j].crit;
   printf("Nr. %3d = %3d  --- Criterion = %7.5f",
         j+1, U->FSV[j].rank+1, U->FSV[j].crit );
   if( feat_description )
     printf("\tName: %s", feature_desc[U->FSV[j].rank] );
   printf("\n");
 }
 U->nrSelFeat = feat;

 save_selected_feat();

 if( protocolFile )
   fclose( ef );
 GNU_error_plot( errInfo, U->nrSelFeat, "Leave-One-Out" );
 FREE( errInfo ); FREE( FSV ); FREE( already );
}

