/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:26 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "def.h"
#include "matrxdef.h"

extern universe *U;
extern str100 outDataDir;
extern boolean is_ascii_file();
extern boolean feat_description;
extern char **feature_desc;

#define UNNORMED	(-1)

#define NORM_0_1	0	/* Linear univariate scale to [0,1] */
#define NORM_M0S1	1	/* Zero mean, unit variance */

#define NUM_NORMS	2		/* Types of normalization */

#define PEEPMAT(M) printf("M=\n");print_Matrix(&M);DBG;      /**/

static int norm_method = UNNORMED;     /* original */

static str80 buf;
static str100 name, outname, rootName, linebuf, nameBuf;
static int menuItem = EMPTY;
static boolean done;

static Matrix Mean = { 0, 0, NULL }, InvCoVar = { 0, 0, NULL };
static Matrix aux1 = { 0, 0, NULL }, aux2 = { 0, 0, NULL };



/*  N O R M A L I Z A T I O N  O F  T H E  F E A T U R E  V A L U E S  */

/*--------------------------------------------------------------*/
static struct norms {
	int index;
	char *name;
} norm_methods[] = {
	{ NORM_0_1,		"Linear to [0,1]" },
	{ NORM_M0S1,		"Zero mean, unit variance" }
};
/*--------------------------------------------------------------*/

char *name_norm()
{
 if( norm_method == UNNORMED )
 return( NULL );
 return( norm_methods[norm_method].name );
}

/* Reset when new universe is loaded */
void reset_norm()
{
 norm_method = UNNORMED;
}

/*
 *  Normalize the feature data using the transformation
 *  T(x)  = (x-min) / (max-min )
 *    This scales all values to the interval [0,1]
*/
static void smp_norm_0_1( smp, dim, min, max )
FeatVector smp;
int dim;
float *min, *max;
{
 int k;

 /* printf("Before normalization:" ); showFV( dim, smp ); DBG;	/**/
 for( k = 0; k < dim; k++ )
 {
   if( min[k] != max[k] )
     smp[k] = (smp[k] - min[k]) / (max[k] - min[k]);
   else
     smp[k] = 0.5;
 }   
 /* printf("After normalization:" ); showFV( dim, smp ); DBG;	/**/
}



/*
 *  Normalize the feature data using the transformation
 *  T(x)  = (x-mean) / standard_deviation
 *  The effect is that the feature is centralized around 0 (mean=0)
 *   and has a variance of 1
*/
static void smp_norm_M0S1( smp, dim, m, s )
FeatVector smp;
int dim;
float *m, *s;
{
 int k;

 /* printf("Before normalization:" ); showFV( dim, smp ); /**/
 for( k = 0; k < dim; k++ )
 {
   if( s[k] != 0.0 )
     smp[k] = (smp[k] - m[k]) / s[k];
   else
   {
     printf("WARNING: had a zero variance for feature %d", k+1 ); NL;
     smp[k] = (smp[k] - m[k]);
   }
 }
 /* printf("After normalization:" ); showFV( dim, smp ); DBG;	/**/
}


static void norm_one_sample( vec, dim )
FeatVector vec;
int dim;
{
 switch( norm_method )
 {
   case NORM_0_1:	smp_norm_0_1( vec, dim, U->stat.min, U->stat.max);
	break;
   case NORM_M0S1:	smp_norm_M0S1( vec, dim, U->stat.mean, U->stat.stddev );
	break;
   default: fprintf(stderr,"What is method %d? - exit...\n", norm_method );
 }
}


#define MAXSTRLEN 100

static void norm_files()
{
 boolean ascii_file, nf, done = FALSE;
 FILE *f = NULL, *outf = NULL;
 int dim, j, k;
 FeatVector featBuf = NULL;

 printf("\n --- ATTENTION ---\n");
 printf("\tIt is now recommended to normalize all those files\n");
 printf("\tthat later will be used as test data.\n");
 printf("\tDo you want to do that?" );
 get_boolean( TRUE, &nf );
 if( ! nf ) return;

 printf("Enter file names - exit by entering empty name!\n");
 do
 {
   printf("File to normalize? "); gets( name );
   if( name[0] == '\0' )
     done = TRUE;
   else
   {
     ascii_file = is_ascii_file( name );
     if( ascii_file )
       f = fopen( name, f_open_text_r );
     else
       f = fopen( name, f_open_bin_r );
     if( f == NULL )
       { printf("Cannot open file '%s'!", name ); WAIT; }
     else
     {
       root_name( name, rootName );
       printf("Save the normalized test data to?\n\t%s", outDataDir );
       gets( nameBuf );
       if( strcmp( rootName, nameBuf )==0 )
       {
	 printf("Filenames %s and %s identical - choose other name",
		rootName, nameBuf ); WAIT;
       }
       else
       {
	 strcpy( outname, outDataDir );
	 strcat( outname, nameBuf ); 
	 if( ascii_file )
	   outf = fopen( outname, f_open_text_w );
         else
	   outf = fopen( outname, f_open_bin_w );
	 if( outf == NULL )
	 {
           printf("Cannot open \"%s\" for writing!", outname ); WAIT;
           fclose(f);
         }
	 else
	 {
	   dataline( f, linebuf ); sscanf( linebuf, "%d", &dim );
	   if( U->nrFeat != dim )
	   {
	     printf("ERROR: Dimension %d of %s is different from %d",
		dim, name, U->nrFeat );
             printf(" of current universe %s", U->name ); WAIT;
	   }
	   else
	   {
	     featBuf = (FeatVector)malloc(sizeof(FeatVector*) * U->nrFeat);
	     CHKPTR(featBuf);
             while( !feof( f ) )
             {
               for( k = 0; k < U->nrFeat; k++ )
               {
                 if( ascii_file )
                   fscanf( f, "%f", &(featBuf[k]) );
                 else
                   fread( &(featBuf[k]), sizeof(float), 1, f );
               }
               if( ascii_file )
                 fscanf( f, "%s", nameBuf );
               else
               {
                 j = 0;
                 do
                 {
                   fread( &(nameBuf[j]), sizeof(char), 1, f );
                   if( nameBuf[j] != '\n' )
                     j++;
                 }
                 while( nameBuf[j] != '\n' && j < MAXSTRLEN );
                 nameBuf[j] = '\0';
               }
               if( nameBuf[0] == '\0' )   /* empty name, error */
                  { fprintf( stderr, "Found an empty class name. Exit...\n");
		  exit(1); }

               if( !feof( f ) )
               {
		 norm_one_sample( featBuf, dim );
		 if( ascii_file )
		 {
		   for( k = 0; k < dim; k++ )
		     fprintf( outf, "%f ", featBuf[k] );
                   fprintf( outf, "%s\n", nameBuf );
		 }
		 else
		 {
		   fwrite( featBuf, sizeof(FeatVector*), dim, outf );
		   fprintf( outf, "%s\n", nameBuf );
		 }
               }
             }
             FREE( featBuf );
	   }
           fclose( outf );
	 }
       }
       fclose( f );
     }
   }
 }
 while( ! done );
}


/*
 *  The actual feature values have been normalized.
 *  Change the name of the features as well
*/
static void change_feat_names_normed()
{
 int i;

 printf("\n >>> Changing the universe name from \"%s\" to", U->name);
 switch( norm_method )
 {
   case NORM_0_1:	strcat( U->name, "*NORM_0_1"); break;
   case NORM_M0S1:	strcat( U->name, "*NORM_M0S1"); break;
   default: fprintf(stderr,"What is method %d? - exit...\n", norm_method );
 }
 printf(" \"%s\" <<<\n", U->name );

 if( ! feat_description ) return;

 printf("\t>>> Changing the actual feature names <<<\n");
 for( i = 0; i < U->nrFeat; i++ )
 {
   strcpy( nameBuf, feature_desc[i] );
   printf("Feature # %4d: %20s  to  ", i+1, nameBuf );
   switch( norm_method )
   {
     case NORM_0_1:		strcat( nameBuf, "*NORM_0_1"); break;
     case NORM_M0S1:		strcat( nameBuf, "*NORM_M0S1"); break;
     default: fprintf(stderr,"What is method %d? - exit...\n", norm_method );
   }
   FREE(feature_desc[i]);
   feature_desc[i] = (char*) malloc( (1+strlen(nameBuf)) * sizeof(char) );
   CHKPTR(feature_desc[i]);
   strcpy( feature_desc[i], nameBuf );
   printf("%20s\n", nameBuf );
 }
 NL;
}


static void norm_0_1()
{
 float min, max;
 int i, k, c;
 boolean go_on = FALSE;

 printf("Normalizing each feature to the range [0,1]\n");

 for( k = 0; k < U->nrFeat; k++ )
 {
   min = U->stat.min[k];
   max = U->stat.max[k];
   if( min == max )
   {
     printf("WARNING: Feature %d has the same min and max value %f\n",
	k+1, min );
     if( ! go_on )
     {
       printf("Do you wish to continue?");
       get_boolean( TRUE, &go_on );
       if( ! go_on ) return;
     }
   }
 }
 for( c = 0; c < U->nrClass; c++ )
   for( i = 0; i < U->C[c].numSampl; i++ )
     smp_norm_0_1( &(U->C[c].S[i*U->nrFeat]), U->nrFeat,
	U->stat.min, U->stat.max );

 printf(" done.\n");
 norm_method = NORM_0_1;
 change_feat_names_normed();
 norm_files();
 generate_stats( FALSE );
}


static void norm_M0S1()
{
 int i, c;

 printf("Normalizing each feature to zero mean and unit variance\n");

 for( c = 0; c < U->nrClass; c++ )
   for( i = 0; i < U->C[c].numSampl; i++ )
     smp_norm_M0S1( &(U->C[c].S[i*U->nrFeat]), U->nrFeat,
	U->stat.mean, U->stat.stddev );

 printf(" done.\n");
 norm_method = NORM_M0S1;
 change_feat_names_normed();
 norm_files();
 generate_stats( FALSE );
}


static void gen_mean_cov()
{
 FeatSelectVector FSV = NULL;
 MatElem dummy;
 int i;

 FSV = (feature*) malloc( U->nrFeat * sizeof(struct feature_) );
 for( i = 0; i < U->nrFeat; i++ )
 {
   FSV[i].rank = i; FSV[i].crit = (float)EMPTY;
 }
 Matrix_Free( &Mean );
 Mean.row = U->nrFeat; Mean.col = 1;
 Matrix_Alloc( &Mean );
 for( i = 0; i < U->nrFeat; i++ )
   Mean.Elem[i] = U->stat.mean[i];
 Matrix_Free( &InvCoVar );
 scatter_mats_dev_kit( FSV, U->nrFeat, &InvCoVar, &aux1, &aux2 );
 Invert_Matrix( &InvCoVar, &dummy );
 /* print_Matrix( &InvCoVar ); print_Matrix( &Mean ); DBG;	/**/
 Matrix_Free( &aux1 ); Matrix_Free( &aux2 );
 /* Use aux1 and aux2 as   Dx1   matrices */
 aux1.row = U->nrFeat; aux1.col = 1;
 Matrix_Alloc( &aux1 );
 FREE( FSV );
}


static void free_mean_cov()
{
 Matrix_Free( &Mean ); Matrix_Free( &InvCoVar );
 Matrix_Free( &aux1 ); Matrix_Free( &aux2 );
}


static void execMenuItem()
{
 switch( menuItem )
 {
   case NORM_0_1:	norm_0_1(); break;
   case NORM_M0S1:	norm_M0S1(); break;
   default: fprintf(stderr,"What the heck is item %d? - exit...\n", menuItem );
 }
}

static char sl[] =
     "-----------------";


void normMain()
{
 int i, d; char *str = NULL;

 set_all_samples_train();
 done = FALSE;
 do
 {
   printf("\n%s  D A T A  N O R M A L I Z A T I O N   %s\n", sl, sl );
   for( i = 0; i < NUM_NORMS; i++ )
   {
     str = norm_methods[i].name;
     printf("(%d)\t%s\n", i+1, str );
   }

   printf("Choice: "); gets(buf);
   switch( buf[0] )
   {
     /* special cases */
     case '?': help( MENU_NORM, buf ); break;
     case '\0': done = TRUE; showUniv( stdout ); break;
     default:
       sscanf(buf,"%d", &d);
       if( d >= 1 && d <= NUM_NORMS )
       {
         menuItem = d-1;
         execMenuItem();
	 done = TRUE;
       }
   }
 }
 while( ! done );
}


void normalize()
{
 /* do only normalize if not yet done */
 if( norm_method != UNNORMED )
 {
   printf("Data is already normalized!"); WAIT;
   return;
 }
 if( U->nrClass >= 1 )
   do
   {
     normMain();
   }
   while( !done );
 else
   { printf("Please load universe first !...");gets( buf );}
}


/**********  H E L P  *****************/

void helpMENU_NORM( cmd )
char *cmd;
{
 int item;

 sscanf( &(cmd[1]), "%d", &item );
 item--;	/* discount one since in the menu it is one more */
 switch( item )
 {
   case NORM_0_1:	item_help("NORM_0_1");  break;
   case NORM_M0S1:	item_help("NORM_M0S1"); break;
   default: ; break;
 }
}
