/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:26 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "def.h"

extern universe *U;
extern str100 inDataDir;
extern str100 dataFile;
extern str100 featNameFile;
extern str100 a_prioriFile;
extern str100 sfName;
extern int feat_dim;
extern boolean loadUnivFile(), loadUnivDir();


boolean verbose = FALSE, feat_description = FALSE;
FILE *infile = NULL;

static boolean load = FALSE;
static char *in_data_dir = NULL;
static char *in_data_file = NULL;
static char *sel_file = NULL;
static char *a_priori_file = NULL;
static char *feat_name_file = NULL;
static char *feat_dim_str = NULL;
static boolean *visited = NULL;  /* register all pararamters that were detected */

#define ALWAYS 1
#define OPTION 0

#define IN_DATA_DIR	"-dir"		/* name of input data directory */
#define IN_DATA_FILE	"-file"		/* name of input data file */
#define SEL_FILE	"-sel"	/* name of file with selected features */
#define A_PRIORI_FILE	"-app"	/* name of file with a priori probabilities */
#define VERBOSE		"-v"
#define FEAT_NAME	"-fnam"		/* name of the feature description */
#define FEAT_DIM	"-dim"		/* optional number of features */


#define MAXSTRLEN 100
static void chkStrLen( str, len )
char *str;
int len;
{
 if( strlen(str) > len )
   fprintf(stderr,
	"WARNING: String %s has been truncated to maximum length %d!\n",
	str, len );
}


static char *parse_input_param( argc, argv, param, when )
int argc;
char **argv;
char *param;
int when;
{
 int i = 0;

 while ((i < argc) && (strcmp(param, argv[i])))
   i++;
 if (i < argc-1) {
   visited[i] = TRUE;
   if( argv[i+1][0] != '-' )
     visited[i+1] = TRUE;
   else
   {
     fprintf( stderr,
	 "Could not find the argument for parameter %s - exit...\n", param );
     exit( 1 );
   }
   return(argv[i+1]);
 }
 else	/* Did not find parameter */
 {
   if (when == ALWAYS)
   {
     fprintf( stderr, "Can't find asked option %s", param );
     exit(1);
   }
 }
 return((char *) NULL);
}


static boolean find_input_param( argc, argv, param, when )
int argc;
char **argv;
char *param;
int when;
{
 boolean found = FALSE;
 int i = 0;

 while( ! found && i < argc )
 {
   found = (strcmp(argv[i],param)==0);
   if( ! found )
     i++;
 }
 if (when == ALWAYS)
 {
   fprintf( stderr, "Can't find asked option %s", param );
   exit(1);
 }
 if( found )
   visited[ i ] = TRUE;
 return( found );
} 


static void usage( argc, argv )
int argc;
char **argv;
{
 fprintf( stderr, "usage: %s [-dir <data-directory> |", argv[0] );
 fprintf( stderr, " -file <data-file> [-dim <# of features>]]\n" );
 fprintf( stderr, "\t[-sel <selected-features-file>] [-v]");
 fprintf( stderr, " [-fnam <feature-names-file>]\n" );
 fprintf( stderr, "\t[-app <a_priori_probabilities-file>]\n\n" ); WAIT;
}


void parse_input( argc, argv )
int argc;
char **argv;
{
 int i;

 visited = (boolean*) malloc( argc*sizeof(boolean) );
 for( i = 0; i < argc; i++ )
   visited[i] = FALSE;
 visited[ 0 ] = TRUE;	/* argv[0] */

 if( argc == 1 )
   usage( argc, argv );
 verbose = find_input_param( argc, argv, VERBOSE, OPTION );
 in_data_dir = parse_input_param( argc, argv, IN_DATA_DIR, OPTION );
 in_data_file = parse_input_param( argc, argv, IN_DATA_FILE, OPTION );
 if( in_data_dir != NULL && in_data_file != NULL )
 {
   fprintf( stderr, "Choose only one option -file or -dir !; Exitus...\n");
   exit(1);
 }

 if( in_data_dir != NULL )
 {
   chkStrLen( in_data_dir, MAXSTRLEN );
   strncpy( inDataDir, in_data_dir, MAXSTRLEN );
   if( verbose )
     printf("Loading data files...\n");
   load = loadUnivDir();
 }

 if( in_data_file  != NULL )
 {
   chkStrLen( in_data_file, MAXSTRLEN );
   strncpy( dataFile, in_data_file, MAXSTRLEN );

   /* look for dimension option */
   feat_dim_str = parse_input_param( argc, argv, FEAT_DIM, OPTION );
   if( feat_dim_str != 0 )
   {
     sscanf( feat_dim_str, "%d", &feat_dim );
     if( feat_dim < 1 || feat_dim > MAXFEAT )
     {
       printf("Wrong dimension %d for option \"%s\" --- ignoring",
	feat_dim, FEAT_DIM ); WAIT;
       feat_dim = EMPTY;
     }
   }
   load = loadUnivFile();
 }

 if( load )
 {
   feat_name_file = parse_input_param( argc, argv, FEAT_NAME, OPTION );
   if( feat_name_file != 0 && !(strcmp(U->name,"UNDEFINED")==0) )
   {
     chkStrLen( feat_name_file, MAXSTRLEN );
     strncpy( featNameFile, feat_name_file, MAXSTRLEN );
     load_feat_names( featNameFile );
   }
   sel_file = parse_input_param( argc, argv, SEL_FILE, OPTION );
   if( sel_file != 0 && !(strcmp(U->name,"UNDEFINED")==0) )
   {
     chkStrLen( sel_file, MAXSTRLEN );
     if( strcmp(sel_file,"all")==0 )
       set_all_selected();
     else
     {
       strncpy( sfName, sel_file, MAXSTRLEN );
       load_sel( sfName );
     }
   }
   else
     set_all_selected(); /* Set by default all features as selected */

   a_priori_file = parse_input_param( argc, argv, A_PRIORI_FILE, OPTION );
   if( a_priori_file != 0 && !(strcmp(U->name,"UNDEFINED")==0) )
   {
     chkStrLen( a_priori_file, MAXSTRLEN );
     strncpy( a_prioriFile, a_priori_file, MAXSTRLEN );
     load_a_priori_probs( a_prioriFile );
   }
 }

 showUniv( stdout );
 /* check if all parameters and arguments were visited */
 for( i = 0; i < argc; i++ )
   if( ! visited[i] )
   {
     fprintf(stderr,"Invalid option \"%s\" or missing parameter - Exit...\n",
		argv[i] );
     FREE( visited );
     exit(1);
   }
 FREE( visited );
}
