/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:26 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "def.h"
#include "clasdef.h"
#include "matrxdef.h"

extern universe *U;
extern boolean *train_smp, *test_smp;
extern int nrSamplesTest, nrSamplesTrain;
extern boolean verbose_reason;
extern float Euclidian_Distance();
extern float *vector();

extern FILE *cpf;               /* Classification protocol file */
extern boolean have_label;      /* Only deliver the classified class for
        pre-classified data */


static boolean interact = FALSE;
static boolean verbose_induction;
static boolean cl_dep_window_width = FALSE;

static int *train_class_nr = NULL;
static float *Pn = NULL;	/* Approximated a posteriori probabilities */
static float *cov_scale = NULL;
static float *window_width = NULL;
static float *test_val = NULL;
static float width_per_cent = 15.0;

#undef PI
#undef PI_2
#define PI	(3.1415926535897931160)
#define PI_2	(6.2831853071795862320)

/* ##### KERNEL TYPES */
#define HYPERCUBIC	0
#define HYPERSPHERIC	1
#define HYPERTRIANGLE	2
#define GAUSSIAN	3
#define EXPONENTIAL	4
#define LORENZ		5

#define NUM_KERNEL_TYPES 6

static int kernel_type = GAUSSIAN;	/* default */

static struct kernels {
	int index;
	str80 name;
} kernel[] = {
	{ HYPERCUBIC,		"Hypercubic" },
	{ HYPERSPHERIC,		"Hyperspheric" },
	{ HYPERTRIANGLE,	"Hypertriangle" },
	{ GAUSSIAN,		"Gaussian" },
	{ EXPONENTIAL,		"Exponential" },
	{ LORENZ,		"Lorenz" }
};

/* ************************************************************ */
static void get_kernel_type()
{
 int i, k; boolean chosen = FALSE;
 str80 buf;

 do
 {
   printf("KERNEL TYPE\n" );
   for( i = 0; i < NUM_KERNEL_TYPES; i++ )
   {
     printf("(%d) %s", i+1, kernel[i].name );
     if( i == kernel_type ) printf("\t\t<--- Current");
     printf("\n");
   }
   printf("Choice: "); gets(buf);
   switch( buf[0] )
   {
     case '?': printf("No help available yet"); WAIT; break;
     /* case '?': help( MENU_FEATSEL_CRIT_GROUP, buf ); break;	/**/
     case '\0':
       chosen = TRUE;
       break;
     default:
       sscanf(buf,"%d", &k);
       if( k >= 1 && k <= NUM_KERNEL_TYPES )
       {
         chosen = TRUE; kernel_type = k-1;
       }
       else
       {
         printf("Invalid kernel for Parzen window! Specify again\n");
       }
   }
 }
 while( ! chosen );
}

/* ******************************************************** */
static void gen_matrix(FSV, len)
FeatSelectVector FSV;
int len;
{
FeatVector smp = NULL;
float *mean = NULL;
float **diff_mean = NULL;
int ii, i, j, k, s_1, s_2, count_help, train_vec;
float Euclid_max, diff, new_width;

smp = vector( len );
mean = vector( len );
cov_scale = vector( U->nrClass );

s_1 = s_2 = 0;
for (i = 0; i < U->nrClass; i++)            
{
  count_help = 0;
  for (k = 0; k < len; k++)
    mean[k] = 0.0;
  cov_scale[i] = 0.0;
  diff_mean = (float **) malloc(len * sizeof(float *));
  for (j = 0; j < U->C[i].numSampl; j++) 
    {
    if( train_smp[s_1] )
      {
      provideFV( smp, i, j, len, FSV );
      for (k = 0; k < len; k++)    
        mean[k] = mean[k] + smp[k];
      count_help++;
      }
    s_1++; 
    }
  for (k = 0; k < len; k++)
    mean[k] = mean[k]/(float)count_help; 
  train_vec = s_1;       /* if count_help -> problems if Hold_out */ 
  for (k = 0; k < len; k++)
    {
    s_2 = 0;
    count_help = 0;
    diff_mean[k] = vector(train_vec);
    for (j = 0; j < U->C[i].numSampl; j++) 
      {
      if( train_smp[s_2] )
        {
        provideFV( smp, i, j, len, FSV );
        diff_mean[k][count_help] = (mean[k] - smp[k]);
        diff_mean[k][count_help] = 
		    diff_mean[k][count_help] * diff_mean[k][count_help];
        cov_scale[i] = cov_scale[i] + diff_mean[k][count_help];
	count_help++;
	}
      s_2++;
      }
    }
  cov_scale[i] = cov_scale[i]/(float)(train_class_nr[i]-1);
  /* printf("cov_scale[%d]=%f", i, cov_scale[i]); DBG; /**/
  for (k = 0; k < len; k++)
    FREE(diff_mean[k]);
  FREE(diff_mean);
}

s_1 = s_2 = 0;
for (i = 0; i < U->nrClass; i++)
  {
  Euclid_max = -INFINITY;
  for (j = 0; j < U->C[i].numSampl; j++)
    {
    if( train_smp[s_1] )
      {
      provideFV( smp, i, j, len, FSV );
      for (k = 0; k < len; k++)
	test_val[k] = smp[k];
      for (ii = 0; ii < U->C[i].numSampl; ii++)
        {
	if( train_smp[s_2] )
	  {
          provideFV( smp, i, ii, len, FSV );
	  diff = Euclidian_Distance(smp, test_val, len);
          if (diff >= Euclid_max)
            Euclid_max = diff;
	  }
	s_2++;
	}
      s_2 = 0; 
      }
    s_1++;
    }
  window_width[i] = Euclid_max * width_per_cent * 0.01; 
  }
if (!cl_dep_window_width)
  {
  new_width = 0.0;
  for (i = 0; i < U->nrClass; i++)
    if (window_width[i] > new_width)
      new_width = window_width[i];
  for (i = 0; i < U->nrClass; i++)
    window_width[i] = new_width; 
  }

if (verbose_induction)
  for (i = 0; i < U->nrClass; i++)
    printf("Window width for class %3d(%s) = %f\n",
	i+1, U->C[i].name, window_width[i] );

FREE(mean);
FREE(smp);
}

/* ******************************************************** */
void show_PARZEN( f )
FILE *f;
{
  fprintf(f,"\tParzen Window Kernel Type: %s\n", kernel[kernel_type].name );
  fprintf(f,"\tWindow width (as a fraction of total range of feature values)");
  fprintf(f,": %6.2f%%\n", width_per_cent);
}

/* ********************************************************* */
static float cheb_dist(v1, v2, dim)
float *v1, *v2;
int dim;
{
 float diff, max;
 int i;

 max = -INFINITY;
 for (i = 0; i < dim; i++)
 {
  diff = (v1[i] - v2[i])*(v1[i] - v2[i]);
  diff = (float)sqrt((double)diff);   
  if (diff > max)
    max = diff;
 }
 return(max);
}

/* ********************************************************* */
static float Gauss (v1, v2, dim, hn, Q)
float *v1, *v2;
int dim;
float hn, Q;
{
 float density, dist;
 int i;
 float factor_1, factor_2;

 factor_1 = 1.0;
 i = 1;
 do
 {
   factor_1 = factor_1 * PI_2 * hn * hn;
   i = i + 1;
 } while (dim >= i);
 factor_1 = 1.0/( (float) sqrt( (double) (Q*factor_1)));
 factor_2 = (-1.0) * 1.0/(2.0 * hn * hn);
 dist = Euclidian_Distance (v1, v2, dim);
 density = factor_1 * (float)exp((double)(factor_2 * dist));
 /* printf("Gauss: density=%f\n", density ); DBG	/**/
 return(density);
}

/* ********************************************************* */
static float Lorenz(v1, v2, dim, hn)
float *v1, *v2;
int dim;
float hn;
{
 float density;

 density = Euclidian_Distance (v1, v2, dim);
 /*density = PI/(1+(density*density));*/
 density = (PI/hn)/(1.0 + ((density*density)/(hn*hn)));
 return(density);
}

/* ********************************************************* */
static float Exponential(v1, v2, dim, hn)
float *v1, *v2;
int dim;
float hn;
{
 float density;

 density = Euclidian_Distance (v1, v2, dim);
 density = (1.0/(2.0*hn))*(float)exp((double) ((-1.0)*(density/(hn*hn))));
 return(density);
}

/* ********************************************************* */
static float Hypertriangle(v1, v2, dim, hn)
float *v1, *v2;
int dim;
float hn;
{
 float density;

 density = Euclidian_Distance (v1, v2, dim);
 if (density > hn)
   density = 0.0;
 else
   density = 4.0*(hn-density);
 return(density);
}

/* ********************************************************* */
void initPARZEN()
{
 free_induced();
 printf(" #### Specifications of the Parzen window classifier:\n\n");
 get_kernel_type();
 printf("Class-dependent window width?");
 get_boolean(TRUE, &cl_dep_window_width);
printf("Window width (in percent of total feature range)?  %6.2f\b\b\b\b\b\b\b",
	width_per_cent);
 get_f_range(&width_per_cent, 0.01, 100.0, LEFT_CLOSED__RIGHT_CLOSED);
}

/* *********************************************************
 * Induce a classifier with the actual set of training samples
 * The dimension of the vector is 'nsf'
 * 'FSV' contains the information of the actual subset of features
 * that are used.
 * 'iact' tells if interaction if desired. For instance in error
 * estimation this is not the case
 * ********************************************************* */
void inducePARZEN( nsf, FSV, iact )  
int nsf;                
FeatSelectVector FSV;  
boolean iact;            
{
 int i, j, s;
 FeatVector smp = NULL;
 interact = iact;
 if( interact )
   {
   printf("\n --- Inducing a %d-variate Parzen Classifier ---\n\n", nsf);
   printf("Verbose learning of decision structure?");
   get_boolean( FALSE, &verbose_induction );
   }
 
 smp = vector (nsf); 
 train_class_nr = (int *) malloc(U->nrClass * sizeof(int));
 CHKPTR(train_class_nr);
 Pn = vector(U->nrClass);
 test_val = vector( nsf );
 window_width = vector(U->nrClass);

 /* ***************** HOW MANY TRAINING VECTORS ********************* */
 s = 0;
 for( i = 0; i < U->nrClass; i++ )
   {
   train_class_nr[i] = 0;
   for( j = 0; j < U->C[i].numSampl; j++ )
     {
     if( train_smp[s] )
       {
       provideFV( smp, i, j, nsf, FSV );
       train_class_nr[i]++;
       }
     s++;
     }
   if (train_class_nr[i] == 0)
     {
     printf("for the %dth class ", i+1);
     EXIT("no training vectors");
     }
   else
     if (verbose_induction)
       printf("%d training vectors for class %d(%s)\n",
		train_class_nr[i], i+1, U->C[i].name);
   }
 
  gen_matrix(FSV, nsf);
  verbose_induction = FALSE;
  FREE(smp);
}


/*  *******************************************************
 *  Free all the structures that were dynamically allocated to store
 *  the classifier, for instance matrices of prototypes
 *  ******************************************************* */
void free_inducedPARZEN()
{
  FREE(cov_scale);
  FREE(train_class_nr);
  FREE(window_width);
  FREE(Pn);
  FREE(test_val);
}

/*  ******************************************************
 *  Classify a SINGLE sample 'smp' of LENGTH 'nsf'
 *  The name of the expected class is stored in 'name'
 *  'FSV' is not needed here, but for compatibility reasons passed
 *  ******************************************************* */
int class1testPARZEN( nsf, FSV, smp, name )
int nsf;
FeatSelectVector FSV;
FeatVector smp;
char *name;
{ 
 int k, j, i, s, resulting_class, best_class;
 float diff;
 int *Ki = NULL;
 float px;			/* unconditional density p(x) */
 float *px_joint = NULL;	/* joint prob. dens. of class i and sample x */
 float phi_s;			/* value of density function */
 float apriori;			/* A priori probability for class i */
 int K;
 float *smp_train = NULL;

 smp_train = vector(nsf);
 px_joint = vector(U->nrClass);
 Ki = (int *) malloc(U->nrClass * sizeof(int)); 
 CHKPTR(Ki);

 /* For the kernels with limited support we only COUNT the samples:
    K : all samples within the window
    Ki: samples that belong to class i
    The approxiamted a posteriori probability is then Ki / K
    see Duda & Hart, 1973 pp. 97
 */
 if ( (kernel_type == HYPERSPHERIC) || (kernel_type == HYPERCUBIC) )
   {
   K = 0;
   s = 0;
   for (i = 0; i < U->nrClass; i++)
     {
     Ki[i] = 0;
     for (j = 0; j < U->C[i].numSampl; j++)
       {
       if (train_smp[s])
         {
	 provideFV( smp_train, i, j, nsf, FSV );
	 for (k = 0; k < nsf; k++)
	   test_val[k] = smp_train[k]; /* every single training vector */
         switch (kernel_type) 
	   {
           case HYPERCUBIC:
	     diff = cheb_dist(smp, test_val, nsf);
	     break;
           case HYPERSPHERIC:
	     diff = Euclidian_Distance(smp, test_val, nsf);
	     break;
	   default:
	     EXIT("unknown kernel type\n");
	   }
         /* Found a sample within the window? */
         if (diff <= window_width[i])
           (Ki[i])++;
         }
       s++;
       }
     }
   for (i = 0; i < U->nrClass; i++)
     K = K + Ki[i];
   for (i = 0; i < U->nrClass; i++)
     {
     if (K != 0)
       Pn[i] = (float)Ki[i] / (float)K;
     else
       Pn[i] = 0.0;
     }
   }
 /* For real kernels we have to calculate the density function explicitly
    1.) The class conditional density
    2.) The unconditional density
 */
 else
 {
   px = 0.0;
   s = 0;
   for (i = 0; i < U->nrClass; i++)
   {
     apriori = U->C[i].a_priori_prob;
     px_joint[i] = 0.0;
     for (j = 0; j < U->C[i].numSampl; j++)
     {
       if (train_smp[s])
       {
         provideFV( smp_train, i, j, nsf, FSV );
	 for (k = 0; k < nsf; k++)
	   test_val[k] = smp_train[k]; /* every single training vector */ 
         switch (kernel_type)
	 {
	   case HYPERTRIANGLE:
	     phi_s = apriori *
	       Hypertriangle(smp, test_val, nsf, window_width[i]);
	     break;
           case GAUSSIAN:
	     phi_s = apriori *
		  Gauss(smp, test_val, nsf, window_width[i], cov_scale[i]);
             break;
           case EXPONENTIAL:
	     phi_s = apriori *
	          Exponential(smp, test_val, nsf, window_width[i]);
	     break;
           case LORENZ:
	     phi_s = apriori *
	          Lorenz(smp, test_val, nsf, window_width[i]);
	     break;
           default:
	     EXIT("unknown kernel type\n");
	 }
         px_joint[i] += phi_s;
       }
       s++;
     }
     px += px_joint[i];
   }
   for(i = 0; i < U->nrClass; i++)
   {
     if (px > 0.000000000000000001)
       Pn[i] = px_joint[i] / px;
     else
       Pn[i] = 0.0;
   }
 }
 if( verbose_reason )
 {
   fprintf(cpf,"\n\n ---- DECISION FUNCTIONS FOR UNKNOWN SAMPLE:\n");
   fshowFV( cpf, nsf, smp );
   if ( (kernel_type == HYPERSPHERIC) || (kernel_type == HYPERCUBIC) )
   {
     fprintf(cpf,"Limited kernels: Samples within the window = %d\n", K );
     for(i = 0; i < U->nrClass; i++)
     {
       fprintf(cpf,"  Samples of class %d(%s) within the window = %d\n",
		i+1, U->C[i].name, Ki[i] );
       fprintf(cpf,"    -> A posteriori prob = %d/%d = %.5f\n",Ki[i],K, Pn[i] );
     }
   }
   else
   {
     fprintf(cpf,"Kernel extends over the whole pattern space.\n");
     fprintf(cpf,"  Estimated density p(x) of the sample = %.5f\n", px );
     for(i = 0; i < U->nrClass; i++)
     {
       fprintf(cpf,"Joint prob. density for class %d(%s) = %.5f\n",
		i+1, U->C[i].name, px_joint[i] );
       fprintf(cpf," -> A posteriori prob. = %.5f/%.5f = %.5f\n",
		px_joint[i], px, Pn[i] );
     }
   }
   fprintf(cpf,"\n");
 }
     
 /* Look for the maximum a posteriori probability */
 for (i = 0;  i < U->nrClass; i++ )
   {
   if (i == 0)
     best_class = 0;
   if ( i > 0)
     {
     if (Pn[i] > Pn[best_class])
       best_class = i; 
     }
   }
 
 FREE(px_joint);
 FREE(Ki);
 FREE(smp_train);

 if( ! have_label )
   return( best_class );


 /* 
    First case: correct classification:
    The name of the best class corresponds to the expected class
    return the value 'CORRECT_CLASS'
 */
 if( (strcmp(U->C[best_class].name,name)==0) )
   resulting_class = CORRECT_CLASS;
 else
   {
   /* 
      Second case: wrong classification:
      return the value of the best class: 'best_class'
   */
   resulting_class = best_class;
   if( verbose_reason )
     fprintf( cpf, " *** Confused %s with expected class %s\n",
       U->C[best_class].name, name );
   }
 return( resulting_class );
}
