/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:27 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "def.h"
#include "clasdef.h"

extern float Euclidian_Distance();

extern universe *U;
extern boolean verbose, verbose_reason;
extern str100 outDataDir;
extern boolean *train_smp, *test_smp;
extern int nrSamplesTest, nrSamplesTrain;
extern boolean feat_description;
extern char **feature_desc;
extern float *vector();
extern float **matrix_alloc();

extern FILE *cpf;               /* Classification protocol file */
extern boolean have_label;      /* Only deliver the classified class for
        pre-classified data */


static boolean interact = TRUE;
static boolean first_prototype_random = FALSE,  purge = FALSE;
static str80 buf, pfName;
static char wheel[] = "|/-\\";
static int nrFeat;

/* ---------  UPDATE MODELS FOR THE Q* -ALGORITHM  -----------*/
#define UPDATE_MEAN 0
#define UPDATE_MARGINAL_MEDIAN 1
#define UPDATE_VECTOR_MEDIAN 2


typedef struct ProType_ {
          FeatVector value;
          boolean outlier;
          int nrFriends;
          int *friends;
        } ProType;

typedef struct ProtypeClass_ {
          str100 name;
          int numProtype;
          ProType *PT;
        } ProtypeClass,  *ProtypeClasses;

#define EPSILON (0.0001)  /**/
#define UNDEFINED (-1)

static ProtypeClasses PTC = NULL;
static int sumProtype = 0;
static ProType *PurgedPT = NULL;
static ProType allPT = { NULL, FALSE, 0, NULL };
static FeatVector Samples = NULL;
static int *train_of_class = NULL;

static char bn[2][4] = {"NO", "YES"};

static struct updates {
        int index;
        str80 name;
} update[] = {
        { UPDATE_MEAN,			"MEAN" },
        { UPDATE_MARGINAL_MEDIAN,	"MARGINAL MEDIAN" },
        { UPDATE_VECTOR_MEDIAN,		"VECTOR MEDIAN" }
};

static int update_method = UPDATE_MEAN;	/**/
/* static int update_method = UPDATE_MARGINAL_MEDIAN;	/**/
/* static int update_method = UPDATE_VECTOR_MEDIAN;	/**/


void show_Q_STAR( f )
FILE *f;
{
 fprintf( f, "\tUPDATE METHOD:\t%s\n", update[update_method].name );
 fprintf( f, "\tInitialize first prototype of each class randomly: %s\n",
        bn[first_prototype_random] );
 fprintf( f, "\tPurge all single sample prototypes: %s\n", bn[purge] );
}


void initQ_STAR()
{
 free_induced();
 update_method++;
 /* Ask for the update method */
 printf("\n #### Q* -algorithm specifications: ####\n");
 printf(" ---- Update rule for the prototype ----\n");
 printf("\nUpdate as the MEAN(1), MARGINAL MEDIAN(2) or VECTOR MEDIAN(3)\n");
 printf("of the correctly classified neighbors? %d\b", update_method );
 get_d_range( &update_method, UPDATE_MEAN+1,
	UPDATE_VECTOR_MEDIAN+1, LEFT_CLOSED__RIGHT_CLOSED );
 update_method--;

 printf("Initialize first prototype of each class randomly?");
 get_boolean( first_prototype_random, &first_prototype_random );
 printf("Purge all single sample prototypes?");
 get_boolean( purge, &purge );
 NL;
}


static void init_ProType( PT )
ProType *PT;
{
 PT->value = NULL;
 PT->outlier = TRUE;

 PT->nrFriends = 0;
 PT->friends = NULL;
}


static void free_ProType( PT )
ProType *PT;
{
 FREE( PT->value );
 PT->outlier = TRUE;

 PT->nrFriends = 0;
 FREE( PT->friends );
}


static void join_friend( fr, PT )
int fr;
ProType *PT;
{
 int i, *newFriends = NULL;

 /* if( fr >= nrSamplesTrain ) EXIT("row too high");	/**/

 newFriends = (int*) malloc( (1+PT->nrFriends) * sizeof(int) );
 CHKPTR( newFriends );
 for( i = 0; i < PT->nrFriends; i++ )
   newFriends[ i ] = PT->friends[i];
 newFriends[ PT->nrFriends ] = fr;
 (PT->nrFriends)++;
 FREE( PT->friends );
 PT->friends = newFriends;
}


static void show_friends( PT )
ProType *PT;
{
 int i;

 printf("Prototype has %d friends:\n\t( ", PT->nrFriends ); 
 for( i = 0; i < PT->nrFriends; i++ )
   printf("%d ", PT->friends[i] );
 printf(")\n\n");
}


static void copy_PT( PTsrc, PTdest, dim )
ProType *PTsrc, *PTdest;
int dim;
{
 int i;

 copy_FV( PTsrc->value, PTdest->value, dim ); 
 PTdest->outlier = PTsrc->outlier;

 PTdest->nrFriends = PTsrc->nrFriends;
 if( PTsrc->nrFriends > 0 )
 {
   FREE( PTdest->friends );
   PTdest->friends = (int*) malloc( PTdest->nrFriends * sizeof(int) );
   CHKPTR( PTdest->friends );
   for( i = 0; i < PTdest->nrFriends; i++ )
     PTdest->friends[i] = PTsrc->friends[i];
 }
} 


static void free_PTC()
{
 int i, p;

 if( PTC == NULL )
   return;
 for( i = 0; i < U->nrClass; i++ )
 {
   for( p = 0; p < PTC[i].numProtype; p++ )
     free_ProType( &(PTC[i].PT[p]) );
   PTC[i].numProtype = 0;
   FREE( PTC[i].PT );
 }
 FREE( PTC );
 sumProtype = 0;
}


static void init_PTC()
{
 int i, clss;

 if( PTC != NULL )
   free_PTC();
 PTC = (ProtypeClass*) malloc( U->nrClass * sizeof(struct ProtypeClass_) );
 CHKPTR(PTC);
 for( i = 0; i < U->nrClass; i++ )
 {
   clss = i;
   strcpy( PTC[i].name, U->C[clss].name );
   PTC[i].numProtype = 0;
   PTC[i].PT = NULL;
 }
}


static void gen_selected_samples( nrSelFeat, FSV, verbose )
int nrSelFeat;
FeatSelectVector FSV;
boolean verbose;
{
 int i, j, k, s;
 int row, nrSmp;

 Samples = vector( nrSamplesTrain*nrSelFeat );
 train_of_class = (int*) malloc( U->nrClass * sizeof(int) );
 CHKPTR(train_of_class);

 /* copy the global sample values to the local variable */

 row = 0; s = 0;
 if( verbose )
   printf(" S A M P L E S:\n");
 for( i = 0; i < U->nrClass; i++ )
 {
   train_of_class[i] = 0;
   nrSmp = U->C[i].numSampl;
   if( verbose )
     printf("Class: %d = %s - Nr. Samples: %d\n",i+1,U->C[i].name,nrSmp);
   for( j = 0; j < nrSmp; j++ )
   {
     if( train_smp[s] )
     {
       (train_of_class[i])++;
       if( verbose )
         printf(" Nr. %3d", row );
       for( k = 0; k < nrSelFeat; k++ )
         Samples[ row*nrSelFeat+k ] =
               U->C[i].S[ U->nrFeat * j + FSV[k].rank ];
       if( verbose )
         showFV( nrSelFeat, &(Samples[ row*nrSelFeat ]) ); /**/
       row++;
     }
     s++;
   }
 }
 /* printf("#samples train: %d\n", nrSamplesTrain);
 for( i = 0; i < U->nrClass; i++ )
   printf("class %d has %d training samples\n", i+1, train_of_class[i] ); DBG;*/
}


/*
 *  Calulate the performance index of a clustering.
 *  It is based on the sum of the squared errors
*/
static float performance_index( nrSelFeat )
int nrSelFeat;
{
 int i, j, p, row, s;
 float dist, minDist, sumDist = 0.0;
 FeatVector Sample = NULL;

 /* for all samples of all classes calculate the distance to
    their optimal cluster center
 */
 row = 0; s = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     if( train_smp[s] )
     {
       Sample = &(Samples[nrSelFeat*row]);
       minDist = INFINITY;
       for( p = 0; p < PTC[i].numProtype; p++ )
       {
         dist = Euclidian_Distance( PTC[i].PT[p].value, Sample, nrSelFeat );
         /* printf("Euclidian_Distance=%f between:\n", dist );
         showFV( nrSelFeat, PTC[i].PT[p].value ); showFV(nrSelFeat,Sample); /**/
         if( dist < minDist )
         {
           minDist = dist;
         }
       }
       sumDist += minDist;
       row++;
     }
     s++;
   }
 }
 return( sumDist );
}

 
static void show_PTC( brief, showFriends, nrSelFeat )
boolean brief, showFriends;
int nrSelFeat;
{
 int i, p;

 if( nrSelFeat == 0 || PTC == NULL ) return;
 if( ! brief )
 {
   printf("\n-------------- %d  P R O T O T Y P E S  -------------------\n",
	sumProtype );
   for( i = 0; i < U->nrClass; i++ )
   {
     printf("%4d prototypes for class: %2d = %s\n",
                 PTC[i].numProtype, i+1, U->C[i].name );
     for( p = 0; p < PTC[i].numProtype; p++ )
     {
       printf(" Nr. %3d: ", p+1 );
       if( PTC[i].PT[p].outlier )
         printf("Outlier!");
       else
         printf("        ");
       showFV( nrSelFeat, PTC[i].PT[p].value );
       if( showFriends )
         show_friends( &(PTC[i].PT[p]) );
     }
   }
   printf("PERFORMANCE INDEX:  %20.10f\n", performance_index( nrSelFeat ) );
   printf("  (sum of squared errors of the distances of the\n");
   printf("   samples from their nearest prototype)\n" );
   printf("-------------------------------------------------------\n");
 }
 else
 {
   for( i = 0; i < U->nrClass; i++ )
     printf("C%d:%d ",i+1, PTC[i].numProtype );
   NL;
 }
}


static void purge_outliers( nrSelFeat )
int nrSelFeat;
{
 int i, p, numOutliers, *outlierInd = NULL, newPtr;

 if( interact && verbose )
   printf(" --- Purging outliers ---\n");
 for( i = 0; i < U->nrClass; i++ )
 {
   if( interact && verbose )
     printf(" %4d prototypes for class: %2d = %s\n",
                 PTC[i].numProtype, i+1, U->C[i].name );
   numOutliers = 0;
   outlierInd = (int*) malloc( PTC[i].numProtype * sizeof(int) );
   CHKPTR(outlierInd);

   for( p = 0; p < PTC[i].numProtype; p++ )
   {
     outlierInd[p] = FALSE;
     if( verbose && interact )
     {
       printf("Prototype Nr. %3d: ", p+1 );
       showFV( nrSelFeat, PTC[i].PT[p].value );
     }
     if( PTC[i].PT[p].outlier )
     {
       outlierInd[p] = TRUE;
       numOutliers++;
     }
   }
   /* ouliers are found : purge them from the prototypes */
   /* purge only if the class has at least one prototypes that is no outlier */
   if( numOutliers > 0 && (PTC[i].numProtype > numOutliers) )
   {
     if( interact && verbose )
       printf("\t*** Found %d outliers for class %d\n", numOutliers, i+1 );
     PurgedPT = (ProType*)
       malloc((PTC[i].numProtype-numOutliers)*sizeof(ProType));CHKPTR(PurgedPT);

     newPtr = 0;
     for( p = 0; p < PTC[i].numProtype; p++ )
     {
       if( outlierInd[p] == FALSE )
       {
         /* no outlier: copy */
         init_ProType( &(PurgedPT[newPtr]) );
         PurgedPT[newPtr].value =
		(FeatVector)malloc(nrSelFeat*sizeof(FeatVector*));
         CHKPTR(PurgedPT[newPtr].value);
         copy_PT( &(PTC[i].PT[p]), &(PurgedPT[newPtr]), nrSelFeat );
         newPtr++;
       }
     }
     /* change and free the pointers */
     for( p = 0; p < PTC[i].numProtype; p++ )
       free_ProType( &(PTC[i].PT[p]) );
     FREE( PTC[i].PT );
     PTC[i].numProtype -= numOutliers; sumProtype -= numOutliers;
     PTC[i].PT = PurgedPT;
     PurgedPT = NULL;
   }
   FREE( outlierInd );
 }
}


static void calc_mean( mean, PT )
FeatVector mean;
ProType *PT;
{
 int k, i;

 for( k = 0; k < nrFeat; k++ )
 {
   mean[k] = 0.0;
   for( i = 0; i < PT->nrFriends; i++ )
   {
     mean[k] += Samples[ nrFeat * PT->friends[i] + k ];
   }
   mean[k] /= PT->nrFriends;
 }
}


/* make it compatible to all compilers */
/* Note: this routine sorts in descending order */
static int order_rule( real1, real2 )
#ifdef CONSTANT
const
#endif
void *real1, *real2;
{
 float *r1, *r2;

 r1 = (float*) real1; r2 = (float*) real2;
 if( *r1 > *r2 ) return((int)1);
 if( *r1 < *r2 ) return((int)-1);
 return((int)0);
}


/* #define PEEP_MED 10		/**/
/*
 * Calculate the Vector median
 * This seems to be a quite new approach from Tampere University.
 * The vector median of n multidimensional samples is the sample 
 * with the minimum sum of Euclidean distances to the other samples. 
 * So, for every sample one has to compute the sum of distances to 
 * the other samples and then to choose the minimum. 
*/
static void calc_VectorMedian( median, PT )
FeatVector median;
ProType *PT;
{
 int i, j, k, nrf, minFriend;
 FeatVector EuclidDistsSum = NULL;
 float dist, minDist;

 nrf =  PT->nrFriends;

 /* Two special cases: Only one or two friends */
 if( nrf == 1 )
 {
   copy_FV( &(Samples[nrFeat*PT->friends[0]]), median, nrFeat );
#ifdef PEEP_MED
   printf("\n --- Only 1 friend: MEDIAN[%d] =", PT->friends[0] );
   showFV( nrFeat, median );
   printf("Friend:\n");
   showFV( nrFeat, &(Samples[nrFeat*PT->friends[0]]) ); DBG;
#endif
   return; 
 }
 /* Take the mean of the two friends */
 if( nrf == 2 )
 {
   for( k = 0; k < nrFeat; k++ )
     median[k] = ( Samples[nrFeat*PT->friends[0]+k] +
             Samples[nrFeat*PT->friends[1]+k] ) / 2.0;
#ifdef PEEP_MED
   printf("\n --- Only 2 friends: MEDIAN[%d,%d] =",
		 PT->friends[0], PT->friends[1] );
   showFV( nrFeat, median );
   printf("Friends:\n");
   showFV( nrFeat, &(Samples[nrFeat*PT->friends[0]]) );
   showFV( nrFeat, &(Samples[nrFeat*PT->friends[1]]) ); DBG;
#endif
   return; 
 }

 EuclidDistsSum = vector( nrf );

 /* calculate the sum of all Euclidean distance for each friend from all
    other friends */
 for( i = 0; i < nrf; i++ )
 {
   EuclidDistsSum[ i ] = 0.0;
   for( j = 0; j < nrf; j++ )
   {
     if( i != j )
     {
       dist = Euclidian_Distance( &(Samples[nrFeat*PT->friends[i]]),
                                  &(Samples[nrFeat*PT->friends[j]]), nrFeat );
       /*printf("\tdist(%d,%d)=%f\n", PT->friends[i], PT->friends[j], dist);/**/
       EuclidDistsSum[ i ] += dist;
     }
   }
#ifdef PEEP_MED
   if( nrf < PEEP_MED )
   {
     printf("Distance=%f  Sample=%d  ", EuclidDistsSum[ i ], PT->friends[i] );
     showFV( nrFeat, &(Samples[nrFeat*PT->friends[i]]) ); /**/
   }
#endif
 }
 /* determine the minimum sum of distances */
 minDist = INFINITY;
 for( i = 0; i < nrf; i++ )
 {
   if( EuclidDistsSum[ i ] < minDist )
   {
     minDist = EuclidDistsSum[ i ];
     minFriend = i;
   }
 }
 copy_FV( &(Samples[nrFeat*PT->friends[minFriend]]), median, nrFeat );
#ifdef PEEP_MED
 if( nrf < PEEP_MED )
 {
   printf("VECTOR MEDIAN\n");
   showFV( nrFeat, &(Samples[nrFeat*PT->friends[minFriend]]) ); /**/
   printf("\n");
 }
#endif
 FREE( EuclidDistsSum );
}


/*
 * Calculate the marginal median, i. e. the vector of the univariate
 * medians for each individual feature
*/
static void calc_marginalMedian( median, PT )
FeatVector median;
ProType *PT;
{
 int k, i, nrf;
 FeatVector seq = NULL;
 boolean odd, even;

 nrf =  PT->nrFriends;
 seq = (FeatVector) malloc( nrf * sizeof( FeatVector* ) );
 CHKPTR( seq );

 odd = nrf % 2 == 1;
 even = nrf % 2 == 0; 

 /* univariate vesion */
 for( k = 0; k < nrFeat; k++ )
 {
   for( i = 0; i < nrf; i++ )
     seq[ i ] = Samples[ nrFeat * PT->friends[i] + k ];
   /* showFV( nrf, seq ); /**/
   qsort( seq, nrf, sizeof(float), order_rule );
   /* showFV( nrf, seq ); DBG; /**/

   if( odd )
     median[ k ] = seq[ (nrf-1)/2 ];
   if( even )
     median[ k ] = (seq[ nrf/2 - 1 ] + seq[ nrf/2 ]) / 2.0;
   /* printf("median[%d] = %f", k, median[k] ); DBG; /**/
 }
 FREE( seq );
}


static boolean update_friends( PT )
ProType *PT;
{
 FeatVector newPTvalue = NULL;
 boolean isEqual, change = FALSE;
 float dist;

 if( PT->nrFriends == 0 )
   return( FALSE );

 newPTvalue = (FeatVector) malloc( nrFeat * sizeof( FeatVector* ) );
 CHKPTR( newPTvalue );

 if( PT->nrFriends > 1 )
   PT->outlier = FALSE;
 else
 {
   /* only reset the outlier predicate if the close sample
      it not the prototype itself
      use buffer 'newPTvalue' temporarily
   */
   copy_FV( &(Samples[nrFeat*PT->friends[0]]), newPTvalue, nrFeat );
   dist = Euclidian_Distance( newPTvalue, PT->value, nrFeat );
   isEqual = dist <= EPSILON;
   if( ! isEqual )
     PT->outlier = FALSE;
 }

 switch( update_method )
 {
   case UPDATE_MEAN : calc_mean( newPTvalue, PT ); break;
   case UPDATE_MARGINAL_MEDIAN : calc_marginalMedian( newPTvalue, PT );  break;
   case UPDATE_VECTOR_MEDIAN : calc_VectorMedian( newPTvalue, PT );  break;
   default : fprintf(stderr,"Trouble with update method - exit...\n"); exit(1);
 } 

 /* Check if there was a change */
 dist = Euclidian_Distance( newPTvalue, PT->value, nrFeat );
 change = dist > EPSILON;
 /* printf("update_friends> dist=%.20f  change=%d\n", dist, change ); /**/
 FREE( PT->value );
 PT->value = newPTvalue;

 PT->nrFriends = 0;
 FREE( PT->friends );
 return( change );
}


/*
 *  update the prototypes of this class by attribute them the
 *  mean of all samples that belong to their domain.
*/
static boolean update_Class( clss )
int clss;
{
 int p;
 boolean proType_change = FALSE;

 /* update all prototypes for this class (exept last prototype which
     was just inserted without close samples
 */
 for( p = 0; p < PTC[clss].numProtype; p++ )
 {
   if( PTC[clss].PT[p].nrFriends > 0 )
     proType_change = update_friends( &(PTC[clss].PT[p]) );
 }
 return( proType_change );
}


static char save_comment[] =
"# NOTE: Make sure that a benchmark file is first filtered by the\n\
#\tabove feature selection mask before classifying it by this set\n\
#\tof prototypes!!! (See 'interfacing' menu).\n\
#\tDo not use the above feature selection mask when\n\
#\tloading this file, otherwise probably a wrong ordering of the\n\
#\tfeatures will result. Set all features as selected instead.\n\
#\tThen use a 1-Nearest-Neighbor classifier as the classifier model.\n\
#\n";

static void dropProtypesFile()
{
 FILE *pf = NULL;
 int i, k, p, clss;
 boolean protFile = TRUE;

 printf("\n Write all prototypes to a file (y/n)? y\b"); gets( buf );
 protFile = buf[0] != 'n' && buf[0] != 'N';
 if( ! protFile ) return;

 printf("Save prototypes to file: %s", outDataDir ); gets( buf );
 if( buf[0] == '\0' ) { printf("Nothing saved...\n"); return; }
 strcpy( pfName, outDataDir );
 strcat( pfName, buf );
 
 if( ( pf = fopen( pfName, f_open_text_w  )) == NULL )
 {
   fprintf( stderr, "Could not open %s for writing !!!\n", pfName );
   exit( 1 );
 }
 fprintf( pf, "#\n# Prototype file for universe: %s\n#\n", U->name );
 fprintf( pf, "# Number of classes: %d\n#\n", U->nrClass );
 fprintf( pf, "# Number of features for this universe: %d\n", U->nrFeat );
 fprintf( pf, "# Number of selected features for this universe: %d\n",
	U->nrSelFeat );
 fprintf( pf, "#\n#\tCurrent selected feature set:\n#\t\tINDEX\tCRITERION");
 if( feat_description ) fprintf( pf, "\tNAME" );
 fprintf( pf, "\n#\t\t--------------------------------------\n");
 for( i = 0; i < U->nrSelFeat; i++ )
 {
   fprintf( pf, "#\t\t%d\t%f", U->FSV[i].rank, U->FSV[i].crit );
   if( feat_description )
     fprintf( pf, "\t%s", feature_desc[U->FSV[i].rank] );
   fprintf( pf, "\n" );
 }
 fprintf( pf, "%s", save_comment );
 fprintf( pf, "#\n#\n# Dimension of continuous feature vector:\n");
 fprintf( pf, "%d\n", U->nrSelFeat );

 for( i = 0; i < U->nrClass; i++ )
 {
   clss = i;
   for( p = 0; p < PTC[i].numProtype; p++ )
   {
     for( k = 0; k < U->nrSelFeat; k++ )
       fprintf( pf, "%7.5f ", PTC[i].PT[p].value[k] );
     fprintf( pf, "  %s\n", U->C[clss].name );
   }
 }
 fclose( pf );
}


static void init_prototypes( nrSelFeat, FSV )
int nrSelFeat;
FeatSelectVector FSV;
{
 int i, j, k, row, random_offset = 0;

 nrFeat = nrSelFeat;
 init_PTC();
 row = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   PTC[i].PT = (ProType*) malloc( sizeof(struct ProType_) );
   PTC[i].numProtype = 1; sumProtype++;
   init_ProType( &(PTC[i].PT[0]) );
   PTC[i].PT[0].value = (FeatVector)malloc(nrFeat * sizeof(FeatVector*));

   if( first_prototype_random )
   {
     /* for every class pick one aleatory sample as first prototype */
     get_random( train_of_class[i], &random_offset );	/**/
     /* printf("offset for class %d = %d", i, random_offset ); DBG; /**/
     /* check offset */
     if( random_offset < 0 || random_offset >= train_of_class[i] )
     {
       fprintf(stderr,"WARNING: offset is %d\n", random_offset);
       fprintf(stderr,"Try changing the setup of the random number gen.\n");
       fprintf(stderr,"init_prototypes> Setting to 0\n");
       random_offset = 0;
     }
     copy_FV( &(Samples[nrFeat*(row+random_offset)]),
		 PTC[i].PT[0].value, nrFeat );
   }
   else
   {
     allPT.nrFriends = train_of_class[i];
     allPT.friends = (int*)malloc( allPT.nrFriends * sizeof(int) );
     CHKPTR( allPT.friends );
     for( j = 0; j < train_of_class[i]; j++ )
       allPT.friends[j] = row + j;

     switch( update_method )
     {
       case UPDATE_MEAN :
         for( k = 0; k < nrFeat; k++ )
           PTC[i].PT[0].value[k] = U->C[i].stat.mean[ FSV[k].rank ]; break;
       case UPDATE_MARGINAL_MEDIAN :
         calc_marginalMedian( PTC[i].PT[0].value, &allPT );
         break;
       case UPDATE_VECTOR_MEDIAN :
         calc_VectorMedian( PTC[i].PT[0].value, &allPT );
         break;
       default : fprintf(stderr,"Trouble with update method - exit...\n");
		 exit(1);
     }
     FREE( allPT.friends );
   }
   row += train_of_class[i];
 }
}

/*
 * Only return confirmation if prototype has really been created
*/
static boolean create_new_prototype( clss, row )
int  clss, row;
{
 int p, pInd; 
 ProType *PT_buf = NULL;
 boolean identical; float dist;

 /* do not create a new prototype if it already exists */
 identical = FALSE; p = 0;
 while( ! identical && p <  PTC[clss].numProtype )
 {
   dist = Euclidian_Distance( PTC[clss].PT[p].value,
				&(Samples[nrFeat*row]), nrFeat );
   identical = dist < EPSILON;
   /* printf("Identical=%d... class=%d  p=%d  dist=%.30f\n",
		identical, clss+1, p+1, dist);	/**/
   if( ! identical )
     p++;
 }
 if( identical )
 {
   /* printf("Already have this prototype for class %d", clss+1 ); DBG; /**/
   return( FALSE );
 }


 PT_buf = (ProType*)
   malloc((1+PTC[clss].numProtype) * sizeof(struct ProType_)); CHKPTR(PT_buf);
 for( p = 0; p <= PTC[clss].numProtype; p++ )
 {
   init_ProType( &(PT_buf[p]) );
   PT_buf[p].value = (FeatVector) malloc(nrFeat * sizeof(FeatVector*));
 }
 for( p = 0; p < PTC[clss].numProtype; p++ )
   copy_PT( &(PTC[clss].PT[p]), &(PT_buf[p]), nrFeat );

 pInd = PTC[clss].numProtype;

 copy_FV( &(Samples[nrFeat*row]), PT_buf[pInd].value, nrFeat );
 PT_buf[pInd].outlier = TRUE;
 PT_buf[pInd].nrFriends = 0;
 PT_buf[pInd].friends = NULL;

 /* copy buffer back and update all classes 0 ... class */
 for( p = 0; p < PTC[clss].numProtype; p++ )
   free_ProType( &( PTC[clss].PT[p]) );
 FREE( PTC[clss].PT );
 /* increase the number of prototypes for class 'class' */
 PTC[clss].numProtype++; sumProtype++;
 PTC[clss].PT = PT_buf;
 return( TRUE );
}


#define MAXITER 1000
static void cluster_qstar( nrSelFeat, FSV )
int nrSelFeat;
FeatSelectVector FSV;
{
 boolean changes = TRUE, mis_classed, created, dummy, verbose = FALSE;
 int s, i, j, c, p, row, minClass, clss, minProType, a, w = 0, runs = 1;
 float minDist, dist;
 int strlen_w;

 if( interact )
 {
   if( ! first_prototype_random && verbose )
     printf(" Inializing first prototype conforming update rule.\n\n");
   printf("\n------  Learning prototypes, using the Q* -algorithm  --------\n");
   printf("\tVerbose iterative learning process ?");
   get_boolean( FALSE, &verbose );
   /* start_timer(""); /**/
 }
 gen_selected_samples( nrSelFeat, FSV, verbose );
 init_prototypes( nrSelFeat, FSV );
 if( verbose )
 {
   show_PTC( FALSE, FALSE, nrSelFeat );
   printf("*** Ignore ouliers because of initialization ***\n");
 }

 strlen_w = strlen(wheel);
 /* main loop of cluster */
 /* while the prototypes do change */
 do
 {
   if( ! verbose && interact )
   {
     printf("Learning iteration %d  --- # prototypes %d --- ...   %c  \r",
	runs, sumProtype, wheel[(w++) % strlen_w] );
     fflush( stdout );
   }
   else
     if( verbose )
       { printf("%c\b", wheel[(w++) % strlen_w] ); fflush(stdout); }
   i = 0; mis_classed = FALSE;
   row = 0; s = 0;
   while( (i < U->nrClass) && ! mis_classed )
   {
     clss = i;
     for( j = 0; j < U->C[clss].numSampl; j++ )
     {
       if( train_smp[s] )
       {
         minDist = INFINITY; minClass = UNDEFINED;
         for( c = 0; c < U->nrClass; c++ )
         {
           for( p = 0; p < PTC[c].numProtype; p++ )
           {
             dist = Euclidian_Distance( PTC[c].PT[p].value,
                                      &(Samples[nrFeat*row]), nrFeat );
             if( dist <= minDist )
             {
               if( dist == minDist && c != clss && minClass == clss )
               {
                 ; /* do nothing: keep the minimum the own class */
               }
               else /* dist < minDist */
               {
                 minDist = dist;
                 minClass = c;
                 minProType = p;
               }
             }
           }
         } 
         /* look if the sample was correctly classified */
         if( minClass == clss )
         {
           /* join the correctly classified pattern to the friends */
           join_friend( row, &(PTC[minClass].PT[minProType]) );
         }
         else
         {
           /* printf("mis_clas class=%d minClass=%d row=%d! -> New prototype\n",
                     clss, minClass, row); NL; /**/
           created = create_new_prototype( clss, row );
           /* Only if a new prototype has really been created */
           if( created )
           {
             mis_classed = TRUE;
             for( a = 0; a <= clss; a++ )
               dummy = update_Class( a );
           }
         }
         row++;
       }
       s++;
     }/* for every sample */
     i++; /* for every class */
   }/* while */
   /* if no misclassifications occured update prototypes for all classes */
   if( ! mis_classed )
     for( i = 0; i < U->nrClass; i++ )
       changes = update_Class( i ) && changes;
   if( verbose )	/**/
   {
     printf("\n\t\t +++++ END OF RUN  %d  +++++", runs );
     show_PTC( FALSE, TRUE, nrSelFeat ); /**/
   }
   if( changes )
     runs++;
 } 
 while( changes && runs < MAXITER );

 if( changes )
 {
   printf("Too many trials to organize - maybe data too similar for");
   EXIT(" this feature set!");
 }
 if( verbose )
 {
   printf("\n --- F I N A L  R E S U L T  after %d runs: ---\n", runs );
   show_PTC( FALSE, FALSE, nrSelFeat );
 }

 /* Purge all prototypes which were generated by only one sample
  *  These prototypes are outliers
 */
 if( interact )
 {
   printf("\n --- Successfully clustered in %d runs to %d prototypes ---\n",
	runs, sumProtype );
   /* stop_timer("CPU costs:\n"); /**/
 }
 if( purge )
 {
   purge_outliers( nrSelFeat );
   if( interact )
     printf("Final # of prototypes after purging is %d\n", sumProtype );
 }
 if( verbose && purge )
 {
   printf("\n ---   R E S U L T  A F T E R  P U R G I N G   ---");
   show_PTC( FALSE, FALSE, nrSelFeat );
 }
}


/*
 *  SAMMON PLOT: RAW SAMPLES TOGETHER WITH Q*-PROTOTYPES
*/

static char PT_STR[] = "-prototypes";
static str100 str_gnu;

#define MF_DEFAULT      (0.2)
#define MF_MAX          (1.0)

#define NR_ITER_MAX 32000
#define NR_ITER_DEFAULT 300

#define DIM_DEFAULT 2

static float MF = MF_DEFAULT;           /* Learning rate */
static int dim = DIM_DEFAULT;
static int nrIter = NR_ITER_DEFAULT;


static void sammon_qstar()
{
 int i, j, k, p, row, N, nc, plot_dim, D;
 boolean ok, direct_plot;
 float x = 0, y = 0, z = 0;  /* Initialize because of the floating exception */
 float **Samples = NULL;
 float **Y = NULL;

 nc = U->nrClass;
 D = U->nrSelFeat;

 if( nc == 0 )
   { printf(" Please load universe first !..." ); gets( buf ); return; }
 if( U->nrSelFeat == 0 )
   { printf(" Select features first please!..." ); gets( buf ); return; }
 if( U->nrSelFeat == 1 )
 {
   printf("It does not make much sense to map unidimensional ");
   printf("data to 2 dimensions!..." ); gets( buf ); return;
 }

 /* How many entries? */
 N = U->sumSampl+sumProtype;

 /* Load samples with selected features only in buffer */
 if( sumProtype == 0 )
   EXIT("Error: no prototypes");

 /* Load samples with selected features only in buffer 'Samples' */
 Samples = matrix_alloc( N, U->nrSelFeat );
 CHKPTR(Samples);

 GNU_init_gnu_file();

 get_plot_dim( D, &plot_dim );

 direct_plot = D == plot_dim;

 axes_gnu( D, plot_dim, direct_plot );

 GNU_scatter_init( 2 * nc, N, plot_dim );
 /* The classes */
 row = 0;
 for( i = 0; i < nc; i++ )
 {
   GNU_scatter_set_name( U->C[i].name );
   GNU_scatter_set_smp( i, U->C[i].numSampl );

   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     for( k = 0; k < U->nrSelFeat; k++ )
       Samples[row][k] = U->C[i].S[j*U->nrFeat+U->FSV[k].rank];
     GNU_scatter_set_index( row, i );
     row++;
   }
 }

 /* The prototypes of the classes */
 for( i = 0; i < nc; i++ )
 {
   sprintf( str_gnu, "%s%s", U->C[i].name, PT_STR );
   GNU_scatter_set_name( str_gnu );
   GNU_scatter_set_smp( i+nc, PTC[i].numProtype );

   for( p = 0; p < PTC[i].numProtype; p++ )
   {
     for( k = 0; k < U->nrSelFeat; k++ )
       Samples[row][k] = PTC[i].PT[p].value[k];
     GNU_scatter_set_index( row, i+nc );
     row++;
   }
 }


 if( direct_plot )
 {
   for( i = 0; i < N; i++ )
   {
     x = Samples[i][0]; y = Samples[i][1]; if( plot_dim==3 ) z = Samples[i][2];
     GNU_scatter_set_point( x, y, z, i, plot_dim );
   }
 }
 else
 {
   printf("\tNumber of learning steps (actual=%d)? ", nrIter );
   get_d_range( &nrIter, 1, NR_ITER_MAX, LEFT_CLOSED__RIGHT_CLOSED );

   printf("\tLearning rate of Steepest Descent (actual=%4.2f)? ", MF );
   get_f_range( &MF, 0.0, MF_MAX, LEFT_OPEN__RIGHT_OPEN );

   Y = matrix_alloc( N, plot_dim );
   ok = map_sammon( Samples, N, D, Y, plot_dim, nrIter, MF ); NL;

   if( ok )
   {
     for( i = 0; i < N; i++ )
     {
       x = Y[i][0]; y = Y[i][1]; if( plot_dim == 3 ) z = Y[i][2];
       GNU_scatter_set_point( x, y, z, i, plot_dim );
     }
   }
   else
     printf("Unable to generate Sammon plot!\n\n");

   matrix_free( &Y, N );
 }
 if( direct_plot || ok )
 {
   GNU_scatter_plot();
   GNU_execute();
 }
 else
 {
   GNU_close_gnu_file();
   GNU_scatter_close();
 }
 matrix_free( &Samples, U->sumSampl );
}

static boolean q_sammon = FALSE;

void induceQ_STAR( nrSelFeat, FSV, iact )
int nrSelFeat;
FeatSelectVector FSV;
boolean iact;
{
 interact = iact;
 cluster_qstar( nrSelFeat, FSV );
 if( interact )
 {
   dropProtypesFile();
   printf("Generate a Sammon plot of the raw samples together with the\n");
   printf("\tinduced prototypes?");
   get_boolean( q_sammon, &q_sammon );
   if( q_sammon )
     sammon_qstar();
 }
}


/*
 * Free the local Samples and the prototypes
*/
void free_inducedQ_STAR()
{
 /* show_PTC(TRUE, TRUE, nrFeat);  /**/
 free_PTC();
 FREE( Samples );
 FREE( train_of_class );
}


int class1testQ_STAR( nsf, FSV, smp, name )
int nsf;
FeatSelectVector FSV;
FeatVector smp;
char *name;
{
 int c, p, minClass, minProType, res_class;
 float dist, minDist;

 minDist = INFINITY; minClass = UNDEFINED;
 if( verbose_reason )
 {
   fprintf(cpf,"\n\n--- CLASSIFICATION RESULT OF 1-NEAREST NEIGHBOR");
   fprintf(cpf," USING Q* -PROTOTYPES ---\n" );
   fprintf(cpf," FOR  UNKNOWN SAMPLE:\n");
   fshowFV( cpf, nsf, smp );
 }
 for( c = 0; c < U->nrClass; c++ )
 {
   for( p = 0; p < PTC[c].numProtype; p++ )
   {
     dist = Euclidian_Distance( PTC[c].PT[p].value, smp, nsf );
     /* printf("Euclidian_Distance=%f between:\n", dist );
        showFV( nsf, PTC[c].PT[p].value ); showFV( nsf, smp); /**/
     if( dist <= minDist )
     {
       minDist = dist;
       minClass = c;
       minProType = p;
     }
   }
 } 
 if( verbose_reason )
 {
    fprintf(cpf, "Minimum distance=%f was to prototype # %d of class %d(%s):\n",
	minDist, minProType+1, minClass+1, U->C[minClass].name );
    fshowFV( cpf, nsf, PTC[minClass].PT[minProType].value );
 }
 if( ! have_label )
   return( minClass );


 if( (strcmp(U->C[minClass].name,name)==0) )
   res_class = CORRECT_CLASS;
 else
 {
   res_class = minClass;
   if( verbose_reason )
     fprintf( cpf, " *** Confused %s with expected class %s\n",
                U->C[minClass].name, name );
 }
 return( res_class );
}
