/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:27 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "def.h"

extern long int my_random();

/*----------------------------------------------*/
/*   Random number generation                   */
/*----------------------------------------------*/

static time_t seed;
static double maxRand;

void init_random_generator()
{
 maxRand = (pow(2.0, 31.0) - 1.0);
 (void)time( &seed );
#ifdef RANDOM_SRC
 my_srandom( (int)seed );
#endif
#ifndef RANDOM_SRC
 srandom( (int)seed );
#endif
}


/* return an integer random number between 0 and 'max-1' */
void get_random( max, ran )
int max, *ran;
{
 int i;
 double r;

#ifdef RANDOM_SRC
 r = (double)my_random();
#endif
#ifndef RANDOM_SRC
 r = (double)random();
#endif
 /* printf("maxRand=%lf r=%lf max=%d", maxRand, r, max );	/**/ 
 r = r / maxRand * (double)max;
 i = (int)(r);
 if( i < 0 || i >= max )
 {
   fprintf(stderr,"get_random> Problems with the random");
   fprintf(stderr," generator - Check setup! Exit...\n"); exit(1);
 }
 *ran = i;
}


/* return a double random number between 0 and 'maxRand' */
void get_ran( ran )
double *ran;
{
 int i;
 double r;

#ifdef RANDOM_SRC
 r = (double)my_random();
#endif
#ifndef RANDOM_SRC
 r = (double)random();
#endif
 *ran = r;
}


/*>>>>>>>>>>>>>>----------------------<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
/*      Generating Gaussian noise using the polar method         */
/*     Knuth: The Art of Computer Programming: Volume II         */
/*>>>>>>>>>>>>>>----------------------<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
void get_zeroMean_sigmaStddev_noise( sigma, r )
float sigma, *r;
{
 int i;
 double U1, U2, V1, V2, S, X1, X2, aux, val;


 do
 {
   get_ran( &U1 ); get_ran( &U2 );
   U1 = U1 / maxRand; U2 = U2 / maxRand;
   /* printf("U1: %.2f     U2: %.2f\n",U1,U2); /**/
   V1 = 2*U1 - 1.0; V2 = 2*U2 - 1.0;
   S = V1*V1 + V2*V2;
 }
 while( S >= 1.0 );

 aux = sqrt( (-2.0)* log(S) / S );
 /*** calculate the Gaussian random numbers X1, X2 ***/
 X1 = V1*aux; X2 = V2*aux;
 /*** the mean is 0; the standard deviation is 1; adapt it ***/

 *r = sigma*X1;
}

