/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:27 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "def.h"
#include "clasdef.h"
#include "matrxdef.h"

extern universe *U;
extern boolean *train_smp, *test_smp;
extern int nrSamplesTest, nrSamplesTrain;
extern boolean verbose_reason;
extern float Euclidian_Distance();
extern boolean error_estimation;

extern float *vector();
/*extern float **matrix_alloc();*/

extern FILE *cpf;               /* Classification protocol file */
extern boolean have_label;      /* Only deliver the classified class for
        pre-classified data */


static boolean verbose, interact = FALSE;
static boolean verbose_induction;
static boolean alg_type_informed;

static Matrix G, Go, *d, w;
static float ***ref_vec = NULL;
static int *nrTrainClass = NULL;
static int *nrRefClass = NULL;
static int allRef, allTrain, which_class;
static float **coeff = NULL;
static float **delta_coeff = NULL;
static float **spread = NULL;
static float **delta_spread = NULL;
static float **var = NULL;
static boolean do_cluster, optimize = FALSE, class_dep_var = TRUE;

static float choice_nr_clusters = 20.0;
static float dist_ind_var = 10.0;
static float Lambda = 1.0;
static float factor = 0.25;

static boolean change_weights=TRUE, change_centres=FALSE, change_spreads=FALSE;
#define REGULARIZATION 1
#define ERROR_CORRECTION 2
static int alg_type = REGULARIZATION;

/* ************************************************************** */
static void choose()
{
 printf("Number of reference vectors (in percent of all samples): ");
 printf("%5.2f\b\b\b\b\b", choice_nr_clusters);
 get_f_range( &choice_nr_clusters, 0.0, 100.0, LEFT_OPEN__RIGHT_CLOSED );
 printf("Cluster the reference vectors? ");
 get_boolean(TRUE, &do_cluster);
 printf("Training algorithm  (Regularization - 1,");
 printf(" Error-Correction Learning - 2) : %1d\b", alg_type);
 get_d_range(&alg_type, REGULARIZATION,
	ERROR_CORRECTION, LEFT_CLOSED__RIGHT_CLOSED);
 if (alg_type == REGULARIZATION)
 {
   printf("Regularization parameter (0 - 100) : %7.2f\b\b\b\b\b\b\b", Lambda);
   get_f_range(&Lambda, 0.0, 100.0, LEFT_CLOSED__RIGHT_CLOSED);
   printf("Individually adjusted variances? ");
   get_boolean(class_dep_var, &class_dep_var);
   if (class_dep_var)
   {
     printf("Overlapping factor (0.0001 - 0.5): %7.2f\b\b\b\b\b\b\b", factor);
     get_f_range(&factor, 0.0001, 0.5, LEFT_CLOSED__RIGHT_CLOSED);
   }
   else  
   {
     printf("Distance-independent variance factor (max. 10^5): ");
     printf("%5.3f\b\b\b\b\b", dist_ind_var);
     get_f_range(&dist_ind_var,0.0, 100000.0, LEFT_OPEN__RIGHT_CLOSED);
   }
 }
 else
  optimize = TRUE;
}

/* ************************************************************** */
void show_RBF( f )
FILE *f;
{
 fprintf(f,"\tPercentage of reference vectors of all samples: %6.2f%%\n",
	choice_nr_clusters);

 fprintf(f,"\tChoice of Hidden Layer Units (Reference Vectors):\n");
 if (do_cluster)
   fprintf(f,"\t\tK-Means Clustering\n");
 else
   fprintf(f,"\t\tRandom Selection\n");
 fprintf(f,"\tTraining algorithm: ");
 switch( alg_type )
 {
   case REGULARIZATION:
     fprintf(f,"REGULARIZATION TECHNIQUE\n");
     fprintf(f,"\t\tRegularization Parameter: %f\n", Lambda);
     if (class_dep_var)
     {
       fprintf(f,"\t\tIndividually Adjusted Variances\n");
       fprintf(f,"\t\t\tOverlapping factor: %.2f\n", factor);
     }
     else
       fprintf(f,"\t\tDistance_independent variance factor: %.2f\n",
		dist_ind_var);
     break;
   case ERROR_CORRECTION:
     fprintf(f,"ERROR-CORRECTION LEARNING\n");
     break;
   default: fprintf(stderr,"Unknown method - exit\n");exit(1);
 }
}

/* ********************************************************* */
static void ref_vec_select(FSV, len)
FeatSelectVector FSV;
int len;
{
float *smp = NULL;
float *old_vec = NULL;
float *mean = NULL;
int store_s, s, s_alt, i, ii, j, k, kk;
int count_help_1, count_help_2, count_help_3, count_help_4, runs;
float dist, alt_dist;
int ***cluster, min_nr, *nr_clusters;
boolean stop_crit_1, not_zero_dist;      
int count_zero_dist;

nrTrainClass = (int *) malloc(U->nrClass * sizeof(int)); 
CHKPTR(nrTrainClass);
nrRefClass = (int *) malloc(U->nrClass * sizeof(int)); 
CHKPTR(nrRefClass);
smp = vector(len);
ref_vec = (float ***) malloc(U->nrClass * sizeof(float **));
CHKPTR(ref_vec);
cluster = (int ***) malloc(U->nrClass * sizeof(int **)); CHKPTR(cluster);
mean = (float*) malloc(len * sizeof(float)); CHKPTR(mean);
old_vec = (float *) malloc(len *sizeof(float)); CHKPTR(old_vec);

if (verbose_induction)
  printf("\n\nRANDOM SELECTION of the set of reference vectors ...\n");

/* ********************************************************************* */
/* ***************** HOW MANY REFERENCE VECTORS ? ********************** */
/* ********************************************************************* */
/* we first need the number of training vectors; second, we calculate the
   the number of reference points according to the percentage parameter
   read in at the beginning; then allocate the necessary space for these
   reference vectors, and finally read in the values for them */
s = 0;
for( i = 0; i < U->nrClass; i++ )
  {
  nrTrainClass[i] = 0;
  for( j = 0; j < U->C[i].numSampl; j++ )
    {
    if( train_smp[s] )   /* s von 0 -> U->nrClass*SUM(C[i].numSampl) */
      nrTrainClass[i]++;
    s++;
    }
  if( nrTrainClass[i] == 0 )
    EXIT("Unable to generate classifier for class");
  if (verbose_induction)
    printf("Class %d(%s) had %d training samples\n", i+1, U->C[i].name,
            nrTrainClass[i]);
  }

nr_clusters = (int *) malloc(U->nrClass * sizeof(int)); CHKPTR(nr_clusters);
for(i = 0; i < U->nrClass; i++)
  nr_clusters[i] = (int)(choice_nr_clusters * (float)nrTrainClass[i]/100.0);
/* nr_clusters[i] is equal to nrRefClass[i]; the following loop has been
   implemented just for "security" reasons */
s = 0;
for( i = 0; i < U->nrClass; i++ )
  {
  count_help_1 = 0;
  nrRefClass[i] = 0;
  for( j = 0; j < U->C[i].numSampl; j++ )
    {
    if (train_smp[s])    
      {
      if (count_help_1 < nr_clusters[i])  /* starting with count = 0 */
	nrRefClass[i]++;
      count_help_1++;
      }
    s++;
    }
  if (nrRefClass[i] == 0)
    EXIT("no reference vectors\n");
  if (verbose_induction)
    printf("Number of reference vectors in class %d(%s)  = %d\n",
            i+1, U->C[i].name, nrRefClass[i]);
  ref_vec[i] = (float **) malloc(nrRefClass[i] * sizeof(float *));
  CHKPTR(ref_vec[i]);
  }

for( i = 0; i < U->nrClass; i++ )
  for( j = 0; j < nrRefClass[i]; j++ )
    ref_vec[i][j] = vector(len);

/* take "nr_clusters" samples out of the training set */ 
s = 0;
for( i = 0; i < U->nrClass; i++ )
  {
  count_help_1 = 0;
  for( j = 0; j < U->C[i].numSampl; j++ )
    {
    if (train_smp[s])
      {
      provideFV(smp, i, j, len, FSV);
      if (count_help_1 < nrRefClass[i])  /* starting with count = 0 */
        for (k = 0; k < len; k++)
          ref_vec[i][count_help_1][k] = smp[k];
      count_help_1++;
      }
    s++;
    }
  }

if (verbose_induction)
  printf("REPLACING reference vectors with zero distances ...\n");

/* ***************************************************************** */
/* ************************ NO CLUSTERING ************************** */
/* ***************************************************************** */
if (!do_cluster)
  {
  runs = 0;
  
  do {
    s = 0;
    count_help_1 = 0;
    for(i = 0; i < U->nrClass; i++)
      {
      for(j = 0; j < nrRefClass[i]; j++)
        {
        for (k = 0; k < nrRefClass[i]; k++)
          {
          dist = Euclidian_Distance(ref_vec[i][k], ref_vec[i][j], len);
          if ((j != k) && (dist == 0))
            { 
            count_help_1++;
            stop_crit_1 = FALSE;
          
            do {
              for (kk = 0; kk < U->C[i].numSampl; kk++)
                {
                if ((train_smp[s]) && (!stop_crit_1))
                  {
                  provideFV(smp, i, kk, len, FSV);
                  if (Euclidian_Distance(ref_vec[i][k],smp,len) != 0.0)
                    {
	            for (ii = 0; ii < len; ii++)
	               ref_vec[i][j][ii] = smp[ii]; 
	            stop_crit_1 = TRUE;
	            }
	          }
                s++;
                } 
              } while (!stop_crit_1);  
	    }
          } 
        }
      }
    runs++;
    } while ((runs < 30) && (count_help_1 != 0));
  
  if (count_help_1 > 0)
    printf("WARNING ! - after 30 runs still equal reference vectors !\n");
  else
    if (verbose_induction)
      printf("no zero difference between any of the reference vectors ...\n");
  }
/* ***************************************************************** */
/* ****************** K-MEANS CLUSTERING *************************** */
/* ***************************************************************** */
else                                                   /* do cluster */
  {
  for( i = 0; i < U->nrClass; i++ )
    {
    cluster[i] = (int **) malloc(nrRefClass[i] * sizeof(int*));
    CHKPTR(cluster[i]);
    for (k = 0; k < nrRefClass[i]; k++)
      {
      cluster[i][k] = (int *) malloc(nrTrainClass[i] * sizeof(int));
      CHKPTR(cluster[i][k]);
      }
    }
  if (verbose_induction)
    printf("CLUSTERING the set of reference vectors (max. 50 runs) ...\n");
  runs = 0;
  stop_crit_1 = FALSE;
  count_help_2 = count_help_3 = count_help_4 = 0;
  
  do {
    for( i = 0; i < U->nrClass; i++ )
      for (k = 0; k < nrRefClass[i]; k++)
        for (j = 0; j < nrTrainClass[i]; j++)
          cluster[i][k][j] = 0;
    s = 0;
    for (i = 0; i < U->nrClass; i++)
      {
      count_help_4 = 0;
      count_help_1 = 0;
      for(j = 0; j < U->C[i].numSampl; j++)   /* all samples of class i */
        {
        alt_dist = INFINITY;
        if (train_smp[s])                     /* is it a training vector */
          {
          provideFV( smp, i, j, len, FSV );
          /* compare this training vector with all reference vectors */
          for (k = 0; k < nrRefClass[i]; k++)
            {
            dist = Euclidian_Distance(ref_vec[i][k], smp, len);
            if (dist < alt_dist)
	      {
	      alt_dist = dist;
	      min_nr = k;
	      }
            }
          /* assign the trainings vector to the cluster [class][ref][sample]*/
          cluster[i][min_nr][count_help_4] = 1;    
          count_help_4++;
	  }
        s++;   
        }
      }
    not_zero_dist = TRUE;
    store_s = 0;
    for( i = 0; i < U->nrClass; i++ )
      {
      s_alt = store_s;            /* the sum of all nrTrainClass[i-1,2,...] */
      count_zero_dist = 0; 
      alt_dist = -INFINITY;
      /* take every reference vector */
      for (k = 0; k < nrRefClass[i]; k++)
        {
        count_help_1 = 1;  
        /* store its location */
        for (ii = 0; ii < len; ii++)
          old_vec[ii] = ref_vec[i][k][ii];
        /* compare every reference vector with every sample */
        count_help_4 = 0;
        s = s_alt;
        for (j = 0; j < U->C[i].numSampl; j++)
          {
          if (train_smp[s])      /* is it a training vector */
            {
            provideFV( smp, i, j, len, FSV );
            /* is the sample assigned to ref_vec number k ? */
            if (cluster[i][k][count_help_4] == 1)
	      {
	      for (ii = 0; ii < len; ii++)
	        ref_vec[i][k][ii] = 
                    (float)count_help_1*ref_vec[i][k][ii] + smp[ii];
              count_help_1++;
	      for (ii = 0; ii < len; ii++)
	        ref_vec[i][k][ii] = ref_vec[i][k][ii]/(float)count_help_1;
	      }
	    count_help_4++;
            }
          s++;
          }
        store_s = s;   
        /* s counted till sum of all nrTrainClass[ii] ,ii = 1,2,...i */
        /* did the reference vector move a lot ? */
        dist = Euclidian_Distance(old_vec, ref_vec[i][k], len);
        /*if (dist > alt_dist)
          alt_dist = dist;*/
        if (dist == 0.0)
          count_zero_dist = count_zero_dist + 1;
        else
	  not_zero_dist = FALSE;
        if (count_help_1 == 0) 
	  count_help_2++;
	if ((count_help_1 == 1) && (runs < 3)) 
	  { 
	  count_help_3++;
          for (ii = 0; ii < len; ii++)
	    mean[ii] = 0.0;
          for (kk = 0; kk < nrRefClass[i]; kk++)
            for (ii = 0; ii < len; ii++)
	      mean[ii] = mean[ii] + ref_vec[i][kk][ii];
          for (ii = 0; ii < len; ii++)
	    {
	    mean[ii] = mean[ii]/(float)kk;
	    ref_vec[i][k][ii] = (ref_vec[i][k][ii] + mean[ii])/2.0;
	    }
	  } 
        }         /* end of all ref_vec[k] */
      }           /* end of all classes */
    if (not_zero_dist)
      stop_crit_1 = TRUE;
    count_help_2 = count_help_3 = 0;
    runs++;
    } while ((!stop_crit_1) && (runs <= 50));
  if (verbose_induction)
    printf("Clustering finished after %d runs\n", runs-1); 
  }  

/* just a final check */
for(i = 0; i < U->nrClass; i++)
   for(j = 0; j < nrRefClass[i]; j++)
     for (k = 0; k < nrRefClass[i]; k++)
       {
       dist = Euclidian_Distance(ref_vec[i][k], ref_vec[i][j], len);
       if ((j != k) && (dist == 0.0))
         EXIT("still equal reference vectors\n");
       }

if (do_cluster)
  {
  for (i = 0; i < U->nrClass; i++)
    if (cluster[i] != NULL)
      {
      for (j = 0; j < nrRefClass[i]; j++)
        if (cluster[i][j] != NULL)
  	  FREE(cluster[i][j]);
	FREE(cluster[i]);
      }
  }

allTrain = 0;
allRef = 0;
for( i = 0; i < U->nrClass; i++ )
  {
  allTrain = allTrain + nrTrainClass[i];
  allRef = nrRefClass[i] + allRef;
  }

FREE(old_vec);
FREE(mean);
FREE(smp);
FREE(nr_clusters);
FREE(cluster);
}

/* ******************************************************************** */
static void gen_var(len, factor)
int len;
float factor;
{
int i, k, kk;
float dist, alt_dist;

var = (float **) malloc(U->nrClass * sizeof(float *));
for (i = 0; i < U->nrClass; i++)
  {
  var[i] = vector(nrRefClass[i]);
  for (k = 0; k < nrRefClass[i]; k++)
     var[i][k] = dist_ind_var;              /* preset or chosen */
  }

if (class_dep_var)      /* i.e. distance dependent */
  {
  factor = (float)fabs(log((double)factor));
  for(i = 0; i < U->nrClass; i++)
    {
    for (k = 0; k < nrRefClass[i]; k++)
      {
      alt_dist = 100.0;          /* infinity */
      for (kk = 0; kk < nrRefClass[i]; kk++)
        {
        dist = Euclidian_Distance(ref_vec[i][k], ref_vec[i][kk], len);
        if ((dist < alt_dist) && (dist != 0.0))
	  {
	  alt_dist = dist;
	  var[i][k] = factor/alt_dist;
	  }
        }
      }
    }
  }
}

/* ******************************************************************** */
static void gen_matrix(FSV,len)
FeatSelectVector FSV;
int len;
{

Matrix *M = NULL;
int i, ii, j, s, s_help;
int count_help_1, count_help_2, count_help_3, count_help_4;
Matrix mat1, mat2, mat3;
float help_, phi, cond_nr, max_1 = -INFINITY, max_2;
MatElem Det;
FeatVector smp = NULL;

smp = vector(len);
init_Matrix( &mat1 );
init_Matrix( &mat2 );
init_Matrix( &mat3 );

init_Matrix( &w );
w.row = allRef; w.col = 1;
Matrix_Alloc( &w ); Reset_Matrix( &w );  /* all zero */
d = (Matrix*) malloc( U->nrClass * sizeof( Matrix ) );
for (i = 0; i < U->nrClass; i++)
  {
  init_Matrix( &(d[i]) );
  d[i].row = allTrain ; d[i].col = 1;
  Matrix_Alloc( &(d[i]) ); Reset_Matrix( &(d[i]) );
  M = &(d[i]);
  count_help_1 = 0;
  for (ii = 0; ii < U->nrClass; ii++)
    {
    if (ii > 0)
      count_help_1 = count_help_1 + nrTrainClass[ii-1];
    for (count_help_2 = 0; count_help_2 < nrTrainClass[ii]; count_help_2++)
      {
      if (ii == i)  
        for (j = 0; j < M->col /* 1 */; j++) 
          M->Elem[(count_help_1 + count_help_2)*M->col+j] = 1.0;
      else
        for (j = 0; j < M->col /* 1 */; j++) 
          M->Elem[(count_help_1 + count_help_2)*M->col+j] = 0.0;
      }
    }
  }

init_Matrix( &G );
G.col = allRef; G.row = allTrain;
Matrix_Alloc( &G ); Reset_Matrix( &G );
M = &G;
count_help_1 = 0;
s = 0;
for (i = 0; i < U->nrClass; i++)
  {
  count_help_4 = 0;
  s_help = 0;
  if (i > 0)
    count_help_1 = count_help_1 + nrTrainClass[i-1];
  for (count_help_2 = 0; count_help_2 < U->C[i].numSampl; count_help_2++)
    { 
    /* provide sample first */
    if (train_smp[s])
      {
      s_help++;
      provideFV(smp, i, count_help_2, len, FSV);       
      count_help_3 = 0;
      for (ii = 0; ii < U->nrClass; ii++)
        {
        if (ii > 0)
	  count_help_3 = count_help_3 + nrRefClass[ii-1];
        for (j = 0; j < /*M->col*/ nrRefClass[ii]; j++)
          {
          phi = Euclidian_Distance(ref_vec[ii][j], smp, len);
  	  phi = phi * phi;
          phi = -phi;
          phi = phi*var[ii][j];
 	  phi = (float)exp((double)phi);
	  if ((count_help_1+count_help_4) > G.row)
	    EXIT("no more rows");
          if ((j+count_help_3) >= G.col)
	    EXIT("no more columns");
	  M->Elem[(count_help_1+count_help_4)*M->col+(j+count_help_3)] = phi;
	  /* Elem[nrTrainClass[k, k==0,..,i-1] + sample_nr[i]] */
	  /* + Column(ref[ii]+ref[nrRefClass[k, k==0,1,..,ii-1]] */
	  }
        }
      count_help_4++;
      }
    s++;
    }
  }     

init_Matrix( &Go );
Go.col = allRef; Go.row = allRef;
Matrix_Alloc( &Go ); Reset_Matrix( &Go );
M = &Go;
count_help_1 = 0;
for (i = 0; i < U->nrClass; i++)
  {
  if (i > 0)
    count_help_1 = count_help_1 + nrRefClass[i-1];
  for (count_help_2 = 0; count_help_2 < nrRefClass[i]; count_help_2++)
    {
    count_help_3 = 0;
    for (ii = 0; ii < U->nrClass; ii++)
      {
      if (ii > 0)
        count_help_3 = count_help_3 + nrRefClass[ii-1];
      for (j = 0; j < nrRefClass[ii]; j++)
        {
        phi = Euclidian_Distance(ref_vec[ii][j],
	                         ref_vec[i][count_help_2], len);
	phi = phi * phi;
        phi = -phi;
	phi = phi*var[ii][j];
	phi = (float)exp((double)phi);
	M->Elem[(count_help_1+count_help_2)*M->col+(j+count_help_3)] = phi;
        }
      }
    }
  }
Transpose_Matrix( &G, &mat1 );
Mult_Matrix(&mat1, &G, &mat2);
Copy_Matrix(&mat2, &G);
ScalarMult_Matrix(Lambda, &Go);
Add_Matrix( &G, &Go, &mat2 );
if (verbose_induction)
  {
  printf("Checking the matrix inversion :\n");
  M = &mat2;
  for (i = 0; i < M->row; i++)
    {
    help_ = 0.0;
    for (j = 0; j < M->col; j++)
      help_ = help_ + M->Elem[i*M->col+j];
    if (help_ > max_1)
      max_1 = help_;
    }
  Copy_Matrix(&mat2, &G);
  Invert_Matrix(&G,&Det);
  M = &G;
  max_2 = 0.0;
  for (i = 0; i < M->row; i++)
    {
    help_ = 0.0;
    for (j = 0; j < M->col; j++)
      help_ = help_ + M->Elem[i*M->col+j];
    if (help_ > max_2)
      max_2 = help_;
    }
  cond_nr = max_1 * max_2;
  printf("Condition Number of the Green's Matrix : %f\n", cond_nr);
  printf("Determinant = %f\n", Det );
  Mult_Matrix( &G, &mat2, &mat3);
  M = &mat3;
  help_ = 0.0;
  for (i = 0; i < M->row; i++)
    for (j = 0; j < M->col; j++)
      help_ = help_ + (float)fabs((double)M->Elem[i*M->col+j]);
  help_ = help_ - M->col;
  printf("Deviation from Identity matrix = %.20f\n", help_);
  }
else
  {
  Copy_Matrix(&mat2, &G);
  Invert_Matrix(&G,&Det);
  }
Mult_Matrix( &G, &mat1, &mat2);
Copy_Matrix(&mat2, &G);

FREE(smp);
Matrix_Free( &mat1 ); 
Matrix_Free( &mat2 );
Matrix_Free( &mat3 );
}

/* *************************************************************** */
static void free_other_stuff()
{
int i, j;

for (i = 0; i < U->nrClass; i++)
  {
  if (var != NULL)
    if (var[i] != NULL)
      FREE(var[i]);
  if (spread != NULL)
    if (spread[i] != NULL)
      FREE(spread[i]);
  if (delta_spread != NULL)
    if (delta_spread[i] != NULL)
      FREE(delta_spread[i]);
  if (coeff != NULL)
    if (coeff[i] != NULL)
      FREE(coeff[i]);
  if (delta_coeff != NULL)
    if (delta_coeff[i] != NULL)
      FREE(delta_coeff[i]);
  if (ref_vec != NULL)
    if (ref_vec[i] != NULL)
      {
      for (j = 0; j < nrRefClass[i]; j++)
	 if (ref_vec[i][j] != NULL)
	   FREE(ref_vec[i][j]);
      FREE(ref_vec[i]);
      }
  }
FREE(var);
FREE(spread);
FREE(delta_spread);
FREE(ref_vec);
FREE(nrTrainClass);
FREE(nrRefClass); 
FREE(coeff);
FREE(delta_coeff);
}

/* *************************************************************** */
static void free_matrix()
{
int i;

Matrix_Free( &G );
Matrix_Free( &w );
Matrix_Free( &Go );
if (d != NULL)
  {
  for (i = 0; i < U->nrClass; i++)
    Matrix_Free(&(d[i]));
  FREE( d );
  }
FREE( d );
}

/* ****************************************************************** */
void induceRBF( nsf, FSV, iact )  
int nsf;  
FeatSelectVector FSV; 
boolean iact;   

{
 int s, i, ii, j, k, help_count, bigrun, help_runs;
 float *sample = NULL;
 float *output = NULL;
 float *delta_ref = NULL;
 float dist, des_output, phi, error, total_error;
 float learn_rate_centres, learn_rate_weights, learn_rate_spreads;
 Matrix *M = NULL;
 boolean more_runs, wait_run, stop_runs, first_run;

 if( error_estimation && alg_type == ERROR_CORRECTION )
   {
   alg_type = REGULARIZATION;
   if( ! alg_type_informed )
     {
     alg_type_informed = TRUE;
     printf("\n\n\t>>> RBF network training algorithm was ");
     printf("\"Error correction\".\n\t>>> Not useful for error estimation.\n");
     printf("\t>>> Resetting to \"Regularization\".\n\n");
     show_class_model( stdout ); WAIT;
     printf("Selecting the reference vectors...\r"); fflush(stdout);
     }
   optimize = FALSE;
   }

 interact = iact;
 verbose = FALSE;
 if( interact )
   {
   printf("\n Inducing a %d-dimensional RBF-Network classifier \n\n", nsf);
   printf("Verbose learning of decision structure?");
   get_boolean( FALSE, &verbose_induction );
   }
 
 sample = vector(nsf);
 output = vector(U->nrClass);
 delta_coeff = (float **) malloc(U->nrClass * sizeof(float *));
 CHKPTR(delta_coeff);
 delta_spread = (float **) malloc(U->nrClass * sizeof(float *));
 CHKPTR(delta_spread);
 coeff = (float **) malloc(U->nrClass * sizeof(float *)); CHKPTR(coeff);
 spread = (float **) malloc(U->nrClass * sizeof(float *)); CHKPTR(spread);

 ref_vec_select(FSV, nsf);    /* select the set of hidden units */
 gen_var(nsf, factor);        /* set the variances */
 if (alg_type == REGULARIZATION)
   {
   if (verbose_induction)
     printf("COMPUTING the Regularization Matrix ...\n"); 
   gen_matrix(FSV,nsf);       /* generate the regularization matrix */
   }
 
 delta_ref = (float *) malloc(allRef * sizeof(float)); CHKPTR(delta_ref);

 for (i = 0; i < U->nrClass; i++)
   {
   coeff[i] = vector(allRef);
   spread[i] = vector(allRef);
   delta_coeff[i] = vector(allRef);
   delta_spread[i] = vector(allRef);
   if (alg_type == REGULARIZATION)
     {
     Mult_Matrix( &G, &(d[i]), &w);
     M = &w;
     }
   for (j = 0; j < allRef; j++)
     {
     delta_ref[j] = 0.0;
     switch (alg_type) {
       case REGULARIZATION   : coeff[i][j] = M->Elem[j];
                               break;
       case ERROR_CORRECTION : coeff[i][j] = 0.0;
	                       break;
       default : 
	   EXIT("type of algorithm undefined !\n");
       }
     spread[i][j] = 1.0;
     }
   }
 
 more_runs = TRUE;
 stop_runs = FALSE;
 wait_run = FALSE;
 help_runs = bigrun = 0;
 first_run = TRUE;
 if (optimize)
   printf("ERROR-CORRECTION LEARNING ...\n");

 do {   /* +++++++++++++++++++++++ */
 
 if (more_runs)
   change_weights = change_spreads = change_centres = FALSE;
 for (i = 0; i < U->nrClass; i++)
   for (j = 0; j < allRef; j++)
     {
     delta_ref[j] = 0.0;
     delta_coeff[i][j] = 0.0;
     delta_spread[i][j] = 0.0;
     }
 
 if (!wait_run)
   {
   total_error = 0.0;
   s = 0; 
   for (i = 0; i < U->nrClass; i++)
    {
    for (k = 0; k < U->C[i].numSampl; k++)  
     {
     if(train_smp[s])
      {
      provideFV(sample, i, k, nsf, FSV);

      /* *********** TEST THE VECTOR FOR ALL OUTPUTS *************** */
      for (which_class = 0; which_class < U->nrClass; which_class++ )
        {
        help_count = 0;
        output[which_class] = 0.0;
        error = 0.0;
        /* ************* WHAT IS THE DESIRED OUTPUT ?  *************** */
        if (i == which_class) /* the class the feature vector belongs to */
          des_output = 1.0;
        else
          des_output = 0.0;
        /* ******* THE VECTOR [i][k] IS TESTED FOR ALL CLASSES ********* */
        for (ii = 0; ii < U->nrClass; ii++)
          {
          /* ****** WITH ALL REFERENCE VECTORS OF A CLASS  ******* */
	  for (j = 0; j < nrRefClass[ii]; j++)
            {
            phi = Euclidian_Distance(ref_vec[ii][j], sample, nsf);
	    phi = phi * phi;
            phi = -phi;
            phi = phi*var[ii][j];
	    phi = phi*spread[ii][j];
	    /* phi = (float)exp((double)phi);	/**/
	    phi = (float)exp((double)phi);
            output[which_class] = output[which_class] 
		  		    + (coeff[which_class][help_count] * phi);
            help_count++;
	    }
          }
        /* **************** WHAT IS THE ERROR ? ******************* */
        error = des_output - output[which_class];
        total_error = total_error + (error * error);  /* the quadratic error */
        help_count = 0;
        for (ii = 0; ii < U->nrClass; ii++)
          {
          for (j = 0; j < nrRefClass[ii]; j++)
	    {
	    phi = Euclidian_Distance(ref_vec[ii][j], sample, nsf);
            dist = phi;
	    phi = phi * phi;
            phi = -phi;
	    phi = (float)exp((double)phi);
	    phi = phi*spread[ii][j];
	    delta_spread[which_class][help_count] = 
  	        delta_spread[which_class][help_count] 
	        + (dist*dist*phi*error*2.0*coeff[which_class][help_count]);
	    delta_coeff[which_class][help_count] =
	            delta_coeff[which_class][help_count] + (phi*error);
	    delta_ref[help_count] = 
	      2.0*coeff[which_class][help_count]*error*phi*dist*spread[ii][j];
	    help_count++;
	    }
          }
        /* *** CALCULATE THE NEW POSITIONS OF THE REFERENCE CENTRES *** */
       
        }             /* end of for all which_classes */
       }
      s++;
      }               /* end of for all training vectors[class i] */
    }                 /* end of for all classes */
   
   if (first_run && optimize)
     {
     total_error = total_error/(float)allTrain;
     printf("total error  = %f\n", total_error); 
     first_run = FALSE;
     }
   } 
 
 if (more_runs && optimize)
   {
   printf("do you wish to adjust the weights ?  : ");
   get_boolean(change_weights, &change_weights);
   printf("do you wish to adjust the centres ?  : ");
   get_boolean(change_centres, &change_centres);
   printf("do you wish to adjust the spreads ?  : ");
   get_boolean(change_spreads, &change_spreads);
   help_runs = 0;
   if ((change_weights) || (change_centres) || (change_spreads))
     {
     printf("How many runs with the given learning parameters ? (max 1000) : ");
     get_d_range(&bigrun, 1, 1000, LEFT_CLOSED__RIGHT_CLOSED);
     }
   else
     stop_runs = TRUE;
   more_runs = FALSE;
   wait_run = FALSE;
   }

 if (change_weights) 
   {
   if (help_runs == 0)
     {
     printf("learning rate weights (0.0 ... 0.005) : ");
     get_f_range(&learn_rate_weights, 0.0, 1.0, LEFT_CLOSED__RIGHT_CLOSED);
     }
   for (which_class = 0; which_class < U->nrClass; which_class++)
     {
     for (j = 0; j < allRef; j++)
       coeff[which_class][j] = coeff[which_class][j]
	 	         + (delta_coeff[which_class][j]*learn_rate_weights);
     }
   }
 
 if (change_spreads)
   {
   if (help_runs == 0)
     {
     printf("learning rate spreads (0 ... 10) : ");
     get_f_range(&learn_rate_spreads, 0.0, 10.0, LEFT_CLOSED__RIGHT_CLOSED);
     }
   for (which_class = 0; which_class < U->nrClass; which_class++)
     {
     for (j = 0; j < allRef; j++)
       spread[which_class][j] =  spread[which_class][j]
		  - (delta_spread[which_class][j]*learn_rate_weights);
     }
   }
 
 if (change_centres)
   {
   if (help_runs == 0)
     {
     printf("learning rate centres (0.0 ... 0.01) : ");
     get_f_range(&learn_rate_centres, 0.0, 1.0, LEFT_CLOSED__RIGHT_CLOSED);
     }
   help_count = 0;
   for (which_class = 0; which_class < U->nrClass; which_class++)
     for (j = 0; j < nrRefClass[which_class]; j++)
       {
       for (k = 0; k < nsf; k++)
	 ref_vec[which_class][j][k] = ref_vec[which_class][j][k] +
                (learn_rate_centres*delta_ref[help_count]); 
       help_count++;
       }
   }
 
 if ((help_runs >= bigrun) && (!stop_runs))
   {
   more_runs = TRUE;
   wait_run = TRUE;
   if (optimize)
     {
     total_error = total_error/(float)allTrain;
     printf("total error  = %f\n", total_error);
     printf("stop optimization ?  : ");
     get_boolean(FALSE, &stop_runs);
     }
   else 
     stop_runs = TRUE;
   }
 help_runs++;
 
 } while (!stop_runs);       /* +++++++++++++++++++++++ */
 
 verbose_induction = FALSE;
 
 FREE(delta_ref);
 FREE(output); 
 FREE(sample);
}

/*  ****************************************************************
 *  Free all the structures that were dynamically allocated to store
 *  the classifier, for instance matrices of prototypes
*/
void free_inducedRBF()
{
free_matrix();
free_other_stuff();
}

void initRBF()
{
 alg_type_informed = FALSE;
 free_induced();
 choose();
}

/*  ***************************************************************
 *  Classify a single sample 'smp' of length 'nsf'
 *  The name of the expected class is stored in 'name'
 *  'FSV' is not needed here, but for compatibility reasons passed
*/
int class1testRBF( nsf, FSV, smp, name )
int nsf;
FeatSelectVector FSV;
FeatVector smp;
char *name;

{
 int i, j, resulting_class, best_class, which_class, help_count;
 float *output = NULL;  
 float phi;

 if( verbose_reason )
 {
   fprintf( cpf,"\n\n ---- DECISION FUNCTIONS FOR UNKNOWN SAMPLE:\n");
   fshowFV( cpf, nsf, smp );
 }
 
 output = (float *) malloc(U->nrClass * sizeof(float));
 CHKPTR(output);
 for (which_class = 0; which_class < U->nrClass; which_class++ )
 {
   help_count = 0;
   output[which_class] = 0.0;
   for (i = 0; i < U->nrClass; i++)
   {
     for (j = 0; j < nrRefClass[i]; j++)
     {
       phi = Euclidian_Distance(ref_vec[i][j], smp, nsf);
       phi = phi * phi;
       phi = -phi;
       phi = phi * var[i][j];
       phi = phi * spread[i][j];
       phi = (float)exp((double)phi);
       output[which_class] =  output[which_class] 
			      + (coeff[which_class][help_count] * phi);
       help_count++;
     }
   }
   if( verbose_reason )
   {
     fprintf( cpf, "\tClass %3d --- %20s = %.5lf\n",
	which_class+1, U->C[which_class].name, output[which_class] );
   }
 }
  
 for (which_class = 0; which_class < U->nrClass; which_class++)
 {
   output[which_class] = output[which_class]*U->C[which_class].a_priori_prob;
   if (which_class == 0)
     best_class = which_class;
   else
     if (output[which_class] > output[best_class])
       best_class = which_class;
 }

 FREE(output);
 if( ! have_label )
   return( best_class );

 /*
   First case: correct classification:
   The name of the best class corresponds to the expected class
   return the value 'CORRECT_CLASS'
 */
 if( (strcmp(U->C[best_class].name,name)==0) )
   resulting_class = CORRECT_CLASS;
 else
  /*
    Second case: wrong classification:
    return the value of the best class: 'best_class'
  */
   {
   resulting_class = best_class;
   if( verbose_reason )
     fprintf( cpf, " *** Confused %s with expected class %s\n",
		U->C[best_class].name, name );
   }

 return( resulting_class );
}
