/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:28 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include "def.h"
#include "ftsldef.h"

extern float selectMultivariate();

extern universe *U;
extern boolean verbose;
extern boolean return_log;


static str80 buf;
static FeatSelectVector FSV = NULL;


static void initSFS( group, crit )
int group, crit;
{
 switch( crit )
 {
   case PATRICK_FISHER:
     return_log = FALSE;
     printf("Return logaritm of selection criterion instead criterion?");
     printf(" (y/n)n\b"); gets( buf );
     if( buf[0] == 'y' ) return_log = TRUE;
     break;
 }
}


void selectFeaturesSFS( group, crit )
int group, crit;
{
 int i, j, maxJ, nrSelectFeat, nrAlreadySelected = 0;
 boolean *already_selected = NULL;
 float merit, maxMerit, maxMeritOld;

 initSFS( group, crit );

 FSV = (feature*) malloc( sizeof(feature) * U->nrFeat ); CHKPTR(FSV);
 already_selected = (boolean*) malloc( sizeof(boolean) * U->nrFeat );
 CHKPTR(already_selected);
 for( j = 0; j < U->nrFeat; j++ )
   already_selected[ j ] = FALSE;

 printf("How many features do you wish to select ? (1-%d): ", U->nrFeat );
 get_d_range( &nrSelectFeat, 1, U->nrFeat, LEFT_CLOSED__RIGHT_CLOSED );

 maxMeritOld = -INFINITY;
 for( i = 0; i < nrSelectFeat; i++ )
 {
   /* for all available features claculate the criterion together with
      the already selected features */
   printf("Selecting feature nr. %4d  \r", i+1 ); fflush( stdout );
   maxMerit = -INFINITY;
   for( j = 0; j < U->nrFeat; j++ )
   {
     if( ! already_selected[ j ] )
     {
       /* set up the feature vector */
       FSV[ nrAlreadySelected ].rank = j;

       merit = selectMultivariate( group, crit, FSV, nrAlreadySelected+1 );

       BLANK_LINE;
       printf("Feature nr. %4d of %4d-%4d is candidate nr. %4d",
		j+1, U->nrFeat, nrAlreadySelected, i+1 );
       printf("   Criterion=%.4f\r", merit ); fflush( stdout );
       if( verbose ) printf("\n");
       if( merit > maxMerit )
       {
         maxMerit = merit;
         maxJ = j;
       }
     }
   }
   BLANK_LINE;
   printf("Set together with new feature nr. %4d had maximum criterion %.4f\n",
	 maxJ+1, maxMerit );
   /* the feature with the maximum merit is determined */
   if( already_selected[ maxJ ] )
     EXIT("\nThe feature was already selected - maybe a precision problem!");
   already_selected[ maxJ ] = TRUE;
   FSV[ nrAlreadySelected ].rank = maxJ;
   FSV[ nrAlreadySelected ].crit = maxMerit;
   nrAlreadySelected++; 
   if( group == MIN_EST_ERR && maxMerit == maxMeritOld )
   {
     printf("\t!!! Warning: Could not increase accuracy in this step\t!!!\n");
     printf("\t!!! \tChoice of feature might be arbitrary\t\t!!!\n");
   }
   maxMeritOld = maxMerit;
 }

 printf("\n");
 U->nrSelFeat = nrSelectFeat;
 FREE( U->FSV ); FREE( already_selected );
 U->FSV = (feature*) malloc( U->nrSelFeat * sizeof(struct feature_) );
 for( i = 0; i < U->nrSelFeat; i++ )
   copy_feature( &(FSV[i]), &(U->FSV[i]) );
 show_selected_feats( stdout, U->FSV, U->nrSelFeat );
 FREE( FSV );
}
