/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:28 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "def.h"

#ifdef UNIX
#include <unistd.h>
#endif


extern universe *U;
extern boolean verbose;
extern str100 outDataDir;
extern boolean feat_description;
extern char **feature_desc;


static boolean done;
static str80 buf, name;
static str100 snnsFile, snnsFileNet, snnsFilePat;
static boolean augmented = TRUE;
static int aug = 1, menuItem = EMPTY;


static struct digit_names {
        int len;
        str20 name;
} digit_name[] = {
        { 4,   "ZERO" },
        { 3,   "ONE" },
        { 3,   "TWO" },
        { 5,   "THREE" },
        { 4,   "FOUR" },
        { 4,   "FIVE" },
        { 3,   "SIX" },
        { 5,   "SEVEN" },
        { 5,   "EIGHT" },
        { 4,   "NINE" }
};

static char strange_char = '\0';

static void check_action( c )
char c;
{
  int digit;

  if( isdigit( c ) )
  {
    digit = (int)c-48;
    printf("# Will substitute the digit %d by its name '%s'",
	digit, digit_name[digit].name );
    printf(" to make it acceptable to SNNS.\n");
  }
  if( c == ' ' )
    printf("# Will substitute a blank by '_' to make it acceptable to SNNS.\n");
}

/*****************************************************************************
  FUNCTION : snns_symbolCheck

  PURPOSE  : spell checker  (check identifiers for matching [A-Za-z]^[|, ]*
******************************************************************************/
static boolean snns_symbolCheck(char *symbol)
{
  boolean ok = TRUE;
  char  c; int pos = 0;

  if( symbol == '\0' )
    return( FALSE );

  c = *symbol;
  check_action( c );
  if (!isalpha( c ))
  {
    strange_char = c;
    ok = isdigit( c );
  }

  while ( symbol[pos+1] != '\0' )
  {
    c = symbol[pos+1]; pos++;
    check_action( c );
    if (!isgraph( c ))
    {
      strange_char = c;
      if( isdigit( c ) || c == ' ' )
        ok = ok && TRUE;
      else
        ok = ok && FALSE;
    }

    if ( c == '|' || c == ',')
    {
      strange_char = c;
      if( isdigit( c ) || c == ' ' )
        ok = ok && TRUE;
      else
        ok = ok && FALSE;
    }
  }
  return( ok );
}



static void get_augmented()
{
 printf("Use augmented input vector (0-th component is fixed to value 1)?");
 get_boolean( augmented, &augmented );
}


static boolean gen_snns_network_file( f )
FILE *f;
{
 str200 timeStr, featName, className;
 char *featstr = NULL, *classNameStr = NULL;
 int u, i, j, k, x, y, feat, hidden, c, off, len, digit;
 boolean ok;

 aug = (int) (augmented == TRUE);
 /* Number of input units : Number of features */
 feat = U->nrSelFeat + aug;
 /* Number of hidden units : 2 * feat + 1 (KOLMOGOROV-Recommendation) */
 hidden = 2 * feat + 1;
 /* Number of output units : Number of classes */
 c = U->nrClass;
 /* Network : fully interconnected 3-layer feedforward */
 /* Learning method : Standard Error Backpropagation */

 fprintf( f, "SNNS network definition file V1.4-3D\n" );
 time_string( timeStr );
 fprintf( f, "generated at %s\n", timeStr );
 fprintf( f, "network name : TOOLDIAG\n" );
 fprintf( f, "source files :\n" );
 fprintf( f, "no. of units : %d\n", feat + hidden + c );
 fprintf( f, "no. of connections : %d\n", feat*hidden + hidden*c );
 fprintf( f, "no. of unit types : %d\n", 0 );
 fprintf( f, "no. of site types : %d\n\n\n", 0 );
 fprintf( f, "learning function : Std_Backpropagation\n" );
 fprintf( f, "update function   : Topological_Order\n\n\n" );

 fprintf( f, "unit definition section :\n\n" );
 fprintf( f, "no.    | typeName | unitName        | act      | bias     | ");
     fprintf( f, "st | position | act func | out func | sites\n" );
 fprintf( f, "-------|----------|-----------------|----------|----------|");
   fprintf( f, "----|----------|----------|----------|-------\n" );
 u = 1; x = 3; y = 2;
 if( augmented )
 {
   if( feat_description )
     strcpy( featName, "AUGMENTED=1" );
   fprintf( f, " %5d |          | %15s | %f | %f | i  | %d, %d, 0 |||\n",
     u, featName, 0.00000, 0.00000, x, y );
   x += 2; 
   u++;
 }
 for( i = 0; i < feat - aug; i++ )
 {
   off = 0;
   featName[0] = '\0';
   if( feat_description )
   {
     featstr = feature_desc[U->FSV[i].rank];
     ok = snns_symbolCheck(featstr);
     if( ! ok )
     {
       fprintf(stderr,
		"Feature or class name invalid (must match [A-Za-z]^[|, ]*)\n");
       fprintf(stderr,"Will try to substitute blanks by underscores and ");
       fprintf(stderr,"digits by their names\n\n");
     }
     for( j = 0; j < strlen(featstr); j++ )
     {
       if( isdigit(featstr[j]) )
       {
         digit = (int)featstr[j] - 48;
         len = digit_name[digit].len;
         for( k = 0; k < len; k++ )
           featName[j+off+k] = digit_name[digit].name[k];
         off += len-1;
       }
       else
       {
         /* printf("---> featstr[%d]=>%c<", j, featstr[j] ); WAIT;	/**/
         if( ! ok )
         {
           fprintf(stderr,"Found some for SNNS unacceptable character '%c'\n",
		strange_char );
           fprintf(stderr,"Change the feature names!\n");
           return( FALSE );
         }
         else
         {
           if( featstr[j] == ' ' )
             featName[j+off] = '_';
           else
             featName[j+off] = featstr[j];
         }
       }
     }
     featName[strlen(featstr)+off] = '\0';
     /* printf("$$$$ FINAL FEATURE NAME = >>>%s<<<\n", featName ); WAIT;/**/
   }
   fprintf( f, " %5d |          | %15s | %f | %f | i  | %d, %d, 0 |||\n",
     u, featName, 0.00000, 0.00000, x, y );
   x += 2; 
   u++;
 }
 x = 2; y += 3;
 for( i = 0; i < hidden; i++ )
 {
   fprintf( f,
	" %5d |          |                 | %f | %f | h  | %d, %d, 0 |||\n",
		u, 0.00000, 0.00000, x, y );
   x += 1;
   u++;
 }
 x = 4; y += 3;


 for( i = 0; i < c; i++ )
 {
   off = 0;
   classNameStr = U->C[i].name;
   ok = snns_symbolCheck(classNameStr);
   for( j = 0; j < strlen(classNameStr); j++ )
   {
     if( isdigit(classNameStr[j]) )
     {
       digit = (int)classNameStr[j] - 48;
       len = digit_name[digit].len;
       for( k = 0; k < len; k++ )
         className[j+off+k] = digit_name[digit].name[k];
       off += len-1;
     }
     else
     {
       if( ! ok )
       {
         fprintf(stderr,"Found some for SNNS unacceptable character '%c'\n",
		strange_char );
         fprintf(stderr,"Change the class names!\n");
         return( FALSE );
       }
       else
       {
         if( classNameStr[j] == ' ' )
           className[j+off] = '_';
         else
           className[j+off] = classNameStr[j];
       }
     }
   }
   /* printf("$$$$ FINAL CLASS NAME = >>>%s<<<\n", className ); WAIT; /**/

   fprintf( f, " %5d |          | %15s | %f | %f | o  | %d, %d, 0 |||\n",
     u, className, 0.00000, 0.00000, x, y );
   x += 2;
   u++;
 }
 fprintf( f, "-------|----------|-----------------|----------|----------|");
   fprintf( f, "----|----------|----------|----------|-------\n" );

 fprintf( f, "\n\nconnection definition section :\n\n" );
 fprintf( f, "target | site | source:weight\n" );
 fprintf( f, "-------|------|---------------------------------------------\n" );
 for( u = feat+1; u < feat+1+hidden; u++ )
 {
   fprintf( f, " %5d |      | ", u );
   for( i = feat; i > 0; i-- )
   {
     if( i > 1 )
       fprintf( f, " %d: 0.00000,", i );
     else
       fprintf( f, " %d: 0.00000", i );
   }
   fprintf( f, "\n" );
 }
 for( u = feat+1+hidden; u < feat+1+hidden+c; u++ )
 {
   fprintf( f, " %5d |      | ", u );
   for( i = feat+hidden; i > feat; i-- )
   {
     if( i > feat+1 )
       fprintf( f, " %d: 0.00000,", i );
     else
       fprintf( f, " %d: 0.00000", i );
   }
   fprintf( f, "\n" );
 }
 fprintf( f, "-------|------|---------------------------------------------\n" );
 return( TRUE );
}


static void gen_snns_pattern_file( f )
FILE *f;
{
 int i, j, k, c;
 str80 timeStr;

 aug = (int) (augmented == TRUE);
 fprintf( f, "SNNS pattern definition file V1.4\n" );
 time_string( timeStr );
 fprintf( f, "generated at %s\n\n", timeStr );
 fprintf( f, "No. of patterns     : %d\n", U->sumSampl );
 fprintf( f, "No. of input units  : %d\n", U->nrSelFeat+aug );
 fprintf( f, "No. of output units : %d\n\n\n", U->nrClass );
 fprintf( f, "#\n#\tPattern file for the universe: %s\n#\n", U->name );

 for( i = 0; i < U->nrClass; i++ )
 {
   fprintf( f, "#\n# %s\n#\n", U->C[i].name );
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     if( augmented )
       fprintf( f, "1.0 ");
     for( k = 0; k < U->nrSelFeat; k++ )
       fprintf( f, "%f ", U->C[i].S[j*U->nrFeat+U->FSV[k].rank] );
     for( c = 0; c < U->nrClass; c++ )
       if( c == i )
         fprintf( f, "1 " );
       else
         fprintf( f, "0 " );
     fprintf( f, "\n" );
   }
 }
}


static void gen_snns_data()
{
 FILE *snnsNet = NULL, *snnsPat = NULL;
 boolean ok;

 if( U->nrSelFeat == 0 )
   { printf(" Select features first please!..." ); gets( buf ); return; }

 strcpy( snnsFile, outDataDir );
 printf("Saving the net and pattern file in (without extension):\n\t\t%s",
	 snnsFile );
 gets( name );
 if( name[0] != '\0' )
 {
   strcat( snnsFile, name );
   strcpy( snnsFileNet, snnsFile ); strcat( snnsFileNet, ".net" );
   strcpy( snnsFilePat, snnsFile ); strcat( snnsFilePat, ".pat" );
   snnsNet = fopen( snnsFileNet, f_open_text_w );
   if( snnsNet == NULL )
     { printf("Cannot open %s! Exitus...\n", snnsFileNet ); exit(1); }
   snnsPat = fopen( snnsFilePat, f_open_text_w );
   if( snnsPat == NULL )
     { printf("Cannot open %s! Exitus...\n", snnsFilePat ); exit(1); }
   if( verbose )
   {
     printf("Generating Network File: %s\n", snnsFileNet );
     printf("Generating Pattern File: %s\n", snnsFilePat );
   }
   get_augmented();
   gen_snns_pattern_file( snnsPat );
   ok = gen_snns_network_file( snnsNet );
   fclose( snnsNet ); fclose( snnsPat );
#ifdef UNIX
   if( ! ok )
   {
     fprintf(stderr,"Deleting the two files\n'%s'\n'%s'\n",
		snnsFileNet, snnsFilePat );
     unlink( snnsFileNet ); unlink( snnsFilePat );
   }
#endif
 }
}


static void gen_snns_test()
{
 str80 raw, filtered;
 FILE *r, *f;
 str100 nameBuf;
 int i, k, c, clss, featDim, nrPat;
 str80 timeStr;
 FeatVector featBuf = NULL;
 boolean found;

 if( U->nrSelFeat == 0 )
   { printf(" Select features first please!..." ); gets( buf ); return; }

 printf("Name of the file to be filtered to test file? "); gets( raw );
 r = fopen( raw, f_open_text_r );
 if( r == NULL )
   { printf(" Cannot open %s...", raw ); gets( buf ); return; }

 strcpy( filtered, outDataDir );
 printf("Name of the output file with filtered features?\n\t\t%s", filtered );
 gets( buf );
 if( (strcmp(raw,buf) == 0) )
   { printf(" Sorry file names are the same!..." ); gets( buf ); return; }
 strcat( filtered, buf );
 strcat( filtered, ".pat" );
 f = fopen( filtered, f_open_text_w );
 if( f == NULL )
   { printf(" Cannot open %s...", raw ); gets( buf ); return; }

 if( verbose )
   printf("Filtering\n\t\t%s\nTO\n\t\t%s\n", raw, filtered );

 get_augmented();
 /* try to read the dimension */
 dataline( r, buf ); sscanf( buf, "%d", &featDim );
 if( featDim != U->nrFeat )
   { printf("Universe and file have different feature dimensions: %d != %d...",
	U->nrFeat, featDim  ); gets( buf ); fclose( r ); fclose( f ); return; }
 featBuf = (FeatVector) malloc(sizeof(FeatVector*) * U->nrFeat);

 /* first count the number of samples */
 nrPat = 0;
 while( !feof( r ) )
 {
   for( k = 0; k < U->nrFeat; k++ )
     fscanf( r, "%f", &(featBuf[k]) );
   fscanf( r, "%s", nameBuf );
   if( !feof( r ) )
     nrPat++;
 }
 fclose( r );
 r = fopen( raw, f_open_text_r );
 dataline( r, buf );

 fprintf( f, "SNNS pattern definition file V1.4\n" );
 time_string( timeStr );
 fprintf( f, "generated at %s\n\n", timeStr );
 fprintf( f, "No. of patterns     : %d\n", nrPat );
 fprintf( f, "No. of input units  : %d\n", U->nrSelFeat+aug );
 fprintf( f, "No. of output units : %d\n\n\n", U->nrClass );
 fprintf( f, "#\n#\tTest pattern file for the universe: %s\n#\n", U->name );

 nrPat = 0;
 while( !feof( r ) )
 {
   for( k = 0; k < U->nrFeat; k++ )
     fscanf( r, "%f", &(featBuf[k]) );
   fscanf( r, "%s", nameBuf );
   if( !feof( r ) )
   {
     nrPat++;
     fprintf( f, "# Pattern nr. %d - class=%s\n", nrPat, nameBuf );
     if( augmented )
       fprintf( f, "1.0 ");
     for( i = 0; i < U->nrSelFeat; i++ )
       fprintf( f, "%f ", featBuf[ U->FSV[i].rank ] );
     /* search the class that matches the string */
     found = FALSE; clss = 0;
     while( ! found && clss < U->nrClass )
     {
       found = strcmp(nameBuf,U->C[clss].name)==0;
       if( ! found )
         clss++;
     }
     if( clss >= U->nrClass )
     {
       printf("\nERROR: Unknown class %s at line %d - exitus...\n",
             nameBuf, nrPat ); exit(1);
     }

     for( c = 0; c < U->nrClass; c++ )
       if( c == clss )
         fprintf( f, "1 " );
       else
         fprintf( f, "0 " );
     fprintf( f, "\n" );
   }
 }
 FREE( featBuf );
 fclose( r ); fclose( f );
}



#define GEN_DATA_FILE_SNNS	0
#define FILTER_SNNS		1

#define NUM_SNNS_ITEMS 2


static struct SNNS_MenuItems {
        int index;
        str80 name;
} SNNS_MenuItem[] = {
        { GEN_DATA_FILE_SNNS,	"Generate network and training pattern file" },
        { FILTER_SNNS,
 "Filter only selected features from a file and generate test pattern file" } };


static void execMenuItem()
{
 switch( menuItem )
 {
   case GEN_DATA_FILE_SNNS:	gen_snns_data(); break;
   case FILTER_SNNS:		gen_snns_test(); break;
   default: fprintf(stderr,"What the heck is item %d? - exit...\n", menuItem );
 }
}


static char sl[] = "-----------------";

void snnsMenu()
{
 int i, d; char *str = NULL;

 do
 {
   printf("\n%s  STUTTGART NEURAL NETWORK SIMULATOR (SNNS)  %s\n", sl, sl );
   for( i = 0; i < NUM_SNNS_ITEMS; i++ )
   {
     str = SNNS_MenuItem[i].name;
     printf("(%d)\t%s\n", i+1, str );
   }
   printf("Choice: ");
   gets(buf); done = FALSE;
   switch( buf[0] )
   {
     /* special cases */
     case '?': help( MENU_SNNS, buf ); break;
     case '\0': done = TRUE; break;
     default:
       sscanf(buf,"%d", &d);
       if( d >= 1 && d <= NUM_SNNS_ITEMS )
       {
         menuItem = d-1;
         execMenuItem();
       }
   }
 }
 while (! done);
}


/**********  H E L P  *****************/

void helpMENU_SNNS( cmd )
char *cmd;
{
 int item;

 sscanf( &(cmd[1]), "%d", &item );
 item--;        /* discount one since in the menu it is one more */
 switch( item )
 {
   case GEN_DATA_FILE_SNNS:	item_help("GEN_DATA_FILE_SNNS"); break;
   case FILTER_SNNS:		item_help("FILTER_SNNS"); break;
   default: ; break;
 }
}
