/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:28 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "def.h"

extern universe *U;


boolean *train_smp = NULL, *test_smp = NULL; /* Lookup for samples */
int nrSamplesTest, nrSamplesTrain;

/* A table with an index for each training sample, useful for
   random presentation of the samples to the training algorithm:
	1.) Class label
	2.) Sample nr. of that class
*/
int *train_vec = NULL, *train_vecClass = NULL, *train_vecSmp = NULL;


/*
 *  Shuffle the array 'orig' of size 'dim' randomly
 *  and put the result into 'shuffled'
*/
void shuffle( orig, shuffled, dim )
int *orig, *shuffled, dim;
{
 int i, j;

 for( i = 0; i < dim; i++ )
 {
   get_random( dim-i, &j ); 
   shuffled[i] = orig[j];
   orig[j] = orig[dim-1-i];
 }
}


/* Generate a training vector containing the class index and sample nr */
void gen_train_vec( random_order )
boolean random_order;
{
 int i, j, row, s, *shuffled_train = NULL;

 if( nrSamplesTrain == 0 )
   EXIT("Had no training samples");

 s = 0; row = 0;
 for( i = 0; i < U->nrClass; i++ )
 {
   for( j = 0; j < U->C[i].numSampl; j++ )
   {
     if( train_smp[row] )
     {
       train_vec[s] = s;
       train_vecClass[s] = i;
       train_vecSmp[s] = j;
       s++;
     }
     row++;
   }
 }
 if( s != nrSamplesTrain )
 {
   fprintf(stderr, "Inconsistence in training samples %d != %d. Exitus...\n",
	s, nrSamplesTrain );
   exit(1);
 }
 if( random_order )
 {
   shuffled_train = (int*) malloc( s * sizeof(int) ); CHKPTR(shuffled_train);
   shuffle( train_vec, shuffled_train, nrSamplesTrain );
   FREE( train_vec );
   train_vec = shuffled_train;
 }
}



static char intValStr[201];
static str80 buf;

#define MAX_TRIALS 1000


static void peep_debug_tt( nrSamples, str )
int nrSamples;
char *str;
{
 int s, c = 0, sum = 0, train = 0, test = 0;
 fprintf( stderr, "%s\n", str);
 for( s = 0; s < nrSamples; s++ )
 {
   if( s == U->C[c].numSampl+sum )
   {
     sum += U->C[c].numSampl; c++;
     fprintf( stderr,"----------------------------------------------------\n");
   }
   fprintf( stderr,
	"%3d of %3d  train: %d  test: %d ---  %3d of %3d of class=%d(%s)\n",
        s+1, nrSamples, train_smp[s], test_smp[s], s-sum+1, U->C[c].numSampl,
	c+1, U->C[c].name ); /**/
   if( !train_smp[s] && !test_smp[s] )
        {fprintf(stderr,"Sample type error; exit...\n"); exit(1);}
   train += (int)train_smp[s]; test += (int)test_smp[s];
 }
 fprintf( stderr, "No. train=%d(%7.2f%%)  No. test=%d(%7.2f%%)  All=%d\n",
	train, 100.0*(float)train/(float)nrSamples,
	test, 100.0*(float)test/(float)nrSamples, nrSamples );
 for( s = 0; s < 70; s++ )
   fprintf( stderr, "*");
 fprintf( stderr, "\n");
}


void free_tt_lookups()
{
 nrSamplesTrain = 0; nrSamplesTest = 0;
 FREE( train_smp ); FREE( test_smp );
 /* training sample vector */
 FREE(train_vec); FREE(train_vecClass); FREE(train_vecSmp);
}


static void init_tt_lookups( nrSmp )
int nrSmp;
{
 int i;

 free_tt_lookups();
 train_smp = (boolean*) malloc( nrSmp * sizeof( boolean ) );
 test_smp = (boolean*) malloc( nrSmp * sizeof( boolean ) );
 for( i = 0; i < nrSmp; i++ )
 {
   train_smp[i] = EMPTY;
   test_smp[i] = EMPTY;
 }

 /* training sample vector */
 train_vec = (int*) malloc( nrSmp * sizeof( int ) );
 train_vecClass = (int*) malloc( nrSmp * sizeof( int ) );
 train_vecSmp = (int*) malloc( nrSmp * sizeof( int ) );
 for( i = 0; i < nrSmp; i++ )
 {
   train_vec[i] = i;
   train_vecClass[i] = EMPTY;
   train_vecSmp[i] = EMPTY;
 }
}


/*
 *  Setting all samples in memory as training samples
*/
void set_all_samples_train()
{
 int i;

 init_tt_lookups( U->sumSampl );
 nrSamplesTrain = U->sumSampl;
 for( i = 0; i < U->sumSampl; i++ )
   { train_smp[i] = TRUE; test_smp[i] = FALSE; } 
}


/*
 *  Setting one sample in memory as test sample
*/
void set_one_sample_test( nr )
int nr;
{
 if( nr != EMPTY )
 { 
   /* printf("Setting sample # %d as test\n", nr ); /**/
   if( nr < 0 || nr >= U->sumSampl ) EXIT("sample nr not valid");
   nrSamplesTrain--; nrSamplesTest++;
   if( nrSamplesTrain < 0 || nrSamplesTest > U->sumSampl) EXIT("Invalid index");
   train_smp[nr] = FALSE; test_smp[nr] = TRUE;
 }
 /* peep_debug_tt( U->sumSampl, "Peeping in 'set_one_sample_test'");/**/
}


void reset_one_sample_test( nr )
int nr;
{
 if( nr != EMPTY )
 { 
   if( nr < 0 || nr >= U->sumSampl ) EXIT("Invalid sample nr");
   nrSamplesTrain++; nrSamplesTest--;
   if( nrSamplesTest < 0 || nrSamplesTrain > U->sumSampl) EXIT("Invalid index");
   train_smp[nr] = TRUE; test_smp[nr] = FALSE;
 }
}


void getPercentTrain( pt )
float *pt;
{
 boolean ok; float f = 70.0;

 printf("Split data into ? percent training data? 70\b\b");
 do
 {
   gets( buf );
   if( buf[0] == '\0' )
     ok = TRUE;
   else
   {
     sscanf( buf,"%f", &f );
     ok = ( f > 0.0 && f < 100.0 );
   }
   if( ! ok )
     printf("Invalid value! Again ? ");
   else
     *pt = f;
 }
 while( ! ok );
}


boolean at_least2test()
{
 boolean ok; int i, j, s = 0;
 int *smpClass = NULL;

 smpClass = (int*) malloc( U->nrClass * sizeof(int) ); CHKPTR(smpClass);
 for( i = 0; i < U->nrClass; i++ )
   smpClass[i] = 0;

 /* check if the training set has at least two samples for each class */
 ok = TRUE; s = 0; i = 0;
 while( ok && i < U->nrClass )
 {
   j = 0;
   while( ok && j < U->C[i].numSampl )
   {
     if( train_smp[s] )
       (smpClass[i])++;
     s++; j++;
   }
   ok = (smpClass[i] >= 2) && ok;
   i++;
 }
 FREE(smpClass);
 return( ok );
}


/*
 * random splitting of the entire data set into x% training samples
 *  and 100-x% test samples
*/
static void gen_tt_random( len, percent )
int len;
float percent;
{
 int i, *orig = NULL, *shuffled = NULL;

 orig = (int*) malloc( len * sizeof(int) ); CHKPTR(orig);
 shuffled = (int*) malloc( len * sizeof(int) ); CHKPTR(shuffled);
 for( i = 0; i < len; i++ )
   { orig[i] = i; shuffled[i] = EMPTY; }

 free_tt_lookups();
 init_tt_lookups( len );

 nrSamplesTrain = (int)(len * percent / 100.0);

 shuffle( orig, shuffled, len );
 for( i = 0; i < nrSamplesTrain; i++ )
   train_smp[shuffled[i]] = TRUE;

 for( i = 0; i < len; i++ )
   if( train_smp[i] == TRUE )
     test_smp[i] = FALSE;
   else
   {
     train_smp[i] = FALSE;
     test_smp[i] = TRUE;
   }
 nrSamplesTest = len - nrSamplesTrain;
 FREE(orig); FREE(shuffled);
}


void split_tt_random( nrSamples, percent_train )
int nrSamples;
float percent_train;
{
 int trials = 0, maxTrials = MAX_TRIALS;
 boolean ok;

 if( percent_train < 0.0 || percent_train > 100.0 )
   { fprintf( stderr, "Percent error: %5.2f %%\n", percent_train); exit( 1 ); }

 do {
   gen_tt_random( nrSamples, percent_train );
   ok = at_least2test();
   ok = nrSamplesTest >= 1 && ok;
   if( trials > 1)
   {
     fprintf(stderr,"Trying to split randomly. Trial # %d\r", trials );
     /* fprintf(stderr,"nrSamplesTest=%d\n", nrSamplesTest );	/**/
   }
   trials++;
 }
 while( ! ok && trials <= maxTrials );

 if( ! ok )
 {
   fprintf(stderr,"split_tt_random> Unable to split data such that");
   fprintf(stderr," each class\n\thas at least 2 training samples");
   fprintf(stderr," and at least one test sample !");
   EXIT("");
 }
 /* peep_debug_tt( nrSamples, "Peeping..." ); DBG;	/**/
}


/*
 * Fill array 'subsets' randomly with a shuffled sequence
 *
*/
void gen_tt_ROTATION( subsets )
int *subsets;
{
 int i, len, *orig = NULL;

 len = U->sumSampl;
 orig = (int*) malloc( len * sizeof(int) ); CHKPTR(orig);
 for( i = 0; i < len; i++ )
   { subsets[i] = EMPTY; orig[i] = i; }

 shuffle( orig, subsets, len );

 FREE(orig);
 /* for( i = 0; i < len; i++ ) printf("ss[%3d]=%d\n", i, subsets[i] ); /**/
}


void split_tt_bootstrap( len )
int len;
{
 int i, j, trials = 0, maxTrials = MAX_TRIALS;
 boolean ok;
 
 free_tt_lookups();
 init_tt_lookups( len );

 do {
   nrSamplesTrain = 0;
   for( i = 0; i < len; i++ )
     { test_smp[i] = TRUE; train_smp[i] = FALSE; }
   for( i = 0; i < len; i++ )
   {
     get_random( len, &j );
     train_smp[j] = TRUE;
     test_smp[j] = FALSE;
   }
   for( i = 0; i < len; i++ )
     if( train_smp[i] )
       nrSamplesTrain++;
   nrSamplesTest = len - nrSamplesTrain;

   /* check if the training set has at least tw( samples for each class */
   ok = at_least2test();
   /* check if there is at least one test sample */
   ok = nrSamplesTest >= 1 && ok;
   trials++;
   if( trials > 1)
     fprintf(stderr,"Trying to split for bootstrap. Trial # %d\r", trials );
 }
 while ( ! ok && trials <= maxTrials );

 if( ! ok )
 {
   fprintf(stderr,"split_tt_bootstrap> Unable to split data such that");
   fprintf(stderr," each class\n\thas at least 2 training samples");
   fprintf(stderr," and at least one test sample !");
   EXIT("");
 }
 /* peep_debug_tt( len, "split_tt_bootstrap: Peeping..." );   /**/
}


static boolean parseIntValStr( str, row, maxSmp )
char *str;
int row, maxSmp;
{
 boolean valid = TRUE;
 int nrGroups = 1, ptr = 0, ptrEnd, len, i, j, max = 0, iniNum, endNum;
 int nrTrain = 0, nrTest = 0;
 str80 gr, iniGr, endGr;

 /* printf("str=>>>%s<<< row=%d maxSmp=%d\n",  str, row, maxSmp );/**/
 len = strlen( str );
 if( len == 0 ) return( FALSE );

 /* reset first */
 for( j = row; j < row+maxSmp; j++ )
   { train_smp[j] = FALSE; test_smp[j] = TRUE; }
   
 /* search for commas in the string */
 for( ptr = 0; ptr < len; ptr++ )
 {
   if( str[ptr] == ',' )
     nrGroups++;
 }
 /* printf("str=>>>%s<<<   len=%d  nrGroups=%d\n", str, len, nrGroups ); /**/
 ptr = 0;
 for( i = 0; i < nrGroups; i++ )
 {
   ptrEnd = ptr;
   while( str[ptrEnd+1] != ',' && str[ptrEnd+1] != '\0' )
     ptrEnd++;
   /* found the next group - analyze it */
   for( j = ptr; j < ptrEnd+1; j++ )
     gr[j-ptr] = str[j];
   gr[ptrEnd-ptr+1] = '\0';
   /* printf("ptr=%d  ptrEnd=%d  gr = >>>%s<<<", ptr, ptrEnd, gr ); DBG; /**/
   /* analyze group string - has to be two integers separated by a hyphen */

   ptr = 0;
   while( gr[ptr] != '-' && gr[ptr] != '\0' )
     ptr++;
   if( gr[ptr] != '-' || gr[0] == '\0' )
     return( FALSE );
   strncpy( iniGr, gr, ptr ); iniGr[ptr] = '\0';
   strncpy( endGr, &(gr[ptr+1]), strlen(gr)-ptr );

   if( iniGr[0] == '\0' || endGr[0] == '\0' )
     return( FALSE );

   /* found two strings with potential numbers */
   if( strlen(endGr) == 1 && endGr[0] == '$' )
     sprintf(endGr, "%d", maxSmp ); 
   if( strlen(iniGr) == 1 && iniGr[0] == '$' )
     sprintf(iniGr, "%d", maxSmp ); 

   sscanf( iniGr, "%d", &iniNum ); sscanf( endGr, "%d", &endNum );
   /* check them */
   if( iniNum > endNum || iniNum <= 0 || iniNum > maxSmp || endNum > maxSmp
	|| iniNum < max || endNum < max )
     return( FALSE );

   /* printf("iniGr=>>>%s<<<  endGr=>>>%s<<< max=%d ###### ini=%d end=%d",
	iniGr, endGr, max, iniNum, endNum ); NL; /**/

   /* finally got the right numbers; labeling training samples */
   for( j = row+iniNum-1; j < row+endNum; j++ )
     { train_smp[j] = TRUE; test_smp[j] = FALSE; }
   
   /* next group */
   max = endNum;
   ptr = ptrEnd+2;
   nrTrain += endNum-iniNum+1; nrTest = maxSmp - nrTrain;
 }
 if( nrTrain == maxSmp )
 {
   printf("WARNING: Really all samples of this class as training (y/n)n\b");
   gets( buf );
   valid = (buf[0] == 'Y') || (buf[0] == 'y');
   if( valid )
   {
     printf("\tThis might cause big trouble during training and test...\n");
     printf("\tReally set all samples of this class as training (y/n)n\b");
     gets( buf );
     valid = (buf[0] == 'Y') || (buf[0] == 'y');
   }
 }
 if( valid )
     { nrSamplesTrain += nrTrain; nrSamplesTest += nrTest; }
 return( valid );
}


void split_tt_explicit( nrSamples )
int nrSamples;
{
 int i, row = 0; boolean ok, equalAll = TRUE, specified = FALSE;

 free_tt_lookups();
 init_tt_lookups( nrSamples );
 /* initialize set all as test samples */
 for( i = 0; i < nrSamples; i++ )
   { train_smp[i] = FALSE; test_smp[i] = TRUE; }
 printf("\nTotal number of all samples for the %d classes is %d\n",
	U->nrClass, nrSamples );
 printf("Specify intervals for the training samples of each class !\n");
 printf("\tFor instance  1-5,9-20,30-$\n");
 printf("\t which means 'from-to' (separated by commas);\n");
 printf("\t '$' means 'last sample of class'\n\n");
 printf("Specification shall be valid for all classes (y/n)y\b");
 gets( buf );
 if( buf[0] == 'N' || buf[0] == 'n')
   equalAll = FALSE;

 if( equalAll )
 {
   for( i = 0; i < U->nrClass; i++ )
     printf("Class %d = %s has %d samples\n",i+1,U->C[i].name,U->C[i].numSampl);
   i = 0;
   while( i < U->nrClass )
   {
     if( ! specified )
     {
       printf("Specify interval(s):\n");
       gets( intValStr );
       specified = TRUE;
     }
     ok = parseIntValStr( intValStr, row, U->C[i].numSampl );
     if( ! ok )
     {
       printf("Invalid specification for class %d ! Type again!\n",i+1);
       specified = FALSE;
     }
     else
     {
       row += U->C[i].numSampl;
       i++;
     }
   }
 }
 else
 {
   for( i = 0; i < U->nrClass; i++ )
   {
     printf("Class %d = %s has %d samples\n",i+1,U->C[i].name,U->C[i].numSampl);
     do
     {
       printf("Specify interval(s):\n");
       gets( intValStr );
       ok = parseIntValStr( intValStr, row, U->C[i].numSampl );
       if( ! ok )
       {
         printf("Invalid specification for class %d ! Type again!\n",i+1);
       }
     }
     while( ! ok );
     row += U->C[i].numSampl;
   }
 }
 /* peep_debug_tt( nrSamples, "Explicit split. Peeping..." );   /**/
 nrSamplesTest = nrSamples - nrSamplesTrain;
 printf("Explicit splitting ok. Train = %d  Test = %d\n",
	nrSamplesTrain, nrSamplesTest );
}
