/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:29 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "def.h"


/*
 * Check if a file is binary or ascii
 *  returns EMPTY if file cannot be found
*/
boolean is_ascii_file( fname )
char *fname;
{
 FILE *f; byte b;
 boolean ascii_file;

 f = fopen( fname, f_open_bin_r );
 if( f == NULL )
    { /* printf("Cannot open %s!...\n", fname ); */ return( EMPTY ); }
 ascii_file = TRUE;
 while( !feof( f ) && ascii_file )
 {
   fread( &b, sizeof(byte), 1, f );
   if( !feof( f ) )
   {
     /* check if there is a non-ascii character */
     ascii_file = ascii_file && (boolean)isascii((int)b);
   }
 }
 fclose( f );
 return( ascii_file );
}


/* determine the minimum of a vector and the index */
void getMin( vec, dim, min, minIndx )
float *vec, *min;
int dim, *minIndx;
{
 int i, minI;
 float m = INFINITY;

 for( i = 0; i < dim; i++ )
   if( vec[i] < m )
   {
     m = vec[i];
     minI = i;
   }
 if( m == INFINITY )
    { fprintf(stderr, "getMin> Problems finding the minimum\n"); }
 *min = m; *minIndx = minI;
}


/* determine the maximum of a vector and the index */
void getMax( vec, dim, max, maxIndx )
float *vec, *max;
int dim, *maxIndx;
{
 int i, maxI;
 float m = -INFINITY;

 for( i = 0; i < dim; i++ )
   if( vec[i] > m )
   {
     m = vec[i];
     maxI = i;
   }
 if( m == -INFINITY )
    { fprintf(stderr, "getMax> Problems finding the maximum\n"); }
 *max = m; *maxIndx = maxI;
}



void copy_FV( src, dest, dim )
FeatVector src, dest;
int dim;
{
 int i;

 for( i = 0; i < dim; i++ )
   dest[i] = src[i];
}


void showFV( numFeat, FV )
int numFeat;
FeatVector FV;
{
 int i;

 printf("    (");
 for( i = 0; i < numFeat; i++ )
   printf("  %.5f  ", FV[i]);
 printf(")^T\n");
}


void fshowFV( f, numFeat, FV )
FILE *f;
int numFeat;
FeatVector FV;
{
 int i;

 fprintf(f,"    (");
 for( i = 0; i < numFeat; i++ )
   fprintf(f,"  %.5f  ", FV[i]);
 fprintf(f,")^T\n");
}


#define LEN 1001
static char buf[LEN];

void dataline( f, line )
FILE *f;
char *line;
{
 int i;

 for( i = 0; i < LEN; i++ )
   buf[i] = '\0';
 do
 {
   fgets( buf, LEN, f );
   /* printf("%s", buf ); /**/
 } while( (buf[0] == '#') && (!feof(f)) );
 i = 0;
 while( (buf[i] != '\0') && (buf[i] != '\n') && (buf[i] != '#') && (!feof(f)) )
   i++;
 if( i >= LEN )
 {
   fprintf(f,"Read a line over %d characters long. Too long!\n", LEN );
   WAIT;
 }
 buf[i] = '\0';
 /* printf("buf=>>>%s<<<", buf ); /**/
 strcpy( line, buf );
}


/* factorial of n */
#define MAX_FACTORIAL 100
double factorial( n )
int n;
{
 double prod = 1.0;
 int i;

 if( n < 0 || n > MAX_FACTORIAL )
   return( (double)EMPTY );
 if( n == 0 )
   return( 1.0 );
 for( i = 1; i <= n; i++ )
   prod *= (double)i;
 return( prod );
}

/* binomial coefficient */
double bin_coeff( n, k )
int n, k;
{
 double bc, nom = 1.0,  denom = 1.0;
 int i, q;

 if( k > n )
   return( (double)EMPTY );;
 if( k == n )
   return( 1.0 );
 if( k > n/2 )
   q = k;
 else
   q = n - k;

 for( i = q+1; i <= n; i++ )
   nom *= (double)i;
 for( i = 1; i <= n-q; i++ )
   denom *= (double)i;
 bc = nom / denom;
 return( bc );
}


static char booleanChr[3] = "ny";

void get_boolean( boolean_default, b )
boolean boolean_default, *b;
{
 char buf[81]; 

 printf(" (y/n)%c\b", booleanChr[boolean_default] );
 buf[0] = '\0'; gets(buf);
 if( buf[0] == '\0' )
   *b = boolean_default;
 else
 {
   switch( boolean_default )
   {
     case FALSE:
       *b = buf[0] == 'y' || buf[0] == 'Y';  break;
     case TRUE:
       *b = !(buf[0] == 'n' || buf[0] == 'N');  break;
     default: fprintf(stderr,"Boolean=%d should never happen", boolean_default );
       EXIT("");
   }
 }
}


/* Checks is 'str' contains only digits */
static boolean is_d( str )
char *str;
{
 int i;

 for( i = 0; i < strlen(str); i++ )
   if( !isdigit( str[i] ) )
     return( FALSE );
 return( TRUE );
}


void get_d_range( d, min, max, bound )
int *d, min, max, bound;
{
 int old;
 boolean ok;
 char buf[81];

 old = *d;

 do
 {
   buf[0] = '\0'; gets(buf);
   if( (buf[0] == '\0') || (! is_d(buf)) )
     *d = old;
   else
     sscanf(buf,"%d",d);
   switch( bound )
   {
     case LEFT_CLOSED__RIGHT_CLOSED : ok = ( *d >= min && *d <= max ); break;
     case LEFT_CLOSED__RIGHT_OPEN : ok = ( *d >= min && *d < max ); break;
     case LEFT_OPEN__RIGHT_CLOSED : ok = ( *d > min && *d <= max ); break;
     case LEFT_OPEN__RIGHT_OPEN : ok = ( *d > min && *d < max ); break;
     default : fprintf(stderr,"Trouble with interval '%d' - exit...\n", bound );
		 exit(1);
   };
   if( ! ok )
     printf("Invalid value! Again ? ");
 }
 while( ! ok );
}


void get_f_range( f, min, max, bound )
float *f, min, max;
int bound;
{
 float old;
 boolean ok;
 char buf[81];

 old = *f;
 do
 {
   buf[0] = '\0'; gets(buf);
   if( buf[0] == '\0' )
     *f = old;
   else
     sscanf(buf,"%f",f);
   switch( bound )
   {
     case LEFT_CLOSED__RIGHT_CLOSED : ok = ( *f >= min && *f <= max ); break;
     case LEFT_CLOSED__RIGHT_OPEN : ok = ( *f >= min && *f < max ); break;
     case LEFT_OPEN__RIGHT_CLOSED : ok = ( *f > min && *f <= max ); break;
     case LEFT_OPEN__RIGHT_OPEN : ok = ( *f > min && *f < max ); break;
     default : fprintf(stderr,"Trouble with interval - exit...\n"); exit(1);
   };
   if( ! ok )
     printf("Invalid value! Again ? ");
 }
 while( ! ok );
}


/*
 * extracts only the filename of a full path name
 *  for instance "../data/file" becomes "file"
 *
 */
#ifdef UNIX
static char slash = '/';
#endif
#ifndef UNIX
static char slash = '\\';
#endif

void root_name( str, rootstr )
char *str, *rootstr;
{
 int i, offset = 1;

 if( str[strlen(str)-1] == slash )
 {
   fprintf(stderr, "Name %s is invalid!\n", str );
   rootstr[0] = '\0';
   return;
 }
 if( str[0] == '\0' )
 {
   rootstr[0] = '\0';
   return;
 }
 i = strlen( str ) - 1;
 while( str[i] != slash && i > 0 )
   i--;
 if( i == 0 )
   offset = 0;
 strncpy( rootstr, &(str[i+offset]), strlen(&(str[i+offset])) );
 rootstr[strlen(&(str[i+offset]))] = '\0';
}


/* For debugging reasons - check memory allocation */
/* #define SPY_MEMORY	/**/

#ifdef SPY_MEMORY
#include <malloc.h>
static struct mallinfo mi;

void debug_alloc_spy()
{
 mi = mallinfo();
 printf("\n\n----  M E M O R Y  A L L O C A T I O N - I N F O   -----\n");
 printf("total space in arena=                 %7d\n", mi.arena );	/**/
 printf("number of ordinary blocks=            %7d\n", mi.ordblks );	/**/
/* printf("number of small blocks=               %7d\n", mi.smblks );	/**/
/* printf("number of holding blocks=             %7d\n", mi.hblks );	/**/
/* printf("space in holding block headers=       %7d\n", mi.hblkhd );	/**/
/* printf("space in small blocks in use=         %7d\n", mi.usmblks );	/**/
/* printf("space in free small blocks=           %7d\n", mi.fsmblks );	/**/
 printf("space in ordinary blocks in use=      %7d\n", mi.uordblks );	/**/
 printf("space in free ordinary blocks=        %7d\n", mi.fordblks );	/**/
/* printf("cost of enabling keep option=         %7d\n", mi.keepcost );	/**/
 printf("--------------------------------------------------------\n");
}
#endif
