/************************************************************************
 *                                                                      *
 *  Program package         T O O L D I A G                             *
 *                                                                      *
 *  Version 2.1.1                                                       *
 *  Date:     Thu Sep 17 11:44:29 1998                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'thomas@inf.ufes.br'.                                *
 *                                                                      *
 ************************************************************************/


/************************************************************************


        BUILD YOUR OWN CLASSIFIER AND INTEGRATE IT INTO THE
	FRAMEWORK OF FEATURE SELECTION AND ERROR ESTIMATION
	IN ORDER TO COMPARE IT TO OTHER ARCHITECTURES


 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "def.h"
#include "clasdef.h"

extern universe *U;
extern boolean verbose_reason;
extern float *vector();
extern int *int_vector();
extern int nrSamplesTrain, nrSamplesTest;
extern int *train_vec, *train_vecClass, *train_vecSmp;

extern FILE *cpf;               /* Classification protocol file */
extern boolean have_label;      /* Only deliver the classified class for
        pre-classified data */


static boolean verbose, interact = FALSE;


void show_YUC( f )
FILE *f;
{
 fprintf( f, "\tSPECIFICATIONS FOR YOUR CLASSIFIER:\n");
}


void initYUC()
{
 free_induced();
}


/* Present the samples in random order to the training algorithm or not */
static boolean stoch_learn = TRUE;	/**/
/* static boolean stoch_learn = FALSE;	/**/

/*
 * Induce a classifier with the actual set of training samples
 * The dimension of the vector is 'nsf'
 * 'FSV' contains the information of the actual subset of features
 * that are used.
 * 'iact' tells if interaction if desired. For instance in error
 * estimation this is not the case
*/
void induceYUC( nsf, FSV, iact )
int nsf;
FeatSelectVector FSV;
boolean iact;
{
 int i, indx, clss, k, p;
 FeatVector smp = NULL;
 int *nrTrainClass = NULL;
 boolean random_order;

 interact = iact;
 verbose = FALSE;
 if( interact )
 {
   printf("\n --- Inducing a %d-dimensional classifier ---\n\n", nsf);
   printf("Verbose learning of decision structure?");
   get_boolean( FALSE, &verbose );
 }

 /* Allocate a buffer for one sample of dimension 'nsf' */
 smp = vector( nsf );
 /* Allocate a buffer for the number of training samples of each class */
 nrTrainClass = int_vector( U->nrClass );
 for( i = 0; i < U->nrClass; i++ )
   nrTrainClass[i] = 0;
 
 printf("The currently used features are:\n");
 for( i = 0; i < nsf; i++ )
   printf("Index of feature nr %5d is %5d\n", i+1, FSV[i].rank );
 printf("\nThe total number of samples is %d\n", U->sumSampl );
 printf("The total number of training samples is %d\n", nrSamplesTrain );
 WAIT;


 if( stoch_learn )
 {
   random_order = TRUE;
   gen_train_vec( random_order );
   for( p = 0; p < nrSamplesTrain; p++ )
   {
     indx = train_vec[p];
     clss = train_vecClass[ indx ];
     k = train_vecSmp[ indx ];
     (nrTrainClass[ clss ])++;

     provideFV( smp, clss, k, nsf, FSV );
     printf("\nSample # %d which is %d of class %d(%s)\n",
		p+1, k+1, clss+1, U->C[clss].name );
     printf("is used as the %d-th training sample of this class.\n",
                nrTrainClass[clss]+1 );
     printf("The sample is: ");
     fshowFV( cpf, nsf, smp ); WAIT;

     /*---------------------------------------------------------------*/
     /*       USE IT FOR THE INDUCTION OF YOUR CLASSIFIER             */
     /*---------------------------------------------------------------*/


   }	/* next pattern */
   for( i = 0; i < U->nrClass; i++ )
     printf("Class %d(%s) has %d trainig samples.\n",
	i+1, U->C[i].name, nrTrainClass[i] );
 
 }

 else	/* BATCH LEARNING */
 {
   random_order = FALSE;
   gen_train_vec( random_order );

   for( p = 0; p < nrSamplesTrain; p++ )
   {
     indx = train_vec[p];
     clss = train_vecClass[ indx ];
     k = train_vecSmp[ indx ];
     (nrTrainClass[ clss ])++;

     provideFV( smp, clss, k, nsf, FSV );
     printf("\nSample # %d which is %d of class %d(%s)\n",
		p+1, k+1, clss+1, U->C[clss].name );
     printf("is used as the %d-th training sample of this class.\n",
                nrTrainClass[clss]+1 );
     printf("The sample is: ");
     fshowFV( cpf, nsf, smp ); WAIT;

     /*---------------------------------------------------------------*/
     /*       USE IT FOR THE INDUCTION OF YOUR CLASSIFIER             */
     /*---------------------------------------------------------------*/



   }	/* next pattern */
   for( i = 0; i < U->nrClass; i++ )
     printf("Class %d(%s) has %d trainig samples.\n",
	i+1, U->C[i].name, nrTrainClass[i] );

 } /* end BATCH */


 FREE(smp);
 FREE(nrTrainClass);
}


/*
 *  Free all the structures that were dynamically allocated to store
 *  the classifier, for instance matrices of prototypes
*/
void free_inducedYUC()
{
}


/*
 *  Classify a single sample 'smp' of length 'nsf'
 *  The name of the expected class is stored in 'name'
 *  'FSV' is not needed here, but for compatibility reasons passed
*/
int class1testYUC( nsf, FSV, smp, name )
int nsf;
FeatSelectVector FSV;
FeatVector smp;
char *name;
{
 int i, resulting_class, best_class;

 if( verbose_reason )
 {
   fprintf(cpf, "\n\n ---- DECISION FUNCTIONS FOR UNKNOWN SAMPLE:\n");
   fshowFV( cpf, nsf, smp );
 }

 /* decide the best class, using your classifier: */
 best_class = 0;

 /* For data without a class label, return the classified label */
 if( ! have_label )
   return( best_class );

 /* 
    First case: correct classification:
    The name of the best class corresponds to the expected class
    return the value 'CORRECT_CLASS'
 */
 if( (strcmp(U->C[best_class].name,name)==0) )
   resulting_class = CORRECT_CLASS;
 else
 {
   /* 
      Second case: wrong classification:
      return the value of the best class: 'best_class'
   */
   resulting_class = best_class;
   if( verbose_reason )
     fprintf( cpf, " *** Confused %s with expected class %s\n",
		U->C[best_class].name, name );
 }
 return( resulting_class );
}
